//*****************************************************************************
//*
//*
//*		Lzma2Class.cpp
//*
//*
//*****************************************************************************
//
//	Copyright Anton Zechner (C) 2007
//
//	Klasse zum Komprimieren von Daten mit dem LZMA2 Algorithmus fr 7Zip.
//
// 	unsigned Value(int)							: Wert  abfargen
// 	unsigned Flags(int)							: Flags setzen/abfragen
// 	unsigned Level(int)							: Level setzen/abfragen
// 	unsigned Error()							: Fehler Code holen
// 	unsigned Mode ()							: Modus setzen/abfragen
//
// 	int		 Compress  (int iLevel=-1);			: Starte   Komprimierung
//	int		 Decompress(int iMode =-2);			: Starte Dekomprimierung
//
// 	void	 SetOut(HANDLE,int,unsigned)		: Datei -Ausgabe bestimmen
// 	void	 SetIn (HANDLE,int,unsigned)		: Datei -Eingabe bestimmen
// 	void	 SetOut(unsigned,void*)				: Memory-Ausgabe bestimmen
// 	void	 SetIn (unsigned,void*)				: Memory-Eingabe bestimmen
// 	void	 SetOut(void*,int(*pProc)(void*,char*,unsigned))	: Externe Ausgabe
// 	void	 SetIn (void*,int(*pProc)(void*,char*,unsigned))	: Externe Eingabe
//

#include	<stdio.h>
#include	<string.h>
#include	"Lzma2Class.h"
#include	"../Filter/Crc32.h"


extern "C"
{
#include	"Lzma/Lzma2Dec.h"
#include	"Lzma/Lzma2Enc.h"
}


#ifdef		 AZ_LIB_ADDAFX
#include	<Afx.h>
#endif

#if			defined(_DEBUG) && defined(_MSC_VER) && !defined(NO_DEBUG_NEW)
#include	"DebugMemLeaks.h"
#endif


#define		LZMA2_INPBUF_SIZE		0x00100000
#define 	LZMA2_OUTBUF_SIZE		0x00400000

static	void * _cdecl Lzma2Alloc(void *pParam,size_t uSize){return new char[uSize];}
static	void   _cdecl Lzma2Free (void *pParam,void   *pPtr){delete pPtr;}

static	ISzAlloc	  sLzmaAlloc = {Lzma2Alloc,Lzma2Free};


//*****************************************************************************
//*
//*		Konstruktor
//*
//*****************************************************************************
CLzma2Class::CLzma2Class()
{

	m_pStreamData	= NULL;							// CLzma2Dec Daten fr Algorithmus
	m_iDataBytes	= -1;

	Clear(1);

}

//*****************************************************************************
//*
//*		Destrucktor
//*
//*****************************************************************************
CLzma2Class::~CLzma2Class()
{
	Clear(0);
}

//*****************************************************************************
//*
//*		Clear
//*
//*****************************************************************************
//  0 :	Alles rcksetzen
//  1 :	Nur Variabeln rcksetzen , Puffer nicht lschen
//  2 :	Nur Puffer lschen
void CLzma2Class::Clear(int iMode)
{

	if(iMode==0 || iMode==2)						// Puffer freiegeben ?
		{
		if(m_pStreamData)
			{
			Lzma2Dec_Free((CLzma2Dec*)m_pStreamData,&sLzmaAlloc);
			delete m_pStreamData;
			m_pStreamData = NULL;
			}
		}

	if(iMode==2)return;

	m_iMode = 0x06;


}

//*****************************************************************************
//*
//*		Free
//*
//*****************************************************************************
//	Puffer freigeben
void CLzma2Class::Free()
{

	Clear(2);

}

//*****************************************************************************
//*
//*		SetProbs
//*
//*****************************************************************************
//	Die Propedaten einstellen
//	pData	: Zeiger auf die Daten
//					[0] : lc=[0]%9  lp=([0]/9)%5  pb=([0]/9/5)%5
//				 [1..4]	: DIC-Size
//	uSize	: Gre der Daten
//	Ergibt die kopierten Bytes bzw. 0 bei einem Fehler
int CLzma2Class::SetProbs(const void *pData,unsigned uSize)
{

	if(uSize>sizeof(m_aProbs))return 0;

	memcpy(m_aProbs,pData,uSize);
	memset(m_aProbs+uSize,0,sizeof(m_aProbs)-uSize);


return uSize;
}

//*****************************************************************************
//*
//*		Decompress
//*
//*****************************************************************************
//	Startet die Endkomprimierung eines Streams
//	iMode		:  Modus zum entpacken
//					-2	  = Default Modus verwenden
//	Ergibt die Anzahl der Entkomprimierten Bytes
iocint64 CLzma2Class::Decompress(int iMode)
{
CLzma2Dec	   *pStream;
ELzmaStatus		iState;
ELzmaFinishMode iFinish;
iocint64		iOutput;
UInt32			uDoneRd;
SizeT			uNext;
SizeT			uWrPos;
SizeT			uCurLen;
SizeT			uOldPos;
SRes			iReturn;
bool			bEndFlag;
bool			bStoped;
int				iShift;
int				iCount;
int				iBytes;



	IOC_ASSERT(m_uOutMode<CL2_NONE && m_uInpMode<CL2_NONE);

	if(iMode==-2)
		{
		iMode = m_iMode;
		}

	iShift			= 0;
	uWrPos			= 0;
	iOutput			= 0;
	m_uCompresLen	= 0;
	m_uDecompresLen	= 0;
	Reset();

	if(m_pStreamData)								// Alten Stream lschen
		{
		Lzma2Dec_Free((CLzma2Dec*)m_pStreamData,&sLzmaAlloc);
		delete m_pStreamData;
		m_pStreamData=NULL;
		}


		m_pStreamData = (char*)new CLzma2Dec;
	if(!m_pStreamData)
		{
		m_iErrorCode = CL2_ERROR_MEMORY;
		throw "Zu wenig Speicher";
		}

	Lzma2Dec_Construct((CLzma2Dec*)m_pStreamData);
	pStream = (CLzma2Dec*)m_pStreamData;


	if(!InitRd(LZMA2_INPBUF_SIZE))					// Lesepuffer initialisieren
		{
		m_iErrorCode = CL2_ERROR_MEMORY;
		throw "Zu wenig Speicher";
		}


	if(Lzma2Dec_Allocate(pStream,m_aProbs[0],&sLzmaAlloc))
		{
		m_iErrorCode = CL2_ERROR_PARAM;
		throw "Falscher Parameter";
		}

	Lzma2Dec_Init(pStream);



													// Enkodingende im Ausgabepuffer berechnen
	if(pStream->decoder.dicBufSize-pStream->decoder.dicPos>=LZMA2_OUTBUF_SIZE)
			uNext = pStream->decoder.dicPos+LZMA2_OUTBUF_SIZE;
	else	uNext = pStream->decoder.dicBufSize;


	for(;;)
		{
		if(m_iBufLenRd-m_iBufOffRd<=0)				// Keine Daten mehr im Lesepuffer
			{
			if(!Fill() && !Eof())
				{
				m_iErrorCode = CL2_ERROR_READ;
				throw "Fehler beim Lesen";
				}
			}

		uOldPos = pStream->decoder.dicPos;
		uCurLen	= uNext-uOldPos;
		iFinish = LZMA_FINISH_ANY;

	    if(uCurLen>0x00400000)
			{
			uCurLen = 0x00400000;
			}

		if(m_iDataBytes>=0)							// Ist die Lnge des Streams bekannt
			{
			UInt64 uRest = m_iDataBytes-m_uDecompresLen;

			if(uRest<=uCurLen)
				{
				uCurLen = (SizeT)uRest;
				iFinish = LZMA_FINISH_END;			// Stream wird abgebrochen
				}
			}

		uDoneRd = m_iBufLenRd-m_iBufOffRd;			// Restliche Bytes im Lesebuffer

		iReturn = Lzma2Dec_DecodeToDic(pStream,uOldPos+uCurLen,(unsigned char*)m_pBufferRd+m_iBufOffRd,&uDoneRd,iFinish,&iState);

		iCount				= pStream->decoder.dicPos-uOldPos;
		m_iBufOffRd        += (UInt32)uDoneRd;
		m_uCompresLen      += uDoneRd;
		iOutput			   += iCount;

		bEndFlag			= (uDoneRd==0 && iCount==0);
		bStoped				= (m_iDataBytes>=0 && iOutput>=m_iDataBytes);



													// Gibt es Daten zum Ausgeben
		if(iReturn!=0 || pStream->decoder.dicPos==uNext || bEndFlag || bStoped)
			{
			iBytes = pStream->decoder.dicPos-uWrPos;

			if(Write(pStream->decoder.dic+uWrPos,iBytes)!=iBytes)
				{
				m_iErrorCode = CL2_ERROR_WRITE;
				throw "Fehler beim Speichern";
				}

			m_uDecompresLen    += iBytes;

			if(pStream->decoder.dicPos==pStream->decoder.dicBufSize)
				{
				pStream->decoder.dicPos = 0;
				uWrPos					= 0;
				}
			else{
				uWrPos = pStream->decoder.dicPos;
				}

													// Enkodingende im Ausgabepuffer berechnen
			if(pStream->decoder.dicBufSize-pStream->decoder.dicPos>=LZMA2_OUTBUF_SIZE)
					uNext = pStream->decoder.dicPos+LZMA2_OUTBUF_SIZE;
			else	uNext = pStream->decoder.dicBufSize;


			if(iReturn!=0)							// Fehler im Stream
				{
				m_iErrorCode = CL2_ERROR_BADBLZMA;
				throw "Formatfehler";
				}

			if(bStoped)
				{
				break;
				}

			if(bEndFlag)
				{
				if(iState==LZMA_STATUS_FINISHED_WITH_MARK)break;
				m_iErrorCode = CL2_ERROR_BADBLZMA;
				throw "Formatfehler";
				}
			}
		}

	Lzma2Dec_FreeProbs(pStream,&sLzmaAlloc);

	if(m_uOutMode!=CL2_MEM)							// Das Ende signalisieren
		{
		Write(NULL,0);
		}


return m_uDecompresLen;
}


//*****************************************************************************
//*
//*		CompressRead
//*
//*****************************************************************************
int CLzma2Class::CompressRead(void *pParam,void *pData,size_t *pSize)
{
ISeqInStream   *pStream=(ISeqInStream*)pParam;
CLzma2Class	   *pThis=(CLzma2Class*)pStream->ptr;
int				iCount;


	iCount = *pSize;
	iCount =  pThis->Read(pData,iCount);

	if(iCount>=0)
		{
		pThis->m_uDecompresLen += iCount;
		pSize[0] = iCount;
		return S_OK;
		}

	pSize[0] = 0;


return SZ_ERROR_READ;
}

//*****************************************************************************
//*
//*		CompressWrite
//*
//*****************************************************************************
size_t CLzma2Class::CompressWrite(void *pParam,const void *pData,size_t uSize)
{
ISeqOutStream  *pStream=(ISeqOutStream*)pParam;
CLzma2Class	   *pThis=(CLzma2Class*)pStream->ptr;
int				iCount;
int				iBytes;


	iCount = uSize;
	iBytes = pThis->Write((void*)pData,iCount);

	if(iBytes==iCount)
		{
		pThis->m_uCompresLen += iBytes;
		return iCount;
		}

	if(iBytes>0)
		{
		pThis->m_uCompresLen += iBytes;
		}


return 0;
}

//*****************************************************************************
//*
//*		Compress
//*
//*****************************************************************************
//	Startet die Komprimierung eines Streams
//	iLevel		: Bestimmt den Kompimierungslevel (1..
//					-1	= aktuellen Level verwenden
//					 9	= maximale Kompimierung
//					 0	= minimale Kompimierung
//	Ergibt die Anzahl der komprimierten Bytes
iocint64 CLzma2Class::Compress(int iLevel)
{
unsigned char	aData[8];
ISeqOutStream	sOutStream;
ISeqInStream	sInStream;
CLzma2EncHandle	pStream;
CLzma2EncProps	sProps;
SizeT			uSize;
SRes 			iRet;



	if(iLevel     != -1)m_iModeLevel=iLevel;			// Kompimierungslevel
	if(m_iModeLevel<  0)m_iModeLevel=0;
	if(m_iModeLevel>  9)m_iModeLevel=9;

		iLevel=m_iModeLevel;
	if(!iLevel)iLevel=1;

	m_uDecompresLen	= 0;
	m_uCompresLen	= 0;
	m_iErrorCode	= CL2_ERROR_OK;

	Reset();


		pStream = Lzma2Enc_Create(&sLzmaAlloc,&sLzmaAlloc);
	if(!pStream)
		{
		m_iErrorCode = CL2_ERROR_MEMORY;
		throw "Zu wenig Speicher";
		}

	Lzma2EncProps_Init(&sProps);

	sProps.lzmaProps.level	= m_iModeLevel;

	if(m_iDataBytes<0)
		{
		m_iDataBytes					= -1;
		sProps.lzmaProps.writeEndMark	=  1;
		}


	if(Lzma2Enc_SetProps(pStream,&sProps)!=S_OK)		// Porterties bernehmen
		{
		Lzma2Enc_Destroy(pStream);
		m_iErrorCode = CL2_ERROR_PARAM;
		throw "Falscher Parameter";
		}


	if(m_iMode&0x01)									// ZIP Header schreiben
		{
		aData[0]	= 9;								// Version 9.20
		aData[1]	= 20;
		aData[2]	= LZMA_PROPS_SIZE;
		aData[4]	= 0;

		if(Write(aData,4)!=4)
			{
			Lzma2Enc_Destroy(pStream);
			m_iErrorCode = CL2_ERROR_WRITE;
			throw "Fehler beim Speichern";
			}

		m_uCompresLen += 4;
		}

	if(m_iMode&0x02)									// Porterties Speicheren
		{
		uSize = LZMA_PROPS_SIZE;

		if(Lzma2Enc_WriteProperties(pStream)!=S_OK)
			{
			Lzma2Enc_Destroy(pStream);
			m_iErrorCode = CL2_ERROR_PARAM;
			throw "Falscher Parameter";
			}

		if(Write(aData,LZMA_PROPS_SIZE)!=LZMA_PROPS_SIZE)
			{
			Lzma2Enc_Destroy(pStream);
			m_iErrorCode = CL2_ERROR_WRITE;
			throw "Fehler beim Speichern";
			}

		m_uCompresLen += LZMA_PROPS_SIZE;
		}

	if(m_iMode&0x04)									// Dateilnge speichern
		{
		if(Write(&m_iDataBytes,8)!=8)
			{
			Lzma2Enc_Destroy(pStream);
			m_iErrorCode = CL2_ERROR_WRITE;
			throw "Fehler beim Speichern";
			}

		m_uCompresLen += 8;
		}



	sInStream .ptr		= this;
	sInStream .Read		= CompressRead;
	sOutStream.ptr		= this;
	sOutStream.Write	= CompressWrite;

	iRet = Lzma2Enc_Encode(pStream,&sOutStream,&sInStream,NULL);

	Lzma2Enc_Destroy(pStream);


	switch(iRet)										// Rckgabewert auswerten
		{
	case S_OK:						break;
	case SZ_ERROR_MEM:			m_iErrorCode = CL2_ERROR_MEMORY;	throw "Zu wenig Speicher";
	case SZ_ERROR_PARAM:		m_iErrorCode = CL2_ERROR_PARAM;		throw "Falscher Parameter";
	case SZ_ERROR_READ:
	case SZ_ERROR_INPUT_EOF:	m_iErrorCode = CL2_ERROR_READ;		throw "Fehler beim Einlesen";
	case SZ_ERROR_WRITE:
	case SZ_ERROR_OUTPUT_EOF: 	m_iErrorCode = CL2_ERROR_WRITE;		throw "Fehler beim Speichern";
	default:					m_iErrorCode = CL2_ERROR_BADBLZMA;	throw "Formatfehler";
		}


return m_uCompresLen;
}

