//*****************************************************************************
//*
//*
//*		LzmaClass.cpp
//*
//*
//*****************************************************************************
//
//	Copyright Anton Zechner (C) 2007
//
//	Klasse zum Komprimieren von Daten mit dem LZMA Algorithmus von 7Zip.
//
// 	unsigned Value(int)							: Wert  abfargen
// 	unsigned Flags(int)							: Flags setzen/abfragen
// 	unsigned Level(int)							: Level setzen/abfragen
// 	unsigned Error()							: Fehler Code holen
// 	unsigned Mode ()							: Modus setzen/abfragen
//
// 	int		 Compress  (int iLevel=-1);			: Starte   Komprimierung
//	int		 Decompress(int iMode =-2);			: Starte Dekomprimierung
//
// 	void	 SetOut(HANDLE,int,unsigned)		: Datei -Ausgabe bestimmen
// 	void	 SetIn (HANDLE,int,unsigned)		: Datei -Eingabe bestimmen
// 	void	 SetOut(unsigned,void*)				: Memory-Ausgabe bestimmen
// 	void	 SetIn (unsigned,void*)				: Memory-Eingabe bestimmen
// 	void	 SetOut(void*,int(*pProc)(void*,char*,unsigned))	: Externe Ausgabe
// 	void	 SetIn (void*,int(*pProc)(void*,char*,unsigned))	: Externe Eingabe
//

#include	<stdio.h>
#include	<string.h>
#include	"LzmaClass.h"
#include	"../Filter/Crc32.h"

extern "C"
{
#include	"Lzma/LzmaDec.h"
#include	"Lzma/LzmaEnc.h"
}


#ifdef		 AZ_LIB_ADDAFX
#include	<Afx.h>
#endif

#if			defined(_DEBUG) && defined(_MSC_VER) && !defined(NO_DEBUG_NEW)
#include	"DebugMemLeaks.h"
#endif


#define		LZMA_INPBUF_SIZE		0x00100000
#define 	LZMA_OUTBUF_SIZE		0x00400000

static	void * _cdecl LzmaAlloc(void *pParam,size_t uSize)
{
return new char[uSize];
}
static	void   _cdecl LzmaFree (void *pParam,void   *pPtr)
{
delete pPtr;
}

static	ISzAlloc	  sLzmaAlloc = {LzmaAlloc,LzmaFree};




//*****************************************************************************
//*
//*		Konstruktor
//*
//*****************************************************************************
CLzmaClass::CLzmaClass()
{

	m_pStreamData	= NULL;							// CLzmaDec Daten fr Algorithmus
	m_iDataBytes	= -1;

	Clear(1);

}

//*****************************************************************************
//*
//*		Destrucktor
//*
//*****************************************************************************
CLzmaClass::~CLzmaClass()
{
	Clear(0);
}

//*****************************************************************************
//*
//*		Clear
//*
//*****************************************************************************
//  0 :	Alles rcksetzen
//  1 :	Nur Variabeln rcksetzen , Puffer nicht lschen
//  2 :	Nur Puffer lschen
void CLzmaClass::Clear(int iMode)
{

	if(iMode==0 || iMode==2)						// Puffer freiegeben ?
		{
		if(m_pStreamData)
			{
			LzmaDec_Free((CLzmaDec*)m_pStreamData,&sLzmaAlloc);
			delete m_pStreamData;
			m_pStreamData = NULL;
			}
		}

	if(iMode==2)return;

	m_iMode = 0x06;


}

//*****************************************************************************
//*
//*		Free
//*
//*****************************************************************************
//	Puffer freigeben
void CLzmaClass::Free()
{

	Clear(2);

}

//*****************************************************************************
//*
//*		SetProbs
//*
//*****************************************************************************
//	Die Propedaten einstellen
//	pData	: Zeiger auf die Daten
//					[0] : lc=[0]%9  lp=([0]/9)%5  pb=([0]/9/5)%5
//				 [1..4]	: DIC-Size
//	uSize	: Gre der Daten
//	Ergibt die kopierten Bytes bzw. 0 bei einem Fehler
int CLzmaClass::SetProbs(const void *pData,unsigned uSize)
{

	if(uSize>sizeof(m_aProbs))return 0;

	memcpy(m_aProbs,pData,uSize);
	memset(m_aProbs+uSize,0,sizeof(m_aProbs)-uSize);


return uSize;
}

//*****************************************************************************
//*
//*		Decompress
//*
//*****************************************************************************
//	Startet die Endkomprimierung eines Streams
//	iMode		:  Modus zum entpacken
//					Bit0  = Zip-Header Laden (4 Byte)
//					Bit1  = Properties  von Stream holen (5 Byte)
//					Bit2  = Streamlnge von Stream holen (8 Byte)
//					-2	  = Default Modus verwenden
//	Ergibt die Anzahl der Entkomprimierten Bytes
iocint64 CLzmaClass::Decompress(int iMode)
{
unsigned char	aData[256];
CLzmaDec	   *pStream;
ELzmaStatus		iState;
ELzmaFinishMode iFinish;
iocint64		iOutput;
UInt32			uDoneRd;
SizeT			uNext;
SizeT			uWrPos;
SizeT			uCurLen;
SizeT			uOldPos;
SRes			iReturn;
bool			bEndFlag;
bool			bStoped;
int				iShift;
int				iCount;
int				iBytes;



	IOC_ASSERT(m_uOutMode<CLA_NONE && m_uInpMode<CLA_NONE);

	if(iMode==-2)
		{
		iMode = m_iMode;
		}

	iShift			= 0;
	uWrPos			= 0;
	iOutput			= 0;
	m_uCompresLen	= 0;
	m_uDecompresLen	= 0;
	Reset();

	if(m_pStreamData)								// Alten Stream lschen
		{
		LzmaDec_Free((CLzmaDec*)m_pStreamData,&sLzmaAlloc);
		delete m_pStreamData;
		m_pStreamData=NULL;
		}


	m_pStreamData = (char*)new CLzmaDec;

	if(!m_pStreamData)
		{
		m_iErrorCode = CLA_ERROR_MEMORY;
		throw "Zu wenig Speicher";
		}

	LzmaDec_Construct((CLzmaDec*)m_pStreamData);
	pStream = (CLzmaDec*)m_pStreamData;

	if(iMode&0x01)									// Zip-Header Laden (1:VersionMaj,1:VersionMin,2:PropSize)
		{
		if(Read(aData,4)!=4)
			{
			m_iErrorCode = CLA_ERROR_READ;
			throw "Fehler beim Lesen";
			}

		   iShift	= aData[2]+aData[3]*256;
		if(iShift<LZMA_PROPS_SIZE)
			{
			m_iErrorCode = CLA_ERROR_BADBLZMA;
			throw "Formatfehler";
			}

		iMode  |= 2;
		iShift -= 5;
		}


	if(iMode&0x02)									// Properties von Stream holen
		{
		if(Read(m_aProbs,LZMA_PROPS_SIZE)!=LZMA_PROPS_SIZE)
			{
			m_iErrorCode = CLA_ERROR_READ;
			throw "Fehler beim Lesen";
			}
		}

	while(iShift>0)									// Restliche Proterty-Daten berspringen
		{
		   iCount = iShift;
		if(iCount>sizeof(aData))iCount=sizeof(aData);

		if(Read(aData,iCount)!=iCount)
			{
			m_iErrorCode = CLA_ERROR_READ;
			throw "Fehler beim Lesen";
			}

		iShift -= iCount;
		}

	if(iMode&0x04)									// Ausgabelnge vom Stream holen
		{
		if(Read(&m_iDataBytes,8)!=8)
			{
			m_iErrorCode = CLA_ERROR_READ;
			throw "Fehler beim Lesen";
			}
		}




	LzmaDec_Allocate(pStream,m_aProbs,5,&sLzmaAlloc);
	LzmaDec_Init(pStream);


	if(!InitRd(LZMA_INPBUF_SIZE))					// Lesepuffer initialisieren
		{
		m_iErrorCode = CLA_ERROR_MEMORY;
		throw "Zu wenig Speicher";
		}

													// Enkodingende im Ausgabepuffer berechnen
	if(pStream->dicBufSize-pStream->dicPos>=LZMA_OUTBUF_SIZE)
			uNext = pStream->dicPos+LZMA_OUTBUF_SIZE;
	else	uNext = pStream->dicBufSize;

	for(;;)
		{
		if(m_iBufLenRd-m_iBufOffRd<=0)				// Keine Daten mehr im Lesepuffer
			{
			if(!Fill() && !Eof())
				{
				m_iErrorCode = CLA_ERROR_READ;
				throw "Fehler beim Lesen";
				}
			}

		uOldPos = pStream->dicPos;
		uCurLen	= uNext-uOldPos;
		iFinish = LZMA_FINISH_ANY;

		if(m_iDataBytes>=0)							// Ist die Lnge des Streams bekannt
			{
			UInt64 uRest = m_iDataBytes-m_uDecompresLen;

			if(uRest<=uCurLen)
				{
				uCurLen = (SizeT)uRest;
				iFinish = LZMA_FINISH_END;			// Stream wird abgebrochen
				}
			}


		uDoneRd = m_iBufLenRd-m_iBufOffRd;			// Restliche Bytes im Lesebuffer

		iReturn = LzmaDec_DecodeToDic(pStream,uOldPos+uCurLen,(unsigned char*)m_pBufferRd+m_iBufOffRd,&uDoneRd,iFinish,&iState);

		iCount				= pStream->dicPos-uOldPos;
		m_iBufOffRd        += (UInt32)uDoneRd;
		m_uCompresLen      += uDoneRd;
		iOutput			   += iCount;

		bEndFlag			= (uDoneRd==0 && iCount==0);
		bStoped				= (m_iDataBytes>=0 && iOutput>=m_iDataBytes);



													// Gibt es Daten zum ausgeben
		if(iReturn!=0 || pStream->dicPos==uNext || bEndFlag || bStoped)
			{
			iBytes				= pStream->dicPos-uWrPos;

			if(Write(pStream->dic+uWrPos,iBytes)!=iBytes)
				{
				m_iErrorCode = CLA_ERROR_WRITE;
				throw "Fehler beim Speichern";
				}

			m_uDecompresLen    += iBytes;

			if(pStream->dicPos==pStream->dicBufSize)
				{
				pStream->dicPos = 0;
				uWrPos          = 0;
				}
			else{
				uWrPos = pStream->dicPos;
				}

													// Enkodingende im Ausgabepuffer berechnen
			if(pStream->dicBufSize-pStream->dicPos>=LZMA_OUTBUF_SIZE)
					uNext = pStream->dicPos+LZMA_OUTBUF_SIZE;
			else	uNext = pStream->dicBufSize;


			if(iReturn!=0)							// Fehler im Stream
				{
				m_iErrorCode = CLA_ERROR_BADBLZMA;
				throw "Formatfehler";
				}

			if(bStoped)
				{
				break;
				}

			if(bEndFlag)
				{
				if(iState==LZMA_STATUS_FINISHED_WITH_MARK)break;
				m_iErrorCode = CLA_ERROR_BADBLZMA;
				throw "Formatfehler";
				}
			}
		}


	if(m_uOutMode!=CLA_MEM)							// Das Ende signalisieren
		{
		Write(NULL,0);
		}


return m_uDecompresLen;
}


//*****************************************************************************
//*
//*		CompressRead
//*
//*****************************************************************************
int CLzmaClass::CompressRead(void *pParam,void *pData,size_t *pSize)
{
ISeqInStream   *pStream=(ISeqInStream*)pParam;
CLzmaClass	   *pThis=(CLzmaClass*)pStream->ptr;
int				iCount;


	iCount = *pSize;
	iCount =  pThis->Read(pData,iCount);

	if(iCount>=0)
		{
		pThis->m_uDecompresLen += iCount;
		pSize[0] = iCount;
		return S_OK;
		}

	pSize[0] = 0;


return SZ_ERROR_READ;
}

//*****************************************************************************
//*
//*		CompressWrite
//*
//*****************************************************************************
size_t CLzmaClass::CompressWrite(void *pParam,const void *pData,size_t uSize)
{
ISeqOutStream  *pStream=(ISeqOutStream*)pParam;
CLzmaClass	   *pThis=(CLzmaClass*)pStream->ptr;
int				iCount;
int				iBytes;


	iCount = uSize;
	iBytes = pThis->Write((void*)pData,iCount);

	if(iBytes==iCount)
		{
		pThis->m_uCompresLen += iBytes;
		return iCount;
		}

	if(iBytes>0)
		{
		pThis->m_uCompresLen += iBytes;
		}


return 0;
}

//*****************************************************************************
//*
//*		Compress
//*
//*****************************************************************************
//	Startet die Komprimierung eines Streams
//	iLevel		: Bestimmt den Kompimierungslevel (1..
//					-1	= aktuellen Level verwenden
//					 9	= maximale Kompimierung
//					 0	= minimale Kompimierung
//	Ergibt die Anzahl der komprimierten Bytes
iocint64 CLzmaClass::Compress(int iLevel)
{
unsigned char	aData[8];
ISeqOutStream	sOutStream;
ISeqInStream	sInStream;
CLzmaEncHandle	pStream;
CLzmaEncProps	sProps;
SizeT			uSize;
SRes 			iRet;



	if(iLevel     != -1)m_iModeLevel=iLevel;			// Kompimierungslevel
	if(m_iModeLevel<  0)m_iModeLevel=0;
	if(m_iModeLevel>  9)m_iModeLevel=9;

		iLevel=m_iModeLevel;
	if(!iLevel)iLevel=1;

	m_uDecompresLen	= 0;
	m_uCompresLen	= 0;
	m_iErrorCode	= CLA_ERROR_OK;

	Reset();


		pStream = LzmaEnc_Create(&sLzmaAlloc);
	if(!pStream)
		{
		m_iErrorCode = CLA_ERROR_MEMORY;
		throw "Zu wenig Speicher";
		}

	LzmaEncProps_Init(&sProps);

	sProps.level	= m_iModeLevel;

	if(m_iDataBytes<0)
		{
		m_iDataBytes		= -1;
		sProps.writeEndMark	=  1;
		}


	if(LzmaEnc_SetProps(pStream,&sProps)!=S_OK)			// Porterties bernehmen
		{
		LzmaEnc_Destroy(pStream,&sLzmaAlloc,&sLzmaAlloc);
		m_iErrorCode = CLA_ERROR_PARAM;
		throw "Falscher Parameter";
		}


	if(m_iMode&0x01)									// ZIP Header schreiben
		{
		aData[0]	= 9;								// Version 9.20
		aData[1]	= 20;
		aData[2]	= LZMA_PROPS_SIZE;
		aData[4]	= 0;

		if(Write(aData,4)!=4)
			{
			LzmaEnc_Destroy(pStream,&sLzmaAlloc,&sLzmaAlloc);
			m_iErrorCode = CLA_ERROR_WRITE;
			throw "Fehler beim Speichern";
			}

		m_uCompresLen += 4;
		}

	if(m_iMode&0x02)									// Porterties Speicheren
		{
		uSize = LZMA_PROPS_SIZE;

		if(LzmaEnc_WriteProperties(pStream,aData,&uSize)!=S_OK)
			{
			LzmaEnc_Destroy(pStream,&sLzmaAlloc,&sLzmaAlloc);
			m_iErrorCode = CLA_ERROR_PARAM;
			throw "Falscher Parameter";
			}

		if(Write(aData,LZMA_PROPS_SIZE)!=LZMA_PROPS_SIZE)
			{
			LzmaEnc_Destroy(pStream,&sLzmaAlloc,&sLzmaAlloc);
			m_iErrorCode = CLA_ERROR_WRITE;
			throw "Fehler beim Speichern";
			}

		m_uCompresLen += LZMA_PROPS_SIZE;
		}

	if(m_iMode&0x04)									// Dateilnge speichern
		{
		if(Write(&m_iDataBytes,8)!=8)
			{
			LzmaEnc_Destroy(pStream,&sLzmaAlloc,&sLzmaAlloc);
			m_iErrorCode = CLA_ERROR_WRITE;
			throw "Fehler beim Speichern";
			}

		m_uCompresLen += 8;
		}



	sInStream .ptr		= this;
	sInStream .Read		= CompressRead;
	sOutStream.ptr		= this;
	sOutStream.Write	= CompressWrite;

	iRet = LzmaEnc_Encode(pStream,&sOutStream,&sInStream,NULL,&sLzmaAlloc,&sLzmaAlloc);

	LzmaEnc_Destroy(pStream,&sLzmaAlloc,&sLzmaAlloc);


	switch(iRet)										// Rckgabewert auswerten
		{
	case S_OK:						break;
	case SZ_ERROR_MEM:			m_iErrorCode = CLA_ERROR_MEMORY;	throw "Zu wenig Speicher";
	case SZ_ERROR_PARAM:		m_iErrorCode = CLA_ERROR_PARAM;		throw "Falscher Parameter";
	case SZ_ERROR_READ:
	case SZ_ERROR_INPUT_EOF:	m_iErrorCode = CLA_ERROR_READ;		throw "Fehler beim Einlesen";
	case SZ_ERROR_WRITE:
	case SZ_ERROR_OUTPUT_EOF: 	m_iErrorCode = CLA_ERROR_WRITE;		throw "Fehler beim Speichern";
	default:					m_iErrorCode = CLA_ERROR_BADBLZMA;	throw "Formatfehler";
		}


return m_uCompresLen;
}

