//*****************************************************************************
//*
//*
//*		LzwClass.cpp
//*
//*
//*****************************************************************************
//
//	Copyright Anton Zechner (C) 2007
//
//	Klasse zum Komprimieren von Daten mit dem LZX Algorithmus
//
// 	unsigned Value(int)							: Wert  abfargen
// 	unsigned Flags(int)							: Flags setzen/abfragen
// 	unsigned Level(int)							: Level setzen/abfragen
// 	unsigned Error()							: Fehler Code holen
// 	unsigned Mode ()							: Modus setzen/abfragen
//
// 	int		 Compress  (int iLevel=-1);			: Starte   Komprimierung
//	int		 Decompress(int iOutLen,int iMode);	: Starte Dekomprimierung
//
// 	void	 SetOut(HANDLE,int,unsigned)		: Datei -Ausgabe bestimmen
// 	void	 SetIn (HANDLE,int,unsigned)		: Datei -Eingabe bestimmen
// 	void	 SetOut(unsigned,void*)				: Memory-Ausgabe bestimmen
// 	void	 SetIn (unsigned,void*)				: Memory-Eingabe bestimmen
// 	void	 SetOut(void*,int(*pProc)(void*,char*,unsigned))	: Externe Ausgabe
// 	void	 SetIn (void*,int(*pProc)(void*,char*,unsigned))	: Externe Eingabe
//
#if		   _MSC_VER > 1299
#ifndef		WINVER
#define		WINVER 	0x0501
#endif
#endif

#include	<stdio.h>
#include	<string.h>
#include	<malloc.h>
#include	"LzwClass.h"
#include	"../Filter/Crc32.h"



#ifdef		 AZ_LIB_ADDAFX
#include	<Afx.h>
#endif

#if			defined(_DEBUG) && defined(_MSC_VER) && !defined(NO_DEBUG_NEW)
#include	"DebugMemLeaks.h"
#endif

#define 	CLW_FILE_BUF				0x2000
#define		INITBITS					9
#define		MINBITS 					12
#define		MAXBITS						16
#define		NULLPTR(type)				((type*)NULL)

#define		UNUSED_CODE					((CODE)0)				// Indicates hash table value unused
#define		CLEAR						((CODE)256) 			// Code requesting table to be cleared
#define		FIRSTFREE					((CODE)257) 			// First free code for token encoding
#define		MAXTOKLEN					512 					// Max chars in token size of buffer
#define		OK							0						// Result codes from functions:

#define		SIGNAL_ERROR				-1						//	 signal function error
#define		NOMEM						2						//	 Ran out of memory
#define		TOKTOOBIG					3						//	 Token longer than MAXTOKLEN chars
#define		READERR 					4						//	 I/O error on input
#define		WRITEERR					5						//	 I/O error on output
#define		INFILEBAD					6						//	 Infile not in compressed format
#define		CODEBAD 					7						//	 Infile contained a bad token code
#define		TABLEBAD					8						//	 The tables got corrupted (!)
#define		NOSAVING					9						//	 no saving in file size
#define		NOTOPENED					10						//	 output file couldn't be opened
#define		YES 						1
#define		NO							0


// Defines for third byte of header

#define		BIT_MASK					0x1F
#define		BLOCK_MASK					0x80
#define		OB_BUF_SIZE 				0x10000

// Masks 0x40 and 0x20 are free.  I think 0x20 should mean that there is a fourth header byte (for expansion).

const CLzwClass::BYTE CLzwClass::m_aMagicHeader[] =
	{
	0x1F, 0x9D
	};

const CLzwClass::BYTE CLzwClass::m_aRmask[9] =
	{
	0x00, 0x01, 0x03,
	0x07, 0x0F, 0x1F,
	0x3F, 0x7F, 0xFF
	};

const CLzwClass::HASH CLzwClass::m_aHsValues[] =
	{
	0x13FF,														// 12-bit codes, 75% occupancy
	0x26C3,														// 13-bit codes, 80% occupancy
	0x4A1D,														// 14-bit codes, 85% occupancy
	0x8D0D,														// 15-bit codes, 90% occupancy
	0xFFD9														// 16-bit codes, 94% occupancy, 6% of code values unused
	};

const CLzwClass::INTCODE CLzwClass::m_aMcValues[] =
	{
	0x0FFF,														// 12-bit codes
	0x1FFF,														// 13-bit codes
	0x3FFF,														// 14-bit codes
	0x7FFF,														// 15-bit codes
	0xEFFF														// 16-bit codes, 6% of code values unused
	};

#define		Hashsize(maxb)				(m_aHsValues[(maxb) -MINBITS])
#define		Maxcode(maxb)				(m_aMcValues[(maxb) -MINBITS])
#define		allocx(type,ptr,size)		(((ptr) = (type*)malloc((size)*sizeof(type)))==NULLPTR(type)? NOMEM:OK)
#define		free_array(type,ptr,off)	if(ptr!=NULLPTR(type)) 						\
										{ 											\
										delete ((char *)((ptr) + (off)));			\
										(ptr) = NULLPTR(type); 						\
										}

#define		alloc_array(type,ptr,size,off)											\
										( allocx(type, ptr,(size) - (off)) == OK 	\
										? (((ptr) -= (off)), OK) 					\
										: NOMEM 									\
										)

#define		suffix(code)				(m_pSfxPtr[code])
#define		prefix(code)				(m_pPfxPtr[code])
#define		probe(hash) 				(m_pHtPtr [hash])
#define		init_tables(hash)			{ 														\
										hash = m_uHashSize; 									\
										memset(m_pHtPtr,0,m_uHashSize*sizeof(m_pHtPtr[0]));		\
										m_uHighCode = ~(~(INTCODE)0 << (m_iBits = INITBITS)); 	\
										m_uNextFree =  (m_iBlockCompres ? FIRSTFREE : 256);		\
										}



//*****************************************************************************
//*
//*		Konstruktor
//*
//*****************************************************************************
CLzwClass::CLzwClass()
{

	m_pTokens		= NULL;
	m_pSfxPtr		= NULL;
	m_pPfxPtr		= NULL;
	m_pHtPtr		= NULL;

	Clear(1);
}

//*****************************************************************************
//*
//*		Destruktor
//*
//*****************************************************************************
CLzwClass::~CLzwClass()
{

	Clear(0);

}

//*****************************************************************************
//*
//*		InitTables
//*
//*****************************************************************************
void CLzwClass::InitTables(unsigned uHash)
{
	uHash = m_uHashSize;

	memset(m_pHtPtr,0,m_uHashSize*sizeof(m_pHtPtr[0]));

	m_uHighCode = ~(~(INTCODE)0 << (m_iBits = INITBITS));
	m_uNextFree = (m_iBlockCompres ? FIRSTFREE : 256);
}


//*****************************************************************************
//*
//*		Free
//*
//*****************************************************************************
//	Alle Puffer lschen
void CLzwClass::Free()
{
	Clear(2);
}


//*****************************************************************************
//*
//*		Clear
//*
//*****************************************************************************
//  0 :	Alles rcksetzen
//  1 :	Nur Variabeln rcksetzen , Puffer nicht lschen
//  2 :	Nur Puffer lschen
void CLzwClass::Clear(int iMode)
{


	CIoCallbacks::Clear(iMode);


	m_uMaxCode		= 0;
	m_uOldMaxCode	= 0;
	m_uHashSize		= 0;
	m_uOldHashSize	= 0;


	if(iMode==0 || iMode==2)							// Puffer freiegeben ?
		{
		free_array(CODE,m_pHtPtr ,  0);
		free_array(CODE,m_pPfxPtr,256);
		free_array(char,m_pSfxPtr,256);

		if(m_pTokens)
			{
			delete 	m_pTokens;
			m_pTokens = NULL;
			}
		}


	if(iMode==2)return;

	m_iMaxBits 		= MAXBITS;
	m_iBlockCompres	= BLOCK_MASK;
	m_uMaxIn		= 0;
	m_uNextIn		= 0;
	m_iBits			= 0;
	m_uPrefxCode	= 0;
	m_uNextFree		= 0;
	m_uHighCode		= 0;
	m_pSfxPtr		= 0;
	m_pPfxPtr		= 0;
	m_pHtPtr		= 0;

	m_uDecompresLen	= 0;								// Anzahl der Entkomprimierten Bytes
	m_uCompresLen	= 0;								// Anzahl der Komprimierten Bytes
	m_iErrorCode	= 0;								// Aktuelle Fehlernummer (siehe CLW_ERROR_???)

}


//*****************************************************************************
//*
//*		AllocTables
//*
//*****************************************************************************
int CLzwClass::AllocTables(INTCODE m_uMaxCode,HASH m_uHashSize)
{

	if(m_uHashSize>m_uOldHashSize)
		{
		free_array(CODE,m_pHtPtr,0);
		m_uOldHashSize = 0;
		}

	if(m_uMaxCode>m_uOldMaxCode)
		{
		free_array(CODE,m_pPfxPtr,256);
		free_array(char,m_pSfxPtr,256);

		if(alloc_array(char,m_pSfxPtr,m_uMaxCode+1,256))
			{
			return 0;
			}

		if(alloc_array(char,m_pSfxPtr,m_uMaxCode+1,256) || alloc_array(CODE,m_pPfxPtr,m_uMaxCode+1,256))
			{
			m_uOldMaxCode	= 0;
			return NOMEM;
			}

		m_uOldMaxCode = m_uMaxCode;
		}

	if(m_uHashSize>m_uOldHashSize)
		{
		if(alloc_array(CODE,m_pHtPtr,m_uHashSize,0) != 0)
			{
			m_uOldHashSize	= 0;
			return NOMEM;
			}

		m_uOldHashSize = m_uHashSize;
		}


return OK;
}

//*****************************************************************************
//*
//*		GetChar
//*
//*****************************************************************************
inline int CLzwClass::GetChar()
{
unsigned char	ucByte;


	if(GetByte(ucByte))									// Fehler oder keine Bytes
		{
		if(Eof())return EOF;
		m_iErrorCode = CLW_ERROR_READ;
		throw "Fehler beim Lesen";
		}


return ucByte;
}


//*****************************************************************************
//*
//*		PutCode
//*
//*****************************************************************************
//	Den nchsten Code speichern.
//	uCode		: Ist der Wert des Codes
//	iBits		: Ist die Lnge des Codes in Bit
void CLzwClass::PutCode(INTCODE uCode,int iBits)
{


	m_ucBitMemory	&= (1<<m_iRestBits)-1;
	uCode		   <<= m_iRestBits;
	uCode			|= m_ucBitMemory;
	iBits		    += m_iRestBits;

	while(iBits>=8)
		{
		if(!PutByte(uCode))								// Das nchste Byte speichern	
			{
			uCode >>= 8;
			iBits  -= 8;
			continue;
			}

		m_iErrorCode = CLW_ERROR_WRITE;
		throw "Fehler beim Schreiben";
		return;
		}

	m_ucBitMemory	= uCode;
	m_iRestBits		= iBits;

}



//*****************************************************************************
//*
//*		NextCode
//*
//*****************************************************************************
//	Ergibt 1 wenn ein Code geladen wurde, bzw. 0 wenn nicht.
int CLzwClass::NextCode(INTCODE *pCodePtr)
{
int			iShift;
int			iBits;
INTCODE		uCode;



	iBits	= m_iBits;
	iShift	= 0;
	uCode	= 0;


	while(iBits>0)
		{
		if(m_iRestBits>0)								// Sind noch bits im Speicher
			{
			uCode  |= m_ucBitMemory<<iShift;
			iShift += m_iRestBits;

			if(m_iRestBits>=iBits)
				{
				m_iRestBits    -= iBits;
				m_ucBitMemory >>= iBits;
				break;
				}
			
			iBits  -= m_iRestBits;
			}

		if(!GetByte(m_ucBitMemory))						// Nchstes Byte einlesen
			{
			m_iRestBits  = 8;
			continue;
			}

		if(Eof())return NO;
		m_iErrorCode = CLW_ERROR_READ;
		throw "Fehler beim Lesen";
		}


	uCode      &= (1<<m_iBits)-1;
	pCodePtr[0]	= uCode;
	


return 1;
}

//*****************************************************************************
//*
//*		Compress
//*
//*****************************************************************************
//	Startet die Komprimierung eines Streams
//	iLevel		: Bestimmt den Kompimierungslevel (1..
//					-1	= aktuellen Level verwenden
//					 7	= maximale Framegre
//					 0	= minimale Framegre
//	Ergibt die Anzahl der komprimierten Bytes
iocint64 CLzwClass::Compress(int iLevel)
{
unsigned char	aData[3];
HASH			aList[256];
HASH			uHashDelta;
HASH			uHash;
INTCODE			uCode;
int				iAdjBits;
int				i;



	IOC_ASSERT(m_uOutMode<CLW_NONE && m_uInpMode<CLW_NONE);


	if(iLevel     != -1)m_iModeLevel=iLevel;			// Kompimierungslevel
	if(m_iModeLevel<  1)m_iModeLevel=1;
	if(m_iModeLevel>  7)m_iModeLevel=7;

	Clear(2);
	Reset ();
	InitRd();
	InitWr();

	m_uDecompresLen	= 0;
	m_uCompresLen	= 0;
	m_iRestBits		= 0;
	m_iErrorCode	= CLW_ERROR_OK;

	m_iMaxBits		= m_iModeLevel+9;
	if(m_iMaxBits<12)m_iMaxBits=12;

	m_uHashSize		= Hashsize(m_iMaxBits);
	m_uMaxCode 		= Maxcode (m_iMaxBits);
	iAdjBits 		= m_iMaxBits-10;


	for(i=255;i>=0;i--)									// Hash-Werte initialisieren
		{
		aList[i] = (((i&0x07)<<7)^i) << iAdjBits;
		}



	if(AllocTables(m_uMaxCode,m_uHashSize)!=OK)
		{
		m_iErrorCode = CLW_ERROR_MEMORY;
		throw "Zuwenig Speicher";
		}

	init_tables(uHash);


	aData[0]	=  0x1F;								// Header schreiben
	aData[1]	=  0x9D;
	aData[2]	= (unsigned char)(m_iMaxBits|m_iBlockCompres);



	if(PutData(aData,3))								// Header abspeichern
		{
		m_iErrorCode = CLW_ERROR_WRITE;
		throw "Fehler beim Schreiben";
		}

	i = GetChar();

	if(i==EOF)											// Ende vom Eingans-Stream erreicht
		{
		if(Flush())
			{
			m_iErrorCode = CLW_ERROR_WRITE;
			throw "Fehler beim Schreiben";
			}

		m_uDecompresLen	= GetPosInp();
		m_uCompresLen	= GetPosOut();

		return 0;
		}

	m_uPrefxCode = (INTCODE)i;

	for(;;)
		{
		i = GetChar();
		if(i==EOF)break;

		uHash = m_uPrefxCode^aList[i];

		if(uHash>=m_uHashSize)
			{
			uHash -= m_uHashSize;
			}

		uCode = (INTCODE)probe(uHash);

		if(uCode!=UNUSED_CODE)
			{
			if(suffix(uCode) != (char)i || (INTCODE)prefix(uCode) != m_uPrefxCode)
				{
				uHashDelta = (0x0120 - i) << (iAdjBits);


				do	{									// rehash and keep looking
					if(uHash>=uHashDelta)
						{
						uHash -= uHashDelta;
						}
					else{
						uHash += m_uHashSize-uHashDelta;
						}

					uCode = (INTCODE)probe(uHash);

					if(uCode==UNUSED_CODE)
						{
						goto NewCode;
						}
					}

				while(suffix(uCode)!=(char)i || (INTCODE)prefix(uCode)!=m_uPrefxCode);
				}

			m_uPrefxCode = uCode;
			}
		else{

		NewCode:

			PutCode(m_uPrefxCode,m_iBits);
			uCode = m_uNextFree;

			if(uCode<=m_uMaxCode)
				{
				probe (uHash) = (CODE)uCode;
				prefix(uCode) = (CODE)m_uPrefxCode;
				suffix(uCode) = (char)i;

				if(uCode > m_uHighCode)
					{
					m_uHighCode += uCode;
					m_iBits++;
					}

				m_uNextFree = uCode+1;
				}
			else if(m_iBlockCompres)
				{
				PutCode((INTCODE)i,m_iBits);
				PutCode(CLEAR     ,m_iBits);

				init_tables(uHash);

				i = GetChar();

				if(i==EOF)
					{
					break;
					}
				}

			m_uPrefxCode = (INTCODE)i;
			}
		}


	PutCode(m_uPrefxCode,m_iBits);
	PutCode(CLEAR,8-m_iRestBits);

	if(Flush())
		{
		m_iErrorCode = CLW_ERROR_WRITE;
		throw "Fehler beim Schreiben";
		}

	m_uDecompresLen	= GetPosInp();
	m_uCompresLen	= GetPosOut();



return m_uCompresLen;
}


//*****************************************************************************
//*
//*		Decompress
//*
//*****************************************************************************
//	Startet die Endkomprimierung eines Streams
//	iMode		: Bestimmt die Endkomprimierungsart
//					-2			= aktuellen Modus verwenden
//					CBZ_STORE	= Daten eins zu eins laden
//					CBZ_DEFLATE	= Daten entkomprimieren
//	Ergibt die Anzahl der Entkomprimierten Bytes
iocint64 CLzwClass::Decompress(int iMode)
{
unsigned char		aTemp[3];
char			   *pTempPtr;
char				cSuffix;
FLAG				iFullTable;
FLAG				iClearTable;
INTCODE				uSaveCode;
INTCODE				uCode;
int					i;



	IOC_ASSERT(m_uOutMode<CLW_NONE && m_uInpMode<CLW_NONE);


	Clear(2);
	Reset ();
	InitRd();
	InitWr();


	m_uDecompresLen	= 0;
	m_uCompresLen	= 0;
	m_iRestBits		= 0;
	m_iErrorCode	= CLW_ERROR_OK;
	uSaveCode		= CLEAR;
	iClearTable		= 1;


	if(!m_pTokens)										// Token-Speicher allokieren
		{
			m_pTokens = new char[MAXTOKLEN];
		if(!m_pTokens)
			{
			m_iErrorCode = CLW_ERROR_MEMORY;
			throw "Zuwenig Speicher";
			}

		m_iMaxToklen = MAXTOKLEN;
		}



	if(GetData(aTemp,3)!=3)								// Header einlesen
		{
		m_iErrorCode = CLW_ERROR_READ;
		throw "Fehler beim Lesen";
		}

	if(m_aMagicHeader[0]!=aTemp[0] || m_aMagicHeader[1]!=aTemp[1])
		{
		m_iErrorCode = CLW_ERROR_BADFORMAT;
		throw "Ungltiges Dateiformat";
		}

	m_iBlockCompres	 = aTemp[2]&BLOCK_MASK;
	m_iMaxBits 		 = aTemp[2]&BIT_MASK;
	m_uMaxCode		 = ~(~(INTCODE)0<<m_iMaxBits);

	if(AllocTables(m_uMaxCode,0))
		{
		m_iErrorCode = CLW_ERROR_MEMORY;
		throw "Zuwenig Speicher";
		}



	do	{												// Hauptschleife
		uCode = uSaveCode;

		if(uCode==CLEAR && iClearTable)
			{
			m_iBits		=  INITBITS;
			m_uHighCode	= ~(~(INTCODE)0<<INITBITS);
			iFullTable	=  0;
			iClearTable =  m_iBlockCompres;
			m_uNextFree	= (m_iBlockCompres==0)? 256:FIRSTFREE;

			if(!NextCode(&m_uPrefxCode))
				{
				m_iErrorCode = CLW_ERROR_BADFORMAT;
				throw "Ungltiges Dateiformat";
				}

			cSuffix     = (char)m_uPrefxCode;

			if(PutByte(cSuffix))
				{
				m_iErrorCode = CLW_ERROR_WRITE;
				throw "Fehler beim Schreiben";
				}

			continue;
			}

		i = 0;

		if(uCode>=m_uNextFree && !iFullTable)
			{
			if(uCode!=m_uNextFree)
				{
				m_iErrorCode = CLW_ERROR_BADFORMAT;
				throw "Ungltiges Dateiformat";
				}

			uCode			= m_uPrefxCode;
			m_pTokens[i++]	= cSuffix;
			}


		while(uCode>=256)
			{
			if(i>=m_iMaxToklen)							// tokens buffer too small
				{
					pTempPtr = new char[m_iMaxToklen*2];
				if(!pTempPtr)
					{
					m_iErrorCode = CLW_ERROR_MEMORY;
					throw "Zuwenig Speicher";
					}

				memcpy(pTempPtr,m_pTokens,m_iMaxToklen);
				delete m_pTokens;
				m_pTokens	  = pTempPtr;
				m_iMaxToklen *= 2;
				}

			m_pTokens[i++]	=  suffix(uCode);
			uCode			= (INTCODE)prefix(uCode);
			}


		cSuffix = (char)uCode;

		if(PutByte(cSuffix))
			{
			m_iErrorCode = CLW_ERROR_WRITE;
			throw "Fehler beim Schreiben";
			}

		for(i--;i>=0;i--)
			{
			if(PutByte(m_pTokens[i]))
				{
				m_iErrorCode = CLW_ERROR_WRITE;
				throw "Fehler beim Schreiben";
				}
			}

		if(!iFullTable)
			{
			uCode			= m_uNextFree;
			prefix(uCode)	= (CODE)m_uPrefxCode;
			suffix(uCode)	= cSuffix;
			m_uPrefxCode	= uSaveCode;

			if(uCode++ == m_uHighCode)
				{
				if(m_uHighCode >= m_uMaxCode)
					{
					iFullTable = 1;
					uCode--;
					}
				else{
					m_iBits     += 1;
					m_uHighCode += uCode;				// m_uNextFree == m_uHighCode + 1
					}
				}

			m_uNextFree = uCode;
			}

		}while(NextCode(&uSaveCode));


	if(Flush())
		{
		m_iErrorCode = CLW_ERROR_WRITE;
		throw "Fehler beim Schreiben";
		}

	m_uDecompresLen	= GetPosOut();
	m_uCompresLen	= GetPosInp();



return m_uDecompresLen;
}

