//*****************************************************************************
//*
//*
//*		ZipCoder.cpp
//*
//*
//*****************************************************************************
//
//	Copyright (c) 2002, Dr Brian Gladman <                 >, Worcester, UK.
//	All rights reserved.
//
//	LICENSE TERMS
//
//	The free distribution and use of this software in both source and binary
//	form is allowed (with or without changes) provided that:
//
//	  1. distributions of this source code include the above copyright
//	     notice, this list of conditions and the following disclaimer;
//
//	  2. distributions in binary form include the above copyright
//	     notice, this list of conditions and the following disclaimer
//	     in the documentation and/or other associated materials;
//
//	  3. the copyright holder's name is not used to endorse products
//	     built using this software without specific written permission.
//
//	ALTERNATIVELY, provided that this notice is retained in full, this product
//	may be distributed under the terms of the GNU General Public License (GPL),
//	in which case the provisions of the GPL apply INSTEAD OF those given above.
//
//	DISCLAIMER
//
//	This software is provided 'as is' with no explicit or implied warranties
//	in respect of its properties, including, but not limited to, correctness
//	and/or fitness for purpose.
//
//	Issue Date: 24/01/2003
//
//	This file implements password based file encryption and authentication
//	using AES in CTR mode, HMAC-SHA1 authentication and RFC2898 password
//	based key derivation.
//
#include <memory.h>
#include "ZipCoder.h"


//*****************************************************************************
//*
//*		DeriveKey
//*
//*****************************************************************************
//	pwd			: password
//	pwd_len		: password length
//	salt[]		: the SALT and its
//	salt_len	: salt length
//	iter		: the number of iterations
//	key[]		: space for the output key
//	key_len		: and its required length
static void DeriveKey(const unsigned char *pwd,unsigned pwd_len,const unsigned char *salt,unsigned int salt_len,unsigned int iter,unsigned char *key,unsigned int key_len)
{
unsigned int    i, j, k, n_blk;
unsigned char	uu[HASH_OUTPUT_SIZE];
unsigned char	ux[HASH_OUTPUT_SIZE];
HmacContext		c1;
HmacContext		c2;
HmacContext		c3;


    HmacShaInit(&c1);									// set HMAC context (c1) for password
    HmacShaKey (&c1,pwd, pwd_len);


    memcpy     (&c2,&c1 ,sizeof(HmacContext));				// set HMAC context (c2) for password and salt
    HmacShaData(&c2,salt,salt_len);


    n_blk = 1+(key_len-1)/HASH_OUTPUT_SIZE;				// find the number of SHA blocks in the key


    for(i=0;i<n_blk;i++)								// for each block in key
		{
        memset(ux,0,HASH_OUTPUT_SIZE);					// ux[] holds the running xor value

        memcpy(&c3,&c2,sizeof(HmacContext));				// set HMAC context (c3) for password and salt


        uu[0] = (unsigned char)((i + 1) >> 24);			// enter additional data for 1st block into uu
        uu[1] = (unsigned char)((i + 1) >> 16);
        uu[2] = (unsigned char)((i + 1) >>  8);
        uu[3] = (unsigned char)((i + 1)      );


        for(j=0,k=4;j<iter;j++)							// this is the key mixing iteration
			{
            HmacShaData(&c3,uu,k);						// add previous round data to HMAC
            HmacShaExit(&c3,uu,HASH_OUTPUT_SIZE);		// obtain HMAC for uu[]


            for(k=0;k<HASH_OUTPUT_SIZE;k++)				// xor into the running xor block
				{
                ux[k] ^= uu[k];
				}

            memcpy(&c3,&c1,sizeof(HmacContext));			// set HMAC context (c3) for password
	        }


        j = 0;
		k = i*HASH_OUTPUT_SIZE;							// compile key blocks into the key output

        while(j<HASH_OUTPUT_SIZE && k<key_len)
			{
            key[k++] = ux[j++];
			}
		}
}


//*****************************************************************************
//*
//*		EncryptData
//*
//*****************************************************************************
//	subroutine for data encryption/decryption
//	this could be speeded up a lot by aligning
//	buffers and using 32 bit operations
static void EncryptData(ZipCrypt *pContext,unsigned char *pData,unsigned uDataLen)
{
unsigned	pos;
unsigned	j;
unsigned	i;

	i   = 0;
	pos = pContext->encr_pos;

    while(i<uDataLen)
	    {
        if(pos==AES_BLOCK_SIZE)
			{
			j = 0;

            while(j<8 && !++pContext->nonce[j])			// increment encryption nonce
				{
                j++;
				}
														// encrypt the nonce to form next xor buffer
            AesEncrypt(pContext->nonce, pContext->encr_bfr, pContext->encr_ctx);
            pos = 0;
			}

        pData[i++] ^= pContext->encr_bfr[pos++];
		}

    pContext->encr_pos = pos;
}


//*****************************************************************************
//*
//*		ZipCryptInit
//*
//*****************************************************************************
//	pContext	: the file encryption context (output)
//	iMode		: the mode to be used 1..3 (input)
//	pPassword	: the user specified password (input)
//	uPwdLen		: the length of the password (input)
//	pSalt		: the salt (random input)
//	pPwVerifyer	: 2 byte password verifier (output)
int ZipCryptInit(ZipCrypt *pContext,int iMode,const unsigned char *pPassword,unsigned uPwdLen,const unsigned char *pSalt,unsigned char *pPwVerifyer)
{
unsigned char	aBuffer[2*MAX_KEY_LENGTH+PWD_VER_LENGTH];



    if(uPwdLen>MAX_PWD_LENGTH)
		{
        return PASSWORD_TOO_LONG;
		}

    if(iMode<1 || iMode>3)
		{
        return BAD_MODE;
		}

    pContext->mode    = iMode;
    pContext->pwd_len = uPwdLen;

														// derive the encryption and authentication keys and the password verifier
    DeriveKey(pPassword,uPwdLen,pSalt,SALT_LENGTH(iMode),KEYING_ITERATIONS,aBuffer,2*KEY_LENGTH(iMode)+PWD_VER_LENGTH);


    pContext->encr_pos = AES_BLOCK_SIZE;				// initialise the encryption nonce and buffer pos
														// if we need a random component in the encryption
														// nonce, this is where it would have to be set
    memset(pContext->nonce, 0, AES_BLOCK_SIZE * sizeof(unsigned char));

														// initialise for encryption using key 1
    AesEncryptKey(aBuffer, KEY_LENGTH(iMode),pContext->encr_ctx);


    HmacShaInit(pContext->auth_ctx);					// initialise for authentication using key 2
    HmacShaKey (pContext->auth_ctx,aBuffer+KEY_LENGTH(iMode),KEY_LENGTH(iMode));

    memcpy(pPwVerifyer,aBuffer+2*KEY_LENGTH(iMode),PWD_VER_LENGTH);



return GOOD_RETURN;
}

//*****************************************************************************
//*
//*		ZipEncrypt
//*
//*****************************************************************************
//	perform 'in place' encryption and authentication
void ZipEncrypt(unsigned char *pData, unsigned int uDataLen, ZipCrypt *pContext)
{

    EncryptData(pContext,pData,uDataLen);
    HmacShaData(pContext->auth_ctx,pData,uDataLen);

}

//*****************************************************************************
//*
//*		ZipDecrypt
//*
//*****************************************************************************
//	perform 'in place' authentication and decryption
void ZipDecrypt(ZipCrypt *pContext,unsigned char *pData,unsigned int uDataLen)
{

    HmacShaData(pContext->auth_ctx,pData,uDataLen);
    EncryptData(pContext,pData,uDataLen);

}


//*****************************************************************************
//*
//*		ZipCryptExit
//*
//*****************************************************************************
//	close encryption/decryption and return the MAC value
int ZipCryptExit(ZipCrypt *pContext,unsigned char *pMac)
{
    HmacShaExit(pContext->auth_ctx,pMac,MAC_LENGTH(pContext->mode));

return MAC_LENGTH(pContext->mode);					// return MAC length in bytes
}

