//*****************************************************************************
//*
//*
//*		Hmac.c
//*
//*
//*****************************************************************************
//	
//	Copyright (c) 2002, Dr Brian Gladman, Worcester, UK.   All rights reserved.
//	
//	LICENSE TERMS
//	
//	The free distribution and use of this software in both source and binary
//	form is allowed (with or without changes) provided that:
//	
//	  1. distributions of this source code include the above copyright
//	     notice, this list of conditions and the following disclaimer;
//	
//	  2. distributions in binary form include the above copyright
//	     notice, this list of conditions and the following disclaimer
//	     in the documentation and/or other associated materials;
//	
//	  3. the copyright holder's name is not used to endorse products
//	     built using this software without specific written permission.
//	
//	ALTERNATIVELY, provided that this notice is retained in full, this product
//	may be distributed under the terms of the GNU General Public License (GPL),
//	in which case the provisions of the GPL apply INSTEAD OF those given above.
//	
//	DISCLAIMER
//	
//	This software is provided 'as is' with no explicit or implied warranties
//	in respect of its properties, including, but not limited to, correctness
//	and/or fitness for purpose.
//	
//	Issue Date: 26/08/2003
//	
//	This is an implementation of HMAC, the FIPS standard keyed hash function
//

#include "hmac.h"
#include "brg_types.h"

#if defined(__cplusplus)
extern "C"
{
#endif

//*****************************************************************************
//*
//*		HmacShaInit
//*
//*****************************************************************************
// initialise the HMAC context to zero 
void HmacShaInit(HmacContext *pData)
{
    memset(pData,0,sizeof(HmacContext));
}


//*****************************************************************************
//*
//*		HmacShaKey
//*
//*****************************************************************************
//	input the HMAC key (can be called multiple times)   
int HmacShaKey(HmacContext *pData,const unsigned char *pKey,unsigned uKeyLen)
{
    
	if(pData->klen == HMAC_IN_DATA)					// error if further key input   
		{
        return HMAC_BAD_MODE;						// is attempted in data mode    
		}

    if(pData->klen + uKeyLen > HASH_INPUT_SIZE)		// if the key has to be hashed  
		{
        if(pData->klen <= HASH_INPUT_SIZE)			// if the hash has not yet been 
			{                                       // started, initialise it and   
            sha_begin(pData->ctx);					// hash stored key characters   
            sha_hash(pData->ctx,pData->key,pData->klen);
		    }

        sha_hash(pData->ctx,pKey,uKeyLen);	        // hash long key data into hash 
	    }
    else{											// otherwise store key data     
        memcpy(pData->key+pData->klen,pKey,uKeyLen);
		}

    pData->klen += uKeyLen;							// update the key length count  



return HMAC_OK;
}

//*****************************************************************************
//*
//*		HmacShaData	
//*
//*****************************************************************************
//	input the HMAC data (can be called multiple times) - 
//	note that this call terminates the key input phase   
void HmacShaData(HmacContext *pData,const unsigned char *pInput,unsigned uInpLen)
{   
unsigned	uPos;


    if(pData->klen != HMAC_IN_DATA)					// if not yet in data phase 
		{
        if(pData->klen > HASH_INPUT_SIZE)			// if key is being hashed   
			{										// complete the hash and    
            sha_end(pData->ctx,pData->key);         // store the result as the  
            pData->klen = HASH_OUTPUT_SIZE;			// key and set new length   
	        }

													// pad the key if necessary 
        memset(pData->key + pData->klen,0,HASH_INPUT_SIZE-pData->klen);

													// xor ipad into key value  
        for(uPos=0;uPos<HASH_INPUT_SIZE/4;uPos++)
			{
            ((uint_32t*)pData->key)[uPos] ^= 0x36363636;
			}

        
        sha_begin(pData->ctx);						// and start hash operation 
        sha_hash (pData->ctx,pData->key, HASH_INPUT_SIZE);

        pData->klen = HMAC_IN_DATA;					// mark as now in data mode
	    }

    
    if(uInpLen)										// hash the data (if any)       
		{
        sha_hash(pData->ctx,pInput,uInpLen);
		}
}

//*****************************************************************************
//*
//*		HmacShaExit
//*
//*****************************************************************************
//	compute and output the MAC value 
void HmacShaExit(HmacContext *pData,unsigned char *pOutput,unsigned uOutLen)
{   
unsigned char	aBuffer[HASH_OUTPUT_SIZE];
unsigned		uPos;

    
    if(pData->klen != HMAC_IN_DATA)						// if no data has been entered perform a null data phase
		{
        HmacShaData(pData,(const unsigned char*)0, 0);
		}

    sha_end(pData->ctx,aBuffer);						// complete the inner hash       

    
    for(uPos=0;uPos<HASH_INPUT_SIZE/4;uPos++)			// set outer key value using opad and removing ipad 
		{
        ((uint_32t*)pData->key)[uPos] ^= 0x36363636 ^ 0x5C5C5C5C;
		}

    
    sha_begin(pData->ctx);								// perform the outer hash operation 
    sha_hash (pData->ctx,pData->key,HASH_INPUT_SIZE );
    sha_hash (pData->ctx,aBuffer   ,HASH_OUTPUT_SIZE);
    sha_end  (pData->ctx,aBuffer);

    
    for(uPos=0;uPos<uOutLen;uPos++)						// output the hash value            
		{
        pOutput[uPos] = aBuffer[uPos];
		}
}


//*****************************************************************************
//*
//*		HmacShaMem
//*
//*****************************************************************************
//	'do it all in one go' subroutine     
void HmacShaMem(const unsigned char *pKey,unsigned uKeyLen,const unsigned char *pInput,unsigned uInpLen,unsigned char *pOutput,unsigned uOutLen)
{   
HmacContext    sContext;

    HmacShaInit(&sContext);
    HmacShaKey (&sContext,pKey   ,uKeyLen);
    HmacShaData(&sContext,pInput ,uInpLen);
    HmacShaExit(&sContext,pOutput,uOutLen);
}


#if defined(__cplusplus)
}
#endif
