//*****************************************************************************
//*
//*
//*		BitField.cpp
//*
//*
//*****************************************************************************
//
//	Copyright Anton Zechen (C) 2013
//
//	Die Klasse CBitField    erzeugt ein Bitfeld das sich automatisch dynamisch vergrert.
//
//	Funktionen:
//
//		int 	Read  (unsigned uPos) const;		Bit lesen
//		int 	Set   (unsigned uPos);				Bit setzen
//		void	Set   (unsigned uPos,int iCount);	Bit setzen
//		int 	Clear (unsigned uPos);				Bit lschen
//		void	Clear (unsigned uPos,int iCount);	Bit lschen
//		int		Change(unsigned uPos);				Bit invertieren
//		void	Change(unsigned uPos,int iCount);	Bit invertieren
//		int 	Write (unsigned uPos,int iBit);		Bit schreiben
//
//		int		Find0() const;						Suche 0-Bit
//		int		Find0(unsigned uStart) const;		Suche 0-Bit ab Position
//		int		Find1() const; 						Suche 1-Bit
//		int		Find1(unsigned uStart) const;		Suche 1-Bit ab Position
//
//		void	ClearAll();							Alle Bits lschen
//		void	SetAll(unsigned uCount);			Alle Bits setzen
//		uchar  *GetMem() const;						Zeiger auf Bitfeld holen
//		void	GetMem(      void *pMem,int iSize);	Bitfeld kopieren
//		void	SetMem(const void *pMem,int iSize);	Bitfeld setzen
//		int 	operator[](unsigned uPos)const;		Ein Byte auslesen
//
#include	"BitField.h"

#if			defined(_DEBUG) && defined(_MSC_VER) && !defined(NO_DEBUG_NEW)
#include	"DebugMemLeaks.h"
#endif



#ifdef _DEBUG
//*****************************************************************************
//*
//*		Konstruktor
//*
//*****************************************************************************
CBitField::CBitField()
{
	m_pOld = 0;
	m_pMem = 0;
	m_uMax = 0;
}

CBitField::CBitField(unsigned uCount)
{
	m_pOld = 0;
	m_pMem = 0;
	m_uMax = 0;
	IncreaseFixed(uCount);
}

//*****************************************************************************
//*
//*		Destruktor
//*
//*****************************************************************************
CBitField::~CBitField()
{
	if(m_pOld)
		{
		delete m_pOld;
		m_pOld = 0;
		}

	if(m_pMem)
		{
		delete m_pMem;
		m_pMem = 0;
		m_uMax = 0;
		}

}

//*****************************************************************************
//*
//*		IncreaseSize
//*
//*****************************************************************************
//	Vergrert das Bitfeld mit dynamischer Vergrerung
//	uCount		: Ist die neue minimale Gre minus 1
void CBitField::IncreaseSize(unsigned uCount)
{
unsigned	uMax;


	uMax = m_uMax*2+128;

	if(uMax<uCount)uMax=uCount;

	IncreaseFixed(uMax);


}


//*****************************************************************************
//*
//*		Read
//*
//*****************************************************************************
//	Liet ein Bit aus dem Bitfeld aus.
//	uPos	: Nummer des Bits das gelesen werden soll.
//	Ergibt den Wert -1 wenn das Bit auf '1' ist, ansonsten 0.
int CBitField::Read(unsigned uPos) const
{

	if(uPos>=m_uMax)return 0;

return bit_read(m_pMem,uPos);
}


//*****************************************************************************
//*
//*		Set
//*
//*****************************************************************************
//	Setzt ein Bit im Bitfeld.
//	uPos		: Nummer des Bits das gesetzt werden soll.
//	iCount		: Ist die Anzahl der zu setzenden Bits (<=0 keine nderung)
//	Ergibt den Wert -1 wenn das Bit vorher auf '1' war, ansonsten 0.
int CBitField::Set(unsigned uPos)
{

	if(uPos>=m_uMax)IncreaseSize(uPos);

return bit_set(m_pMem,uPos);
}

void CBitField::Set(unsigned uPos,int iCount)
{
unsigned	uEnd;


	if(iCount<=0)return;

	uEnd = uPos+iCount;

	if(uEnd>m_uMax)
		{
		IncreaseSize(uEnd);
		}

	bit_set(m_pMem,uPos,iCount);
}


//*****************************************************************************
//*
//*		Clear
//*
//*****************************************************************************
//	Lscht ein Bit im Bitfeld.
//	uPos		: Nummer des Bits das gelscht werden soll.
//	iCount		: Ist die Anzahl der zu lschenden Bits (<=0 keine nderung)
//	Ergibt den Wert -1 wenn das Bit vorher auf '1' war, ansonsten 0.
int CBitField::Clear(unsigned uPos)
{

	if(uPos>=m_uMax)return 0;

return bit_clear(m_pMem,uPos);
}

void CBitField::Clear(unsigned uPos,int iCount)
{
unsigned	uEnd;


	if(iCount<=0)return;

	uEnd = uPos+iCount;

	if(uEnd>m_uMax)
		{
		IncreaseSize(uEnd);
		}

	bit_clear(m_pMem,uPos,iCount);
}


//*****************************************************************************
//*
//*		Change
//*
//*****************************************************************************
//	Invertiert ein Bit im Bitfeld.
//	uPos		: Nummer des Bits das invertiert werden soll.
//	iCount		: Ist die Anzahl der zu invertierenden Bits (<=0 keine nderung)
//	Ergibt den Wert -1 wenn das Bit vorher auf '1' war, ansonsten 0.
int CBitField::Change(unsigned uPos)
{

	if(uPos>=m_uMax)IncreaseSize(uPos);

return bit_change(m_pMem,uPos);
}

void CBitField::Change(unsigned uPos,int iCount)
{
unsigned	uEnd;


	if(iCount<=0)return;

	uEnd = uPos+iCount;

	if(uEnd>m_uMax)
		{
		IncreaseSize(uEnd);
		}

	bit_change(m_pMem,uPos,iCount);
}


//*****************************************************************************
//*
//*		Write
//*
//*****************************************************************************
//	Schreibt ein Bit im Bitfeld.
//	uPos		: Nummer des Bits das gendert werden soll.
//	iBit		: Ist der neue Wert des Bits (ungleich 0 fr 1)
//	Ergibt den Wert -1 wenn das Bit vorher auf '1' war, ansonsten 0.
int CBitField::Write(unsigned uPos,int iBit)
{

	if(uPos>=m_uMax)IncreaseSize(uPos);

return bit_write(m_pMem,uPos,iBit);
}


//*****************************************************************************
//*
//*		Find0
//*
//*****************************************************************************
//	Sucht im Bitfeld das nchste Bit das auf '0' ist.
//	uStart		: Ist die Bitposition ab der gesucht wird
//	Ergibt die Position bei der ein '0'-Bit gefunden wurde,
//	bzw. 0xFFFFFFFF wenn kein '0'-Bit gefunden wurde.
int	CBitField::Find0() const
{
unsigned	uPos;

	uPos = bit_find_0(m_pMem,m_uMax);

	if(uPos>=m_uMax)return 0xFFFFFFFF;


return uPos;
}

int	CBitField::Find0(unsigned uStart) const
{
unsigned	uPos;

	uPos = bit_find_0(m_pMem,m_uMax,uStart);

	if(uPos>=m_uMax)return 0xFFFFFFFF;


return uPos;
}


//*****************************************************************************
//*
//*		Find1
//*
//*****************************************************************************
//	Sucht im Bitfeld das nchste Bit das auf '1' ist.
//	uStart		: Ist die Bitposition ab der gesucht wird
//	Ergibt die Position bei der ein '1'-Bit gefunden wurde,
//	bzw. 0xFFFFFFFF wenn kein '1'-Bit gefunden wurde.
int	CBitField::Find1() const
{
unsigned	uPos;

	uPos = bit_find_1(m_pMem,m_uMax);

	if(uPos>=m_uMax)return 0xFFFFFFFF;


return uPos;
}

int	CBitField::Find1(unsigned uStart) const
{
unsigned	uPos;

	uPos = bit_find_1(m_pMem,m_uMax,uStart);

	if(uPos>=m_uMax)return 0xFFFFFFFF;


return uPos;
}


//*****************************************************************************
//*
//*		ClearAll
//*
//*****************************************************************************
//	Alle Bits im Bitfeld lschen
void CBitField::ClearAll()
{

	memset(m_pMem,0x00,m_uMax/8);

}


//*****************************************************************************
//*
//*		SetAll
//*
//*****************************************************************************
//	Alle Bits im Bitfeld setzen
//	uCount		: Anzahl der Bits die gesetzt werden sollen
void CBitField::SetAll(unsigned uCount)
{

	if(uCount>m_uMax)									// Gre prfen
		{
		IncreaseFixed(uCount);
		memset(m_pMem,0xFF,m_uMax/8);
		}
	else{
		memset(m_pMem         ,0xFF,uCount/8);
		memset(m_pMem+uCount/8,0x00,m_uMax/8-uCount/8);
		}

}


//*****************************************************************************
//*
//*		GetMem
//*
//*****************************************************************************
//	Den Zeiger auf den Speicher des Bitfeldes holen.
unsigned char *CBitField::GetMem() const
{
return m_pMem;
}


//*****************************************************************************
//*
//*		operator[]
//*
//*****************************************************************************
//	Ein Byte auslesen aus dem Bitfeld auslesen
//	uPos	: Ist die Byteposition
//	Ergibt das aus dem Bitfeld
int CBitField::operator[](unsigned uPos)const
{

	if(uPos>m_uMax/8)return 0;							// Poisition hinter dem Ende


return m_pMem[uPos];
}

#endif


//*****************************************************************************
//*
//*		GetMem
//*
//*****************************************************************************
//	Den Speicher des Bitfeldes holen.
//	pMem		: Zeiger auf den Speicherpuffer.
//	iSize		: Ist die Gre des Speicherpuffers in Byte.
void CBitField::GetMem(void *pMem,int iSize) const
{
char	*pData=(char*)pMem;


	if(iSize<=0)return;

	if((unsigned)iSize<=m_uMax/8)						// Gre prfen
		{
		memcpy(pData,m_pMem,iSize);
		return;
		}

	memcpy(pData,m_pMem,m_uMax/8);
	memset(pData+m_uMax/8,0,iSize-m_uMax/8);

}


//*****************************************************************************
//*
//*		SetMem
//*
//*****************************************************************************
//	Einen Speicherbereich ins Bitfeld kopieren.
//	pMem		: Zeiger auf den Speicherbereich.
//	iSize		: Ist die Gre des Speicherbereich in Byte.
void CBitField::SetMem(const void *pMem,int iSize)
{


	if(iSize<=0)return;

	if((unsigned)iSize>m_uMax/8)						// Gre prfen
		{
		IncreaseFixed(iSize*8);
		}

	memcpy(m_pMem,pMem,iSize);

}


//*****************************************************************************
//*
//*		IncreaseFixed
//*
//*****************************************************************************
//	Vergrert das Bitfeld
//	uMax		: Ist die neue minimale Gre
void CBitField::IncreaseFixed(unsigned uMax)
{
unsigned char  *pMem;


	uMax +=  32;										// Auf 32 Bits aufrunden
	uMax &= ~31;

	if(uMax<=m_uMax)return;

	pMem	= new unsigned char[uMax/8];

	memcpy(pMem,m_pMem           ,m_uMax/8);
	memset(pMem+m_uMax/8,0,uMax/8-m_uMax/8);

	if(m_pMem)											// Alten Speicher freigeben
		{
		delete m_pOld;
		m_pOld = 0;
		}

	m_pOld	= m_pMem;
	m_pMem	= pMem;
	m_uMax	= uMax;

}
