//*****************************************************************************
//*
//*
//*		BitField.h
//*
//*
//*****************************************************************************
//
//	Copyright Anton Zechen (C) 2013
//
//	Die Klasse CBitField    erzeugt ein Bitfeld das sich automatisch dynamisch vergrert.
//
//	Funktionen:
//
//		int 	Read  (unsigned uPos) const;		Bit lesen
//		int 	Set   (unsigned uPos);				Bit setzen
//		void	Set   (unsigned uPos,int iCount);	Bit setzen
//		int 	Clear (unsigned uPos);				Bit lschen
//		void	Clear (unsigned uPos,int iCount);	Bit lschen
//		int		Change(unsigned uPos);				Bit invertieren
//		void	Change(unsigned uPos,int iCount);	Bit invertieren
//		int 	Write (unsigned uPos,int iBit);		Bit schreiben
//															
//		int		Find0() const;						Suche 0-Bit
//		int		Find0(unsigned uStart) const;		Suche 0-Bit ab Position
//		int		Find1() const; 						Suche 1-Bit
//		int		Find1(unsigned uStart) const;		Suche 1-Bit ab Position
//															
//		void	ClearAll();							Alle Bits lschen
//		void	SetAll(unsigned uCount);			Alle Bits setzen
//		uchar  *GetMem() const;						Zeiger auf Bitfeld holen
//		void	GetMem(      void *pMem,int iSize);	Bitfeld kopieren
//		void	SetMem(const void *pMem,int iSize);	Bitfeld setzen 
//		int 	operator[](unsigned uPos)const;		Ein Byte auslesen
//		
#ifndef		__BITFIELD_H__
#define		__BITFIELD_H__

#if 		_MSC_VER >= 1000
#pragma 	once
#endif

#include	<string.h>
#include	"BitFunctions.h"


//*****************************************************************************
//*
//*		Klasse CBitField
//*
//*****************************************************************************
class CBitField
{
public:

	CBitField(unsigned uCount);
	CBitField();
   ~CBitField();

	int 			Read  (unsigned uPos) const;			// Bit lesen
	int 			Set   (unsigned uPos);					// Bit setzen
	void  			Set   (unsigned uPos,int iCount);		// Bit setzen
	int 			Clear (unsigned uPos);					// Bit lschen
	void  			Clear (unsigned uPos,int iCount);		// Bit lschen
	int				Change(unsigned uPos);					// Bit invertieren
	void			Change(unsigned uPos,int iCount);		// Bit invertieren
	int 			Write (unsigned uPos,int iBit);			// Bit schreiben

	int				Find0() const;							// Suche 0-Bit
	int				Find0(unsigned uStart) const;			// Suche 0-Bit ab Position
	int				Find1() const; 							// Suche 1-Bit
	int				Find1(unsigned uStart) const;			// Suche 1-Bit ab Position

	void  			ClearAll();								// Alle Bits lschen
	void 			SetAll(unsigned uCount);				// Alle Bits setzen
	unsigned char  *GetMem() const;							// Zeiger auf Bitfeld holen
	void			GetMem(      void *pMem,int iSize)const;// Bitfeld kopieren
	void			SetMem(const void *pMem,int iSize);		// Bitfeld setzen 
	int 			operator[](unsigned uPos)const;			// Ein Byte auslesen

protected:

	void 			IncreaseSize (unsigned uCount);			// Das Bitfeld vergrern (dynamisch)
	void 			IncreaseFixed(unsigned uCount);			// Das Bitfeld vergrern

	unsigned char  *m_pOld;									// Zeiger auf altes Bitfeld
	unsigned char  *m_pMem;									// Zeiger auf den Bitfeld
	unsigned		m_uMax;									// Maximale Anzahl an Bits


};


#ifndef _DEBUG

//*****************************************************************************
//*
//*		Konstruktor
//*
//*****************************************************************************
inline CBitField::CBitField()
{
	m_pOld = 0;
	m_pMem = 0;
	m_uMax = 0;
}

inline CBitField::CBitField(unsigned uCount)
{
	m_pOld = 0;
	m_pMem = 0;
	m_uMax = 0;
	IncreaseFixed(uCount);
}

//*****************************************************************************
//*
//*		IncreaseSize
//*
//*****************************************************************************
//	Vergrert das Bitfeld mit dynamischer Vergrerung
//	uCount		: Ist die neue minimale Gre minus 1
inline void CBitField::IncreaseSize(unsigned uCount)
{
unsigned	uMax;
	

	uMax = m_uMax*2+128;

	if(uMax<uCount)uMax=uCount;

	IncreaseFixed(uMax);


}

//*****************************************************************************
//*
//*		Read
//*
//*****************************************************************************
//	Liet ein Bit aus dem Bitfeld aus.
//	uPos	: Nummer des Bits das gelesen werden soll.
//	Ergibt den Wert -1 wenn das Bit auf '1' ist, ansonsten 0. 
inline int CBitField::Read(unsigned uPos) const
{
	
	if(uPos>=m_uMax)return 0;

return bit_read(m_pMem,uPos);
}

//*****************************************************************************
//*
//*		Set
//*
//*****************************************************************************
//	Setzt ein Bit im Bitfeld.
//	uPos		: Nummer des Bits das gesetzt werden soll.
//	iCount		: Ist die Anzahl der zu setzenden Bits (<=0 keine nderung)
//	Ergibt den Wert -1 wenn das Bit vorher auf '1' war, ansonsten 0.  
inline int CBitField::Set(unsigned uPos)
{
	
	if(uPos>=m_uMax)IncreaseSize(uPos);

return bit_set(m_pMem,uPos);
}

inline void CBitField::Set(unsigned uPos,int iCount)
{
unsigned	uEnd;


	if(iCount<=0)return;
	
	uEnd = uPos+iCount;
	
	if(uEnd>m_uMax)										// Gre prfen
		{
		IncreaseSize(uEnd);
		}

	bit_set(m_pMem,uPos,iCount);			
}

//*****************************************************************************
//*
//*		Clear
//*
//*****************************************************************************
//	Lscht ein Bit im Bitfeld.
//	uPos		: Nummer des Bits das gelscht werden soll.
//	iCount		: Ist die Anzahl der zu lschenden Bits (<=0 keine nderung)
//	Ergibt den Wert -1 wenn das Bit vorher auf '1' war, ansonsten 0.  
inline int CBitField::Clear(unsigned uPos)
{
	
	if(uPos>=m_uMax)return 0;

return bit_clear(m_pMem,uPos);
}

inline void CBitField::Clear(unsigned uPos,int iCount)
{
unsigned	uEnd;


	if(iCount<=0)return;
	
	uEnd = uPos+iCount;
	
	if(uEnd>m_uMax)										// Gre prfen
		{
		IncreaseSize(uEnd);
		}

	bit_clear(m_pMem,uPos,iCount);			
}

//*****************************************************************************
//*
//*		Change
//*
//*****************************************************************************
//	Invertiert ein Bit im Bitfeld.
//	uPos		: Nummer des Bits das invertiert werden soll.
//	iCount		: Ist die Anzahl der zu invertierenden Bits (<=0 keine nderung)
//	Ergibt den Wert -1 wenn das Bit vorher auf '1' war, ansonsten 0.  
inline int CBitField::Change(unsigned uPos)
{
	
	if(uPos>=m_uMax)IncreaseSize(uPos);

return bit_change(m_pMem,uPos);
}

inline void CBitField::Change(unsigned uPos,int iCount)
{
unsigned	uEnd;


	if(iCount<=0)return;
	
	uEnd = uPos+iCount;
	
	if(uEnd>m_uMax)										// Gre prfen
		{
		IncreaseSize(uEnd);
		}

	bit_change(m_pMem,uPos,iCount);			
}


//*****************************************************************************
//*
//*		Write
//*
//*****************************************************************************
//	Schreibt ein Bit im Bitfeld.
//	uPos		: Nummer des Bits das gendert werden soll.
//	iBit		: Ist der neue Wert des Bits (ungleich 0 fr 1)
//	Ergibt den Wert -1 wenn das Bit vorher auf '1' war, ansonsten 0.  
inline int CBitField::Write(unsigned uPos,int iBit)
{
	
	if(uPos>=m_uMax)IncreaseSize(uPos);

return bit_write(m_pMem,uPos,iBit);
}

//*****************************************************************************
//*
//*		Find0
//*
//*****************************************************************************
//	Sucht im Bitfeld das nchste Bit das auf '0' ist. 
//	uStart		: Ist die Bitposition ab der gesucht wird 
//	Ergibt die Position bei der ein '0'-Bit gefunden wurde, 
//	bzw. 0xFFFFFFFF wenn kein '0'-Bit gefunden wurde.
inline int	CBitField::Find0() const
{
unsigned	uPos;

	uPos = bit_find_0(m_pMem,m_uMax); 

	if(uPos>=m_uMax)return 0xFFFFFFFF;


return uPos;
}
 
inline int	CBitField::Find0(unsigned uStart) const
{
unsigned	uPos;

	uPos = bit_find_0(m_pMem,m_uMax,uStart); 

	if(uPos>=m_uMax)return 0xFFFFFFFF;


return uPos;
}

//*****************************************************************************
//*
//*		Find1
//*
//*****************************************************************************
//	Sucht im Bitfeld das nchste Bit das auf '1' ist. 
//	uStart		: Ist die Bitposition ab der gesucht wird 
//	Ergibt die Position bei der ein '1'-Bit gefunden wurde, 
//	bzw. 0xFFFFFFFF wenn kein '1'-Bit gefunden wurde.
inline int	CBitField::Find1() const
{
unsigned	uPos;

	uPos = bit_find_1(m_pMem,m_uMax); 

	if(uPos>=m_uMax)return 0xFFFFFFFF;


return uPos;
}
 
inline int	CBitField::Find1(unsigned uStart) const
{
unsigned	uPos;

	uPos = bit_find_1(m_pMem,m_uMax,uStart); 

	if(uPos>=m_uMax)return 0xFFFFFFFF;


return uPos;
}

//*****************************************************************************
//*
//*		ClearAll
//*
//*****************************************************************************
//	Alle Bits im Bitfeld lschen
inline void CBitField::ClearAll()
{

	memset(m_pMem,0x00,m_uMax/8);							

}

//*****************************************************************************
//*
//*		SetAll
//*
//*****************************************************************************
//	Alle Bits im Bitfeld setzen
inline void CBitField::SetAll(unsigned uCount)
{
	
	if(uCount>m_uMax)
		{												// Bitfeld vergrern
		IncreaseFixed(uCount);		
		memset(m_pMem,0xFF,m_uMax/8);
		}
	else{
		memset(m_pMem         ,0xFF,uCount/8);
		memset(m_pMem+uCount/8,0x00,m_uMax/8-uCount/8);
		}

}

//*****************************************************************************
//*
//*		GetMem
//*
//*****************************************************************************
//	Den Speicher des Bitfeldes holen.
//	pMem		: Zeiger auf den Speicherpuffer.
//	iSize		: Ist die Gre des Speicherpuffers in Byte.
inline unsigned char *CBitField::GetMem() const
{
return m_pMem;
}

//*****************************************************************************
//*
//*		operator[]
//*
//*****************************************************************************
//	Ein Byte auslesen aus dem Bitfeld auslesen
//	uPos	: Ist die Byteposition
//	Ergibt das aus dem Bitfeld
inline int CBitField::operator[](unsigned uPos)const
{

	if(uPos>m_uMax/8)return 0;

return m_pMem[uPos];
}

#endif		
#endif
