//*****************************************************************************
//*
//*
//*		exfat_cluster.c
//*
//*
//*****************************************************************************
/*
	cluster.c (03.09.09)
	exFAT file system implementation library.

	Free exFAT implementation.
	Copyright (C) 2010-2016  Andrew Nayenko

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc.,
	51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include	"exfat.h"
#include	<errno.h>
#include	<string.h>


static int shrink_file(struct exfat* ef, struct exfat_node* node,uint32_t current, uint32_t difference);


//*****************************************************************************
//*
//*		s2o
//*
//*****************************************************************************
// 	Sector to absolute offset.
static EXFAT_INLINE exfat_off_t s2o(const struct exfat *ef,exfat_off_t sector)
{

return sector << ef->sb->sector_bits;
}


//*****************************************************************************
//*
//*		c2s
//*
//*****************************************************************************
// 	Cluster to sector.
static EXFAT_INLINE exfat_off_t c2s(const struct exfat *ef,cluster_t cluster)
{
	if(cluster < EXFAT_FIRST_DATA_CLUSTER)
		{
		exfat_bug1("invalid cluster number %u", cluster);
		}

return le32_to_cpu(ef->sb->cluster_sector_start) + ((exfat_off_t) (cluster - EXFAT_FIRST_DATA_CLUSTER) << ef->sb->spc_bits);
}


//*****************************************************************************
//*
//*		exfat_c2o
//*
//*****************************************************************************
// 	Cluster to absolute offset.
EXFAT_STATIC exfat_off_t exfat_c2o(const struct exfat *ef,cluster_t cluster)
{

return s2o(ef, c2s(ef, cluster));
}


//*****************************************************************************
//*
//*		s2c
//*
//*****************************************************************************
// 	Sector to cluster.
static EXFAT_INLINE cluster_t s2c(const struct exfat *ef,exfat_off_t sector)
{
return (cluster_t)((sector - le32_to_cpu(ef->sb->cluster_sector_start)) >> ef->sb->spc_bits) + EXFAT_FIRST_DATA_CLUSTER;
}


//*****************************************************************************
//*
//*		bytes2clusters
//*
//*****************************************************************************
// 	Size in bytes to size in clusters (rounded upwards).
static EXFAT_INLINE uint32_t bytes2clusters(const struct exfat *ef,uint64_t bytes)
{
uint64_t	cluster_size = CLUSTER_SIZE(*ef->sb);

return (uint32_t)((bytes + cluster_size - 1) / cluster_size);
}

//*****************************************************************************
//*
//*		exfat_load_segment
//*
//*****************************************************************************
EXFAT_STATIC void exfat_load_segment(const struct exfat *ef,size_t index)
{
exfat_off_t	offset;
size_t		place;
size_t		size;


	index >>= ef->cmap.seg_size;

	if(BMAP_GET(ef->cmap.valid,index))			// cluster all ready loaded	
		{
		return;
		}

	size	= (1<<ef->cmap.seg_size)/8;
	place	= index*size/sizeof(bitmap_t);
	offset	= exfat_c2o(ef,ef->cmap.start_cluster+index);

	if(exfat_pread(ef->dev, ef->cmap.chunk+place,size,offset) < 0 )
		{
		exfat_error2("failed to read clusters bitmap %u bytes starting at cluster %#x)", size, ef->cmap.start_cluster+index);
		return; 
		}

	BMAP_SET(ef->cmap.valid,index);	

}


//*****************************************************************************
//*
//*		exfat_next_cluster
//*
//*****************************************************************************
EXFAT_STATIC cluster_t exfat_next_cluster(const struct exfat *ef,const struct exfat_node *node,cluster_t cluster)
{
exfat_off_t		fat_offset;
le32_t			next;


	if(cluster < EXFAT_FIRST_DATA_CLUSTER)
		{
		exfat_bug1("bad cluster 0x%x", cluster);
		}

	if(IS_CONTIGUOUS(*node))
		{
		return cluster + 1;
		}

	fat_offset = s2o(ef, le32_to_cpu(ef->sb->fat_sector_start))	+ cluster * sizeof(cluster_t);
	
	if(exfat_pread(ef->dev, &next, sizeof(next), fat_offset) < 0)
		{
		return EXFAT_CLUSTER_BAD;				// the caller should handle this and print appropriate error message 
		}


return le32_to_cpu(next);
}

//*****************************************************************************
//*
//*		exfat_next_clusters
//*
//*****************************************************************************
EXFAT_STATIC int exfat_next_clusters(const struct exfat *ef,const struct exfat_node *node,cluster_t cluster,cluster_t *entry)
{
exfat_off_t		fat_offset;
le32_t			array[128];
int				size;
int				num;


	if(cluster < EXFAT_FIRST_DATA_CLUSTER)
		{
		exfat_bug1("bad cluster 0x%x", cluster);
		}

	if(IS_CONTIGUOUS(*node))
		{					    
		*entry = cluster+1;
		return 128;
		}

	fat_offset	= s2o(ef, le32_to_cpu(ef->sb->fat_sector_start))	+ cluster * sizeof(cluster_t);
	size		= (512-((int)fat_offset&0x1FF));
	
	
	if(exfat_pread(ef->dev, array , size , fat_offset) < 0)
		{
		return 0;								// the caller should handle this and print appropriate error message 
		}

	size   /= sizeof(cluster_t);
	cluster = le32_to_cpu(array[0]);
	*entry  = cluster;

	for(num=1;num<size;num++)					// count the continous clusters
		{
		cluster++;
		if(cluster^le32_to_cpu(array[num]))break;
		}
	

return num;
}

//*****************************************************************************
//*
//*		exfat_advance_cluster
//*
//*****************************************************************************
EXFAT_STATIC cluster_t exfat_advance_cluster(const struct exfat *ef,struct exfat_node *node,uint32_t count)
{
uint32_t	i;


	if(node->fptr_index > count)
		{
		node->fptr_index	= 0;
		node->fptr_cluster	= node->start_cluster;
		}

	for(i=node->fptr_index;i<count;i++)
		{
		node->fptr_cluster = exfat_next_cluster(ef, node, node->fptr_cluster);

		if(CLUSTER_INVALID(node->fptr_cluster))
			{
			break;								// the caller should handle this and print appropriate error message 
			}		          
		}

	node->fptr_index = count;


return node->fptr_cluster;
}

//*****************************************************************************
//*
//*		find_bit_and_set
//*
//*****************************************************************************
static EXFAT_INLINE cluster_t find_bit_and_set(const struct exfat *ef,size_t start,size_t end)
{
bitmap_t	   *bitmap=ef->cmap.chunk;
size_t			start_bitindex;
size_t			end_bitindex;
size_t			start_index;
size_t			end_index;
size_t			index;
size_t			pos;



	end_index   = DIV_ROUND_UP(end,sizeof(bitmap_t)*8);
	start_index = start          /(sizeof(bitmap_t)*8);

	for(index=start_index;index<end_index;index++)
		{
		exfat_load_segment(ef,index*(sizeof(bitmap_t)*8));

		if(bitmap[index]==(bitmap_t)~0)
			{
			continue;
			}
		
		start_bitindex	= MAX((index  )*(sizeof(bitmap_t)*8),start);
		end_bitindex	= MIN((index+1)*(sizeof(bitmap_t)*8),end  );

		for(pos=start_bitindex;pos<end_bitindex;pos++)
			{
			if(BMAP_GET(bitmap,pos)==0)
				{
				BMAP_SET(bitmap, pos);
				return pos + EXFAT_FIRST_DATA_CLUSTER;
				}
			}
		}


return EXFAT_CLUSTER_END;
}

//*****************************************************************************
//*
//*		flush_nodes
//*
//*****************************************************************************
static int flush_nodes(struct exfat *ef,struct exfat_node *node)
{
struct exfat_node  *p;
int					rc;


	for(p=node->child;p!=NULL;p=p->next)
		{
		rc = flush_nodes(ef, p);

		if(rc != 0)
			{
			return rc;
			}
		}


return exfat_flush_node(ef, node);
}

//*****************************************************************************
//*
//*		exfat_flush_nodes
//*
//*****************************************************************************
EXFAT_STATIC int exfat_flush_nodes(struct exfat *ef)
{

return flush_nodes(ef, ef->root);
}

//*****************************************************************************
//*
//*		exfat_flush
//*
//*****************************************************************************
EXFAT_STATIC int exfat_flush(struct exfat *ef)
{
exfat_off_t		offset;
uint32_t		index;
uint32_t		place;
uint32_t		size;


	if(ef->cmap.dirty)
		{
		index	= ef->cmap.first_changed;
		size	= (1<<ef->cmap.seg_size)/8;
		place	= index*size/sizeof(bitmap_t);
		offset	= exfat_c2o(ef,ef->cmap.start_cluster+index);

		for(;index<ef->cmap.last_changed;index++)
			{
			if(BMAP_GET(ef->cmap.changed,index))
				{
				if(exfat_pwrite(ef->dev,ef->cmap.chunk+place,size,offset) < 0)
					{
					exfat_error0("failed to write clusters bitmap");
					return -EIO;
					}
				
				BMAP_CLR(ef->cmap.changed,index);
				}
			
			offset += size;
			place  += size/sizeof(bitmap_t);
			}
		
		ef->cmap.first_changed	= 0;
		ef->cmap.last_changed	= 0;
		ef->cmap.dirty			= false;
		}


return 0;
}

//*****************************************************************************
//*
//*		set_next_cluster
//*
//*****************************************************************************
static bool set_next_cluster(const struct exfat *ef,bool contiguous,cluster_t current,cluster_t next)
{
exfat_off_t		fat_offset;
le32_t			next_le32;


	if(contiguous)
		{
		return true;
		}

	fat_offset = s2o(ef, le32_to_cpu(ef->sb->fat_sector_start))	+ current * sizeof(cluster_t);
	next_le32  = cpu_to_le32(next);

	if(exfat_pwrite(ef->dev, &next_le32, sizeof(next_le32), fat_offset) < 0 )
		{
		exfat_error2("failed to write the next cluster %#x after %#x", next,current);
		return false;
		}


return true;
}

//*****************************************************************************
//*
//*		allocate_cluster
//*
//*****************************************************************************
static cluster_t allocate_cluster(struct exfat *ef,cluster_t hint)
{
cluster_t	cluster;
uint32_t	index;


	   hint-= EXFAT_FIRST_DATA_CLUSTER;
	if(hint>= ef->cmap.chunk_size)
		{
		hint = 0;
		}

	   cluster = find_bit_and_set(ef,hint,ef->cmap.chunk_size);
	if(cluster == EXFAT_CLUSTER_END)
		{
		cluster = find_bit_and_set(ef, 0, hint);
		}

	if(cluster == EXFAT_CLUSTER_END)
		{
		exfat_error0("no free space left");
		return EXFAT_CLUSTER_END;
		}

	index   = cluster-EXFAT_FIRST_DATA_CLUSTER;
	index >>= ef->cmap.seg_size;

	BMAP_SET(ef->cmap.changed,index);

	if(ef->cmap.last_changed)
		{
		if(ef->cmap.first_changed>index)
			{
			ef->cmap.first_changed = index;
			}

		if(ef->cmap.last_changed<=index)
			{
			ef->cmap.last_changed  = index+1;
			}
		}
	else{
		ef->cmap.first_changed = index;
		ef->cmap.last_changed  = index+1;
		}

	ef->cmap.dirty = true;


return cluster;
}

//*****************************************************************************
//*
//*		free_cluster
//*
//*****************************************************************************
static void free_cluster(struct exfat *ef,cluster_t cluster)
{
uint32_t	index;


	if(CLUSTER_INVALID(cluster))
		{
		exfat_bug1("freeing invalid cluster 0x%x", cluster);
		}

	if(cluster - EXFAT_FIRST_DATA_CLUSTER >= ef->cmap.size)
		{
		exfat_bug2("freeing non-existing cluster 0x%x (0x%x)", cluster,ef->cmap.size);
		}


	index = cluster-EXFAT_FIRST_DATA_CLUSTER;
	exfat_load_segment(ef,index);
	
	BMAP_CLR(ef->cmap.chunk,index);
	
	index >>= ef->cmap.seg_size;

	BMAP_SET(ef->cmap.changed,index);

	if(ef->cmap.last_changed)
		{
		if(ef->cmap.first_changed>index)
			{
			ef->cmap.first_changed = index;
			}

		if(ef->cmap.last_changed<=index)
			{
			ef->cmap.last_changed  = index+1;
			}
		}
	else{
		ef->cmap.first_changed = index;
		ef->cmap.last_changed  = index+1;
		}

	ef->cmap.dirty = true;

}

//*****************************************************************************
//*
//*		make_noncontiguous
//*
//*****************************************************************************
static bool make_noncontiguous(const struct exfat *ef,cluster_t first,cluster_t last)
{
cluster_t	c;


	for(c=first;c<last;c++)
		{
		if(!set_next_cluster(ef, false, c, c + 1))
			{
			return false;
			}
		}


return true;
}


//*****************************************************************************
//*
//*		grow_file
//*
//*****************************************************************************
static int grow_file(struct exfat *ef,struct exfat_node *node,uint32_t current,uint32_t difference)
{
uint32_t	allocated = 0;
cluster_t	previous;
cluster_t	next;


	if(difference == 0)
		{
		exfat_bug0("zero clusters count passed");
		}

	if(node->start_cluster != EXFAT_CLUSTER_FREE)
		{
												// get the last cluster of the file 
		previous = exfat_advance_cluster(ef, node, current - 1);

		if(CLUSTER_INVALID(previous))
			{
			exfat_error1("invalid cluster 0x%x while growing", previous);
			return -EIO;
			}
		}
	else{
		if(node->fptr_index != 0)
			{
			exfat_bug1("non-zero pointer index (%u)", node->fptr_index);
			}
												
		
		previous = allocate_cluster(ef, 0);		//	file does not have clusters (i.e. is empty), allocate

		if(CLUSTER_INVALID(previous))
			{
			return -ENOSPC;
			}

		node->fptr_cluster	= node->start_cluster = previous;
		allocated			= 1;
												// file consists of only one cluster, so it's contiguous 
		node->flags |= EXFAT_ATTRIB_CONTIGUOUS;
		}

	while(allocated < difference)
		{
		next = allocate_cluster(ef, previous + 1);

		if(CLUSTER_INVALID(next))
			{
			if(allocated != 0)
				{
				shrink_file(ef, node, current + allocated, allocated);
				}

			return -ENOSPC;
			}

		if(next != previous - 1 && IS_CONTIGUOUS(*node))
			{
												//	it's a pity, but we are not able to keep the file contiguous
			if(!make_noncontiguous(ef, node->start_cluster, previous))
				{
				return -EIO;
				}

			node->flags    &= ~EXFAT_ATTRIB_CONTIGUOUS;
			node->flags    |=  EXFAT_ATTRIB_DIRTY;
			}

		if(!set_next_cluster(ef, IS_CONTIGUOUS(*node), previous, next))
			{
			return -EIO;
			}

		previous = next;
		allocated++;
		}

	if(!set_next_cluster(ef, IS_CONTIGUOUS(*node), previous,EXFAT_CLUSTER_END))
		{
		return -EIO;
		}


return 0;
}

//*****************************************************************************
//*
//*		shrink_file
//*
//*****************************************************************************
static int shrink_file(struct exfat *ef,struct exfat_node *node,uint32_t current,uint32_t difference)
{
cluster_t	previous;
cluster_t	last;
cluster_t	next;

	if(difference == 0)
		{
		exfat_bug0("zero difference passed");
		}

	if(node->start_cluster == EXFAT_CLUSTER_FREE)
		{
		exfat_bug1("unable to shrink empty file (%u clusters)", current);
		}

	if(current < difference)
		{
		exfat_bug2("file underflow (%u < %u)", current, difference);
		}

	if(current > difference)									// crop the file 
		{
		last = exfat_advance_cluster(ef, node,current - difference - 1);

		if(CLUSTER_INVALID(last))
			{
			exfat_error1("invalid cluster 0x%x while shrinking", last);
			return -EIO;
			}

		previous = exfat_next_cluster(ef, node, last);

		if(!set_next_cluster(ef, IS_CONTIGUOUS(*node), last,EXFAT_CLUSTER_END))
			{
			return -EIO;
			}
		}
	else{
		previous			= node->start_cluster;
		node->start_cluster = EXFAT_CLUSTER_FREE;
		node->flags 	   |= EXFAT_ATTRIB_DIRTY;
		}

	node->fptr_index	= 0;
	node->fptr_cluster	= node->start_cluster;

	while(difference--) 										// free remaining clusters 
		{
		if(CLUSTER_INVALID(previous))
			{
			exfat_error1("invalid cluster 0x%x while freeing after shrink",previous);
			return -EIO;
			}

		next = exfat_next_cluster(ef, node, previous);

		if(!set_next_cluster(ef, IS_CONTIGUOUS(*node), previous,EXFAT_CLUSTER_FREE))
			{
			return -EIO;
			}

		free_cluster(ef, previous);
		previous = next;
		}

return 0;
}

//*****************************************************************************
//*
//*		erase_raw
//*
//*****************************************************************************
static bool erase_raw(struct exfat *ef,size_t size,exfat_off_t offset)
{

	if(exfat_pwrite(ef->dev, ef->zero_cluster, size, offset) < 0 )
		{
		exfat_error2("failed to erase %zu bytes at %"PRId64, size, offset);
		return false;
		}


return true;
}

//*****************************************************************************
//*
//*		erase_range
//*
//*****************************************************************************
static int erase_range(struct exfat *ef,struct exfat_node *node,uint64_t begin,uint64_t end)
{
uint64_t	cluster_boundary;
cluster_t	cluster;


	if(begin >= end)
		{
		return 0;
		}

	cluster_boundary	= (begin | (CLUSTER_SIZE(*ef->sb) - 1)) + 1;
	cluster 			= exfat_advance_cluster(ef, node,(cluster_t)(begin / CLUSTER_SIZE(*ef->sb)));
	
	if(CLUSTER_INVALID(cluster))
		{
		exfat_error1("invalid cluster 0x%x while erasing", cluster);
		return -EIO;
		}

														// erase from the beginning to the closest cluster boundary 
	if(!erase_raw(ef, (size_t)(MIN(cluster_boundary, end) - begin),exfat_c2o(ef, cluster) + begin % CLUSTER_SIZE(*ef->sb)))
		{
		return -EIO;
		}

	
	while(cluster_boundary < end)						// erase whole clusters 
		{
		cluster = exfat_next_cluster(ef, node, cluster);
		
		if(CLUSTER_INVALID(cluster))					// the cluster cannot be invalid because we have just allocated it 
			{
			exfat_bug1("invalid cluster 0x%x after allocation", cluster);
			}

		if(!erase_raw(ef, CLUSTER_SIZE(*ef->sb), exfat_c2o(ef, cluster)))
			{
			return -EIO;
			}

		cluster_boundary += CLUSTER_SIZE(*ef->sb);
		}


return 0;
}

//*****************************************************************************
//*
//*		exfat_truncate
//*
//*****************************************************************************
EXFAT_STATIC int exfat_truncate(struct exfat *ef,struct exfat_node *node,uint64_t size,bool erase)
{
uint32_t	c1 = bytes2clusters(ef,node->size);
uint32_t	c2 = bytes2clusters(ef,size);
int			rc = 0;


	if(node->references == 0 && node->parent)
		{
		exfat_bug0("no references, node changes can be lost");
		}

	if(node->size == size)
		{
		return 0;
		}

	if(c1 < c2)
		{
		rc = grow_file(ef, node, c1, c2 - c1);
		}
	else if(c1 > c2)
		{
		rc = shrink_file(ef, node, c1, c1 - c2);
		}

	if(rc != 0)
		{
		return rc;
		}

	if(erase)
		{
		   rc = erase_range(ef, node, node->size, size);
		if(rc!= 0)
			{
			return rc;
			}
		}

	exfat_update_mtime(node);
	node->size		= size;
	node->flags    |= EXFAT_ATTRIB_DIRTY;


return 0;
}

//*****************************************************************************
//*
//*		exfat_count_free_clusters
//*
//*****************************************************************************
EXFAT_STATIC uint32_t exfat_count_free_clusters(const struct exfat *ef)
{
bitmap_t	bit_values;
uint32_t	free_cl=0;
uint32_t	index;
uint32_t	mask;
uint32_t	pos;


	mask = (1<<ef->cmap.seg_size)-1;

	for(index=0;index<ef->cmap.size;index+=sizeof(bitmap_t)*8)
		{
		if(!(index&mask))
			{
			exfat_load_segment(ef,index);
			}

		bit_values = ef->cmap.chunk[index/(sizeof(bitmap_t)*8)];
		
		if(bit_values==0)
			{
			free_cl += sizeof(bitmap_t)*8;
			continue;
			}
			
		if(bit_values==(bitmap_t)~0)
			{
			continue;
			}

		for(pos=sizeof(bitmap_t)*8;pos>0;pos--)
			{	
			if(bit_values&1)
				{
				free_cl++;
				}
			
			bit_values >>= 1;
			}
		}


return free_cl;
}

