//*****************************************************************************
//*
//*
//*		exfat_node.c
//*
//*
//*****************************************************************************
/*
	node.c (09.10.09)
	exFAT file system implementation library.

	Free exFAT implementation.
	Copyright (C) 2010-2016  Andrew Nayenko

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc.,
	51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include	"exfat.h"
#include	<errno.h>
#include	<string.h>



struct iterator									// on-disk nodes iterator
	{
	cluster_t		cluster;
	exfat_off_t		offset;
	int				contiguous;
	char		   *chunk;
	};

//*****************************************************************************
//*
//*		exfat_get_node
//*
//*****************************************************************************
EXFAT_STATIC struct exfat_node *exfat_get_node(struct exfat_node *node)
{
	//	if we switch to multi-threaded mode we will need atomic

	node->references++;

return node;
}

//*****************************************************************************
//*
//*		exfat_put_node
//*
//*****************************************************************************
EXFAT_STATIC void exfat_put_node(struct exfat *ef,struct exfat_node *node)
{
char		buffer[UTF8_BYTES(EXFAT_NAME_MAX) + 1];

	   node->references--;
	if(node->references < 0)
		{
		exfat_get_name(node, buffer, sizeof(buffer) - 1);
		exfat_bug1("reference counter of '%s' is below zero", buffer);
		}
	else if(node->references == 0 && node != ef->root)
		{
		if(node->flags & EXFAT_ATTRIB_DIRTY)
			{
			exfat_get_name(node, buffer, sizeof(buffer) - 1);
			exfat_warn1("dirty node '%s' with zero references", buffer);
			}
		}
}

//*****************************************************************************
//*
//*		exfat_cleanup_node
//*
//*****************************************************************************
// 	*
// 	This function must be called on rmdir and unlink (after the last
// 	exfat_put_node()) to free clusters.
EXFAT_STATIC int exfat_cleanup_node(struct exfat *ef,struct exfat_node *node)
{
int			rc = 0;

	if(node->references != 0)
		{
		exfat_bug1("unable to cleanup a node with %d references",node->references);
		}

	if(node->flags & EXFAT_ATTRIB_UNLINKED)
		{
		// free all clusters and node structure itself
		rc = exfat_truncate(ef, node, 0, true);
		// free the node even in case of error or its memory will be lost
		free(node);
		}

return rc;
}

//*****************************************************************************
//*
//*		co2o
//*
//*****************************************************************************
// 	*
// 	Cluster + offset from the beginning of the directory to absolute offset.
static exfat_off_t co2o(struct exfat *ef,cluster_t cluster,exfat_off_t offset)
{

return exfat_c2o(ef, cluster) + offset % CLUSTER_SIZE(*ef->sb);
}

//*****************************************************************************
//*
//*		opendir
//*
//*****************************************************************************
static int opendir(struct exfat *ef,const struct exfat_node *dir,struct iterator *it)
{
	if(!(dir->flags & EXFAT_ATTRIB_DIR))
		{
		exfat_bug0("not a directory");
		}

	it->cluster 	= dir->start_cluster;
	it->offset		= 0;
	it->contiguous	= IS_CONTIGUOUS(*dir);
	it->chunk		= (char*)malloc(CLUSTER_SIZE(*ef->sb));

	if(it->chunk == NULL)
		{
		exfat_error0("out of memory");
		return -ENOMEM;
		}

	if(exfat_pread(ef->dev, it->chunk, CLUSTER_SIZE(*ef->sb),exfat_c2o(ef, it->cluster)) < 0 )
		{
		exfat_error1("failed to read directory cluster %#x", it->cluster);
		return -EIO;
		}

return 0;
}

//*****************************************************************************
//*
//*		closedir
//*
//*****************************************************************************
static void closedir(struct iterator *it)
{
	it->cluster 	= 0;
	it->offset		= 0;
	it->contiguous	= 0;
	free(it->chunk);
	it->chunk = NULL;
}

//*****************************************************************************
//*
//*		fetch_next_entry
//*
//*****************************************************************************
static bool fetch_next_entry(struct exfat *ef,const struct exfat_node *parent,struct iterator *it)
{
	
	it->offset += sizeof(struct exfat_entry);					// move iterator to the next entry in the directory
	
	if((it->offset & (CLUSTER_SIZE(*ef->sb) - 1)) == 0)			// fetch the next cluster if needed
		{
		if(it->offset >= parent->size)							//	reached the end of directory; the caller should check this
			{
			return true;
			}

		it->cluster = exfat_next_cluster(ef, parent, it->cluster);
		if(CLUSTER_INVALID(it->cluster))
			{
			exfat_error1("invalid cluster 0x%x while reading directory",it->cluster);
			return false;
			}

		if(exfat_pread(ef->dev, it->chunk, CLUSTER_SIZE(*ef->sb),exfat_c2o(ef, it->cluster)) < 0 )
			{
			exfat_error1("failed to read the next directory cluster %#x",it->cluster);
			return false;
			}
		}

return true;
}

//*****************************************************************************
//*
//*		allocate_node
//*
//*****************************************************************************
static struct exfat_node *allocate_node(void)
{
struct exfat_node  *node = (struct exfat_node*)malloc(sizeof(struct exfat_node));




	if(node == NULL)
		{
		exfat_error0("failed to allocate node");
		return NULL;
		}

	memset(node, 0, sizeof(struct exfat_node));

return node;
}

//*****************************************************************************
//*
//*		init_node_meta1
//*
//*****************************************************************************
static void init_node_meta1(struct exfat_node *node,const struct exfat_entry_meta1 *meta1)
{
	node->flags = le16_to_cpu(meta1->attrib);
	node->mtime = exfat_exfat2unix(meta1->mdate , meta1->mtime ,meta1->mtime_cs );
	node->ctime = exfat_exfat2unix(meta1->crdate, meta1->crtime,meta1->crtime_cs);
	node->atime = exfat_exfat2unix(meta1->adate , meta1->atime , 0);	// there is no centiseconds field for atime
}

//*****************************************************************************
//*
//*		init_node_meta2
//*
//*****************************************************************************
static void init_node_meta2(struct exfat_node *node,const struct exfat_entry_meta2 *meta2)
{
	node->size			= le64_to_cpu(meta2->size);
	node->start_cluster = le32_to_cpu(meta2->start_cluster);
	node->fptr_cluster	= node->start_cluster;

	if(meta2->flags & EXFAT_FLAG_CONTIGUOUS)
		{
		node->flags |= EXFAT_ATTRIB_CONTIGUOUS;
		}
}

//*****************************************************************************
//*
//*		get_entry_ptr
//*
//*****************************************************************************
static const struct exfat_entry *get_entry_ptr(const struct exfat *ef,const struct iterator *it)
{
	return (const struct exfat_entry*)(it->chunk + it->offset % CLUSTER_SIZE(*ef->sb));
}

//*****************************************************************************
//*
//*		check_node
//*
//*****************************************************************************
static bool check_node(const struct exfat_node *node,uint16_t actual_checksum,uint16_t reference_checksum,uint64_t valid_size)
{
char		buffer[UTF8_BYTES(EXFAT_NAME_MAX) + 1];

	/*
	   Validate checksum first. If it's invalid all other fields probably
	   contain just garbage.
	*/
	if(actual_checksum != reference_checksum)
		{
		exfat_get_name(node, buffer, sizeof(buffer) - 1);
		exfat_error3("'%s' has invalid checksum (%#hx != %#hx)", buffer,actual_checksum, reference_checksum);
		return false;
		}

	/*
	   exFAT does not support sparse files but allows files with uninitialized
	   clusters. For such files valid_size means initialized data size and
	   cannot be greater than file size. See SetFileValidData() function
	   description in MSDN.
	*/
	if(valid_size > node->size)
		{
		exfat_get_name(node, buffer, sizeof(buffer) - 1);
		exfat_error3("'%s' has valid size (%"PRIu64") greater than size ""(%"PRIu64")", buffer, valid_size, node->size);
		return false;
		}

return true;
}

//*****************************************************************************
//*
//*		decompress_upcase
//*
//*****************************************************************************
static void decompress_upcase(uint16_t *output,const le16_t *source,size_t size)
{
size_t		si;
size_t		oi;

	for(oi=0;oi<EXFAT_UPCASE_CHARS;oi++)
		{
		output[oi] = oi;
		}

	for(si=0,oi=0;si<size&&oi<EXFAT_UPCASE_CHARS;si++)
		{
		uint16_t	ch = le16_to_cpu(source[si]);

		if(ch == 0xFFFF && si + 1 < size)						// indicates a run
			{
			oi += le16_to_cpu(source[++si]);
			}
		else{
			output[oi++] = ch;
			}
		}
}


//*****************************************************************************
//*
//*		readdir
//*
//*****************************************************************************
// 	Reads one entry in directory at position pointed by iterator and fills
// 	node structure.
static int readdir(struct exfat *ef,const struct exfat_node *parent,struct exfat_node **node,struct iterator *it)
{
const struct exfat_entry        *entry;
const struct exfat_entry_meta1  *meta1;
const struct exfat_entry_meta2  *meta2;
const struct exfat_entry_name   *file_name;
const struct exfat_entry_upcase *upcase;
const struct exfat_entry_bitmap *bitmap;
const struct exfat_entry_label  *label;
uint16_t	reference_checksum = 0;
uint16_t	actual_checksum = 0;
uint64_t	valid_size    = 0;
uint64_t	upcase_size   = 0;
uint8_t		continuations = 0;
le16_t	   *upcase_comp = NULL;
le16_t	   *namep = NULL;
uint32_t	data_bytes;
uint32_t	seg_bytes;
int			rc = -EIO;


	*node = NULL;

	for(;;)
		{
		if(it->offset >= parent->size)
			{
			if(continuations != 0)
				{
				exfat_error1("expected %u continuations", (unsigned)continuations);
				goto error;
				}

			return -ENOENT; 									// that's OK, means end of directory
			}

		entry = get_entry_ptr(ef, it);

		switch(entry->type)
			{
		case EXFAT_ENTRY_FILE:

			if(continuations != 0)
				{
				exfat_error1("expected %u continuations before new entry",(unsigned)continuations);
				goto error;
				}

			meta1			= (const struct exfat_entry_meta1*) entry;
			continuations	=  meta1->continuations;
			

			if(continuations < 2)								//	each file entry must have at least 2 continuations:
				{
				exfat_error1("too few continuations (%u)", (unsigned)continuations);
				goto error;
				}

			if(continuations > 1 +DIV_ROUND_UP(EXFAT_NAME_MAX, EXFAT_ENAME_MAX))
				{
				exfat_error1("too many continuations (%u)", (unsigned)continuations);
				goto error;
				}

			reference_checksum	= le16_to_cpu(meta1->checksum);
			actual_checksum 	= exfat_start_checksum(meta1);
			*node				= allocate_node();

			if(*node == NULL)
				{
				rc = -ENOMEM;
				goto error;
				}
			
			(*node)->entry_cluster = it->cluster;				// new node has zero reference counter
			(*node)->entry_offset  = it->offset;
			init_node_meta1(*node, meta1);
			namep = (*node)->name;
			break;

		case EXFAT_ENTRY_FILE_INFO:

			if(continuations < 2)
				{
				exfat_error1("unexpected continuation (%u)",(unsigned)continuations);
				goto error;
				}

			   meta2 = (const struct exfat_entry_meta2*) entry;
			if(meta2->flags & ~(EXFAT_FLAG_ALWAYS1 | EXFAT_FLAG_CONTIGUOUS))
				{
				exfat_error1("unknown flags in meta2 (0x%02X)",(unsigned)meta2->flags);
				goto error;
				}

			init_node_meta2(*node, meta2);
			actual_checksum = exfat_add_checksum(entry, actual_checksum);
			valid_size		= le64_to_cpu(meta2->valid_size);
																// empty files must be marked as non-contiguous
			if((*node)->size == 0 && (meta2->flags & EXFAT_FLAG_CONTIGUOUS))
				{
				exfat_error1("empty file marked as contiguous (0x%02X)",(unsigned)meta2->flags);
				goto error;
				}
																// directories must be aligned on at cluster boundary
			if(((*node)->flags & EXFAT_ATTRIB_DIR) && (*node)->size % CLUSTER_SIZE(*ef->sb) != 0)
				{
				exfat_error1("directory has invalid size %"PRIu64" bytes",(*node)->size);
				goto error;
				}

			continuations--;
			break;

		case EXFAT_ENTRY_FILE_NAME:

			if(continuations == 0)
				{
				exfat_error0("unexpected continuation");
				goto error;
				}

			file_name		= (const struct exfat_entry_name*) entry;
			actual_checksum =  exfat_add_checksum(entry, actual_checksum);

			memcpy(namep, file_name->name,MIN(EXFAT_ENAME_MAX,((*node)->name + EXFAT_NAME_MAX - namep)) *sizeof(le16_t));
			namep += EXFAT_ENAME_MAX;
			if(--continuations == 0)
				{
				if(!check_node(*node, actual_checksum, reference_checksum,valid_size))
					{
					goto error;
					}

				if(!fetch_next_entry(ef, parent, it))
					{
					goto error;
					}

				return 0;										// entry completed
				}

			break;

		case EXFAT_ENTRY_UPCASE:

			if(ef->upcase != NULL)
				{
				break;
				}

			upcase = (const struct exfat_entry_upcase*) entry;
			if(CLUSTER_INVALID(le32_to_cpu(upcase->start_cluster)))
				{
				exfat_error1("invalid cluster 0x%x in upcase table",le32_to_cpu(upcase->start_cluster));
				goto error;
				}

			   upcase_size = le64_to_cpu(upcase->size);
			if(upcase_size == 0 || upcase_size > EXFAT_UPCASE_CHARS * sizeof(uint16_t) || upcase_size % sizeof(uint16_t) != 0)
				{
				exfat_error1("bad upcase table size (%"PRIu64" bytes)",upcase_size);
				goto error;
				}

			   upcase_comp = (le16_t*)malloc((int)upcase_size);
			if(upcase_comp == NULL)
				{
				exfat_error1("failed to allocate upcase table (%"PRIu64" bytes)",upcase_size);
				rc = -ENOMEM;
				goto error;
				}

																// read compressed upcase table
			if(exfat_pread(ef->dev, upcase_comp,(size_t)upcase_size,exfat_c2o(ef, le32_to_cpu(upcase->start_cluster))) < 0 )
				{
				free(upcase_comp);
				exfat_error2("failed to read upper case table ""(%"PRIu64" bytes starting at cluster %#x)",upcase_size,le32_to_cpu(upcase->start_cluster));
				goto error;
				}

																// decompress upcase table
			   ef->upcase = (uint16_t*)calloc(EXFAT_UPCASE_CHARS, sizeof(uint16_t));
			if(ef->upcase == NULL)
				{
				free(upcase_comp);
				exfat_error0("failed to allocate decompressed upcase table");
				rc = -ENOMEM;
				goto error;
				}

			decompress_upcase(ef->upcase,upcase_comp,(size_t)(upcase_size/sizeof(uint16_t)));
			free(upcase_comp);
			break;

		case EXFAT_ENTRY_BITMAP:

			bitmap				  = (const struct exfat_entry_bitmap*) entry;
			ef->cmap.start_cluster = le32_to_cpu(bitmap->start_cluster);
			
			if(CLUSTER_INVALID(ef->cmap.start_cluster))
				{
				exfat_error1("invalid cluster 0x%x in clusters bitmap",ef->cmap.start_cluster);
				goto error;
				}

			ef->cmap.size = le32_to_cpu(ef->sb->cluster_count) - EXFAT_FIRST_DATA_CLUSTER;

			if(le64_to_cpu(bitmap->size) < DIV_ROUND_UP(ef->cmap.size, 8))
				{
				exfat_error2("invalid clusters bitmap size: %"PRIu64" (expected at least %u)",le64_to_cpu(bitmap->size),DIV_ROUND_UP(ef->cmap.size, 8));
				goto error;
				}

			ef->cmap.seg_size	= ef->sb->sector_bits+ef->sb->spc_bits+3;
			ef->cmap.chunk_size = ef->cmap.size;				// FIXME bitmap can be rather big, up to 512 MB
			data_bytes			= BMAP_SIZE(ef->cmap.chunk_size);
			ef->cmap.seg_byte	= (data_bytes>>ef->cmap.seg_size)+1;
			seg_bytes			= ROUND_UP(ef->cmap.seg_byte,8);
			ef->cmap.chunk		= (bitmap_t*)malloc(data_bytes+seg_bytes*2);
			ef->cmap.valid		= ef->cmap.chunk+data_bytes/sizeof(bitmap_t);
			ef->cmap.changed	= ef->cmap.valid+seg_bytes /sizeof(bitmap_t);

			if(ef->cmap.chunk == NULL)
				{
				exfat_error1("failed to allocate clusters bitmap chunk ""(%"PRIu64" bytes)", le64_to_cpu(bitmap->size));
				rc = -ENOMEM;
				goto error;
				}

			memset(ef->cmap.valid  ,0,seg_bytes);
			memset(ef->cmap.changed,0,seg_bytes);

			break;

		case EXFAT_ENTRY_LABEL:

			label = (const struct exfat_entry_label*) entry;
			if(label->length > EXFAT_ENAME_MAX)
				{
				exfat_error1("too long label (%u chars)",(unsigned)label->length);
				goto error;
				}

			if(utf16_to_utf8(ef->label, label->name,sizeof(ef->label) - 1, EXFAT_ENAME_MAX) != 0)
				{
				goto error;
				}

			break;

		default:

			if(!(entry->type & EXFAT_ENTRY_VALID))
				{
				break;											// deleted entry, ignore it
				}

			if(!(entry->type & EXFAT_ENTRY_OPTIONAL))
				{
				exfat_error1("unknown entry type 0x%02X",(unsigned)entry->type);
				goto error;
				}

																// optional entry, warn and skip
			exfat_warn1("unknown entry type 0x%02X",(unsigned)entry->type);

			if(continuations == 0)
				{
				exfat_error0("unexpected continuation");
				goto error;
				}

			continuations--;
			break;
			}

		if(!fetch_next_entry(ef, parent, it))
			{
			goto error;
			}
		}
																// we never reach here
error:

	free(*node);
	*node = NULL;

return rc;
}

//*****************************************************************************
//*
//*		exfat_cache_directory
//*
//*****************************************************************************
EXFAT_STATIC int exfat_cache_directory(struct exfat *ef,struct exfat_node *dir)
{
struct exfat_node  *current = NULL;
struct exfat_node  *node;
struct iterator		it;
int					rc;


	if(dir->flags & EXFAT_ATTRIB_CACHED)						// already cached
		{
		if(dir->cache_turn==ef->cache_turn)
			{
			return 0;
			}

		goto Update;											
		}

	   rc = opendir(ef, dir, &it);
	if(rc != 0)
		{
		return rc;
		}

	while((rc = readdir(ef, dir, &node, &it)) == 0)
		{
		node->parent = dir;
		if(current != NULL)
			{
			current->next	= node;
			node->prev		= current;
			}
		else{
			dir->child = node;
			}

		current = node;
		}

	closedir(&it);

	if(rc != -ENOENT)
		{
		for(current=dir->child;current;current=node)			// rollback
			{
			node = current->next;
			free(current);
			}

		dir->child = NULL;
		return rc;
		}

	dir->flags |= EXFAT_ATTRIB_CACHED;
	
Update:
	
	ef ->cache_turn += 1;
	dir->cache_turn  = ef->cache_turn;

	for(;;)														// set cache turn
		{
		dir = dir->parent;

		if(!dir || !(dir->flags & EXFAT_ATTRIB_CACHED))
			{
			break;
			}
		
		dir->cache_turn = ef->cache_turn;
		}

	ef->cache_dirs++;


return 0;
}

//*****************************************************************************
//*
//*		tree_attach
//*
//*****************************************************************************
static void tree_attach(struct exfat_node *dir,struct exfat_node *node)
{
	
	node->parent = dir;

	if(dir->child)
		{
		dir->child->prev	= node;
		node->next			= dir->child;
		}

	dir->child = node;

}

//*****************************************************************************
//*
//*		tree_detach
//*
//*****************************************************************************
static void tree_detach(struct exfat_node *node)
{
	if(node->prev)
		{
		node->prev->next = node->next;
		}
	else{														// this is the first node in the list
		node->parent->child = node->next;
		}

	if(node->next)
		{
		node->next->prev = node->prev;
		}

	node->parent	= NULL;
	node->prev		= NULL;
	node->next		= NULL;
}

//*****************************************************************************
//*
//*		reset_cache
//*
//*****************************************************************************
static void reset_cache(struct exfat *ef,struct exfat_node *node)
{
char	buffer[UTF8_BYTES(EXFAT_NAME_MAX) + 1];


	while(node->child)
		{
		struct exfat_node  *p = node->child;

		reset_cache(ef, p);
		tree_detach(p);
		free(p);
		}

	if(node->flags & EXFAT_ATTRIB_CACHED)
		{
		node->flags &= ~EXFAT_ATTRIB_CACHED;

		if(ef->cache_dirs)
			{
			ef->cache_dirs--;
			}
		}

	if(node->references != 0)
		{
		exfat_get_name(node,buffer,sizeof(buffer) - 1);
		exfat_warn2   ("non-zero reference counter (%d) for '%s'",node->references,buffer);
		}

	if(node!=ef->root && (node->flags&EXFAT_ATTRIB_DIRTY))
		{
		exfat_get_name(node, buffer, sizeof(buffer) - 1);
		exfat_bug1("node '%s' is dirty", buffer);
		}

	while(node->references>0)
		{
		exfat_put_node(ef, node);
		}

}

//*****************************************************************************
//*
//*		exfat_reset_cache
//*
//*****************************************************************************
EXFAT_STATIC void exfat_reset_cache(struct exfat *ef)
{
	reset_cache(ef, ef->root);
}


//*****************************************************************************
//*
//*		exfat_cache_directory
//*
//*****************************************************************************
static int cache_release(struct exfat *ef,struct exfat_node *node,unsigned diff)
{
struct exfat_node  *next;
unsigned 			value;
unsigned 			count;

	
	if(!node || !(node->flags&EXFAT_ATTRIB_CACHED))
		{
		return 0;
		}

	next   = node->child;
	count  = 0;

	while(next)
		{
		count += cache_release(ef,next,diff);

		if(next->references>0 || next->child)
			{
			count |= 0x80000000;
			}
		
		next = next->next;
		}

	if(count<0x80000000 && node->references<=0 && !(node->flags&EXFAT_ATTRIB_DIRTY))
		{
		   value = ef->cache_turn - node->cache_turn;
		if(value>diff)
			{
			reset_cache(ef,node);
			count++;
			}
		}


return count;
}

//*****************************************************************************
//*
//*		exfat_cache_directory
//*
//*****************************************************************************
EXFAT_STATIC int exfat_cache_release(struct exfat *ef,unsigned diff)
{
return cache_release(ef,ef->root,diff);
}

//*****************************************************************************
//*
//*		next_entry
//*
//*****************************************************************************
static bool next_entry(struct exfat *ef,const struct exfat_node *parent,cluster_t *cluster,exfat_off_t *offset)
{

	*offset += sizeof(struct exfat_entry);

	if(*offset % CLUSTER_SIZE(*ef->sb) == 0)
		{
		*cluster = exfat_next_cluster(ef, parent, *cluster);

		if(CLUSTER_INVALID(*cluster))
			{
			exfat_error1("invalid cluster %#x while getting next entry",*cluster);
			return false;
			}
		}


return true;
}

//*****************************************************************************
//*
//*		exfat_flush_node
//*
//*****************************************************************************
EXFAT_STATIC int exfat_flush_node(struct exfat *ef,struct exfat_node *node)
{
cluster_t	cluster;
exfat_off_t	offset;
exfat_off_t	meta1_offset, meta2_offset;
struct exfat_entry_meta1 meta1;
struct exfat_entry_meta2 meta2;


	if(!(node->flags & EXFAT_ATTRIB_DIRTY))
		{
		return 0; // no need to flush
		}

	if(ef->ro)
		{
		exfat_bug0("unable to flush node to read-only FS");
		}

	if(node->parent == NULL)
		{
		return 0; // do not flush unlinked node
		}

	cluster 		= node->entry_cluster;
	offset			= node->entry_offset;
	meta1_offset	= co2o(ef, cluster, offset);

	if(!next_entry(ef, node->parent, &cluster, &offset))
		{
		return -EIO;
		}

	meta2_offset = co2o(ef, cluster, offset);

	if(exfat_pread(ef->dev, &meta1, sizeof(meta1), meta1_offset) < 0 )
		{
		exfat_error0("failed to read meta1 entry on flush");
		return -EIO;
		}

	if(meta1.type != EXFAT_ENTRY_FILE)
		{
		exfat_bug1("invalid type of meta1: 0x%0XX", (unsigned)meta1.type);
		}

	meta1.attrib = cpu_to_le16((uint16_t)node->flags);
	exfat_unix2exfat(node->ctime, &meta1.crdate,&meta1.crtime,&meta1.crtime_cs);
	exfat_unix2exfat(node->mtime, &meta1.mdate, &meta1.mtime, &meta1.mtime_cs );
	exfat_unix2exfat(node->atime, &meta1.adate, &meta1.atime, NULL);

	if(exfat_pread(ef->dev, &meta2, sizeof(meta2), meta2_offset) < 0 )
		{
		exfat_error0("failed to read meta2 entry on flush");
		return -EIO;
		}

	if(meta2.type != EXFAT_ENTRY_FILE_INFO)
		{
		exfat_bug1("invalid type of meta2: 0x%02X",(unsigned)meta2.type);
		}

	meta2.size			= meta2.valid_size = cpu_to_le64(node->size);
	meta2.start_cluster = cpu_to_le32(node->start_cluster);
	meta2.flags 		= EXFAT_FLAG_ALWAYS1;

	if(node->size != 0 && IS_CONTIGUOUS(*node))				// empty files must not be marked as contiguous
		{
		meta2.flags |= EXFAT_FLAG_CONTIGUOUS;
		}

															// name hash remains unchanged, no need to recalculate it
	meta1.checksum = exfat_calc_checksum(&meta1, &meta2, node->name);

	if(exfat_pwrite(ef->dev, &meta1, sizeof(meta1), meta1_offset) < 0 )
		{
		exfat_error0("failed to write meta1 entry on flush");
		return -EIO;
		}

	if(exfat_pwrite(ef->dev, &meta2, sizeof(meta2), meta2_offset) < 0 )
		{
		exfat_error0("failed to write meta2 entry on flush");
		return -EIO;
		}

	node->flags &= ~EXFAT_ATTRIB_DIRTY;


return exfat_flush(ef);
}

//*****************************************************************************
//*
//*		erase_entry
//*
//*****************************************************************************
static bool erase_entry(struct exfat *ef,struct exfat_node *node)
{
cluster_t	cluster = node->entry_cluster;
exfat_off_t	offset  = node->entry_offset;
int			name_entries;
uint8_t		entry_type;


	name_entries = DIV_ROUND_UP(utf16_length(node->name), EXFAT_ENAME_MAX);
	entry_type   = EXFAT_ENTRY_FILE & ~EXFAT_ENTRY_VALID;

	if(exfat_pwrite(ef->dev, &entry_type, 1, co2o(ef, cluster, offset)) < 0 )
		{
		exfat_error0("failed to erase meta1 entry");
		return false;
		}

	if(!next_entry(ef, node->parent, &cluster, &offset))
		{
		return false;
		}

	entry_type = EXFAT_ENTRY_FILE_INFO & ~EXFAT_ENTRY_VALID;

	if(exfat_pwrite(ef->dev, &entry_type, 1, co2o(ef, cluster, offset)) < 0 )
		{
		exfat_error0("failed to erase meta2 entry");
		return false;
		}

	while(name_entries--)
		{
		if(!next_entry(ef, node->parent, &cluster, &offset))
			{
			return false;
			}

		entry_type = EXFAT_ENTRY_FILE_NAME & ~EXFAT_ENTRY_VALID;

		if(exfat_pwrite(ef->dev, &entry_type, 1,co2o(ef, cluster, offset)) < 0 )
			{
			exfat_error0("failed to erase name entry");
			return false;
			}
		}

return true;
}

//*****************************************************************************
//*
//*		shrink_directory
//*
//*****************************************************************************
static int shrink_directory(struct exfat *ef,struct exfat_node *dir,exfat_off_t deleted_offset)
{
const struct exfat_node *node;
const struct exfat_node *last_node;
uint64_t	 entries = 0;
uint64_t	 new_size;


	if(!(dir->flags & EXFAT_ATTRIB_DIR))
		{
		exfat_bug0("attempted to shrink a file");
		}

	if(!(dir->flags & EXFAT_ATTRIB_CACHED))
		{
		exfat_bug0("attempted to shrink uncached directory");
		}

	for(last_node=node=dir->child;node;node=node->next)
		{
		if(deleted_offset < node->entry_offset)
			{
			return 0;											// there are other entries after the removed one, no way to shrink
			}

		if(last_node->entry_offset < node->entry_offset)
			{
			last_node = node;
			}
		}

	if(last_node)
		{														// offset of the last entry
		entries += last_node->entry_offset / sizeof(struct exfat_entry);
																// two subentries with meta info
		entries += 2;
																// subentries with file name
		entries += DIV_ROUND_UP(utf16_length(last_node->name),EXFAT_ENAME_MAX);
		}

	   new_size = DIV_ROUND_UP(entries * sizeof(struct exfat_entry),CLUSTER_SIZE(*ef->sb)) * CLUSTER_SIZE(*ef->sb);
	if(new_size == 0)											// directory always has at least 1 cluster
		{
		new_size = CLUSTER_SIZE(*ef->sb);
		}

	if(new_size == dir->size)
		{
		return 0;
		}


return exfat_truncate(ef, dir, new_size, true);
}

//*****************************************************************************
//*
//*		delete_node
//*
//*****************************************************************************
static int delete_node(struct exfat *ef,struct exfat_node *node)
{
struct exfat_node  *parent;
exfat_off_t			offset;
int					rc;


	parent	= node->parent;
	offset	= node->entry_offset;

	exfat_get_node(parent);

	if(!erase_entry(ef, node))
		{
		exfat_put_node(ef, parent);
		return -EIO;
		}

	exfat_update_mtime(parent);
	tree_detach(node);

	rc				= shrink_directory(ef,parent,offset);
	node->flags    |= EXFAT_ATTRIB_UNLINKED;

	if(rc != 0)
		{
		exfat_flush_node(ef,parent);
		exfat_put_node  (ef,parent);
		return rc;
		}

	rc = exfat_flush_node(ef, parent);
	exfat_put_node(ef, parent);


return rc;
}

//*****************************************************************************
//*
//*		exfat_unlink
//*
//*****************************************************************************
EXFAT_STATIC int exfat_unlink(struct exfat *ef,struct exfat_node *node)
{

	if(node->flags & EXFAT_ATTRIB_DIR)
		{
		return -EISDIR;
		}

return delete_node(ef, node);
}

//*****************************************************************************
//*
//*		exfat_rmdir
//*
//*****************************************************************************
EXFAT_STATIC int exfat_rmdir(struct exfat *ef,struct exfat_node *node)
{
int			rc;

	if(!(node->flags & EXFAT_ATTRIB_DIR))
		{
		return -ENOTDIR;
		}

	   rc = exfat_cache_directory(ef, node);					// check that directory is empty
	if(rc != 0)
		{
		return rc;
		}

	if(node->child)
		{
		return -ENOTEMPTY;
		}

return delete_node(ef, node);
}

//*****************************************************************************
//*
//*		grow_directory
//*
//*****************************************************************************
static int grow_directory(struct exfat *ef,struct exfat_node *dir,uint64_t asize,uint32_t difference)
{

return exfat_truncate(ef, dir,DIV_ROUND_UP(asize + difference, CLUSTER_SIZE(*ef->sb))* CLUSTER_SIZE(*ef->sb), true);
}

//*****************************************************************************
//*
//*		find_slot
//*
//*****************************************************************************
static int find_slot(struct exfat *ef,struct exfat_node *dir,cluster_t *cluster,exfat_off_t *offset,int subentries)
{
const struct exfat_entry *entry;
int					contiguous = 0;
struct iterator		it;
int					rc;


	   rc = opendir(ef, dir, &it);
	if(rc != 0)
		{
		return rc;
		}

	for(;;)
		{
		if(contiguous == 0)
			{
			*cluster	= it.cluster;
			*offset 	= it.offset;
			}

		   entry = get_entry_ptr(ef, &it);
		if(entry->type & EXFAT_ENTRY_VALID)
			{
			contiguous = 0;
			}
		else{
			contiguous++;
			}

		if(contiguous == subentries)
			{
			break;												// suitable slot is found
			}

		if(it.offset + sizeof(struct exfat_entry) >= dir->size)
			{
			   rc = grow_directory(ef, dir, dir->size,(subentries - contiguous) * sizeof(struct exfat_entry));
			if(rc != 0)
				{
				closedir(&it);
				return rc;
				}
			}

		if(!fetch_next_entry(ef, dir, &it))
			{
			closedir(&it);
			return -EIO;
			}
		}

	closedir(&it);


return 0;
}

//*****************************************************************************
//*
//*		write_entry
//*
//*****************************************************************************
static int write_entry(struct exfat *ef,struct exfat_node *dir,const le16_t *name,cluster_t cluster,exfat_off_t offset,uint16_t attrib)
{
const size_t			 name_length = utf16_length(name);
const int				 name_entries = DIV_ROUND_UP(name_length, EXFAT_ENAME_MAX);
struct exfat_node		*node;
struct exfat_entry_meta1 meta1;
struct exfat_entry_meta2 meta2;
int						 i;


	   node = allocate_node();
	if(node == NULL)
		{
		return -ENOMEM;
		}

	node->entry_cluster = cluster;
	node->entry_offset	= offset;
	memcpy(node->name, name, name_length * sizeof(le16_t));

	memset(&meta1,0,sizeof(meta1));

	meta1.type			= EXFAT_ENTRY_FILE;
	meta1.continuations = 1 + name_entries;
	meta1.attrib		= cpu_to_le16(attrib);
	exfat_unix2exfat(time(NULL), &meta1.crdate, &meta1.crtime,&meta1.crtime_cs);
	meta1.adate 		= meta1.mdate = meta1.crdate;
	meta1.atime 		= meta1.mtime = meta1.crtime;
	meta1.mtime_cs		= meta1.crtime_cs;						// there is no atime_cs

	memset(&meta2,0,sizeof(meta2));

	meta2.type			= EXFAT_ENTRY_FILE_INFO;
	meta2.flags 		= EXFAT_FLAG_ALWAYS1;
	meta2.name_length	= name_length;
	meta2.name_hash 	= exfat_calc_name_hash(ef, node->name);
	meta2.start_cluster = cpu_to_le32(EXFAT_CLUSTER_FREE);
	meta1.checksum		= exfat_calc_checksum(&meta1, &meta2, node->name);

	if(exfat_pwrite(ef->dev, &meta1, sizeof(meta1),co2o(ef, cluster, offset)) < 0 )
		{
		exfat_error0("failed to write meta1 entry");
		return -EIO;
		}

	if(!next_entry(ef, dir, &cluster, &offset))
		{
		return -EIO;
		}

	if(exfat_pwrite(ef->dev, &meta2, sizeof(meta2),co2o(ef, cluster, offset)) < 0 )
		{
		exfat_error0("failed to write meta2 entry");
		return -EIO;
		}

	for(i=0;i<name_entries;i++)
		{
		struct exfat_entry_name name_entry = {EXFAT_ENTRY_FILE_NAME, 0};

		memcpy(name_entry.name, node->name + i * EXFAT_ENAME_MAX,MIN(EXFAT_ENAME_MAX, EXFAT_NAME_MAX - i * EXFAT_ENAME_MAX) *sizeof(le16_t));

		if(!next_entry(ef, dir, &cluster, &offset))
			{
			return -EIO;
			}

		if(exfat_pwrite(ef->dev, &name_entry, sizeof(name_entry),co2o(ef, cluster, offset)) < 0 )
			{
			exfat_error0("failed to write name entry");
			return -EIO;
			}
		}

	init_node_meta1(node, &meta1);
	init_node_meta2(node, &meta2);

	tree_attach(dir, node);
	exfat_update_mtime(dir);


return 0;
}

//*****************************************************************************
//*
//*		create
//*
//*****************************************************************************
static int create(struct exfat *ef,const char *path,uint16_t attrib)
{
struct exfat_node  *dir;
struct exfat_node  *existing;
cluster_t			cluster = EXFAT_CLUSTER_BAD;
exfat_off_t			offset = -1;
le16_t				name[EXFAT_NAME_MAX + 1];
int					rc;

	   rc = exfat_split(ef, &dir, &existing, name, path);
	if(rc != 0)
		{
		return rc;
		}
//PrintfLogger("crate A");
	if(existing != NULL)
		{
		exfat_put_node(ef, existing);
		exfat_put_node(ef, dir);
		return -EEXIST;
		}

	   rc = find_slot(ef, dir, &cluster, &offset,2 + DIV_ROUND_UP(utf16_length(name), EXFAT_ENAME_MAX));
//PrintfLogger("crate B");
	if(rc != 0)
		{
		exfat_put_node(ef, dir);
		return rc;
		}

	   rc = write_entry(ef, dir, name, cluster, offset, attrib);
//PrintfLogger("crate C");
	if(rc != 0)
		{
		exfat_put_node(ef, dir);
		return rc;
		}

	rc = exfat_flush_node(ef, dir);
//PrintfLogger("crate D");
	exfat_put_node(ef, dir);
//PrintfLogger("crate E");

return rc;
}

//*****************************************************************************
//*
//*		exfat_mknod
//*
//*****************************************************************************
EXFAT_STATIC int exfat_mknod(struct exfat *ef,const char *path)
{

return create(ef, path, EXFAT_ATTRIB_ARCH);
}

//*****************************************************************************
//*
//*		exfat_mkdir
//*
//*****************************************************************************
EXFAT_STATIC int exfat_mkdir(struct exfat *ef,const char *path)
{
struct exfat_node  *node;
int					rc;


	   rc = create(ef, path, EXFAT_ATTRIB_DIR);
	if(rc != 0)
		{
		return rc;
		}

	   rc = exfat_lookup(ef, &node, path);
	if(rc != 0)
		{
		return 0;
		}
																// directories always have at least one cluster
	   rc = exfat_truncate(ef, node, CLUSTER_SIZE(*ef->sb), true);
	if(rc != 0)
		{
		delete_node(ef, node);
		exfat_put_node(ef, node);
		return rc;
		}

	   rc = exfat_flush_node(ef, node);
	if(rc != 0)
		{
		delete_node(ef, node);
		exfat_put_node(ef, node);
		return rc;
		}

	exfat_put_node(ef, node);


return 0;
}

//*****************************************************************************
//*
//*		rename_entry
//*
//*****************************************************************************
static int rename_entry(struct exfat *ef,struct exfat_node *dir,struct exfat_node *node,const le16_t *name,cluster_t new_cluster,exfat_off_t new_offset)
{
struct exfat_entry_meta1 meta1;
struct exfat_entry_meta2 meta2;
cluster_t		old_cluster  = node->entry_cluster;
exfat_off_t		old_offset   = node->entry_offset;
const size_t	name_length  = utf16_length(name);
const int		name_entries = DIV_ROUND_UP(name_length, EXFAT_ENAME_MAX);
int				i;

	if(exfat_pread(ef->dev, &meta1, sizeof(meta1),co2o(ef, old_cluster, old_offset)) < 0 )
		{
		exfat_error0("failed to read meta1 entry on rename");
		return -EIO;
		}

	if(!next_entry(ef, node->parent, &old_cluster, &old_offset))
		{
		return -EIO;
		}

	if(exfat_pread(ef->dev, &meta2, sizeof(meta2),co2o(ef, old_cluster, old_offset)) < 0 )
		{
		exfat_error0("failed to read meta2 entry on rename");
		return -EIO;
		}

	meta1.continuations = 1 + name_entries;
	meta2.name_hash 	= exfat_calc_name_hash(ef, name);
	meta2.name_length	= name_length;
	meta1.checksum		= exfat_calc_checksum(&meta1, &meta2, name);

	if(!erase_entry(ef, node))
		{
		return -EIO;
		}

	node->entry_cluster = new_cluster;
	node->entry_offset	= new_offset;

	if(exfat_pwrite(ef->dev, &meta1, sizeof(meta1),co2o(ef, new_cluster, new_offset)) < 0 )
		{
		exfat_error0("failed to write meta1 entry on rename");
		return -EIO;
		}

	if(!next_entry(ef, dir, &new_cluster, &new_offset))
		{
		return -EIO;
		}

	if(exfat_pwrite(ef->dev, &meta2, sizeof(meta2),co2o(ef, new_cluster, new_offset)) < 0 )
		{
		exfat_error0("failed to write meta2 entry on rename");
		return -EIO;
		}

	for(i=0;i<name_entries;i++)
		{
		struct exfat_entry_name name_entry =
			{
			EXFAT_ENTRY_FILE_NAME, 0
			};
		memcpy(name_entry.name, name + i * EXFAT_ENAME_MAX,EXFAT_ENAME_MAX * sizeof(le16_t));
		if(!next_entry(ef, dir, &new_cluster, &new_offset))
			{
			return -EIO;
			}

		if(exfat_pwrite(ef->dev, &name_entry, sizeof(name_entry),co2o(ef, new_cluster, new_offset)) < 0 )
			{
			exfat_error0("failed to write name entry on rename");
			return -EIO;
			}
		}

	memcpy(node->name, name, (EXFAT_NAME_MAX + 1) * sizeof(le16_t));
	tree_detach(node);
	tree_attach(dir, node);


return 0;
}

//*****************************************************************************
//*
//*		exfat_rename
//*
//*****************************************************************************
EXFAT_STATIC int exfat_rename(struct exfat *ef,const char *old_path,const char *new_path)
{
le16_t				name[EXFAT_NAME_MAX + 1];
struct exfat_node  *node;
struct exfat_node  *existing;
struct exfat_node  *now;
struct exfat_node  *dir;
cluster_t			cluster;
exfat_off_t			offset;
int					rc;



	   rc = exfat_lookup(ef, &node, old_path);
	if(rc!= 0)
		{
		return rc;
		}

	   rc = exfat_split(ef, &dir, &existing, name, new_path);
	if(rc!= 0)
		{
		exfat_put_node(ef, node);
		return rc;
		}

	if(node->flags & EXFAT_ATTRIB_DIR)						// check that target is not a subdirectory of the source
		{
		for(now=dir;now;now=now->parent)
			{
			if(node == now)
				{
				if(existing != NULL)
					{
					exfat_put_node(ef, existing);
					}

				exfat_put_node(ef, dir);
				exfat_put_node(ef, node);
				return -EINVAL;
				}
			}
		}

	if(existing != NULL)
		{
		if(existing != node)								// remove target if it's not the same node as source
			{
			if(existing->flags & EXFAT_ATTRIB_DIR)
				{
				if(node->flags & EXFAT_ATTRIB_DIR)
					{
					rc = exfat_rmdir(ef, existing);
					}
				else{
					rc = -ENOTDIR;
					}
				}
			else{
				if(!(node->flags & EXFAT_ATTRIB_DIR))
					{
					rc = exfat_unlink(ef, existing);
					}
				else{
					rc = -EISDIR;
					}
				}

			exfat_put_node(ef, existing);

			if(rc != 0)										//	free clusters even if something went wrong; overwise they
				{
				exfat_cleanup_node(ef, existing);
				exfat_put_node(ef, dir);
				exfat_put_node(ef, node);
				return rc;
				}

			   rc = exfat_cleanup_node(ef, existing);
			if(rc != 0)
				{
				exfat_put_node(ef, dir);
				exfat_put_node(ef, node);
				return rc;
				}
			}
		else{
			exfat_put_node(ef, existing);
			}
		}
	
	cluster = EXFAT_CLUSTER_BAD;
	offset  = -1;

	   rc = find_slot(ef, dir, &cluster, &offset,2 + DIV_ROUND_UP(utf16_length(name), EXFAT_ENAME_MAX));
	if(rc!= 0)
		{
		exfat_put_node(ef, dir);
		exfat_put_node(ef, node);
		return rc;
		}

	rc = rename_entry(ef, dir, node, name, cluster, offset);
	exfat_put_node(ef, dir);
	exfat_put_node(ef, node);


return rc;
}

//*****************************************************************************
//*
//*		exfat_utimes
//*
//*****************************************************************************
EXFAT_STATIC void exfat_utimes(struct exfat_node *node,const struct timespec tv[2])
{
	node->atime 	= tv[0].tv_sec;
	node->mtime 	= tv[1].tv_sec;
	node->flags    |= EXFAT_ATTRIB_DIRTY;
}

//*****************************************************************************
//*
//*		exfat_update_atime
//*
//*****************************************************************************
EXFAT_STATIC void exfat_update_atime(struct exfat_node *node)
{
	node->atime 	= time(NULL);
	node->flags    |= EXFAT_ATTRIB_DIRTY;
}

//*****************************************************************************
//*
//*		exfat_update_mtime
//*
//*****************************************************************************
EXFAT_STATIC void exfat_update_mtime(struct exfat_node *node)
{
	node->mtime 	= time(NULL);
	node->flags    |= EXFAT_ATTRIB_DIRTY;
}

//*****************************************************************************
//*
//*		exfat_get_label
//*
//*****************************************************************************
EXFAT_STATIC const char *exfat_get_label(struct exfat *ef)
{

return ef->label;
}

//*****************************************************************************
//*
//*		find_label
//*
//*****************************************************************************
static int find_label(struct exfat *ef,cluster_t *cluster,exfat_off_t *offset)
{
struct iterator		it;
int					rc;


	   rc = opendir(ef, ef->root, &it);
	if(rc != 0)
		{
		return rc;
		}

	for(;;)
		{
		if(it.offset >= ef->root->size)
			{
			closedir(&it);
			return -ENOENT;
			}

		if(get_entry_ptr(ef, &it)->type == EXFAT_ENTRY_LABEL)
			{
			*cluster	= it.cluster;
			*offset 	= it.offset;
			closedir(&it);
			return 0;
			}

		if(!fetch_next_entry(ef, ef->root, &it))
			{
			closedir(&it);
			return -EIO;
			}
		}
}

//*****************************************************************************
//*
//*		exfat_set_label
//*
//*****************************************************************************
EXFAT_STATIC int exfat_set_label(struct exfat *ef,const char *label)
{
le16_t		label_utf16[EXFAT_ENAME_MAX + 1];
struct      exfat_entry_label entry;
cluster_t	cluster;
exfat_off_t	offset;
int			rc;


	memset(label_utf16,0,sizeof(label_utf16));

	   rc = utf8_to_utf16(label_utf16, label, EXFAT_ENAME_MAX, strlen(label));
	if(rc!= 0)
		{
		return rc;
		}

	   rc = find_label(ef, &cluster, &offset);
	if(rc == -ENOENT)
		{
		rc = find_slot(ef, ef->root, &cluster, &offset, 1);
		}

	if(rc != 0)
		{
		return rc;
		}

	entry.type		= EXFAT_ENTRY_LABEL;
	entry.length	= utf16_length(label_utf16);
	memcpy(entry.name, label_utf16, sizeof(entry.name));

	if(entry.length == 0)
		{
		entry.type ^= EXFAT_ENTRY_VALID;
		}

	if(exfat_pwrite(ef->dev, &entry, sizeof(struct exfat_entry_label),co2o(ef, cluster, offset)) < 0 )
		{
		exfat_error0("failed to write label entry");
		return -EIO;
		}

	strcpy(ef->label, label);

return 0;
}
