//*****************************************************************************
//*
//*
//*		exfat_time.c
//*
//*
//*****************************************************************************
/*
	time.c (03.02.12)
	exFAT file system implementation library.

	Free exFAT implementation.
	Copyright (C) 2010-2016  Andrew Nayenko

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc.,
	51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include	"exfat.h"

//	timezone offset from UTC in seconds; positive for western timezones,

static long exfat_timezone;
static int 	exfat_daylight;

#define		SEC_IN_MIN					TYPE_64(60)
#define		SEC_IN_HOUR 				(60 * SEC_IN_MIN)
#define		SEC_IN_DAY					(24 * SEC_IN_HOUR)
#define		SEC_IN_YEAR 				(365 * SEC_IN_DAY)		// not leap year 
// Unix epoch started at 0:00:00 UTC 1 January 1970 
#define		UNIX_EPOCH_YEAR 			1970
// exFAT epoch started at 0:00:00 UTC 1 January 1980 
#define		EXFAT_EPOCH_YEAR			1980
// number of years from Unix epoch to exFAT epoch 
#define		EPOCH_DIFF_YEAR 			(EXFAT_EPOCH_YEAR - UNIX_EPOCH_YEAR)
// number of days from Unix epoch to exFAT epoch (considering leap days) 
#define		EPOCH_DIFF_DAYS 			(EPOCH_DIFF_YEAR * 365 + EPOCH_DIFF_YEAR / 4)
// number of seconds from Unix epoch to exFAT epoch (considering leap days) 
#define		EPOCH_DIFF_SEC				(EPOCH_DIFF_DAYS * SEC_IN_DAY)
//	number of leap years passed from exFAT epoch to the specified year
#define		LEAP_YEARS(year)			((EXFAT_EPOCH_YEAR + (year) - 1) / 4 - (EXFAT_EPOCH_YEAR - 1) / 4)
// checks whether the specified year is leap 
#define		IS_LEAP_YEAR(year)			((EXFAT_EPOCH_YEAR + (year)) % 4 == 0)

static const time_t days_in_year[] =
	{
	// Jan	Feb	 Mar  Apr  May	Jun	 Jul  Aug  Sep	Oct	 Nov  Dec 
	0,	 0,	 31,  59,  90, 120, 151, 181, 212, 243, 273, 304, 334
	};

struct exfat_time_data
	{
	unsigned short			usYear;					// Year (ie. 2006)
	unsigned char			ucMonth;				// Mounth (1..12)
	unsigned char			ucDay;					// Day (1..31)
	unsigned char			ucHour;					// Hour (0..24)
	unsigned char			ucMinute;				// Minute (0..59)
	unsigned char			ucSecond;				// Secound (0..59)
	unsigned char			ucIsDst;				// Daylight saving in minutes
	};


#define YEAR_DAYS			(365)					// days in   1 year
#define	YEAR_DAYS_4			(YEAR_DAYS*4+1)			// days in   4 years
#define	YEAR_DAYS_100		(YEAR_DAYS_4*25-1)		// days in 100 years
#define	YEAR_DAYS_400		(YEAR_DAYS_100*4+1)		// days in 400 years

//*****************************************************************************
//*
//*		exfat_weekday
//*
//*****************************************************************************
static unsigned exfat_weekday(unsigned uDay,unsigned uMonth,unsigned uYear)
{
static const unsigned char	ucYearDays[16] = {0%7,306%7,337%7,0%7,31%7,61%7,92%7,122%7,153%7,184%7,214%7,245%7,275%7,365%7,365%7,365%7};
unsigned					uVal;

	uMonth   = uMonth&0x0F;
	uDay    &= 0x1F;

	if(uMonth<3)uYear--;

	uVal     =  1;										// 1.March.0000 = Mo
	uVal	+= (uYear% 4)*YEAR_DAYS;	uYear/=4;
	uVal	+= (uYear%25)*YEAR_DAYS_4;	uYear/=25;
	uVal	+= (uYear% 4)*YEAR_DAYS_100;uYear/=4;
	uVal	+= (uYear   )*YEAR_DAYS_400;
	uVal    += ucYearDays[uMonth];
	uVal	+= uDay;
	uVal	%= 7;


return	uVal+1;
}



//*****************************************************************************
//*
//*		exfat_daylightsaving
//*
//*****************************************************************************
static time_t exfat_daylightsaving(const struct exfat_time_data *pData)
{
unsigned	uVal;


	if(pData->ucMonth<=3)								// check begin
        {
        if(pData->ucMonth<3 || pData->ucDay<25)
			{
			return  0;
			}

        uVal = exfat_weekday(25,3,pData->usYear);
        uVal = (7-uVal)+25;

        if(pData->ucDay>uVal)return 3600;
        if(pData->ucDay<uVal)return 0;

		uVal = (pData->ucIsDst)? 1:2;

        if(pData->ucHour<uVal)return 0;

		return 3600;
        }

	if(pData->ucMonth<=10)								// check end
        {
        if(pData->ucMonth<10 || pData->ucDay<25)
			{
			return 3600;
			}

        uVal=exfat_weekday(25,10,pData->usYear);
        uVal=(7-uVal)+25;

        if(pData->ucDay>uVal)return 0;
        if(pData->ucDay<uVal)return 3600;

		uVal = (pData->ucIsDst)? 1:2;

        if(pData->ucHour<uVal)return 3600;

		return 0;
        }


return 0;
}

//*****************************************************************************
//*
//*		exfat_exfat2unix
//*
//*****************************************************************************
EXFAT_STATIC time_t exfat_exfat2unix(le16_t date,le16_t time,uint8_t centisec)
{
time_t		unix_time = EPOCH_DIFF_SEC;
uint16_t	ndate	= le16_to_cpu(date);
uint16_t	ntime	= le16_to_cpu(time);
uint16_t	day		= ndate & 0x1F; 							// 5 bits, 1-31 
uint16_t	month	= ndate >> 5 & 0x0F;						// 4 bits, 1-12 
uint16_t	year	= ndate >> 9;								// 7 bits, 1-127 (+1980) 
uint16_t	twosec	= ntime & 0x1F;								// 5 bits, 0-29 (2 sec granularity) 
uint16_t	min		= ntime >> 5 & 0x3F;						// 6 bits, 0-59 
uint16_t	hour	= ntime >> 11;								// 5 bits, 0-23 


	if(day == 0 || month == 0 || month > 12)
		{
		exfat_error3("bad date %04u-%02u-%02u",year + EXFAT_EPOCH_YEAR,(unsigned)month,(unsigned)day);
		return 0;
		}

	if(hour > 23 || min > 59 || twosec > 29)
		{
		exfat_error3("bad time %02u:%02u:%02u",(unsigned)hour,(unsigned)min,twosec*2);
		return 0;
		}

	if(centisec > 199)
		{
		exfat_error1("bad centiseconds count %u",(unsigned)centisec);
		return 0;
		}

	// every 4th year between 1904 and 2096 is leap 
	unix_time += (time_t)(year * SEC_IN_YEAR + LEAP_YEARS(year) * SEC_IN_DAY);
	unix_time += (days_in_year[month] * (time_t)SEC_IN_DAY);

	// if it's leap year and February has passed we should add 1 day 
	if((EXFAT_EPOCH_YEAR + year) % 4 == 0 && month > 2)
		{
		unix_time += SEC_IN_DAY;
		}

	unix_time += (day - 1) * (time_t)SEC_IN_DAY;
	unix_time += hour      * (time_t)SEC_IN_HOUR;
	unix_time += min       * (time_t)SEC_IN_MIN;
	// exFAT represents time with 2 sec granularity 
	unix_time += twosec * 2;
	unix_time += centisec / 100;

	// exFAT stores timestamps in local time, so we correct it to UTC 
	unix_time += exfat_timezone;

	if(exfat_daylight)
		{
		struct exfat_time_data	data;

		data.usYear		= (short)(year+1980);
		data.ucMonth	= (char )(month);
		data.ucDay		= (char )(day  );
		data.ucHour		= (char )(hour );
		data.ucIsDst	= 0;
		
		unix_time -= exfat_daylightsaving(&data);
		}



return unix_time;
}

//*****************************************************************************
//*
//*		exfat_unix2exfat
//*
//*****************************************************************************
EXFAT_STATIC void exfat_unix2exfat(time_t unix_time,le16_t *date,le16_t *time,uint8_t *centisec)
{
time_t		shift = (time_t)EPOCH_DIFF_SEC + exfat_timezone;
uint16_t	day, month, year;
uint16_t	twosec, min, hour;
int			days;
int			i;


	if(unix_time < shift)							// time before exFAT epoch cannot be represented 
		{
		unix_time = shift;
		}

	unix_time -= shift;

	days	=  unix_time / (time_t)SEC_IN_DAY;
	year	= (4 * days) / (4 * 365 + 1);
	days   -=  year * 365 + LEAP_YEARS(year);
	month	=  0;

	for(i=1;i<=12;i++)
		{
		int		leap_day = (IS_LEAP_YEAR(year) && i == 2);
		int		leap_sub = (IS_LEAP_YEAR(year) && i >= 3);

		if(i == 12 || days - leap_sub < days_in_year[i + 1] + leap_day)
			{
			month	= i;
			days   -= days_in_year[i] + leap_sub;
			break;
			}
		}

	day = days + 1;

	hour	= (uint16_t)((unix_time % (time_t)SEC_IN_DAY ) / (time_t)SEC_IN_HOUR);
	min 	= (uint16_t)((unix_time % (time_t)SEC_IN_HOUR) / (time_t)SEC_IN_MIN);
	twosec	= (uint16_t)((unix_time % (time_t)SEC_IN_MIN ) / 2);

	if(exfat_daylight)
		{
		static const unsigned char 	mon_end[]={31,31,28,31,30,31,30,31,31,30,31,30,31,31,31,31};
		struct exfat_time_data		data;
		unsigned					end;

		data.usYear		= (short)(year+1980);
		data.ucMonth	= (char )(month);
		data.ucDay		= (char )(day  );
		data.ucHour		= (char )(hour );
		data.ucIsDst	= 0xFF;
		
		if(exfat_daylightsaving(&data))
			{
			   hour += 1;
			if(hour>=24)
				{
				hour = 0;
				day += 1;
				end  = mon_end[month];

				if(day>end) 
					{
					if(month==2 && !((year+EXFAT_EPOCH_YEAR)&3))
						{
						end++;
						}
					
					if(day>=end) 
						{
						day    = 1;
						month += 1;

						if(month>12)
							{
							month  = 1;
							year  += 1;
							}
						}
					}
				}
			}
		}


	*date	=  cpu_to_le16((uint16_t)(day    | (month << 5) | (year <<  9)));
	*time	=  cpu_to_le16((uint16_t)(twosec | (min   << 5) | (hour << 11)));
	
	if(centisec)
		{
		*centisec = (unix_time % 2) * 100;
		}
}

//*****************************************************************************
//*
//*		exfat_tzset
//*
//*****************************************************************************
EXFAT_STATIC void exfat_tzset(void)
{
time_t		now;


	tzset();

	now 			= time(NULL);
	exfat_timezone	= mktime(gmtime(&now)) - now;
	exfat_daylight	= (exfat_timezone==-3600);

}
