//*****************************************************************************
//*
//*
//*		exfat_utf.c
//*
//*
//*****************************************************************************
/*
	utf.c (13.09.09)
	exFAT file system implementation library.

	Free exFAT implementation.
	Copyright (C) 2010-2016  Andrew Nayenko

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc.,
	51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include	"exfat.h"
#include	<errno.h>

//*****************************************************************************
//*
//*		wchar_to_utf8
//*
//*****************************************************************************
static char *wchar_to_utf8(char *output,unsigned wc,size_t outsize)
{
	if(wc <= 0x7F)
		{
		if(outsize < 1)
			{
			return NULL;
			}

		*output++ = (char) wc;
		}
	else if(wc <= 0x07FF)
		{
		if(outsize < 2)
			{
			return NULL;
			}

		*output++ = 0xC0 | (wc >> 6);
		*output++ = 0x80 | (wc & 0x3F);
		}
	else if(wc <= 0xFFFF)
		{
		if(outsize < 3)
			{
			return NULL;
			}

		*output++ = 0xE0 | (wc >> 12);
		*output++ = 0x80 | ((wc >> 6) & 0x3F);
		*output++ = 0x80 | (wc & 0x3F);
		}
	else if(wc <= 0x001FFFFF)
		{
		if(outsize < 4)
			{
			return NULL;
			}

		*output++ = 0xF0 | (wc >> 18);
		*output++ = 0x80 | ((wc >> 12) & 0x3F);
		*output++ = 0x80 | ((wc >> 6) & 0x3F);
		*output++ = 0x80 | (wc & 0x3F);
		}
	else if(wc <= 0x03FFFFFF)
		{
		if(outsize < 5)
			{
			return NULL;
			}

		*output++ = 0xF8 | (wc >> 24);
		*output++ = 0x80 | ((wc >> 18) & 0x3F);
		*output++ = 0x80 | ((wc >> 12) & 0x3F);
		*output++ = 0x80 | ((wc >> 6) & 0x3F);
		*output++ = 0x80 | (wc & 0x3F);
		}
	else if(wc <= 0x7FFFFFFF)
		{
		if(outsize < 6)
			{
			return NULL;
			}

		*output++ = 0xFC | (wc >> 30);
		*output++ = 0x80 | ((wc >> 24) & 0x3F);
		*output++ = 0x80 | ((wc >> 18) & 0x3F);
		*output++ = 0x80 | ((wc >> 12) & 0x3F);
		*output++ = 0x80 | ((wc >> 6) & 0x3F);
		*output++ = 0x80 | (wc & 0x3F);
		}
	else{
		return NULL;
		}

return output;
}

//*****************************************************************************
//*
//*		utf16_to_wchar
//*
//*****************************************************************************
static const le16_t *utf16_to_wchar(const le16_t *input,unsigned *wc,size_t insize)
{

	if((le16_to_cpu(input[0]) & 0xFC00) == 0xD800)
		{
		if(insize < 2 || (le16_to_cpu(input[1]) & 0xFC00) != 0xDC00)
			{
			return NULL;
			}

		*wc 	= ((wchar_t) (le16_to_cpu(input[0]) & 0x03FF) << 10);
		*wc    |= (le16_to_cpu(input[1]) & 0x03FF);
		*wc    +=  0x00010000;

		return input + 2;
		}
	
	*wc = le16_to_cpu(*input);
	

return input + 1;
}

//*****************************************************************************
//*
//*		utf16_to_utf8
//*
//*****************************************************************************
EXFAT_STATIC int utf16_to_utf8(char *output,const le16_t *input,size_t outsize,size_t insize)
{
const le16_t   *inp = input;
char		   *outp = output;
unsigned 		wc;


	while((size_t)(inp-input)<insize && le16_to_cpu(*inp))
		{
		   inp = utf16_to_wchar(inp, &wc, insize - (inp - input));
		if(inp == NULL)
			{
			exfat_error0("illegal UTF-16 sequence");
			return -EILSEQ;
			}

		   outp = wchar_to_utf8(outp, wc, outsize - (outp - output));
		if(outp == NULL)
			{
			exfat_error0("name is too long");
			return -ENAMETOOLONG;
			}
		}

	*outp = '\0';

return 0;
}

//*****************************************************************************
//*
//*		utf8_to_wchar
//*
//*****************************************************************************
static const char *utf8_to_wchar(const char *input,wchar_t *wc,size_t insize)
{
	if((input[0] & 0x80) == 0 && insize >= 1)
		{
		*wc = (wchar_t) input[0];
		return input + 1;
		}

	if((input[0] & 0xE0) == 0xC0 && insize >= 2)
		{
		*wc = (((wchar_t) input[0] & 0x1F) << 6) |
		((wchar_t) input[1] & 0x3F);
		return input + 2;
		}

	if((input[0] & 0xF0) == 0xE0 && insize >= 3)
		{
		*wc = (((wchar_t) input[0] & 0x0F) << 12) |
		(((wchar_t) input[1] & 0x3F) << 6) |
		((wchar_t) input[2] & 0x3F);
		return input + 3;
		}

	if((input[0] & 0xF8) == 0xF0 && insize >= 4)
		{
		*wc = (((wchar_t) input[0] & 0x07) << 18) |
		(((wchar_t) input[1] & 0x3F) << 12) |
		(((wchar_t) input[2] & 0x3F) << 6) |
		((wchar_t) input[3] & 0x3F);
		return input + 4;
		}

	if((input[0] & 0xFC) == 0xF8 && insize >= 5)
		{
		*wc = (((wchar_t) input[0] & 0x03) << 24) |
		(((wchar_t) input[1] & 0x3F) << 18) |
		(((wchar_t) input[2] & 0x3F) << 12) |
		(((wchar_t) input[3] & 0x3F) << 6) |
		((wchar_t) input[4] & 0x3F);
		return input + 5;
		}

	if((input[0] & 0xFE) == 0xFC && insize >= 6)
		{
		*wc = (((wchar_t) input[0] & 0x01) << 30) |
		(((wchar_t) input[1] & 0x3F) << 24) |
		(((wchar_t) input[2] & 0x3F) << 18) |
		(((wchar_t) input[3] & 0x3F) << 12) |
		(((wchar_t) input[4] & 0x3F) << 6) |
		((wchar_t) input[5] & 0x3F);
		return input + 6;
		}

return NULL;
}

//*****************************************************************************
//*
//*		wchar_to_utf16
//*
//*****************************************************************************
static le16_t *wchar_to_utf16(le16_t *output,unsigned wc,size_t outsize)
{

	if(wc <= 0xFFFF)											// if character is from BMP 
		{
		if(outsize == 0)
			{
			return NULL;
			}

		output[0] = cpu_to_le16((uint16_t)wc);

		return output + 1;
		}

	if(outsize < 2)
		{
		return NULL;
		}

	wc		   -= 0x00010000;
	output[0]	= cpu_to_le16((uint16_t)(0xD800 | ((wc >> 10) & 0x03FF)));
	output[1]	= cpu_to_le16((uint16_t)(0xDC00 | ((wc      ) & 0x03FF)));


return output + 2;
}

//*****************************************************************************
//*
//*		utf8_to_utf16
//*
//*****************************************************************************
EXFAT_STATIC int utf8_to_utf16(le16_t *output,const char *input,size_t outsize,size_t insize)
{
le16_t		   *outp = output;
const char	   *inp  = input;
wchar_t			wc;

	
	while((size_t)(inp-input)<insize && *inp)
		{
		   inp = utf8_to_wchar(inp, &wc, insize - (inp - input));
		if(inp == NULL)
			{
			exfat_error0("illegal UTF-8 sequence");
			return -EILSEQ;
			}

		   outp = wchar_to_utf16(outp, wc, outsize - (outp - output));
		if(outp == NULL)
			{
			exfat_error0("name is too long");
			return -ENAMETOOLONG;
			}
		}

	*outp = cpu_to_le16(0);

return 0;
}

//*****************************************************************************
//*
//*		utf16_length
//*
//*****************************************************************************
EXFAT_STATIC size_t utf16_length(const le16_t *str)
{
size_t		i = 0;

	while(le16_to_cpu(str[i]))
		{
		i++;
		}

return i;
}
