//*****************************************************************************
//*
//*
//*		ext4.h
//*
//*
//*****************************************************************************
//	Copyright (c) 2013 Grzegorz Kostka (kostka.grzegorz@gmail.com)
//	All rights reserved.
//	
//	Redistribution and use in source and binary forms, with or without
//	modification, are permitted provided that the following conditions
//	are met:
//	
//	- Redistributions of source code must retain the above copyright
//	  notice, this list of conditions and the following disclaimer.
//	- Redistributions in binary form must reproduce the above copyright
//	  notice, this list of conditions and the following disclaimer in the
//	  documentation and/or other materials provided with the distribution.
//	- The name of the author may not be used to endorse or promote products
//	  derived from this software without specific prior written permission.
//	
//	THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
//	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
//	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
//	IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
//	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
//	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
//	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//	* @addtogroup lwext4
//	@{
//	*
//	@file  ext4.h
//	@brief Ext4 high level operations (files, directories, mount points...).
//	       Client has to include only this file.

#ifndef		EXT4_H_ 					
#define		EXT4_H_ 					

#include	"ext4_config.h"
#include	"ext4_compiler.h"
#include	"ext4_types.h"
#include	"ext4_errno.h"
#include	"ext4_oflags.h"
#include	"ext4_debug.h"
#include	"ext4_blockdev.h"

#include	<stddef.h>

#ifdef		__cplusplus 				
extern "C" 	{
#endif



//*******************************OS LOCK INFERFACE**************************


struct ext4_lock					//*@brief	OS dependent lock interface.
	{
	void (*lock  )(void);			//*@brief	Lock access to mount point
	void (*unlock)(void);			//*@brief	Unlock access to mount point
	};

//*******************************FILE DESCRIPTOR****************************

typedef struct ext4_file			//*@brief	File descriptor
	{
	struct ext4_mountpoint *mp;		//*@brief	Mount point handle.
	uint32_t		inode;			//*@brief	File inode id	
	uint32_t		flags;			//*@brief	Open flags.
	uint64_t		fsize;			//*@brief	File size.
	uint64_t		fpos;			//*@brief	File position
	#if CONFIG_EXT4_HAS_NAMEBUF
	char			nbuf[256];		//*@brief	name buffer
	uint8_t			nlen;			//*@brief	size of name
	#endif 	
	#if CONFIG_EXT4_HAS_FTYPE
	uint8_t			ftype;			//*@brief	file type
	#endif 	
	#if CONFIG_EXT4_SET_TIMES
	uint8_t			ftset;			//*@brief	Set file times
	uint32_t		ctime;			//*@brief	Create time	(bit0=modify bit1=create bit2=access)
	#endif 
	}ext4_file; 

//****************************DIRECTORY DESCRIPTOR**************************


typedef struct ext4_direntry		//*@brief	Directory entry descriptor. Copy from ext4_types.h	
	{
	uint32_t		inode;
	uint16_t		entry_length;
	uint8_t			name_length;
	uint8_t			inode_type;
	uint8_t			name[256];
	}ext4_direntry; 

typedef struct ext4_dir 
	{
	ext4_file		f;				//*@brief	File descriptor
	ext4_direntry	de;				//*@brief	Current directory entry.
	uint64_t		next_off;		//*@brief	Next entry offset
	}ext4_dir; 

struct ext4_mount_stats				//*@brief	Some of the filesystem stats.
	{
	uint32_t		inodes_count;
	uint32_t		free_inodes_count;
	uint64_t		blocks_count;
	uint64_t		free_blocks_count;
	uint32_t		block_size;
	uint32_t 		block_group_count;
	uint32_t 		blocks_per_group;
	uint32_t 		inodes_per_group;
	char			volume_name[16];
	};

struct ext4_attributes
	{
	uint64_t		fsize;
	uint32_t		ctime;
	uint32_t		atime;
	uint32_t		mtime;
	uint32_t		fmode;			//*@brief	are the attributes
	uint32_t		fmask;			//*@brief	mask for set attributes 
	#if CONFIG_EXT4_HAS_NAMEBUF
	char			nbuf[256];		//*@brief	name buffer
	uint8_t			nlen;			//*@brief	size of name
	#endif 	
	};

#if CONFIG_EXT4_MOUNTPOINTS_COUNT>0
#define 	EX4_HANDLE		
#define 	EX4_PARAM(a)		a
#define 	EX4_PARAM_EX(a)		a
#define 	EX4_HANDLE_PUSH	
#else 	
#define 	EX4_HANDLE			void  *ex4_hanldle,
#define 	EX4_PARAM(a)		void  *ex4_hanldle
#define 	EX4_PARAM_EX(a)		void **ex4_hanldle
#define 	EX4_HANDLE_PUSH		       ex4_hanldle,	
#endif 	

#if CONFIG_EXT4_BLOCKDEVS_COUNT>0
#define 	EX4_DEVICE(a)		a
#else 	
#define 	EX4_DEVICE(a)		struct ext4_blockdev *dev_block
#endif 	


//*****************************************************************************
//*
//*		ext4_device_register
//*
//*****************************************************************************
//	*@brief	Register a block device to a name.
//	         @warning Block device has to be filled by
//	         Block cache may by created automatically when bc parameter is NULL.
//	@param   bd block device
//	@param   bd block device cache
//	@param   dev_name register name
#if CONFIG_EXT4_BLOCKDEVS_COUNT>0
CONFIG_EXT4_STATIC int ext4_device_register(struct ext4_blockdev *bd, struct ext4_bcache *bc,const char *dev_name);
#endif 	

//*****************************************************************************
//*
//*		ext4_mount
//*
//*****************************************************************************
//	*@brief	Mount a block device with EXT4 partition to the mount point.
//	@param   dev_name block device name (@ref ext4_device_register)
//	@param   dev_block pointer to the block device
//	@param   mount_point mount point, for example
//	@param   ex4_hanldle handle of the ext4 data
//	         -   /
//	         -   /my_partition/
//	         -   /my_second_partition/
//	@param   read_only mount as read-only mode.
CONFIG_EXT4_STATIC int ext4_mount(EX4_DEVICE(const char *dev_name),EX4_PARAM_EX(const char *mount_point),int read_only);

//*****************************************************************************
//*
//*		ext4_umount
//*
//*****************************************************************************
//	*@brief	Umount operation.
//	@param   mount_point mount name
CONFIG_EXT4_STATIC int ext4_umount(EX4_PARAM(const char *mount_point));

//*****************************************************************************
//*
//*		ext4_journal_start
//*
//*****************************************************************************
//	*@brief	Start journaling. Journaling start/stop functions are transparent
//	         and might be used on filesystems without journaling support.
//	@warning Usage:
//	             ext4_mount("sda1", "/");
//	             ext4_journal_start("/");
//	
//	             //File operations here...
//	
//	             ext4_journal_stop("/");
//	             ext4_umount("/");
//	@param   mount_point mount name
CONFIG_EXT4_STATIC int ext4_journal_start(EX4_PARAM(const char *mount_point));

//*****************************************************************************
//*
//*		ext4_journal_stop
//*
//*****************************************************************************
//	*@brief	Stop journaling. Journaling start/stop functions are transparent
//	         and might be used on filesystems without journaling support.
//	@param   mount_point mount name
CONFIG_EXT4_STATIC int ext4_journal_stop(EX4_PARAM(const char *mount_point));


//*****************************************************************************
//*
//*		ext4_recover
//*
//*****************************************************************************
//	*@brief	Journal recovery.
//	@param   mount_point mount point
//	@warning Must be called after @ref ext4_mount
CONFIG_EXT4_STATIC int ext4_recover(EX4_PARAM(const char *mount_point));

//*****************************************************************************
//*
//*		ext4_mount_point_stats
//*
//*****************************************************************************
//	*@brief	Get file system params.
//	@param   mount_point mount path
//	@param   stats ext fs stats
CONFIG_EXT4_STATIC int ext4_mount_point_stats(EX4_PARAM(const char *mount_point),struct ext4_mount_stats *stats);

//*****************************************************************************
//*
//*		ext4_mount_setup_locks
//*
//*****************************************************************************
//	*@brief	Setup OS lock routines.
//	@param   mount_point mount path
//	@param   locks - lock and unlock functions
CONFIG_EXT4_STATIC int ext4_mount_setup_locks(EX4_PARAM(const char *mount_point),const struct ext4_lock *locks);

//*****************************************************************************
//*
//*		ext4_get_sblock
//*
//*****************************************************************************
//	*@brief	Acquire the filesystem superblock pointer of a mp.
//	@param   mount_point mount path
//	@param   superblock pointer
CONFIG_EXT4_STATIC int ext4_get_sblock(EX4_PARAM(const char *mount_point), struct ext4_sblock **sb);

//*****************************************************************************
//*
//*		ext4_cache_write_back		
//*
//*****************************************************************************
//	*@brief	Enable/disable write back cache mode.
//	@warning Default model of cache is write trough. It means that when You do:
//	
//	         ext4_fopen(...);
//	         ext4_fwrie(...);
//	                          < --- data is flushed to physical drive
//	
//	         When you do:
//	         ext4_cache_write_back(..., 1);
//	         ext4_fopen(...);
//	         ext4_fwrie(...);
//	                          < --- data is NOT flushed to physical drive
//	         ext4_cache_write_back(..., 0);
//	                          < --- when write back mode is disabled all
//	                                cache data will be flushed
//	To enable write back mode permanently just call this function
//	once after ext4_mount (and disable before ext4_umount).
//	
//	Some of the function use write back cache mode internally.
//	If you enable write back mode twice you have to disable it twice
//	to flush all data:
//	
//	     ext4_cache_write_back(..., 1);
//	     ext4_cache_write_back(..., 1);
//	
//	     ext4_cache_write_back(..., 0);
//	     ext4_cache_write_back(..., 0);
//	
//	Write back mode is useful when you want to create a lot of empty
//	files/directories.
//	
//	@param   mount_point  mount point path
//	@param   on			  enable/disable
CONFIG_EXT4_STATIC int ext4_cache_write_back(EX4_PARAM(const char *mount_point), int on);


//*****************************************************************************
//*
//*		ext4_fremove
//*
//*****************************************************************************
//	*@brief	Remove file by path.
//	@param   path path to file
CONFIG_EXT4_STATIC int ext4_fremove(EX4_HANDLE const char *path);


//*****************************************************************************
//*
//*		ext4_flink
//*
//*****************************************************************************
//	*@brief	create a hardlink for a file.
//	@param   path path to file
//	@param   hardlink_path path of hardlink
CONFIG_EXT4_STATIC int ext4_flink(EX4_HANDLE const char *path, const char *hardlink_path);

//*****************************************************************************
//*
//*		ext4_frename
//*
//*****************************************************************************
//	*@brief Rename file
//	@param path source
//	@param new_path destination
CONFIG_EXT4_STATIC int ext4_frename(EX4_HANDLE const char *path, const char *new_path);

//*****************************************************************************
//*
//*		ext4_fopen
//*
//*****************************************************************************
//	*@brief	File open function.
//	@param   path filename (has to start from mount point)
//	         /my_partition/my_file
//	@param   flags open file flags
//	|---------------------------------------------------------------|
//	|   r or rb                 O_RDONLY                            |
//	|---------------------------------------------------------------|
//	|   w or wb                 O_WRONLY|O_CREAT|O_TRUNC            |
//	|---------------------------------------------------------------|
//	|   a or ab                 O_WRONLY|O_CREAT|O_APPEND           |
//	|---------------------------------------------------------------|
//	|   r+ or rb+ or r+b        O_RDWR                              |
//	|---------------------------------------------------------------|
//	|   w+ or wb+ or w+b        O_RDWR|O_CREAT|O_TRUNC              |
//	|---------------------------------------------------------------|
//	|   a+ or ab+ or a+b        O_RDWR|O_CREAT|O_APPEND             |
//	|---------------------------------------------------------------|
//	
CONFIG_EXT4_STATIC int ext4_fopen(EX4_HANDLE ext4_file *f, const char *path, const char *flags);

//*****************************************************************************
//*
//*		ext4_fopen2
//*
//*****************************************************************************
//	*@brief	Alternate file open function.
//	@param   filename, (has to start from mount point)
//	         /my_partition/my_file
//	@param   flags open file flags
CONFIG_EXT4_STATIC int ext4_fopen2(EX4_HANDLE ext4_file *f, const char *path, int flags);

//*****************************************************************************
//*
//*		ext4_fopen_ex
//*
//*****************************************************************************
//	*@brief	file/directory open function.
//	@param   filename, (has to start from mount point)
//	         /my_partition/my_file
CONFIG_EXT4_STATIC int ext4_fopen_ex(EX4_HANDLE ext4_file *f,const char *path);

//*****************************************************************************
//*
//*		ext4_fclose
//*
//*****************************************************************************
//	*@brief	File close function.
//	@param   f file handle
CONFIG_EXT4_STATIC int ext4_fclose(ext4_file *f);

//*****************************************************************************
//*
//*		ext4_fill_raw_inode
//*
//*****************************************************************************
//	*@brief	Fill in the ext4_inode buffer.
//	@param   path fetch inode data of the path
//	@param   ret_ino the inode id of the path
//	@param   ext4_inode buffer
CONFIG_EXT4_STATIC int ext4_fill_raw_inode(EX4_HANDLE const char *path, uint32_t *ret_ino,struct ext4_inode *inode);

//*****************************************************************************
//*
//*		ext4_ftruncate
//*
//*****************************************************************************
//	*@brief	File truncate function.
//	@param   f file handle
//	@param   new file size
CONFIG_EXT4_STATIC int ext4_ftruncate(ext4_file *f, uint64_t size);

//*****************************************************************************
//*
//*		ext4_fread
//*
//*****************************************************************************
//	*@brief	Read data from file.
//	@param   f file handle
//	@param   buf output buffer
//	@param   size bytes to read
//	@param   rcnt bytes read (may be NULL)
CONFIG_EXT4_STATIC int ext4_fread(ext4_file *f, void *buf, size_t size, size_t *rcnt);

//*****************************************************************************
//*
//*		ext4_fwrite
//*
//*****************************************************************************
//	*@brief	Write data to file.
//	@param   f file handle
//	@param   buf data to write
//	@param   size write length
//	@param   wcnt bytes written (may be NULL)
CONFIG_EXT4_STATIC int ext4_fwrite(ext4_file *f, const void *buf, size_t size, size_t *wcnt);

//*****************************************************************************
//*
//*		ext4_fseek
//*
//*****************************************************************************
//	*@brief	File seek operation.
//	@param   f file handle
//	@param   offset offset to seek
//	@param   origin seek type:
//	             @ref SEEK_SET
//	             @ref SEEK_CUR
//	             @ref SEEK_END
CONFIG_EXT4_STATIC int ext4_fseek(ext4_file *f, uint64_t offset, uint32_t origin);

//*****************************************************************************
//*
//*		ext4_ftell
//*
//*****************************************************************************
//	*@brief	Get file position.
//	@param   f file handle
CONFIG_EXT4_STATIC uint64_t ext4_ftell(ext4_file *f);

//*****************************************************************************
//*
//*		ext4_fsize
//*
//*****************************************************************************
//	*@brief	Get file size.
//	@param   f file handle
CONFIG_EXT4_STATIC uint64_t ext4_fsize(ext4_file *f);

//*****************************************************************************
//*
//*		ext4_file_get_attributes
//*
//*****************************************************************************
//	*@brief	Get several file attributes.
//	@param   f file handle
//	@param   attrib pointer to attribute buffer
CONFIG_EXT4_STATIC int ext4_file_get_attributes(ext4_file *f,struct ext4_attributes *attrib);

//*****************************************************************************
//*
//*		ext4_file_set_attributes
//*
//*****************************************************************************
//	*@brief	Set several file attributes.
//	@param   f file handle
//	@param   attrib pointer to attribute buffer
CONFIG_EXT4_STATIC int ext4_file_set_attributes(ext4_file *f,struct ext4_attributes *attrib);

//*****************************************************************************
//*
//*		ext4_chmod
//*
//*****************************************************************************
//	*@brief Change file/directory/link mode bits
//	@param path to file/dir/link
//	@param mode new mode bits (for example 0777)
CONFIG_EXT4_STATIC int ext4_chmod(EX4_HANDLE const char *path, uint32_t mode);

//*****************************************************************************
//*
//*		ext4_chown
//*
//*****************************************************************************
//	*@brief Change file owner and group
//	@param path to file/dir/link
//	@param uid user id
//	@param gid group id
CONFIG_EXT4_STATIC int ext4_chown(EX4_HANDLE const char *path, uint32_t uid, uint32_t gid);

//*****************************************************************************
//*
//*		ext4_file_set_atime
//*
//*****************************************************************************
//	*@brief Set file access time
//	@param path to file/dir/link
//	@param atime access timestamp
CONFIG_EXT4_STATIC int ext4_file_set_atime(EX4_HANDLE const char *path, uint32_t atime);

//*****************************************************************************
//*
//*		ext4_file_set_mtime
//*
//*****************************************************************************
//	*@brief Set file modify time
//	@param path to file/dir/link
//	@param mtime modify timestamp
CONFIG_EXT4_STATIC int ext4_file_set_mtime(EX4_HANDLE const char *path, uint32_t mtime);

//*****************************************************************************
//*
//*		ext4_file_set_ctime
//*
//*****************************************************************************
//	*@brief Set file change time
//	@param path to file/dir/link
//	@param ctime change timestamp
CONFIG_EXT4_STATIC int ext4_file_set_ctime(EX4_HANDLE const char *path, uint32_t ctime);

//*****************************************************************************
//*
//*		ext4_fsymlink
//*
//*****************************************************************************
//	*@brief Create symbolic link
//	@param target destination path
//	@param path source entry
CONFIG_EXT4_STATIC int ext4_fsymlink(EX4_HANDLE const char *target, const char *path);


//*****************************************************************************
//*
//*		ext4_readlink
//*
//*****************************************************************************
//	*@brief Read symbolic link payload
//	@param path to symlink
//	@param buf output buffer
//	@param bufsize output buffer max size
//	@param rcnt bytes read
CONFIG_EXT4_STATIC int ext4_readlink(EX4_HANDLE const char *path, char *buf, size_t bufsize, size_t *rcnt);

//*****************************************************************************
//*
//*		ext4_readlink
//*
//*****************************************************************************
//	*@brief Read symbolic link payload
//	@param path to symlink
//	@param buf output buffer
//	@param bufsize output buffer max size
//	@param rcnt bytes read
//	@param attrib receive attributes of file if no link
#if CONFIG_EXT4_HAS_FTYPE
CONFIG_EXT4_STATIC int ext4_readlink_ex(EX4_HANDLE const char *path, char *buf, size_t bufsize, size_t *rcnt,ext4_attributes *attrib);
#endif 	

//*****************************************************************************
//*
//*		ext4_setxattr
//*
//*****************************************************************************
//	*@brief Set extended attribute
//	@param path to file/directory
//	@param name name of the entry to add
//	@param name_len length of @name in bytes
//	@param data data of the entry to add
//	@param data_size size of data to add
//	@param replace whether existing entries should be replaced
CONFIG_EXT4_STATIC int ext4_setxattr(EX4_HANDLE const char *path, const char *name, size_t name_len,const void *data, size_t data_size, int replace);

//*****************************************************************************
//*
//*		ext4_getxattr
//*
//*****************************************************************************
//	*@brief Get extended attribute
//	@param path to file/directory
//	@param name name of the entry to get
//	@param name_len length of @name in bytes
//	@param data data of the entry to get
//	@param data_size size of data to get
CONFIG_EXT4_STATIC int ext4_getxattr(EX4_HANDLE const char *path, const char *name, size_t name_len,void *buf, size_t buf_size, size_t *data_size);

//*****************************************************************************
//*
//*		ext4_listxattr
//*
//*****************************************************************************
//	*@brief List extended attributes
//	@param path to file/directory
//	@param list list to hold the name of entries
//	@param size size of @list in bytes
//	@param ret_size used bytes of @list
CONFIG_EXT4_STATIC int ext4_listxattr(EX4_HANDLE const char *path, char *list, size_t size, size_t *ret_size);

//*****************************************************************************
//*
//*		ext4_removexattr
//*
//*****************************************************************************
//	*@brief Remove extended attribute
//	@param path to file/directory
//	@param name name of the entry to remove
//	@param name_len length of @name in bytes
CONFIG_EXT4_STATIC int ext4_removexattr(EX4_HANDLE const char *path, const char *name, size_t name_len);

//*****************************************************************************
//*
//*		ext4_dir_rm
//*
//*****************************************************************************
//	*@brief	Recursive directory remove.
//	@param   path directory path to remove
CONFIG_EXT4_STATIC int ext4_dir_rm(EX4_HANDLE const char *path,int recursive);

//*****************************************************************************
//*
//*		ext4_dir_mk
//*
//*****************************************************************************
//	*@brief	Create new directory.
//	@param   name new directory name
CONFIG_EXT4_STATIC int ext4_dir_mk(EX4_HANDLE const char *path);

//*****************************************************************************
//*
//*		ext4_dir_open
//*
//*****************************************************************************
//	*@brief	Directory open.
//	@param   d directory handle
//	@param   path directory path
CONFIG_EXT4_STATIC int ext4_dir_open(EX4_HANDLE ext4_dir *d, const char *path);

//*****************************************************************************
//*
//*		ext4_dir_close
//*
//*****************************************************************************
//	*@brief	Directory close.
//	@param   d directory handle
CONFIG_EXT4_STATIC int ext4_dir_close(ext4_dir *d);

//*****************************************************************************
//*
//*		ext4_dir_entry_next
//*
//*****************************************************************************
//	*@brief	Return next directory entry.
//	@param   d directory handle
//	@param   id entry id
CONFIG_EXT4_STATIC const ext4_direntry *ext4_dir_entry_next(ext4_dir *d);

//*****************************************************************************
//*
//*		ext4_dir_entry_rewind
//*
//*****************************************************************************
//	*@brief	Rewine directory entry offset.
CONFIG_EXT4_STATIC void ext4_dir_entry_rewind(ext4_dir *d);

//*****************************************************************************
//*
//*		ext4_dir_get_attributes
//*
//*****************************************************************************
//	*@brief	decodes a directory entry
//	@param   d directory handle
//	@param   attrip pointer to info buffer
CONFIG_EXT4_STATIC int ext4_dir_get_attributes(ext4_dir *d,struct ext4_attributes *attrib);

#ifdef		__cplusplus 				
}
#endif

#endif // EXT4_H_ 

//	*
//	@}
