//*****************************************************************************
//*
//*
//*		ext4_blockdev.c
//*
//*
//*****************************************************************************
//	Copyright (c) 2013 Grzegorz Kostka (kostka.grzegorz@gmail.com)
//	All rights reserved.
//
//	Redistribution and use in source and binary forms, with or without
//	modification, are permitted provided that the following conditions
//	are met:
//
//	- Redistributions of source code must retain the above copyright
//	  notice, this list of conditions and the following disclaimer.
//	- Redistributions in binary form must reproduce the above copyright
//	  notice, this list of conditions and the following disclaimer in the
//	  documentation and/or other materials provided with the distribution.
//	- The name of the author may not be used to endorse or promote products
//	  derived from this software without specific prior written permission.
//
//	THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
//	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
//	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
//	IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
//	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
//	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
//	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//	* @addtogroup lwext4
//	@{
//	*
//	@file  ext4_blockdev.c
//	@brief Block device module.

#include	"ext4_config.h"
#include	"ext4_compiler.h"
#include	"ext4_types.h"
#include	"ext4_misc.h"
#include	"ext4_errno.h"
#include	"ext4_debug.h"

#include	"ext4_blockdev.h"

#include	<string.h>
#include	<stdlib.h>

//*****************************************************************************
//*
//*		ext4_bdif_lock
//*
//*****************************************************************************
static void ext4_bdif_lock(struct ext4_blockdev *bdev)
{
int			r;

	if(!bdev->bdif->lock)
		{
		return;
		}

	r = bdev->bdif->lock(bdev);
	ext4_assert(r == EOK);
}

//*****************************************************************************
//*
//*		ext4_bdif_unlock
//*
//*****************************************************************************
static void ext4_bdif_unlock(struct ext4_blockdev *bdev)
{
int			r;


	if(!bdev->bdif->unlock)
		{
		return;
		}

	r = bdev->bdif->unlock(bdev);
	ext4_assert(r == EOK);
}

//*****************************************************************************
//*
//*		ext4_bdif_bread
//*
//*****************************************************************************
static int ext4_bdif_bread(struct ext4_blockdev *bdev,void *buf,uint64_t blk_id,uint32_t blk_cnt)
{
int			r;


	ext4_bdif_lock(bdev);
	r = bdev->bdif->bread(bdev, buf, blk_id, blk_cnt);
	bdev->bdif->bread_ctr++;
	ext4_bdif_unlock(bdev);

return r;
}

//*****************************************************************************
//*
//*		ext4_bdif_bwrite
//*
//*****************************************************************************
static int ext4_bdif_bwrite(struct ext4_blockdev *bdev,const void *buf,uint64_t blk_id,uint32_t blk_cnt)
{
int			r;


	ext4_bdif_lock(bdev);
	r = bdev->bdif->bwrite(bdev, buf, blk_id, blk_cnt);
	bdev->bdif->bwrite_ctr++;
	ext4_bdif_unlock(bdev);

return r;
}

//*****************************************************************************
//*
//*		ext4_block_init
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_init(struct ext4_blockdev *bdev)
{
int			rc;


	ext4_assert(bdev);
	ext4_assert(bdev->bdif);
	ext4_assert(bdev->bdif->open && bdev->bdif->close && bdev->bdif->bread && bdev->bdif->bwrite);

	if(bdev->bdif->ph_refctr)
		{
		bdev->bdif->ph_refctr++;
		return EOK;
		}

	   rc = bdev->bdif->open(bdev);							// Low level block init
	if(rc != EOK)
		{
		return rc;
		}

	bdev->bdif->ph_refctr = 1;

return EOK;
}

//*****************************************************************************
//*
//*		ext4_block_bind_bcache
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_bind_bcache(struct ext4_blockdev *bdev,struct ext4_bcache *bc)
{
	ext4_assert(bdev && bc);
	bdev->bc = bc;
	bc->bdev = bdev;

return EOK;
}

//*****************************************************************************
//*
//*		ext4_block_set_lb_size
//*
//*****************************************************************************
CONFIG_EXT4_STATIC void ext4_block_set_lb_size(struct ext4_blockdev *bdev,uint32_t lb_bsize)
{
	
	ext4_assert(!(lb_bsize % bdev->bdif->ph_bsize));		// Logical block size has to be multiply of physical

	bdev->lg_bsize	= lb_bsize;
	bdev->lg_bcnt	= bdev->part_size / lb_bsize;
}

//*****************************************************************************
//*
//*		ext4_block_fini
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_fini(struct ext4_blockdev *bdev)
{
	ext4_assert(bdev);

	if(!bdev->bdif->ph_refctr)
		{
		return EOK;
		}

	   bdev->bdif->ph_refctr--;
	if(bdev->bdif->ph_refctr)
		{
		return EOK;
		}

	//Low level block fini

return bdev->bdif->close(bdev);
}

//*****************************************************************************
//*
//*		ext4_block_flush_buf
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_flush_buf(struct ext4_blockdev *bdev,struct ext4_buf *buf)
{
int			r;
struct ext4_bcache *bc = bdev->bc;

	if(ext4_bcache_test_flag(buf, BC_DIRTY) &&
	   ext4_bcache_test_flag(buf, BC_UPTODATE))
		{
		   r = ext4_blocks_set_direct(bdev, buf->data, buf->lba, 1);
		if(r)
			{
			if(buf->end_write)
				{
				bc->dont_shake = true;
				buf->end_write(bc, buf, r, buf->end_write_arg);
				bc->dont_shake = false;
				}

			return r;
			}

		ext4_bcache_remove_dirty_node(bc, buf);
		ext4_bcache_clear_flag(buf, BC_DIRTY);
		if(buf->end_write)
			{
			bc->dont_shake = true;
			buf->end_write(bc, buf, r, buf->end_write_arg);
			bc->dont_shake = false;
			}
		}

return EOK;
}

//*****************************************************************************
//*
//*		ext4_block_flush_lba
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_flush_lba(struct ext4_blockdev *bdev,uint64_t lba)
{
int					r = EOK;
struct ext4_buf	   *buf;
struct ext4_block	b;


	   buf = ext4_bcache_find_get(bdev->bc, &b, lba);
	if(buf)
		{
		r = ext4_block_flush_buf(bdev, buf);
		ext4_bcache_free(bdev->bc, &b);
		}

return r;
}

//*****************************************************************************
//*
//*		ext4_block_cache_shake
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_cache_shake(struct ext4_blockdev *bdev)
{
int					r = EOK;
struct ext4_buf	   *buf;


	if(bdev->bc->dont_shake)
		{
		return EOK;
		}

	while(!RB_EMPTY(&bdev->bc->lru_root) &&
	      ext4_bcache_is_full(bdev->bc))
		{

		buf = ext4_buf_lowest_lru(bdev->bc);
		ext4_assert(buf);
		if(ext4_bcache_test_flag(buf, BC_DIRTY))
			{
			   r = ext4_block_flush_buf(bdev, buf);
			if(r != EOK)
				{
				break;
				}
			}

		ext4_bcache_drop_buf(bdev->bc, buf);
		}

return r;
}

//*****************************************************************************
//*
//*		ext4_block_get_noread
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_get_noread(struct ext4_blockdev *bdev,struct ext4_block *b,uint64_t lba)
{
int		is_new;
int		r;


	ext4_assert(bdev && b);

	if(!bdev->bdif->ph_refctr)
		{
		return EIO;
		}

	if(!(lba < bdev->lg_bcnt))
		{
		return ERANGE;
		}

	b->lb_id = lba;
	
	   r = ext4_block_cache_shake(bdev);						// If cache is full we have to (flush and) drop it anyway :(
	if(r != EOK)
		{
		return r;
		}

	   r = ext4_bcache_alloc(bdev->bc, b, &is_new);
	if(r != EOK)
		{
		return r;
		}

	if(!b->data)
		{
		return ENOMEM;
		}

return EOK;
}

//*****************************************************************************
//*
//*		ext4_block_get
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_get(struct ext4_blockdev *bdev,struct ext4_block *b,uint64_t lba)
{
int			r = ext4_block_get_noread(bdev, b, lba);


	if(r != EOK)
		{
		return r;
		}

	if(ext4_bcache_test_flag(b->buf, BC_UPTODATE))
		{
		//	Data in the cache is up-to-date.

		return EOK;
		}

	   r = ext4_blocks_get_direct(bdev, b->data, lba, 1);
	if(r != EOK)
		{
		ext4_bcache_free(bdev->bc, b);
		b->lb_id = 0;
		return r;
		}

	//	Mark buffer up-to-date, since

	ext4_bcache_set_flag(b->buf, BC_UPTODATE);

return EOK;
}

//*****************************************************************************
//*
//*		ext4_block_set
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_set(struct ext4_blockdev *bdev,struct ext4_block *b)
{
	ext4_assert(bdev && b);
	ext4_assert(b->buf);

	if(!bdev->bdif->ph_refctr)
		{
		return EIO;
		}

return ext4_bcache_free(bdev->bc, b);
}

//*****************************************************************************
//*
//*		ext4_blocks_get_direct
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_blocks_get_direct(struct ext4_blockdev *bdev,void *buf,uint64_t lba,uint32_t cnt)
{
uint64_t	pba;
uint32_t	pb_cnt;

	ext4_assert(bdev && buf);

	pba 	= (lba * bdev->lg_bsize + bdev->part_offset) / bdev->bdif->ph_bsize;
	pb_cnt	=  bdev->lg_bsize / bdev->bdif->ph_bsize;

return ext4_bdif_bread(bdev, buf, pba, pb_cnt * cnt);
}

//*****************************************************************************
//*
//*		ext4_blocks_set_direct
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_blocks_set_direct(struct ext4_blockdev *bdev,const void *buf,uint64_t lba,uint32_t cnt)
{
uint64_t	pba;
uint32_t	pb_cnt;

	ext4_assert(bdev && buf);

	pba 	= (lba * bdev->lg_bsize + bdev->part_offset) / bdev->bdif->ph_bsize;
	pb_cnt	=  bdev->lg_bsize / bdev->bdif->ph_bsize;

return ext4_bdif_bwrite(bdev, buf, pba, pb_cnt * cnt);
}

//*****************************************************************************
//*
//*		ext4_block_writebytes
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_writebytes(struct ext4_blockdev *bdev,uint64_t off,const void *buf,uint32_t len)
{
uint64_t		block_idx;
uint32_t		unalg;
uint32_t		blen;
int				r = EOK;
const uint8_t  *p = (uint8_t*)buf;


	ext4_assert(bdev && buf);

	if(!bdev->bdif->ph_refctr)
		{
		return EIO;
		}

	if(off + len > bdev->part_size)
		{
		return EINVAL;											// Ups. Out of range operation
		}

	block_idx = ((off + bdev->part_offset) / bdev->bdif->ph_bsize);

																
	unalg = ((uint32_t)off & (bdev->bdif->ph_bsize - 1));		// OK lets deal with the first possible unaligned block

	if(unalg)
		{
		uint32_t	wlen = (bdev->bdif->ph_bsize-unalg > len)? len : bdev->bdif->ph_bsize-unalg;

		   r = ext4_bdif_bread(bdev, bdev->bdif->ph_bbuf, block_idx, 1);
		if(r!= EOK)
			{
			return r;
			}

		memcpy(bdev->bdif->ph_bbuf + unalg, p, wlen);
		
		   r = ext4_bdif_bwrite(bdev, bdev->bdif->ph_bbuf, block_idx, 1);
		if(r!= EOK)
			{
			return r;
			}

		p	   += wlen;
		len    -= wlen;
		block_idx++;
		}

	
	blen	= len / bdev->bdif->ph_bsize;						// Aligned data
	
	   r = ext4_bdif_bwrite(bdev, p, block_idx, blen);
	if(r!= EOK)
		{
		return r;
		}

	p		   += bdev->bdif->ph_bsize * blen;
	len 	   -= bdev->bdif->ph_bsize * blen;

	block_idx  += blen;

	if(len) 													//Rest of the data
		{
		   r = ext4_bdif_bread(bdev, bdev->bdif->ph_bbuf, block_idx, 1);
		if(r!= EOK)
			{
			return r;
			}

		memcpy(bdev->bdif->ph_bbuf, p, len);
		
		   r = ext4_bdif_bwrite(bdev, bdev->bdif->ph_bbuf, block_idx, 1);
		if(r!= EOK)
			{
			return r;
			}
		}

return r;
}

//*****************************************************************************
//*
//*		ext4_block_readbytes
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_readbytes(struct ext4_blockdev *bdev,uint64_t off,void *buf,uint32_t len)
{
uint64_t	block_idx;
uint32_t	blen;
uint32_t	unalg;
uint8_t	   *p = (uint8_t*)buf;
int			r = EOK;


	ext4_assert(bdev && buf);

	if(!bdev->bdif->ph_refctr)
		{
		return EIO;
		}

	if(off + len > bdev->part_size)
		{
		return EINVAL;											// Ups. Out of range operation
		}

	block_idx = ((off + bdev->part_offset) / bdev->bdif->ph_bsize);
	unalg     = ((uint32_t)off & (bdev->bdif->ph_bsize - 1));	// OK lets deal with the first possible unaligned block

	if(unalg)
		{
		uint32_t	rlen;

		rlen = (bdev->bdif->ph_bsize-unalg > len)? len : bdev->bdif->ph_bsize-unalg;

		   r = ext4_bdif_bread(bdev, bdev->bdif->ph_bbuf, block_idx, 1);
		if(r!= EOK)
			{
			return r;
			}

		memcpy(p, bdev->bdif->ph_bbuf + unalg, rlen);

		p	   += rlen;
		len    -= rlen;
		block_idx++;
		}

	
	blen = len / bdev->bdif->ph_bsize;							// Aligned data

	   r = ext4_bdif_bread(bdev, p, block_idx, blen);
	if(r != EOK)
		{
		return r;
		}

	p		   += bdev->bdif->ph_bsize * blen;
	len 	   -= bdev->bdif->ph_bsize * blen;

	block_idx  += blen;

	if(len) 													//Rest of the data
		{
		   r = ext4_bdif_bread(bdev, bdev->bdif->ph_bbuf, block_idx, 1);
		if(r!= EOK)
			{
			return r;
			}

		memcpy(p, bdev->bdif->ph_bbuf, len);
		}


return r;
}

//*****************************************************************************
//*
//*		ext4_block_cache_flush
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_cache_flush(struct ext4_blockdev *bdev)
{

	while(!SLIST_EMPTY(&bdev->bc->dirty_list))
		{
		int					r;
		struct ext4_buf	   *buf = SLIST_FIRST(&bdev->bc->dirty_list);


		ext4_assert(buf);
		r = ext4_block_flush_buf(bdev, buf);
		if(r != EOK)
			{
			return r;
			}
		}

return EOK;
}

//*****************************************************************************
//*
//*		ext4_block_cache_write_back
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_block_cache_write_back(struct ext4_blockdev *bdev,uint8_t on_off)
{

	if(on_off)
		{
		bdev->cache_write_back++;
		}

	if(!on_off && bdev->cache_write_back)
		{
		bdev->cache_write_back--;
		}

	if(bdev->cache_write_back)
		{
		return EOK;
		}

	

return ext4_block_cache_flush(bdev);							// Flush data in all delayed cache blocks
}

//	*
//	@}
