//*****************************************************************************
//*
//*
//*		ext4_crc32.c
//*
//*
//*****************************************************************************
//	Copyright (c) 2014 Grzegorz Kostka (kostka.grzegorz@gmail.com)
//	All rights reserved.
//	
//	Based on FreeBSD.
//	
//	Redistribution and use in source and binary forms, with or without
//	modification, are permitted provided that the following conditions
//	are met:
//	
//	- Redistributions of source code must retain the above copyright
//	  notice, this list of conditions and the following disclaimer.
//	- Redistributions in binary form must reproduce the above copyright
//	  notice, this list of conditions and the following disclaimer in the
//	  documentation and/or other materials provided with the distribution.
//	- The name of the author may not be used to endorse or promote products
//	  derived from this software without specific prior written permission.
//	
//	THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
//	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
//	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
//	IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
//	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
//	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
//	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//	* @addtogroup lwext4
//	@{
//	*
//	@file  ext4_crc32c.c
//	@brief Crc32c routine. Taken from FreeBSD kernel.

#include	"ext4_config.h"
#include	"ext4_compiler.h"
#include	"ext4_types.h"
#include	"ext4_misc.h"
#include	"ext4_errno.h"
#include	"ext4_debug.h"

#include	"ext4_crc32.h"

static const uint32_t crc32_tab[] = 
	{
	0x00000000, 0x77073096, 0xEE0E612C, 0x990951BA, 0x076DC419, 0x706AF48F,
	0xE963A535, 0x9E6495A3, 0x0EDB8832, 0x79DCB8A4, 0xE0D5E91E, 0x97D2D988,
	0x09B64C2B, 0x7EB17CBD, 0xE7B82D07, 0x90BF1D91, 0x1DB71064, 0x6AB020F2,
	0xF3B97148, 0x84BE41DE, 0x1ADAD47D, 0x6DDDE4EB, 0xF4D4B551, 0x83D385C7,
	0x136C9856, 0x646BA8C0, 0xFD62F97A, 0x8A65C9EC, 0x14015C4F, 0x63066CD9,
	0xFA0F3D63, 0x8D080DF5, 0x3B6E20C8, 0x4C69105E, 0xD56041E4, 0xA2677172,
	0x3C03E4D1, 0x4B04D447, 0xD20D85FD, 0xA50AB56B, 0x35B5A8FA, 0x42B2986C,
	0xDBBBC9D6, 0xACBCF940, 0x32D86CE3, 0x45DF5C75, 0xDCD60DCF, 0xABD13D59,
	0x26D930AC, 0x51DE003A, 0xC8D75180, 0xBFD06116, 0x21B4F4B5, 0x56B3C423,
	0xCFBA9599, 0xB8BDA50F, 0x2802B89E, 0x5F058808, 0xC60CD9B2, 0xB10BE924,
	0x2F6F7C87, 0x58684C11, 0xC1611DAB, 0xB6662D3D, 0x76DC4190, 0x01DB7106,
	0x98D220BC, 0xEFD5102A, 0x71B18589, 0x06B6B51F, 0x9FBFE4A5, 0xE8B8D433,
	0x7807C9A2, 0x0F00F934, 0x9609A88E, 0xE10E9818, 0x7F6A0DBB, 0x086D3D2D,
	0x91646C97, 0xE6635C01, 0x6B6B51F4, 0x1C6C6162, 0x856530D8, 0xF262004E,
	0x6C0695ED, 0x1B01A57B, 0x8208F4C1, 0xF50FC457, 0x65B0D9C6, 0x12B7E950,
	0x8BBEB8EA, 0xFCB9887C, 0x62DD1DDF, 0x15DA2D49, 0x8CD37CF3, 0xFBD44C65,
	0x4DB26158, 0x3AB551CE, 0xA3BC0074, 0xD4BB30E2, 0x4ADFA541, 0x3DD895D7,
	0xA4D1C46D, 0xD3D6F4FB, 0x4369E96A, 0x346ED9FC, 0xAD678846, 0xDA60B8D0,
	0x44042D73, 0x33031DE5, 0xAA0A4C5F, 0xDD0D7CC9, 0x5005713C, 0x270241AA,
	0xBE0B1010, 0xC90C2086, 0x5768B525, 0x206F85B3, 0xB966D409, 0xCE61E49F,
	0x5EDEF90E, 0x29D9C998, 0xB0D09822, 0xC7D7A8B4, 0x59B33D17, 0x2EB40D81,
	0xB7BD5C3B, 0xC0BA6CAD, 0xEDB88320, 0x9ABFB3B6, 0x03B6E20C, 0x74B1D29A,
	0xEAD54739, 0x9DD277AF, 0x04DB2615, 0x73DC1683, 0xE3630B12, 0x94643B84,
	0x0D6D6A3E, 0x7A6A5AA8, 0xE40ECF0B, 0x9309FF9D, 0x0A00AE27, 0x7D079EB1,
	0xF00F9344, 0x8708A3D2, 0x1E01F268, 0x6906C2FE, 0xF762575D, 0x806567CB,
	0x196C3671, 0x6E6B06E7, 0xFED41B76, 0x89D32BE0, 0x10DA7A5A, 0x67DD4ACC,
	0xF9B9DF6F, 0x8EBEEFF9, 0x17B7BE43, 0x60B08ED5, 0xD6D6A3E8, 0xA1D1937E,
	0x38D8C2C4, 0x4FDFF252, 0xD1BB67F1, 0xA6BC5767, 0x3FB506DD, 0x48B2364B,
	0xD80D2BDA, 0xAF0A1B4C, 0x36034AF6, 0x41047A60, 0xDF60EFC3, 0xA867DF55,
	0x316E8EEF, 0x4669BE79, 0xCB61B38C, 0xBC66831A, 0x256FD2A0, 0x5268E236,
	0xCC0C7795, 0xBB0B4703, 0x220216B9, 0x5505262F, 0xC5BA3BBE, 0xB2BD0B28,
	0x2BB45A92, 0x5CB36A04, 0xC2D7FFA7, 0xB5D0CF31, 0x2CD99E8B, 0x5BDEAE1D,
	0x9B64C2B0, 0xEC63F226, 0x756AA39C, 0x026D930A, 0x9C0906A9, 0xEB0E363F,
	0x72076785, 0x05005713, 0x95BF4A82, 0xE2B87A14, 0x7BB12BAE, 0x0CB61B38,
	0x92D28E9B, 0xE5D5BE0D, 0x7CDCEFB7, 0x0BDBDF21, 0x86D3D2D4, 0xF1D4E242,
	0x68DDB3F8, 0x1FDA836E, 0x81BE16CD, 0xF6B9265B, 0x6FB077E1, 0x18B74777,
	0x88085AE6, 0xFF0F6A70, 0x66063BCA, 0x11010B5C, 0x8F659EFF, 0xF862AE69,
	0x616BFFD3, 0x166CCF45, 0xA00AE278, 0xD70DD2EE, 0x4E048354, 0x3903B3C2,
	0xA7672661, 0xD06016F7, 0x4969474D, 0x3E6E77DB, 0xAED16A4A, 0xD9D65ADC,
	0x40DF0B66, 0x37D83BF0, 0xA9BCAE53, 0xDEBB9EC5, 0x47B2CF7F, 0x30B5FFE9,
	0xBDBDF21C, 0xCABAC28A, 0x53B39330, 0x24B4A3A6, 0xBAD03605, 0xCDD70693,
	0x54DE5729, 0x23D967BF, 0xB3667A2E, 0xC4614AB8, 0x5D681B02, 0x2A6F2B94,
	0xB40BBE37, 0xC30C8EA1, 0x5A05DF1B, 0x2D02EF8D
	};

// 
// CRC LOOKUP TABLE 
// ================ 
// The following CRC lookup table was generated automagically 
// by the Rocksoft^tm Model CRC Algorithm Table Generation 
// Program V1.0 using the following model parameters: 
// 
// Width : 4 bytes. 
// Poly : 0x1EDC6F41L 
// Reverse : TRUE. 
// 
// For more information on the Rocksoft^tm Model CRC Algorithm, 
// see the document titled "A Painless Guide to CRC Error 
// Detection Algorithms" by Ross Williams 
// (ross@guest.adelaide.edu.au.). This document is likely to be 
// in the FTP archive "ftp.adelaide.edu.au/pub/rocksoft". 
// 
static const uint32_t crc32c_tab[256] = 
	{
	0x00000000l, 0xF26B8303l, 0xE13B70F7l, 0x1350F3F4l, 0xC79A971Fl,
	0x35F1141Cl, 0x26A1E7E8l, 0xD4CA64EBl, 0x8AD958CFl, 0x78B2DBCCl,
	0x6BE22838l, 0x9989AB3Bl, 0x4D43CFD0l, 0xBF284CD3l, 0xAC78BF27l,
	0x5E133C24l, 0x105EC76Fl, 0xE235446Cl, 0xF165B798l, 0x030E349Bl,
	0xD7C45070l, 0x25AFD373l, 0x36FF2087l, 0xC494A384l, 0x9A879FA0l,
	0x68EC1CA3l, 0x7BBCEF57l, 0x89D76C54l, 0x5D1D08BFl, 0xAF768BBCl,
	0xBC267848l, 0x4E4DFB4Bl, 0x20BD8EDEl, 0xD2D60DDDl, 0xC186FE29l,
	0x33ED7D2Al, 0xE72719C1l, 0x154C9AC2l, 0x061C6936l, 0xF477EA35l,
	0xAA64D611l, 0x580F5512l, 0x4B5FA6E6l, 0xB93425E5l, 0x6DFE410El,
	0x9F95C20Dl, 0x8CC531F9l, 0x7EAEB2FAl, 0x30E349B1l, 0xC288CAB2l,
	0xD1D83946l, 0x23B3BA45l, 0xF779DEAEl, 0x05125DADl, 0x1642AE59l,
	0xE4292D5Al, 0xBA3A117El, 0x4851927Dl, 0x5B016189l, 0xA96AE28Al,
	0x7DA08661l, 0x8FCB0562l, 0x9C9BF696l, 0x6EF07595l, 0x417B1DBCl,
	0xB3109EBFl, 0xA0406D4Bl, 0x522BEE48l, 0x86E18AA3l, 0x748A09A0l,
	0x67DAFA54l, 0x95B17957l, 0xCBA24573l, 0x39C9C670l, 0x2A993584l,
	0xD8F2B687l, 0x0C38D26Cl, 0xFE53516Fl, 0xED03A29Bl, 0x1F682198l,
	0x5125DAD3l, 0xA34E59D0l, 0xB01EAA24l, 0x42752927l, 0x96BF4DCCl,
	0x64D4CECFl, 0x77843D3Bl, 0x85EFBE38l, 0xDBFC821Cl, 0x2997011Fl,
	0x3AC7F2EBl, 0xC8AC71E8l, 0x1C661503l, 0xEE0D9600l, 0xFD5D65F4l,
	0x0F36E6F7l, 0x61C69362l, 0x93AD1061l, 0x80FDE395l, 0x72966096l,
	0xA65C047Dl, 0x5437877El, 0x4767748Al, 0xB50CF789l, 0xEB1FCBADl,
	0x197448AEl, 0x0A24BB5Al, 0xF84F3859l, 0x2C855CB2l, 0xDEEEDFB1l,
	0xCDBE2C45l, 0x3FD5AF46l, 0x7198540Dl, 0x83F3D70El, 0x90A324FAl,
	0x62C8A7F9l, 0xB602C312l, 0x44694011l, 0x5739B3E5l, 0xA55230E6l,
	0xFB410CC2l, 0x092A8FC1l, 0x1A7A7C35l, 0xE811FF36l, 0x3CDB9BDDl,
	0xCEB018DEl, 0xDDE0EB2Al, 0x2F8B6829l, 0x82F63B78l, 0x709DB87Bl,
	0x63CD4B8Fl, 0x91A6C88Cl, 0x456CAC67l, 0xB7072F64l, 0xA457DC90l,
	0x563C5F93l, 0x082F63B7l, 0xFA44E0B4l, 0xE9141340l, 0x1B7F9043l,
	0xCFB5F4A8l, 0x3DDE77ABl, 0x2E8E845Fl, 0xDCE5075Cl, 0x92A8FC17l,
	0x60C37F14l, 0x73938CE0l, 0x81F80FE3l, 0x55326B08l, 0xA759E80Bl,
	0xB4091BFFl, 0x466298FCl, 0x1871A4D8l, 0xEA1A27DBl, 0xF94AD42Fl,
	0x0B21572Cl, 0xDFEB33C7l, 0x2D80B0C4l, 0x3ED04330l, 0xCCBBC033l,
	0xA24BB5A6l, 0x502036A5l, 0x4370C551l, 0xB11B4652l, 0x65D122B9l,
	0x97BAA1BAl, 0x84EA524El, 0x7681D14Dl, 0x2892ED69l, 0xDAF96E6Al,
	0xC9A99D9El, 0x3BC21E9Dl, 0xEF087A76l, 0x1D63F975l, 0x0E330A81l,
	0xFC588982l, 0xB21572C9l, 0x407EF1CAl, 0x532E023El, 0xA145813Dl,
	0x758FE5D6l, 0x87E466D5l, 0x94B49521l, 0x66DF1622l, 0x38CC2A06l,
	0xCAA7A905l, 0xD9F75AF1l, 0x2B9CD9F2l, 0xFF56BD19l, 0x0D3D3E1Al,
	0x1E6DCDEEl, 0xEC064EEDl, 0xC38D26C4l, 0x31E6A5C7l, 0x22B65633l,
	0xD0DDD530l, 0x0417B1DBl, 0xF67C32D8l, 0xE52CC12Cl, 0x1747422Fl,
	0x49547E0Bl, 0xBB3FFD08l, 0xA86F0EFCl, 0x5A048DFFl, 0x8ECEE914l,
	0x7CA56A17l, 0x6FF599E3l, 0x9D9E1AE0l, 0xD3D3E1ABl, 0x21B862A8l,
	0x32E8915Cl, 0xC083125Fl, 0x144976B4l, 0xE622F5B7l, 0xF5720643l,
	0x07198540l, 0x590AB964l, 0xAB613A67l, 0xB831C993l, 0x4A5A4A90l,
	0x9E902E7Bl, 0x6CFBAD78l, 0x7FAB5E8Cl, 0x8DC0DD8Fl, 0xE330A81Al,
	0x115B2B19l, 0x020BD8EDl, 0xF0605BEEl, 0x24AA3F05l, 0xD6C1BC06l,
	0xC5914FF2l, 0x37FACCF1l, 0x69E9F0D5l, 0x9B8273D6l, 0x88D28022l,
	0x7AB90321l, 0xAE7367CAl, 0x5C18E4C9l, 0x4F48173Dl, 0xBD23943El,
	0xF36E6F75l, 0x0105EC76l, 0x12551F82l, 0xE03E9C81l, 0x34F4F86Al,
	0xC69F7B69l, 0xD5CF889Dl, 0x27A40B9El, 0x79B737BAl, 0x8BDCB4B9l,
	0x988C474Dl, 0x6AE7C44El, 0xBE2DA0A5l, 0x4C4623A6l, 0x5F16D052l,
	0xAD7D5351l
	};

//*****************************************************************************
//*
//*		crc32
//*
//*****************************************************************************
static EXT4_INLINE uint32_t crc32(uint32_t crc,const void *buf,uint32_t size,const uint32_t *tab)
{
const uint8_t  *p = (const uint8_t *)buf;

	while(size--)
		{
		crc = tab[(crc ^ *p++) & 0xFF] ^ (crc >> 8);
		}

return (crc);
}

//*****************************************************************************
//*
//*		ext4_crc32
//*
//*****************************************************************************
CONFIG_EXT4_STATIC uint32_t ext4_crc32(uint32_t crc,const void *buf,uint32_t size)
{

return crc32(crc, buf, size, crc32_tab);
}

//*****************************************************************************
//*
//*		ext4_crc32c
//*
//*****************************************************************************
CONFIG_EXT4_STATIC uint32_t ext4_crc32c(uint32_t crc,const void *buf,uint32_t size)
{

return crc32(crc, buf, size, crc32c_tab);
}

//	*
//	@}
