//*****************************************************************************
//*
//*
//*		ext4_dir.c
//*
//*
//*****************************************************************************
//	Copyright (c) 2013 Grzegorz Kostka (kostka.grzegorz@gmail.com)
//
//
//	HelenOS:
//	Copyright (c) 2012 Martin Sucha
//	Copyright (c) 2012 Frantisek Princ
//	All rights reserved.
//
//	Redistribution and use in source and binary forms, with or without
//	modification, are permitted provided that the following conditions
//	are met:
//
//	- Redistributions of source code must retain the above copyright
//	  notice, this list of conditions and the following disclaimer.
//	- Redistributions in binary form must reproduce the above copyright
//	  notice, this list of conditions and the following disclaimer in the
//	  documentation and/or other materials provided with the distribution.
//	- The name of the author may not be used to endorse or promote products
//	  derived from this software without specific prior written permission.
//
//	THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
//	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
//	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
//	IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
//	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
//	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
//	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//	* @addtogroup lwext4
//	@{
//	*
//	@file  ext4_dir.h
//	@brief Directory handle procedures.

#include	"ext4_config.h"
#include	"ext4_compiler.h"
#include	"ext4_types.h"
#include	"ext4_misc.h"
#include	"ext4_errno.h"
#include	"ext4_debug.h"
#include	"ext4_trans.h"
#include	"ext4_dir.h"
#include	"ext4_dir_idx.h"
#include	"ext4_crc32.h"
#include	"ext4_inode.h"
#include	"ext4_fs.h"

#include	<string.h>


//*****************************************************************************
//*
//*		ext4_dir_get_tail
//*
//*****************************************************************************
//  Walk through a dirent block to find a checksum "dirent" at the tail
static struct ext4_dir_entry_tail *ext4_dir_get_tail(struct ext4_inode_ref *inode_ref,struct ext4_dir_en *de)
{
struct ext4_dir_entry_tail *t;
struct ext4_sblock *sb = &inode_ref->fs->sb;

	t = EXT4_DIRENT_TAIL(de, ext4_sb_get_block_size(sb));

	if(t->reserved_zero1 || t->reserved_zero2)
		{
		return NULL;
		}

	if(to_le16(t->rec_len) != sizeof(struct ext4_dir_entry_tail))
		{
		return NULL;
		}

	if(t->reserved_ft != EXT4_DIRENTRY_DIR_CSUM)
		{
		return NULL;
		}

return t;
}

#if CONFIG_META_CSUM_ENABLE

//*****************************************************************************
//*
//*		ext4_dir_csum
//*
//*****************************************************************************
static uint32_t ext4_dir_csum(struct ext4_inode_ref *inode_ref,struct ext4_dir_en *dirent,int size)
{
uint32_t	csum;
struct ext4_sblock *sb = &inode_ref->fs->sb;
uint32_t	ino_index = to_le32(inode_ref->index);
uint32_t	ino_gen = to_le32(ext4_inode_get_generation(inode_ref->inode));

	// First calculate crc32 checksum against fs uuid
	csum = ext4_crc32c(EXT4_CRC32_INIT, sb->uuid, sizeof(sb->uuid));
	//	Then calculate crc32 checksum against inode number

	csum = ext4_crc32c(csum, &ino_index, sizeof(ino_index));
	csum = ext4_crc32c(csum, &ino_gen, sizeof(ino_gen));
	// Finally calculate crc32 checksum against directory entries
	csum = ext4_crc32c(csum, dirent, size);

return csum;
}
#else
#define		ext4_dir_csum(...)			0
#endif

//*****************************************************************************
//*
//*		ext4_dir_csum_verify
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_dir_csum_verify(struct ext4_inode_ref *inode_ref,struct ext4_dir_en *dirent)
{
	#ifdef		CONFIG_META_CSUM_ENABLE
	struct ext4_dir_entry_tail *t;
	struct ext4_sblock *sb = &inode_ref->fs->sb;
	ptrdiff_t diff;
	uint32_t csum;

	// Compute the checksum only if the filesystem supports it
	if(ext4_sb_feature_ro_com(sb, EXT4_FRO_COM_METADATA_CSUM))
		{
			t = ext4_dir_get_tail(inode_ref, dirent);
		if(!t)
			{
			// There is no space to hold the checksum
			return false;
			}

		diff	= (char *)t - (char *)dirent;
		csum	=  ext4_dir_csum(inode_ref, dirent, diff);

		if(t->checksum != to_le32(csum))
			{
			return false;
			}
		}
	#endif

return true;
}

//*****************************************************************************
//*
//*		ext4_dir_init_entry_tail
//*
//*****************************************************************************
CONFIG_EXT4_STATIC void ext4_dir_init_entry_tail(struct ext4_dir_entry_tail *t)
{
	memset(t, 0, sizeof(struct ext4_dir_entry_tail));
	t->rec_len		= to_le16(sizeof(struct ext4_dir_entry_tail));
	t->reserved_ft	= EXT4_DIRENTRY_DIR_CSUM;
}

//*****************************************************************************
//*
//*		ext4_dir_set_csum
//*
//*****************************************************************************
CONFIG_EXT4_STATIC void ext4_dir_set_csum(struct ext4_inode_ref *inode_ref,struct ext4_dir_en *dirent)
{
struct ext4_dir_entry_tail *t;
struct ext4_sblock *sb = &inode_ref->fs->sb;
ptrdiff_t			diff;
uint32_t			csum;


															// Compute the checksum only if the filesystem supports it
	if(ext4_sb_feature_ro_com(sb, EXT4_FRO_COM_METADATA_CSUM))
		{
			t = ext4_dir_get_tail(inode_ref, dirent);
		if(!t)
			{
			return;											// There is no space to hold the checksum
			}

		diff		= (char *)t - (char *)dirent;
		csum		=  ext4_dir_csum(inode_ref, dirent, diff);
		t->checksum =  to_le32(csum);
		}
}

//*****************************************************************************
//*
//*		ext4_dir_iterator_set
//*
//*****************************************************************************
// 	*@brief Do some checks before returning iterator.
// 	@param it Iterator to be checked
// 	@param block_size Size of data block
// 	@return Error code
static int ext4_dir_iterator_set(struct ext4_dir_iter *it,uint32_t block_size)
{
uint32_t			off_in_block;
uint16_t			length;
struct ext4_sblock *sb;
struct ext4_dir_en *en;


	off_in_block = (uint32_t)it->curr_off % block_size;
	sb			 = &it->inode_ref->fs->sb;
	it->curr	 = NULL;

	if((off_in_block % 4) != 0) 								// Ensure proper alignment
		{
		return EIO;
		}

	
	if(off_in_block > block_size - 8)							// Ensure that the core of the entry does not overflow the block
		{
		return EIO;
		}

	en = (struct ext4_dir_en *)(it->curr_blk.data + off_in_block);

	
	length = ext4_dir_en_get_entry_len(en);			
	if(off_in_block + length > block_size)						// Ensure that the whole entry does not overflow the block
		{
		return EIO;
		}

	
	if(ext4_dir_en_get_name_len(sb, en) > length - 8)			// Ensure the name length is not too large
		{
		return EIO;
		}

	
	it->curr = en;												// Everything OK - "publish" the entry


return EOK;
}

//*****************************************************************************
//*
//*		ext4_dir_iterator_seek
//*
//*****************************************************************************
// 	*@brief Seek to next valid directory entry.
// 	       Here can be jumped to the next data block.
// 	@param it  Initialized iterator
// 	@param pos Position of the next entry
// 	@return Error code
static int ext4_dir_iterator_seek(struct ext4_dir_iter *it,uint64_t pos)
{
struct ext4_sblock *sb = &it->inode_ref->fs->sb;
struct ext4_inode  *inode = it->inode_ref->inode;
struct ext4_blockdev *bdev = it->inode_ref->fs->bdev;
uint64_t			size = ext4_inode_get_size(sb, inode);
uint64_t			current_blk_idx;
uint32_t			next_blk_idx;
uint32_t			block_size;
ext4_fsblk_t		next_blk;
int					r;

	// The iterator is not valid until we seek to the desired position
	it->curr = NULL;

	if(pos >= size) 											// Are we at the end?
		{
		if(it->curr_blk.lb_id)
			{

			r				 = ext4_block_set(bdev, &it->curr_blk);
			it->curr_blk.lb_id = 0;
			if(r != EOK)
				{
				return r;
				}
			}

		it->curr_off = pos;
		return EOK;
		}

	// Compute next block address
	block_size		=  ext4_sb_get_block_size(sb);
	current_blk_idx =  it->curr_off / block_size;
	next_blk_idx	= (uint32_t)(pos / block_size);


	//	If we don't have a block or are moving across block boundary,
	//	we need to get another block
	if((it->curr_blk.lb_id == 0) ||
	   (current_blk_idx != next_blk_idx))
		{
		if(it->curr_blk.lb_id)
			{
			r				 = ext4_block_set(bdev, &it->curr_blk);
			it->curr_blk.lb_id = 0;

			if(r != EOK)
				{
				return r;
				}
			}

		   r = ext4_fs_get_inode_dblk_idx(it->inode_ref, next_blk_idx,&next_blk, false);
		if(r != EOK)
			{
			return r;
			}

		   r = ext4_trans_block_get(bdev, &it->curr_blk, next_blk);
		if(r != EOK)
			{
			it->curr_blk.lb_id = 0;
			return r;
			}
		}

	it->curr_off = pos;

return ext4_dir_iterator_set(it, block_size);
}

//*****************************************************************************
//*
//*		ext4_dir_iterator_init
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_dir_iterator_init(struct ext4_dir_iter *it,struct ext4_inode_ref *inode_ref,uint64_t pos)
{
	it->inode_ref		= inode_ref;
	it->curr			= 0;
	it->curr_off		= 0;
	it->curr_blk.lb_id	= 0;

return ext4_dir_iterator_seek(it, pos);
}

//*****************************************************************************
//*
//*		ext4_dir_iterator_next
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_dir_iterator_next(struct ext4_dir_iter *it)
{
int			r = EOK;
uint16_t	skip;

	while(r == EOK)
		{
		skip	= ext4_dir_en_get_entry_len(it->curr);
		r		= ext4_dir_iterator_seek(it, it->curr_off + skip);

		if(!it->curr)
			{
			break;
			}
		//Skip NULL referenced entry
		if(ext4_dir_en_get_inode(it->curr) != 0)
			{
			break;
			}
		}

return r;
}

//*****************************************************************************
//*
//*		ext4_dir_iterator_fini
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_dir_iterator_fini(struct ext4_dir_iter *it)
{
	it->curr = 0;

	if(it->curr_blk.lb_id)
		{
		return ext4_block_set(it->inode_ref->fs->bdev, &it->curr_blk);
		}

return EOK;
}

//*****************************************************************************
//*
//*		ext4_dir_write_entry
//*
//*****************************************************************************
CONFIG_EXT4_STATIC void ext4_dir_write_entry(struct ext4_sblock *sb,struct ext4_dir_en *en,uint16_t entry_len,struct ext4_inode_ref *child,const char *name,size_t name_len)
{
	ext4_assert(entry_len <= ext4_sb_get_block_size(sb));		// Check maximum entry length

	switch(ext4_inode_type(sb, child->inode))					// Set type of entry
		{
	case EXT4_INODE_MODE_DIRECTORY:

		ext4_dir_en_set_inode_type(sb, en, EXT4_DE_DIR);
		break;

	case EXT4_INODE_MODE_FILE:

		ext4_dir_en_set_inode_type(sb, en, EXT4_DE_REG_FILE);
		break;

	case EXT4_INODE_MODE_SOFTLINK:

		ext4_dir_en_set_inode_type(sb, en, EXT4_DE_SYMLINK);
		break;

	default:

		ext4_assert(0);											// FIXME: right now we only support 3 inode type.
		}

	ext4_dir_en_set_inode(en, child->index);					// Set basic attributes
	ext4_dir_en_set_entry_len(en, entry_len);
	ext4_dir_en_set_name_len(sb, en, (uint16_t)name_len);

	memcpy(en->name, name, name_len);							// Write name
}

//*****************************************************************************
//*
//*		ext4_dir_add_entry
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_dir_add_entry(struct ext4_inode_ref *parent,const char *name,uint32_t name_len,struct ext4_inode_ref *child)
{
int					r;
struct ext4_fs	   *fs = parent->fs;
struct ext4_sblock *sb = &parent->fs->sb;
uint32_t			iblock;
ext4_fsblk_t		fblock;
uint32_t			block_size;
uint64_t			inode_size;
uint32_t			total_blocks;
struct ext4_block	block;
struct ext4_block	b;
struct ext4_dir_en *blk_en;
int					success;


	#if CONFIG_DIR_INDEX_ENABLE
																// Index adding (if allowed)
	if((ext4_sb_feature_com(sb, EXT4_FCOM_DIR_INDEX)) && (ext4_inode_has_flag(parent->inode, EXT4_INODE_FLAG_INDEX)))
		{
		r = ext4_dir_dx_add_entry(parent, child, name);

		
		if(r != EXT4_ERR_BAD_DX_DIR)							// Check if index is not corrupted
			{
			if(r != EOK)
				{
				return r;
				}

			return EOK;
			}

																// Needed to clear dir index flag if corrupted
		ext4_inode_clear_flag(parent->inode, EXT4_INODE_FLAG_INDEX);
		parent->dirty = true;
		}
	#endif

	
	iblock			=  0;										// Linear algorithm
	fblock			=  0;
	block_size		=  ext4_sb_get_block_size(sb);
	inode_size		=  ext4_inode_get_size(sb, parent->inode);
	total_blocks	= (uint32_t)(inode_size / block_size);

	
	success = false;											// Find block, where is space for new entry and try to add
	
	for(iblock=0;iblock<total_blocks;iblock++)
		{
		   r = ext4_fs_get_inode_dblk_idx(parent, iblock, &fblock, false);
		if(r != EOK)
			{
			return r;
			}

		   r = ext4_trans_block_get(fs->bdev, &block, fblock);
		if(r != EOK)
			{
			return r;
			}

		if(!ext4_dir_csum_verify(parent, (struct ext4_dir_en *)block.data))
			{
			ext4_dbg2(DEBUG_DIR,DBG_WARN "Leaf block checksum failed.""Inode: %" PRIu32", ""Block: %" PRIu32"\n",parent->index,iblock);
			}

																// If adding is successful, function can finish
		   r = ext4_dir_try_insert_entry(sb, parent, &block, child,name, name_len);
		if(r == EOK)
			{
			success = true;
			}

		   r = ext4_block_set(fs->bdev, &block);
		if(r != EOK)
			{
			return r;
			}

		if(success)
			{
			return EOK;
			}
		}

															

	iblock	= 0;												// No free block found - needed to allocate next data block
	fblock	= 0;
	
	   r = ext4_fs_append_inode_dblk(parent, &fblock, &iblock);
	if(r!=EOK)
		{
		return r;
		}

	
	   r = ext4_trans_block_get_noread(fs->bdev, &b, fblock);	// Load new block
	if(r != EOK)
		{
		return r;
		}

	memset(b.data, 0, block_size);								// Fill block with zeroes

	blk_en = (struct ext4_dir_en *)b.data;

	if(ext4_sb_feature_ro_com(sb, EXT4_FRO_COM_METADATA_CSUM))	// Save new block
		{
		uint16_t	el = block_size - sizeof(struct ext4_dir_entry_tail);


		ext4_dir_write_entry(sb, blk_en, el, child, name, name_len);
		ext4_dir_init_entry_tail(EXT4_DIRENT_TAIL(b.data, block_size));
		}
	else{
		ext4_dir_write_entry(sb, blk_en, (uint16_t)block_size, child, name, name_len);
		}

	ext4_dir_set_csum(parent, (struct ext4_dir_en *)b.data);
	ext4_trans_set_block_dirty(b.buf);
	
	r = ext4_block_set(fs->bdev, &b);


return r;
}

//*****************************************************************************
//*
//*		ext4_dir_find_entry
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_dir_find_entry(struct ext4_dir_search_result *result,struct ext4_inode_ref *parent,const char *name,uint32_t name_len)
{
int					r;
ext4_fsblk_t		fblock;
struct ext4_block	b;
struct ext4_sblock *sb = &parent->fs->sb;
struct ext4_dir_en *res_entry;
uint64_t			inode_size;
uint32_t			total_blocks;
uint32_t			block_size;
uint32_t			iblock;


	result->block.lb_id = 0;							// Entry clear
	result->dentry		= NULL;

	#if CONFIG_DIR_INDEX_ENABLE
														// Index search
	if((ext4_sb_feature_com(sb, EXT4_FCOM_DIR_INDEX)) && (ext4_inode_has_flag(parent->inode, EXT4_INODE_FLAG_INDEX)))
		{
		r = ext4_dir_dx_find_entry(result, parent, name_len, name);

		if(r != EXT4_ERR_BAD_DX_DIR)					// Check if index is not corrupted
			{
			if(r != EOK)
				{
				return r;
				}

			return EOK;
			}
														// Needed to clear dir index flag if corrupted
		ext4_inode_clear_flag(parent->inode, EXT4_INODE_FLAG_INDEX);
		parent->dirty = true;
		}
	#endif


														// Linear algorithm
	block_size		=  ext4_sb_get_block_size(sb);
	inode_size		=  ext4_inode_get_size(sb, parent->inode);
	total_blocks	= (uint32_t)(inode_size / block_size);

	for(iblock=0;iblock<total_blocks;iblock++)			// Walk through all data blocks
		{
														// Load block address
		   r = ext4_fs_get_inode_dblk_idx(parent, iblock, &fblock, false);
		if(r != EOK)
			{
			return r;
			}
														// Load data block
		   r = ext4_trans_block_get(parent->fs->bdev, &b, fblock);
		if(r != EOK)
			{
			return r;
			}

		if(!ext4_dir_csum_verify(parent, (struct ext4_dir_en *)b.data))
			{
			ext4_dbg2(DEBUG_DIR,DBG_WARN "Leaf block checksum failed.""Inode: %" PRIu32", ""Block: %" PRIu32"\n",parent->index,iblock);
			}
														// Try to find entry in block
		   r = ext4_dir_find_in_block(&b, sb, name_len, name, &res_entry);
		if(r == EOK)
			{
			result->block	= b;
			result->dentry	= res_entry;
			return EOK;
			}

		   r = ext4_block_set(parent->fs->bdev, &b);	// Entry not found - put block and continue to the next block
		if(r != EOK)
			{
			return r;
			}
		}

return ENOENT;
}

//*****************************************************************************
//*
//*		ext4_dir_remove_entry
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_dir_remove_entry(struct ext4_inode_ref *parent,const char *name,uint32_t name_len)
{
struct ext4_sblock *sb = &parent->fs->sb;
struct ext4_dir_search_result result;
uint16_t	del_len;
uint32_t	pos;
int			rc;

																// Check if removing from directory
	if(!ext4_inode_is_type(sb, parent->inode, EXT4_INODE_MODE_DIRECTORY))
		{
		return ENOTDIR;
		}
																// Try to find entry
	   rc = ext4_dir_find_entry(&result, parent, name, name_len);
	if(rc != EOK)
		{
		return rc;
		}

	ext4_dir_en_set_inode(result.dentry, 0);					// Invalidate entry

	
	pos = (uint8_t *)result.dentry - result.block.data;			// Store entry position in block


	if(pos != 0)												// If entry is not the first in block, it must be merged with previous entry
		{
		struct ext4_dir_en *tmp_de;
		uint32_t			offset;
		uint16_t			de_len;
		
		offset	= 0;
		tmp_de	=(struct ext4_dir_en *)result.block.data;		// Start from the first entry in block
		de_len	= ext4_dir_en_get_entry_len(tmp_de);
							
		
		while((offset + de_len) < pos)							// Find direct predecessor of removed entry
			{
			offset += ext4_dir_en_get_entry_len(tmp_de);
			tmp_de	= (struct ext4_dir_en *)(result.block.data + offset);
			de_len	= ext4_dir_en_get_entry_len(tmp_de);
			}

		ext4_assert(de_len + offset == pos);

																// Add to removed entry length to predecessor's length
		del_len = ext4_dir_en_get_entry_len(result.dentry);
		ext4_dir_en_set_entry_len(tmp_de, (uint16_t)(de_len + del_len));
		}

	ext4_dir_set_csum(parent,(struct ext4_dir_en *)result.block.data);
	ext4_trans_set_block_dirty(result.block.buf);

return ext4_dir_destroy_result(parent, &result);
}

//*****************************************************************************
//*
//*		ext4_dir_try_insert_entry
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_dir_try_insert_entry(struct ext4_sblock *sb,struct ext4_inode_ref *inode_ref,struct ext4_block *dst_blk,struct ext4_inode_ref *child,const char *name,uint32_t name_len)
{
	// Compute required length entry and align it to 4 bytes
	uint32_t block_size = ext4_sb_get_block_size(sb);
	uint16_t required_len = sizeof(struct ext4_fake_dir_entry) + name_len;
	struct ext4_dir_en *start;
	struct ext4_dir_en *stop;

	if((required_len % 4) != 0)
		{
		required_len += 4 - (required_len % 4);
		}

																
	start	= (struct ext4_dir_en *)(dst_blk->data);			// Initialize pointers, stop means to upper bound
	stop	= (struct ext4_dir_en *)(dst_blk->data + block_size);

	while(start < stop)											//	Walk through the block and check for invalid entries or entries with free space for new entry
		{
		uint32_t	inode   = ext4_dir_en_get_inode(start);
		uint16_t	rec_len = ext4_dir_en_get_entry_len(start);
		uint8_t		itype   = ext4_dir_en_get_inode_type(sb, start);

																// If invalid and large enough entry, use it
		if((inode == 0) && (itype != EXT4_DIRENTRY_DIR_CSUM) &&	(rec_len >= required_len))
			{
			ext4_dir_write_entry(sb, start, rec_len, child, name,name_len);
			ext4_dir_set_csum(inode_ref, (struct ext4_dir_en *)dst_blk->data);
			ext4_trans_set_block_dirty(dst_blk->buf);

			return EOK;
			}

		if(inode != 0)											// Valid entry, try to split it
			{
			uint16_t	free_space;
			uint16_t	used_len;
			uint16_t	sz;

			used_len = ext4_dir_en_get_name_len(sb, start);

			sz = sizeof(struct ext4_fake_dir_entry) + used_len;

			if((used_len % 4) != 0)
				{
				sz += 4 - (used_len % 4);
				}

			free_space = rec_len - sz;
			 			
			if(free_space >= required_len)						// There is free space for new entry
				{
				struct ext4_dir_en * new_entry;			
																// Cut tail of current entry
				new_entry = (struct ext4_dir_en *)((uint8_t *)start + sz);	
				ext4_dir_en_set_entry_len(start, sz);
				ext4_dir_write_entry(sb, new_entry, free_space,child, name, name_len);

				ext4_dir_set_csum(inode_ref,(struct ext4_dir_en *)dst_blk->data);
				ext4_trans_set_block_dirty(dst_blk->buf);
				return EOK;
				}
			}											 
																// Jump to the next entry																
		start = (struct ext4_dir_en *)((uint8_t *)start + rec_len);			
		}

	

return ENOSPC;													// No free space found for new entry
}

//*****************************************************************************
//*
//*		ext4_dir_find_in_block
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_dir_find_in_block(struct ext4_block *block,struct ext4_sblock *sb,size_t name_len,const char *name,struct ext4_dir_en **res_entry)
{
struct ext4_dir_en *de;
uint8_t 		   *addr_limit;
uint16_t			de_len;

																
	de = (struct ext4_dir_en *)block->data;						// Start from the first entry in block

	addr_limit = block->data + ext4_sb_get_block_size(sb);		// Set upper bound for cycling

	while((uint8_t *)de < addr_limit)							// Walk through the block and check entries
		{
		if((uint8_t *)de + name_len > addr_limit)				// Termination condition
			{
			break;
			}

		if(ext4_dir_en_get_inode(de) != 0)						// Valid entry - check it
			{
			uint16_t	el;
			
			   el = ext4_dir_en_get_name_len(sb, de);
			if(el == name_len)									// For more efficient compare only lengths firstly
				{
				if(memcmp(name, de->name, name_len) == 0)		// Compare names
					{
					*res_entry = de;
					return EOK;
					}
				}
			}

		de_len = ext4_dir_en_get_entry_len(de);

		if(de_len == 0) 										// Corrupted entry
			{
			return EINVAL;
			}

		de = (struct ext4_dir_en *)((uint8_t *)de + de_len);	// Jump to next entry
		}

	
return ENOENT;													// Entry not found
}

//*****************************************************************************
//*
//*		ext4_dir_destroy_result
//*
//*****************************************************************************
CONFIG_EXT4_STATIC int ext4_dir_destroy_result(struct ext4_inode_ref *parent,struct ext4_dir_search_result *result)
{
	if(result->block.lb_id)
		{
		return ext4_block_set(parent->fs->bdev, &result->block);
		}

return EOK;
}

//	*
//	@}
