//*****************************************************************************
//*
//*
//*		io_raw.c
//*
//*
//*****************************************************************************
//	Copyright (c) 2013 Grzegorz Kostka (kostka.grzegorz@gmail.com)
//	All rights reserved.
//	
//	Redistribution and use in source and binary forms, with or without
//	modification, are permitted provided that the following conditions
//	are met:
//	
//	- Redistributions of source code must retain the above copyright
//	  notice, this list of conditions and the following disclaimer.
//	- Redistributions in binary form must reproduce the above copyright
//	  notice, this list of conditions and the following disclaimer in the
//	  documentation and/or other materials provided with the distribution.
//	- The name of the author may not be used to endorse or promote products
//	  derived from this software without specific prior written permission.
//	
//	THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
//	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
//	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
//	IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
//	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
//	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
//	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#include	"ext4_config.h"
#include	"ext4_compiler.h"
#include	"ext4_blockdev.h"
#include	"ext4_errno.h"

#include	<stdio.h>
#include	<string.h>

#ifdef		WIN32						
#include	<windows.h>
#include	<winioctl.h>

//*@brief	Default filename.
static const char *fname = "ext2";

//*@brief	IO block size.
#define		EXT4_IORAW_BSIZE			512

//*@brief	Image file descriptor.
static HANDLE dev_file;

//*********************BLOCKDEV INTERFACE*************************************
static int io_raw_open(struct ext4_blockdev *bdev);
static int io_raw_bread(struct ext4_blockdev *bdev, void *buf, uint64_t blk_id,uint32_t blk_cnt);
static int io_raw_bwrite(struct ext4_blockdev *bdev, const void *buf,uint64_t blk_id, uint32_t blk_cnt);
static int io_raw_close(struct ext4_blockdev *bdev);

//****************************************************************************

static uint8_t _filedev_ph_bbuf[512];

static struct ext4_blockdev_iface _filedev_iface =
	{
	io_raw_open,
	io_raw_bread,
	io_raw_bwrite,
	io_raw_close,
	0,
	0,
	512,
	0,
	_filedev_ph_bbuf,
	};

static struct ext4_blockdev _filedev =
	{
	&_filedev_iface,
	0,
	0,
	};

//*****************************************************************************
//*
//*		io_raw_open
//*
//*****************************************************************************
static int io_raw_open(struct ext4_blockdev *bdev)
{
char			path[64];
DISK_GEOMETRY	pdg;
uint64_t		disk_size;
BOOL			bResult = FALSE;
DWORD			junk;
DWORD			high;

	// REM AZ sprintf(path, "\\\\.\\%s", fname);
	sprintf(path, "%s", fname);

	dev_file =
	CreateFile(path, GENERIC_READ | GENERIC_WRITE,FILE_SHARE_WRITE | FILE_SHARE_READ, NULL, OPEN_EXISTING,FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH, NULL);

	if(dev_file == INVALID_HANDLE_VALUE) 
		{
		return EIO;
		}

	bResult =DeviceIoControl(dev_file, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &pdg, sizeof(pdg), &junk, (LPOVERLAPPED)NULL);

	if(!bResult) 
		{

		disk_size = GetFileSize(dev_file,&high);

		if(disk_size==0xFFFFFFFF) 
			{
			disk_size=0;
			}
		else{ 
			disk_size  |= (uint64_t)high<<32;
			bResult 	=  TRUE;
			}

		_filedev.bdif->ph_bsize = 512;
		_filedev.bdif->ph_bcnt	= disk_size / 512;
		_filedev.part_offset	= 0;
		_filedev.part_size		= disk_size;
		}
	else{ 
		disk_size = pdg.Cylinders.QuadPart * (ULONG)pdg.TracksPerCylinder *
		(ULONG)pdg.SectorsPerTrack * (ULONG)pdg.BytesPerSector;

		_filedev.bdif->ph_bsize = pdg.BytesPerSector;
		_filedev.bdif->ph_bcnt	= disk_size / pdg.BytesPerSector;

		_filedev.part_offset	= 0;
		_filedev.part_size		= disk_size;
		}


	if(bResult == FALSE) 
		{
		CloseHandle(dev_file);
		return EIO;
		}

return EOK;
}

//*****************************************************************************
//*
//*		io_raw_bread
//*
//*****************************************************************************
static int io_raw_bread(struct ext4_blockdev *bdev,void *buf,uint64_t blk_id,uint32_t blk_cnt)
{
long		hipart = (uint32_t)(blk_id >> (32 - 9));
long		lopart = (uint32_t)(blk_id << 9);
long		err;
DWORD		n;

	SetLastError(0);
	lopart = SetFilePointer(dev_file, lopart, &hipart, FILE_BEGIN);

	if(lopart == -1 && NO_ERROR != (err = GetLastError())) 
		{
		return EIO;
		}


	if(!ReadFile(dev_file, buf, blk_cnt * 512, &n, NULL)) 
		{
		err = GetLastError();
		return EIO;
		}

return EOK;
}

//*****************************************************************************
//*
//*		io_raw_bwrite
//*
//*****************************************************************************
static int io_raw_bwrite(struct ext4_blockdev *bdev,const void *buf,uint64_t blk_id,uint32_t blk_cnt)
{
long		hipart = (uint32_t)(blk_id >> (32 - 9));
long		lopart = (uint32_t)(blk_id << 9);
long		err;
DWORD		n;

	SetLastError(0);

	lopart = SetFilePointer(dev_file, lopart, &hipart, FILE_BEGIN);

	if(lopart == -1 && NO_ERROR != (err = GetLastError())) 
		{
		return EIO;
		}


	if(!WriteFile(dev_file, buf, blk_cnt * 512, &n, NULL)) 
		{
		err = GetLastError();
		return EIO;
		}

return EOK;
}

//*****************************************************************************
//*
//*		io_raw_close
//*
//*****************************************************************************
static int io_raw_close(struct ext4_blockdev *bdev)
{
	CloseHandle(dev_file);

return EOK;
}

//*****************************************************************************
//*
//*		ext4_io_raw_dev_get
//*
//*****************************************************************************
struct ext4_blockdev *ext4_io_raw_dev_get(void) 
{

return &_filedev; 
}

//*****************************************************************************
//*
//*		ext4_io_raw_filename
//*
//*****************************************************************************
void ext4_io_raw_filename(const char *n) 
{
	fname = n; 
}

//****************************************************************************
#endif
