//*****************************************************************************
//*
//*
//*		TlsEncoder.cpp
//*
//*
//*****************************************************************************
//
//	(C) Copyright Anton Zechner 2013
//
#include	"Tls.h"
#include	"TlsEncoder.h"
#include	"..\FileBuffer.h"


static const unsigned   uSignature = 'T'|('L'<<8)|('S'<<16)|('P'<<24);

#define 	TLS_MODE_SINGLE		-1
#define 	TLS_MODE_FIRST		 0
#define 	TLS_MODE_NEXT		 1
#define 	TLS_MODE_LAST		 2


//*****************************************************************************
//*
//*		Konstruktor
//*
//*****************************************************************************
TlsEncoder::TlsEncoder()
{
	Initialize();
}

//*****************************************************************************
//*
//*		Destruktor
//*
//*****************************************************************************
TlsEncoder::~TlsEncoder()
{
}


//*****************************************************************************
//*
//*		Initialize
//*
//*****************************************************************************
void TlsEncoder::Initialize()
{

	m_iFormat    = 0;									// Format fr das Speichern
	m_uPicureId	 = 0;									// Einteutige Bild-ID
	m_iSequDone	 = 0;									// Wurden die Bildsequenz eingestellt
	m_iAreasDone = 0;									// Wurden die Bereiche eingestellt

}

//*****************************************************************************
//*
//*		SetFormat
//*
//*****************************************************************************
//	Setl ein Format fr das Speichern ein
//	iFormat		: Ist das Format (0,24 oder 32)
//	Ergibt das eingestellte Format
int TlsEncoder::SetFormat(int iFormat)
{

	if(iFormat!=8 && iFormat!=32)
		{
		m_iFormat = 0;
		}
	else{
		m_iFormat = iFormat;
		}


return m_iFormat;
}

//*****************************************************************************
//*
//*		SetPicureId
//*
//*****************************************************************************
//	Stellt die Bild-ID frs Speichern ein
//	uPicureId		: Ist die neue Bild-ID
//	Ergibt die
unsigned TlsEncoder::SetPicureId(unsigned uPicureId)
{
unsigned	uOld;


	uOld		= m_uPicureId;
	m_uPicureId	= uPicureId;


return uOld;
}

//*****************************************************************************
//*
//*		SetSequence
//*
//*****************************************************************************
//	Stellt den Sequenz-Header fr Sequenzen ein
//	uTime		: Zeit in Millisekunden
//	iNext		: Nummer des nchsten Bildes (-1=immer aufsteigend)
void TlsEncoder::SetSequence(unsigned uTime,int iNext)
{

	if(uTime>0xFFFF)uTime=0xFFFF;

	m_sSequence.usShowTime		= uTime;			// Anzeigezeit in Millisekunden
	m_sSequence.ucNextPicture	= iNext;			// Nchstes Bild
	m_sSequence.ucUnused		= 0;				// Unbenutzt
	m_iSequDone					= (iNext<0)? 2:1;

}

//*****************************************************************************
//*
//*		SetErrorAreas
//*
//*****************************************************************************
//	Stellt die Fehlerbereiche fr das Speichern ein
void TlsEncoder::SetErrorAreas(unsigned *pArea)
{

	memcpy(m_sAreas.aAreaErrors,pArea,16*4);
	m_iAreasDone = 1;

}

//*****************************************************************************
//*
//*		CallProgressFunction
//*
//*****************************************************************************
void TlsEncoder::CallProgressFunction (unsigned int percent)
{
bool	bCancel;


	if(pProgressProc == NULL)return;

	bCancel = false;
	pProgressProc(*this,pProgressData,1,1,percent,bCancel);

	if(bCancel)throw EGraphicsAbort();

}

//*****************************************************************************
//*
//*		WriteImage
//*
//*****************************************************************************
//	Speichert das Bild ab.
//	pName		: Dateiname der Bilddatei
//  sImage		: Klasse in der das Bild gespeichert wird
void TlsEncoder::WriteImage(const char *pName,const BitmapImage &sImage)
{
CFileBuffer	sBuffer;


	if(!sBuffer.Open(pName,CFB_CREATE))throw ETlsFileOpenError();

	try	{
		WriteImage(&sBuffer,sImage,0,1);
		}
	catch(...)
		{
		sBuffer.Close();
		throw;
		}

	sBuffer.Close();

}

void TlsEncoder::WriteImage(const wchar_t *pName,const BitmapImage &sImage)
{
CFileBuffer	sBuffer;


	if(!sBuffer.Open(pName,CFB_CREATE))throw ETlsFileOpenError();

	try	{
		WriteImage(&sBuffer,sImage,0,1);
		}
	catch(...)
		{
		sBuffer.Close();
		throw;
		}

	sBuffer.Close();

}

void TlsEncoder::WriteImage(CFileBuffer *pFile,const BitmapImage &sImage)
{

	WriteImage(pFile,sImage,0,1);

}

void TlsEncoder::WriteImage(const char *pName,const BitmapImage *pImages,unsigned uPicCount,unsigned *pAreas,unsigned *pTimes,unsigned char *pNext)
{
CFileBuffer		sBuffer;


	if(!sBuffer.Open(pName,CFB_CREATE))throw ETlsFileOpenError();

	try	{
		WriteImage(&sBuffer,pImages,uPicCount,pAreas,pTimes,pNext);
		}
	catch(...)
		{
		sBuffer.Close();
		throw;
		}

	sBuffer.Close();

}

void TlsEncoder::WriteImage(const wchar_t *pName,const BitmapImage *pImages,unsigned uPicCount,unsigned *pAreas,unsigned *pTimes,unsigned char *pNext)
{
CFileBuffer		sBuffer;


	if(!sBuffer.Open(pName,CFB_CREATE))throw ETlsFileOpenError();

	try	{
		WriteImage(&sBuffer,pImages,uPicCount,pAreas,pTimes,pNext);
		}
	catch(...)
		{
		sBuffer.Close();
		throw;
		}

	sBuffer.Close();

}

void TlsEncoder::WriteImage(CFileBuffer *pFile,const BitmapImage *pImages,unsigned uPicCount,unsigned *pAreas,unsigned *pTimes,unsigned char *pNext)
{
unsigned	uStart;
unsigned	uPos;


	if(uPicCount>255)throw ETlsDataError();

	uStart = pFile->Tell();

	for(uPos=0;uPos<uPicCount;uPos++)
		{
		if(pAreas)SetErrorAreas(pAreas+16*uPos);
		if(pTimes || pNext)SetSequence((pTimes)? pTimes[uPos]:5000,(pNext)? pNext[uPos]:-1);
		WriteImage(pFile,pImages[uPos],uPos,uPicCount);
		pFile->Seek(uStart);
		}

}

void TlsEncoder::WriteImage(CFileBuffer *pFile,const BitmapImage &sImage,unsigned uPicPos,unsigned uPicCount)
{
static const char	aPads[4]={0,0,0,0,};
unsigned			uLineWidth;
unsigned			uOutWidth;
unsigned			uBitWidth;
unsigned			uFullSize;
unsigned			uOutBits;
unsigned			uPadSize;
unsigned			uMaxRows;
unsigned			uBaseRow;
unsigned			uColor;
unsigned			uStart;
unsigned			uOff;
unsigned			uCol;
unsigned			uRow;
unsigned			uNum;
unsigned			uPos;
unsigned			uBit;
unsigned char	   *pPtr;
int					iCnt;



	if(!sImage.IsOk())
		{
		throw ETlsDataError();
		}


	if(m_iAreasDone==0)									// Fehlerbereiche aus dem Bild holen
		{
		memset(m_sAreas.aAreaErrors,0,sizeof(m_sAreas.aAreaErrors));

		if(sImage.BitCount()<=8)
			{
			for(uPos=0;uPos<11*16;uPos++)				// In den Alpha-Werten der Farben ablegen
				{
				uOff	= (uPos>>4)*3;
				uBit	= (sImage.GetColor(uPos))>>24;
				uNum	= (uPos&0x8F);
				
				if(uNum!=(uBit&0x8F))
					{
					memset(m_sAreas.aAreaErrors,0,sizeof(m_sAreas.aAreaErrors));
					break;
					}

				uBit	= (uBit>>4)&7;
				m_sAreas.aAreaErrors[uPos&0x0F] |= uBit<<uOff;
				}
			}
		else{
			for(uPos=0;uPos<11*16;uPos++)				// In den Alpha-Werten der Pixel ablegen
				{
				uCol	= (uPos%sImage.Width());
				uRow	= (uPos/sImage.Width());
				uNum	= sImage.GetPixel(uCol,uRow);
				uOff	= (uPos>> 4)*3;
				uBit	= (uNum>>28)&7;

				m_sAreas.aAreaErrors[uPos&0x0F] |= uBit<<uOff;
				}

			for(;uPos<11*16+11*2;uPos++)
				{
				uCol	= (uPos%sImage.Width());
				uRow	= (uPos/sImage.Width());
				uNum	= sImage.GetPixel(uCol,uRow);
				uOff	= (uPos>> 4)*3;
				uBit	= (uNum>>28)&7;
				uNum	= (uPos-11*16)&0x07;
				
				if(uBit!=uNum)							// Ungltige CRC
					{
					memset(m_sAreas.aAreaErrors,0,sizeof(m_sAreas.aAreaErrors));
					break;
					}
				}
			}
		}

	if(m_iSequDone==0)									// Sequenzinfos einstellen
		{
		m_sSequence.usShowTime		= 5000;
		m_sSequence.ucUnused		= 0;
		}

	if(m_iSequDone!=1)
		{
		m_sSequence.ucNextPicture	= (uPicPos+1>=uPicCount)? 0:uPicPos+1;
		}
	else if(m_sSequence.ucNextPicture>=uPicCount)
		{
		throw ETlsDataError();
		}


	if(uPicPos>0)										// Ein Bild anhngen
		{
		if(m_sFileHeader.usSizeX!=sImage.Width())
			{
			throw ETlsDataError();
			}

		if(m_sFileHeader.usSizeY!=sImage.Height())
			{
			throw ETlsDataError();
			}

		if(m_sFileHeader.ucSequPics>=255)				// Zuviele Bilder
			{
			throw ETlsDataError();
			}

		uOutBits = m_sFileHeader.ucFormat*8;
		}
	else{
		if(sImage.Width()>TLS_SIZE_MAX_X)
			{
			throw ETlsDataError();
			}

		if(sImage.Height()>TLS_SIZE_MAX_Y)
			{
			throw ETlsDataError();
			}

		if(m_iFormat)
			{
			uOutBits = m_iFormat;
			}
		else if(sImage.BitCount()<=8)
			{
			uOutBits = 8;
			}
		else{
			uOutBits = 32;
			}

		m_sFileHeader.uSignature = uSignature;			// "TLSP"
		m_sFileHeader.usSizeX	 = sImage.Width();		// Bildbreite
		m_sFileHeader.usSizeY	 = sImage.Height();		// Bildhhe
		m_sFileHeader.ucFormat	 = uOutBits/8;			// 1=RGB-3Bit 4=RGB-24Bit
		m_sFileHeader.ucSequPics = uPicCount-1;			// Zustzliche Sequenzbilder
		m_sFileHeader.usCrc		 = 0;					// CRC16 mit Polygon x^16 + x^12 + x^5 + 1
		m_sFileHeader.uPicureId	 = m_uPicureId;			// Eindeutige Bild-ID
		}


	uMaxRows	= m_sFileHeader.usSizeY*uPicCount;
	uBaseRow	= m_sFileHeader.usSizeY*uPicPos;

	uBitWidth   = ((uOutBits   )*sImage.Width());
	uOutWidth	= ((uBitWidth+7)/ 8);
	uPadSize	= ((uOutWidth+3)&~3)-uOutWidth;
	uLineWidth	= ((uOutWidth+uPadSize));
	uFullSize	= ((uLineWidth*m_sFileHeader.usSizeY));
	uFullSize  += sizeof(m_sAreas)+sizeof(m_sSequence);


	if(uPicPos==0)										// Prfe die Gesamtgre
		{
		iCnt	= (uFullSize+0xFFFF)>>16;
		iCnt   *=  uPicCount;

		if(iCnt>0x7800)									// Das Bild wird ber 2GB
			{
			throw ETlsDataError();
			}
		}


//******************** Datei-Header schreiben *********************************


	uStart = pFile->Tell();

	if(uPicPos<=0)										// Erstes Bild
		{
		   iCnt =pFile->Write(&m_sFileHeader,sizeof(m_sFileHeader));
		if(iCnt!=sizeof(m_sFileHeader))throw ETlsFileWriteError();

		m_usCalcCrc = TlsCrcCalc(&m_sFileHeader,sizeof(m_sFileHeader),0);

		CallProgressFunction(0);
		}
	else{
		uOff  = uPicPos*uFullSize;
		uOff += sizeof(m_sFileHeader);

		   pFile->Seek(uOff);
		if(pFile->Tell()!=(int)uOff)
			{
			throw ETlsCorruptFile();
			}
		}



	   iCnt =pFile->Write(&m_sAreas,sizeof(m_sAreas));	// Fehlerbereiche speichern
	if(iCnt!=sizeof(m_sAreas))throw ETlsFileWriteError();

	m_usCalcCrc = TlsCrcCalc(&m_sAreas,sizeof(m_sAreas),m_usCalcCrc);



	if(sImage.BitCount()==1)							// 1-Bit Monchrombild
		{
		for(uPos=0;uPos<sImage.Height();uPos++)
			{
			CallProgressFunction((uBaseRow+uPos)*100/uMaxRows);

			pPtr = sImage.GetRow4Temp(uPos);

			for(uOff=0;uOff<sImage.Width();uOff++)
				{
				uColor		= *(unsigned*)(pPtr+uOff*4);
				pPtr[uOff]  =  (uColor&0x00C0C0C0)? 0x70:0x00;
				pPtr[uOff] |=  (uColor>>24)&0x8F;
				}

			memset(pPtr+uOff,0,4);
			m_usCalcCrc = TlsCrcCalc(pPtr,uLineWidth,m_usCalcCrc);

			if(pFile->Write(pPtr ,uLineWidth)!=(int)uLineWidth)throw ETlsFileWriteError();
			}
		}
	else if(uOutBits<=8)								// 8-Bit Pallette
		{
		for(uPos=0;uPos<sImage.Height();uPos++)
			{
			CallProgressFunction((uBaseRow+uPos)*100/uMaxRows);

			pPtr = sImage.GetRow4Temp(uPos);

			for(uOff=0;uOff<sImage.Width();uOff++)
				{
				uColor		= *(unsigned*)(pPtr+uOff*4);
				pPtr[uOff]  =  (uColor&0x000000C0)? 0x10:0x00;
				pPtr[uOff] |=  (uColor&0x0000C000)? 0x20:0x00;
				pPtr[uOff] |=  (uColor&0x00C00000)? 0x40:0x00;
				pPtr[uOff] |=  (uColor>>24)&0x8F;
				}

			memset(pPtr+uOff,0,4);
			m_usCalcCrc = TlsCrcCalc(pPtr,uLineWidth,m_usCalcCrc);

			if(pFile->Write(pPtr ,uLineWidth)!=(int)uLineWidth)throw ETlsFileWriteError();
			}
		}
	else{												// 24/32-Bit RGB-Bild
		for(uPos=0;uPos<sImage.Height();uPos++)
			{
			CallProgressFunction((uBaseRow+uPos)*100/uMaxRows);

			pPtr = sImage.GetRow4Temp(uPos);

			for(uOff=3;uOff<uOutWidth;uOff+=4)
				{
				pPtr[uOff] &= 0x8F;
				}

			m_usCalcCrc = TlsCrcCalc(pPtr,uOutWidth,m_usCalcCrc);

			if(pFile->Write(pPtr,uOutWidth)!=(int)uOutWidth)throw ETlsFileWriteError();
			}
		}

	if(uPicCount>1)										// Sequenzinfos speichern
		{
		   iCnt =pFile->Write(&m_sSequence,sizeof(m_sSequence));
		if(iCnt!=sizeof(m_sSequence))throw ETlsFileWriteError();

		m_usCalcCrc = TlsCrcCalc(&m_sSequence,sizeof(m_sSequence),m_usCalcCrc);
		}


	if(uPicPos+1>=uPicCount)							// CRC speichern
		{
		pFile->Seek(uStart);
		m_sFileHeader.usCrc = m_usCalcCrc;

		   iCnt =pFile->Write(&m_sFileHeader,sizeof(m_sFileHeader));
		if(iCnt!=sizeof(m_sFileHeader))throw ETlsFileWriteError();

		CallProgressFunction(100);
		}


}


