//*****************************************************************************
//*
//*
//*		FileUtilityLinux.c
//*
//*
//*****************************************************************************
//
//  This is a library fore file and directory access for LINUX,
//
//	Global Macros:
//
//		GLOBAL_CONFIG			Enable the configuration via GlobalConfig.h header file (1=on 0=off)
//
//  Copyright  2018    Anton Zechner
//
//
#ifdef		GLOBAL_CONFIG
#define 	FILEUTILITYLINUX_C
#include	<GlobalConfig.h>
#undef		FILEUTILITYLINUX_C
#endif

#ifndef		_LARGE_FILES
#define 	_LARGE_FILES		1
#endif
#ifndef		__USE_LARGEFILE64
#define 	__USE_LARGEFILE64
#endif
#ifndef		_LARGEFILE64_SOURCE
#define 	_LARGEFILE64_SOURCE
#endif
#ifndef		_LARGEFILE_SOURCE
#define 	_LARGEFILE_SOURCE
#endif


#include	<time.h>
#include	<errno.h>
#include    <fcntl.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<dirent.h>
#include	<unistd.h>
#include	<memory.h>
#include	<sys/stat.h>
#include	<sys/time.h>
#include	<sys/types.h>
#include	"FileUtility.h"

#ifndef		F_GETPATH
#define 	F_GETPATH	(1024+7)
#endif		

#ifndef		FIND_LINKS
#define		FIND_LINKS	1							// resolve links in find functions
#endif
#ifndef		FIND_MAX
#define		FIND_MAX	8							// maximal count of find handles
#endif
#ifndef		FIND_LEN
#define 	FIND_LEN    256							// maximal path size in find handles
#endif
#define		FIND_ONE	0xFFFFFFFD					// handle for single file

typedef struct
   {
   int		    		iFree;						// is the handle in use
   int					iLen;						// is the length of the find path
   int 					iNum;						// is the length of the seach name
   DIR				   *pDir;						// is the pointer to the directory entry
   char			    	cFind[FIND_LEN+4];			// is the seach name with wildchards
   char					cName[FIND_LEN+4];			// is the path name of the search
   }FindMem;

static	FindMem			aFindMem[FIND_MAX];
static	int				iFindInit=0;


//*****************************************************************************
//*
//*		SysError
//*
//*****************************************************************************
//	Converts an Linunx ErrorCode to a System-Error-Code
static int SysError()
{


	switch(errno)
		{
	case EPERM			: return SYS_ERR_ACCESS_DENIED;
	case EBADF			: return SYS_ERR_ACCESS_DENIED;
	case ENOENT			: return SYS_ERR_FILE_NOT_FOUND;
	case ENXIO			: return SYS_ERR_FILE_NOT_FOUND;
	case EACCES			: return SYS_ERR_ACCESS_DENIED;
	case ENOTBLK		: return SYS_ERR_ACCESS_DENIED;
	case EEXIST			: return SYS_ERR_ALREADY_EXISTS;
	case EXDEV			: return SYS_ERR_FILE_NOT_FOUND;
	case ENODEV			: return SYS_ERR_FILE_NOT_FOUND;
	case ENOTDIR		: return SYS_ERR_PATH_NOT_FOUND;
	case EINVAL			: return SYS_ERR_PARAM;
	case ENFILE			: return SYS_ERR_NO_HANDLES;
	case EMFILE			: return SYS_ERR_NO_HANDLES;
	case EFBIG			: return SYS_ERR_DISKFULL;
	case ENOSPC			: return SYS_ERR_DISKFULL;
	case EROFS			: return SYS_ERR_ACCESS_DENIED;
	case ENAMETOOLONG	: return SYS_ERR_INVALID_FILENAME;
	case ENOSTR			: return SYS_ERR_ACCESS_DENIED;
		}



return SYS_ERR_UNKNOWN;
}

//*****************************************************************************
//*
//*		SysClearFindStruct
//*
//*****************************************************************************
static __inline void SysClearFindStruct(SysFindStruct *pFind)
{
	pFind->dwSize=0;
	pFind->cName[0]=0;
}


//*****************************************************************************
//*
//*		SysResolveLink
//*
//*****************************************************************************
//	resolves an sympolic link 
//	pFind		: pointer to file structer of thje symbolic link
//	pPath		: directory path symbolic link 
//	uLen		: length of the directory path 
//	returns 0 if a link was found or an error code below zero.
#if FIND_LINKS
static int SysResolveLink(SysFindStruct *pFind,const char *pPath,unsigned uLen)
{
char 			cPath[256];
struct stat64 	sStat;
int				iAttr;
int 			iMode;
unsigned		uNum;



	if(uLen>=sizeof(cPath)-8)							// is the path too long ?
		{
		return SYS_ERR_PARAM;
		}
		
	if(!uLen)
		{
		uLen	= 2;
		pPath	= "./";
		}
	
	uNum = strlen(pFind->cName);
	
	if(uLen+uNum>=sizeof(cPath)-1)						// is the name too long ?
		{
		return SYS_ERR_PARAM;
		}

	memcpy(cPath,pPath,uLen);
	
	if(cPath[uLen-1]!='/')
		{
		cPath[uLen] = '/';
		uLen++;
		}
		
	memcpy(cPath+uLen,pFind->cName,uNum+1);
	
	if(stat64(cPath,&sStat)==-1)						// get file data
		{
		return SysError();
		}

	pFind->dwTimeCreate = (SysUtc)sStat.st_ctime;
	pFind->dwTimeAccess = (SysUtc)sStat.st_atime;
	pFind->dwTimeWrite  = (SysUtc)sStat.st_mtime;
	pFind->dwSize       = (SysDW )sStat.st_size;

	iAttr  = (sStat.st_size>>(32-SYS_HIGHSHIFT))&SYS_HIGHMASK;
	iAttr |= (sStat.st_mode&~S_IFMT)<<16;
	iMode  = (sStat.st_mode& S_IFMT);

	if(sStat.st_mode&S_IWUSR)iAttr|=SYS_WRITE;
	if(sStat.st_mode&S_IRUSR)iAttr|=SYS_READ;
	if(sStat.st_mode&S_IXUSR)iAttr|=SYS_EXEC;
	if(iMode       ==S_IFDIR)iAttr|=SYS_DIR;

	pFind->dwAttr = iAttr|(S_IFLNK<<16);


return 0;
}

#else
#define		SysResolveLink(pFind,pPath,uLen)
#endif		


//*****************************************************************************
//*
//*		SysFileExist
//*
//*****************************************************************************
//	Checks if a file exists
//	pFilename	: pointer to file cName
//	Returns	: 	0 = file din't exist
//				1 = file exists
//				2 =	file exists and is a directory
int SysFileExist(const char *pFilename)
{
struct stat 	sStat;


	if(stat(pFilename,&sStat)==-1)
		{
		return 0;
		}

	if((sStat.st_mode&S_IFMT)==S_IFDIR)
		{
		return 2;
		}


return 1;
}


//*****************************************************************************
//*
//*		SysGetAttributes
//*
//*****************************************************************************
//	get the the attribute from a file
//	pFilename	: pointer to file cName
//	return		: <0 = error
//				  SYS_READ
//				  SYS_WRITE
//				  SYS_DIR
//				  SYS_HIDDEN
//				  SYS_ARCHIVE
int SysGetAttributes(const char *pFilename)
{
struct stat 	sStat;
int 			iMode;
int				iRet;


	if(stat(pFilename,&sStat)==-1)
		{
		return SysError();
		}

	iRet  = sStat.st_mode<<16;
	iMode = sStat.st_mode&S_IFMT;
	
	if(sStat.st_mode&S_IWUSR)iRet|=SYS_WRITE;
	if(sStat.st_mode&S_IRUSR)iRet|=SYS_READ;
	if(sStat.st_mode&S_IXUSR)iRet|=SYS_EXEC;
	if(iMode       ==S_IFDIR)iRet|=SYS_DIR;


return iRet;
}


//*****************************************************************************
//*
//*		SysSetAttributes
//*
//*****************************************************************************
//	set the the attribute from a file
//	pFilename	: pointer to file cName
//	return		: <0 = error
//				   1 = ok
int SysSetAttributes(const char *pFilename,SysDW dwAttr)
{
struct stat 	sStat;
int 			iAttr;


	if(dwAttr&(SYS_DIR|SYS_DIR))
		{
		return SYS_ERR_PARAM;
		}

	if(stat(pFilename,&sStat)==-1)
		{
		return SysError();
		}

	iAttr  = sStat.st_mode;
	iAttr &= ~(S_IRUSR|S_IWUSR|S_IXUSR);


	if(dwAttr&SYS_WRITE)iAttr|=S_IWUSR;
	if(dwAttr&SYS_READ )iAttr|=S_IRUSR;
	if(dwAttr&SYS_EXEC )iAttr|=S_IXUSR;


	if(chmod(pFilename,iAttr)==-1)
		{
		return SysError();
		}


return SYS_ERR_NONE;
}

//*******************************************************************************
//*
//*		strfcmp
//*
//*******************************************************************************
//	Compares to strings, the second string could contain the wildchar ?.
//	pText		: is the pointer to the first string.
//	pSource		: is the pointer to the second string.
//				  a '?' represents an sigle char.
//	iLen		: is the length of the strings in byte
//	returns			0 if the strings are equal
//		    		1 if the strings are not equal
static int	strfcmp(const char *pText,const char *pSource,int iLen)
{
register int	i;
register char	a,b;


	for(i=0;i<iLen;i++)								// compare strings
		{
		b=pSource[i];
		if(b=='?')continue;
		a=pText[i];
		if(a>='a' && a<='z')a+='A'-'a';
		if(b>='a' && b<='z')b+='A'-'a';
		if(a!=b)return 	1;
		}


return 0;
}

//*******************************************************************************
//*
//*		strnxcmp
//*
//*******************************************************************************
//	Compares to strings, the second string could contain the wildchars * and ?
//	pText		: is the pointer to the first string.
//	iTextLen	: is the length of the first strings in byte
//	pSource		: is the pointer to the second string.
//				  a '?' represents an sigle char.
//				  a '*' represents several chars.
//	iLen		: is the length of the second strings in byte
//	returns			0 if the strings are equal
//		    		1 if the strings are not equal
static int strnxcmp(const char *pText,int iTextLen,const char *pSource,int iSourceLen)
{
int		iTextPos,iFindPos;
int		i,iLen,iStar;



	if(iSourceLen<=0)return 0;


	iFindPos=0;
	iTextPos =0;


	for(;;) 										// Suchen Schleife
		{
		if(iSourceLen<=0)break;
		if(  iTextLen<=0)break;

		iLen=0;
		while(iLen<iSourceLen)						// Wort Lnge holen
			{
			if(pSource[iLen]=='*')break;
			iLen++;
			}
		if(iTextLen<iLen)return 1;
		if(iLen)
			{
			if(strfcmp(pText,pSource,iLen))return 1;
			iSourceLen-=iLen;pSource+=iLen;
			  iTextLen-=iLen;  pText+=iLen;
			continue;
			}

		iStar=0;
		while(iSourceLen>0)							// Wildcard berspringen
			{
			if(*pSource=='?'){iStar++;pSource++;iSourceLen--;continue;}
			if(*pSource!='*')break;
			    pSource++;iSourceLen--;
			}
		if(  iTextLen<iStar)return 1;
		if(iSourceLen<=  0 )return 0;

		iTextLen -= iStar;
		pText    += iStar;
		iLen      = 0;

		while(iLen<iSourceLen)						// Wort Lnge holen
			{
			if(pSource[iLen]=='*')break;
			iLen++;
			}

		while(iTextLen>=iLen)						// Source Wortteil
			{
			if(!strfcmp(pText,pSource,iLen))
				{
				if(iSourceLen==iLen)return iTextLen!=iLen;
				i=strnxcmp( pText  +iLen,  iTextLen -iLen,
							pSource+iLen,iSourceLen -iLen);

				if(i==0)return i;
				}

			iTextLen--;
			pText++;
			}


		return 1;
		}

/********************** Suchen_len <= 0 **************************************/

	if(iSourceLen>0)								// Stern am Wortende
		{
		while(iSourceLen>0)
			{
			if(*pSource!='*')return 1;
			iSourceLen--;
			pSource--;
			}
		return 0;
		}

	if(iTextLen>0)return 1;

return 0;
}

//*****************************************************************************
//*
//*		SysFindFirstEx
//*
//*****************************************************************************
//	search for a file
//	pFilename	: find path (with '*' and '?' )
//	pFind		: pointer to file structer
//	pError		: if this pointer isn't zero, here the error number will be sored
//	returns a find handle or 0 if an error occurs
SysFind	SysFindFirstEx(const char *pFilename,SysFindStruct *pFind,int *pError)
{
unsigned 	uNr;
DIR		   *pDir;
struct stat sStat;
SysFind		hHandle;
int 		iAttr;
int 		iFlag;
int 		iMode;
int			iLen;
int			iNum;
int			iErr;
int			i;


	if(!iFindInit)									// init all
	   {
	   memset(aFindMem,0,sizeof(aFindMem));
	   iFindInit=1;
	   }

	if(!pFind)
		{
		if(pError)*pError=SYS_ERR_PARAM;
		SysClearFindStruct(pFind);
		return 0;
		}

	for(uNr=0;uNr<FIND_MAX;uNr++)					// find empty handle
		{
		if(aFindMem[uNr].iFree==0)break;
		}

	if(uNr>=FIND_MAX)
		{
		if(pError)*pError=SYS_ERR_NO_HANDLES;
		SysClearFindStruct(pFind);
		return 0;
		}

	aFindMem[uNr].iFree=1;

/*****************************************************************************/

	iLen = strlen(pFilename);

	for(i=iLen,iFlag=0;;)							// find start of path
		{
		if(i<=0)break;
		i--;
		if(pFilename[i]=='/')break;
		if(pFilename[i]!='*')
		if(pFilename[i]!='?')continue;
		iFlag = 1;
		}

	if(!iFlag)										// no wildchars
		{
		aFindMem[uNr].iFree=0;

		if(iLen-i>=sizeof(pFind->cName))
			{
			if(pError)*pError=SYS_ERR_PARAM;
			SysClearFindStruct(pFind);
			return 0;
			}

		if(stat(pFilename,&sStat)==-1)				// exists this file
			{
			if(pError)*pError=SysError();
			SysClearFindStruct(pFind);
			return 0;
			}

		if(pFilename[i]=='/')i++;
		
		iNum = iLen-i;
		
		if(iNum>=sizeof(pFind->cName))				// is the name too long ?
			{
			if(pError)*pError=SYS_ERR_PARAM;
			SysClearFindStruct(pFind);
			return 0;
			}
		
		memcpy(pFind->cName,pFilename+i,iNum);

		pFind->cName[iNum]  = 0;
		pFind->dwTimeCreate = (SysUtc)sStat.st_ctime;
		pFind->dwTimeAccess = (SysUtc)sStat.st_atime;
		pFind->dwTimeWrite  = (SysUtc)sStat.st_mtime;
		pFind->dwSize       = (SysDW )sStat.st_size;

		iAttr  = (sStat.st_size>>(32-SYS_HIGHSHIFT))&SYS_HIGHMASK;
		iAttr |= (sStat.st_mode<<16);
		iMode  = (sStat.st_mode&S_IFMT);

		if(sStat.st_mode&S_IWUSR)iAttr|=SYS_WRITE;
		if(sStat.st_mode&S_IRUSR)iAttr|=SYS_READ;
		if(sStat.st_mode&S_IXUSR)iAttr|=SYS_EXEC;
		if(iMode       ==S_IFDIR)iAttr|=SYS_DIR;

		pFind->dwAttr = iAttr;

		if(iMode==S_IFLNK)							// is it a symbolic link
			{
			SysResolveLink(pFind,pFilename,i);
			}

		return (SysFind)(FIND_ONE+1);
		}


	if(i>FIND_LEN)									// is the path too long ?
		{
		if(pError)*pError=SYS_ERR_PARAM;
		SysClearFindStruct(pFind);
		aFindMem[uNr].iFree=0;
		return 0;
		}

	if(pFilename[i]=='/')i++;

	memcpy(aFindMem[uNr].cName,pFilename,i);
	       aFindMem[uNr].cName[i]=0;
	       aFindMem[uNr].iLen =i;

	if(iLen-i>FIND_LEN)								// is the name too long ?
	   {
	   if(pError)*pError=SYS_ERR_PARAM;
	   SysClearFindStruct(pFind);
	   aFindMem[uNr].iFree=0;
	   return 0;
	   }

	memcpy(aFindMem[uNr].cFind,pFilename+i,iLen-i+1);
		   aFindMem[uNr].iNum = iLen-i;

	if(!aFindMem[uNr].cName[0])						// is it the local folder
		{
		aFindMem[uNr].cName[0] = '.';
		aFindMem[uNr].cName[1] = 0;
		aFindMem[uNr].iLen     = 1;
		}

	   pDir=opendir(aFindMem[uNr].cName);			// open directory entries
	if(pDir==NULL)
		{
		if(pError)*pError=SYS_ERR_PATH_NOT_FOUND;
		SysClearFindStruct(pFind);
		aFindMem[uNr].iFree=0;
		return 0;
		}

	i = aFindMem[uNr].iLen;
	aFindMem[uNr].cName[i]='/';i++;
	aFindMem[uNr].cName[i]=0;
	aFindMem[uNr].iLen++;
	aFindMem[uNr].pDir=pDir;

	hHandle = (SysFind)((intptr_t)(uNr+1));
	iErr    = SysFindNext(hHandle,pFind);

	if(iErr)										// seach to first filename
		{
		if(pError)*pError=iErr;
		SysFindClose(hHandle);
		return 0;
		}

	if(pError)*pError=SYS_ERR_NONE;


return hHandle;
}


//*****************************************************************************
//*
//*		SysFindNext
//*
//*****************************************************************************
//	search for the next file
//	hHandle	: find handle
//	pFind	: pointer to file structer
//	returns 0 if a file was found or an error code below zero
int SysFindNext(SysFind	hHandle,SysFindStruct *pFind)
{
struct dirent  *pResult;
struct dirent   sEntry;
struct stat		sStat;
int				iAttr;
int				iMode;
int				iLen;
int				iNum;
char		   *pStr;
DIR			   *pDir;
unsigned 		uNr;



	uNr = (unsigned)((intptr_t)hHandle)-1;

	if(uNr>=FIND_MAX || aFindMem[uNr].iFree==0)
		{
		SysClearFindStruct(pFind);
		return (uNr==FIND_ONE)? SYS_ERR_FILE_NOT_FOUND:SYS_ERR_PARAM;
		}

	pDir=aFindMem[uNr].pDir;
	iLen=aFindMem[uNr].iLen;
	pStr=aFindMem[uNr].cName;

	for(;;)
		{
		readdir_r(pDir,&sEntry,&pResult);			// get next directory entry

		if(!pResult)
			{
			SysClearFindStruct(pFind);
			return SYS_ERR_FILE_NOT_FOUND;
			}


		iNum=strlen(sEntry.d_name);
		if(strnxcmp(sEntry.d_name,iNum,aFindMem[uNr].cFind,aFindMem[uNr].iNum))continue;
		if(iLen+iNum>FIND_LEN)continue;

		memcpy(pStr+iLen,sEntry.d_name,iNum+1);
		if(stat(pStr,&sStat)==-1)continue;

		strcpy(pFind->cName,sEntry.d_name);

		pFind->dwTimeCreate = (SysUtc)sStat.st_ctime;
		pFind->dwTimeAccess = (SysUtc)sStat.st_atime;
		pFind->dwTimeWrite  = (SysUtc)sStat.st_mtime;
		pFind->dwSize       = (SysDW )sStat.st_size;

		iAttr  = (sStat.st_size>>(32-SYS_HIGHSHIFT))&SYS_HIGHMASK;
		iAttr |= (sStat.st_mode<<16);
		iMode  = (sStat.st_mode&S_IFMT);

		if(sStat.st_mode&S_IWUSR)iAttr|=SYS_WRITE;
		if(sStat.st_mode&S_IRUSR)iAttr|=SYS_READ;
		if(sStat.st_mode&S_IXUSR)iAttr|=SYS_EXEC;
		if(iMode       ==S_IFDIR)iAttr|=SYS_DIR;

		pFind->dwAttr = iAttr;

		if(iMode==S_IFLNK)							// is it a symbolic link
			{
			SysResolveLink(pFind,pStr,iLen);
			}

		break;
		}


return 0;
}


//*****************************************************************************
//*
//*		SysFindClose
//*
//*****************************************************************************
//	Closes a find handle
//	hHandle	: find handle
//	Returns 0 if all ok or 0< if an error occurrs
int SysFindClose(SysFind hHandle)
{
unsigned 	uNr;


	uNr = (unsigned)((intptr_t)hHandle)-1;

	if(uNr>=FIND_MAX)
		{
		return (uNr==FIND_ONE)? SYS_ERR_FILE_NOT_FOUND:SYS_ERR_PARAM;
		}

	if(aFindMem[uNr].iFree==0)
		{
		return SYS_ERR_PARAM;
		}


	closedir(aFindMem[uNr].pDir);
	aFindMem[uNr].iFree=0;


return 0;
}

//*****************************************************************************
//*
//*		SysMkdir
//*
//*****************************************************************************
//	create a new directory
//	pFilename	: is the cName of the directory
//	return 0 if all ok or an error number below zero
int	SysMkdir(const char *pFilename)
{

	if(mkdir(pFilename,0777)!=-1)return 0;


return SysError();
}


//*****************************************************************************
//*
//*		SysRmdir
//*
//*****************************************************************************
//	removes a directory
//	pFilename	: is the cName of the directory
//	return 0 if all ok or an error number below zero
int	SysRmdir(const char *pFilename)
{

	if(rmdir(pFilename)!=-1)return 0;
	if(errno==ENOENT)return SYS_ERR_PATH_NOT_FOUND;

return SysError();
}


//*****************************************************************************
//*
//*		SysDelete
//*
//*****************************************************************************
//	delets a file
//	pFilename	: is the cName of the file
//	return 0 if all ok or an error number below zero
int	SysDelete(const char *pFilename)
{

	if(remove(pFilename)!=-1)return 0;


return SysError();
}



//*****************************************************************************
//*
//*		SysRename
//*
//*****************************************************************************
//	renames file
//	pFilename	: is the cName of the file
//	return 0 if all ok or an error number below zero
int SysRename(const char *pFilename,const char *pFilenameNew)
{

	if(rename(pFilename,pFilenameNew)!=-1)return 0;


return SysError();
}


//*****************************************************************************
//*
//*		SysFileLength
//*
//*****************************************************************************
//	get the length of a file
//	pFilename	: is the cName of the file
//	return the file length or an error number below zero
int SysFileLength(const char *pFilename)
{
struct stat 	sStat;


	if(stat(pFilename,&sStat)==-1)
		{
		return SysError();
		}


return sStat.st_size;
}

//*****************************************************************************
//*
//*		SysFileLengthEx
//*
//*****************************************************************************
//	get the length of a file (64 bit version)
//	pFilename	: is the name of the file
//	return the file length or an error number below zero
SysPos SysFileLengthEx(const char *pFilename)
{
struct stat64 	sStat;


	if(stat64(pFilename,&sStat)==-1)
		{
		return SysError();
		}


return sStat.st_size;
}


//*****************************************************************************
//*
//*		SysFileInfo
//*
//*****************************************************************************
//	gets informations of a file
//	pFilename	: is the name of the file
//	pInfo		: here the file info will be stored
//	returns 0 if ok or an error number below zero
int SysFileInfo(const char *pFilename,SysFindStruct *pInfo)
{
struct stat64 	sStat;
int				iAttr;
int 			iMode;


	if(!pInfo)return SYS_ERR_PARAM;

	if(stat64(pFilename,&sStat)==-1)
		{
		return SysError();
		}

	pInfo->cName[0] 	= 0;

	pInfo->dwTimeCreate = (SysUtc)sStat.st_ctime;
	pInfo->dwTimeAccess = (SysUtc)sStat.st_atime;
	pInfo->dwTimeWrite  = (SysUtc)sStat.st_mtime;
	pInfo->dwSize       = (SysDW )sStat.st_size;

	iAttr  = (sStat.st_size>>(32-SYS_HIGHSHIFT))&SYS_HIGHMASK;
	iAttr |= (sStat.st_mode<<16);
	iMode  = sStat.st_mode&S_IFMT;

	if(sStat.st_mode&S_IWUSR)iAttr|=SYS_WRITE;
	if(sStat.st_mode&S_IRUSR)iAttr|=SYS_READ;
	if(sStat.st_mode&S_IXUSR)iAttr|=SYS_EXEC;
	if(iMode       ==S_IFDIR)iAttr|=SYS_DIR;

	pInfo->dwAttr = iAttr;


return SYS_ERR_NONE;
}


//*****************************************************************************
//*
//*		SysFullPath
//*
//*****************************************************************************
//	converts a relative path to a full path
//	pPath		: is the relative path
//	pBuffer		: is the buffer for the full path
//	iMax		: is the size of the buffer
//	returns the length in bytes of the full path cName.
int SysFullPath(const char *pPath,char *pBuffer,int iMax)
{
char	*pStr;
int		 iLen,iSlash;
int		 iPos=0,iAdd;



	if(iMax<=0)return 0;

	pBuffer[0]=0;
	iLen=strlen(pPath);
	if(iLen+1>=iMax)return 0;

	if(pPath[0]!='/')
		{
		pStr=getcwd(pBuffer,iMax);
	    if(pStr==NULL)return 0;
    	iPos=strlen(pBuffer);
		if(iPos+1+iLen>=iMax)return 0;
		pBuffer[iPos]='/';
		iPos++;
		memcpy(pBuffer+iPos,pPath,iLen+1);
		iLen+=iPos;
		}
	else{
		memcpy(pBuffer,pPath,iLen+1);
		iPos=1;
		}


	while(iPos<iLen)
		{
		if(iPos>0 && pBuffer[iPos]=='/')				// filter "//"
			{
			memmove(pBuffer+iPos,pBuffer+iPos+1,iLen-iPos);
			iLen-=1;
			continue;
			}
														// filter "./"
		if(iPos>0 && pBuffer[iPos]=='.' && pBuffer[iPos+1]=='/')
			{
			memmove(pBuffer+iPos,pBuffer+iPos+2,iLen-iPos-1);
			iLen-=2;
			continue;
			}
														// filter "."
		if(iPos>1 && pBuffer[iPos]=='.' && pBuffer[iPos+1]==0)
			{
			iPos--;
			pBuffer[iPos]=0;
			iLen=iPos;
			break;
			}

														// filter "../"
		if(iPos>0 && pBuffer[iPos]=='.' && pBuffer[iPos+1]=='.' && pBuffer[iPos+2]=='/')
			{
			iSlash=iPos;

			for(iPos--;iPos>0;iPos--)
				{
				if(pBuffer[iPos-1]=='/')break;
				}

			iAdd = iSlash-iPos+3;

			memmove(pBuffer+iPos,pBuffer+iSlash+3,iLen-iSlash-2);
			iLen-=iAdd;
			continue;
			}
														// filter ".."
		if(iPos>1 && pBuffer[iPos]=='.' && pBuffer[iPos+1]=='.' && pBuffer[iPos+2]==0)
			{
			for(iPos-=2;iPos>0;iPos--)
				{
				if(pBuffer[iPos]=='/')break;
				}

			if(iPos<=0)
				{
				pBuffer[0]='/';
				iPos=1;
				}

			pBuffer[iPos]=0;
			iLen=iPos;
			break;
			}

		for(;pBuffer[iPos];iPos++)						// to next slash
			{
			if(pBuffer[iPos]!='/')continue;
			iPos++;
			break;
			}
		}



return iLen;
}

//******************************************************************************
//*
//*		SysFileOpenEx
//*
//******************************************************************************
//	ffnet eine Datei.
//	pFilename		: Ist der Name der Datei.
//	iMode			: Bestimmt wie die Datei geffnet werden soll.
//						SYS_READ	= Datei zum Lesen ffnen
//						SYS_WRITE	= Datei zum Schreiben ffnen und nicht auf 0 krzen
//						SYS_CREATE	= Datei zum Schreiben ffnen und auf 0 krzen
//	pError			: Hier wird die Fehlernummer gespeichert wenn der Zeiger
//					  ungleich NULL ist.
//	Ergibt ein File-Handle bzw. 0 bei einem Fehler
SysFile SysFileOpenEx(const char *pFilename,int iMode,int *pError)
{
FILE		*pFile;
const char 	*pMode;



	pMode = "rb";



	if(iMode&(SYS_WRITE|SYS_CREATE))
		{
		if(iMode&SYS_READ)
				pMode = (iMode&SYS_TRUNC)? "w+b":"r+b";
		else	pMode = (iMode&SYS_TRUNC)? "wb" :"r+b";
		}

	pFile = fopen(pFilename,pMode);

	if(!pFile && (iMode&SYS_CREATE) && !(iMode&SYS_TRUNC))
		{
		pFile = fopen(pFilename,(iMode&SYS_READ)? "w+b":"wb");
		}


    if(pError)
        {
		if(!pFile)
				*pError=SysError();
        else	*pError=SYS_ERR_NONE;
        }


return (SysFile)pFile;
}

//******************************************************************************
//*
//*		SysFileRead
//*
//******************************************************************************
//	Liest Daten aus einer Datei
//	hFile		: File-Handle der Datei
//	pData		: Zeiger auf den Lesepuffer.
//	iSize		: Ist die Anzahl der zuladenden Bytes
//	Ergibt die Anzahl der geladenen Bytes, oder einen Fehlercode kleiner 0.
int SysFileRead(SysFile hFile,void *pData,int iSize)
{
int 	iByte;


	if(hFile==0)return SYS_ERR_PARAM;
	if(iSize==0)return 0;

	iByte = fread(pData,1,iSize,(FILE*)hFile);

	if(iByte<=0 && !feof((FILE*)hFile) && ferror((FILE*)hFile))
		{
		return SysError();
		}



return iByte;
}


//******************************************************************************
//*
//*		SysFileWrite
//*
//******************************************************************************
//	Speichert Daten in einer Datei
//	hFile		: File-Handle der Datei
//	pData		: Zeiger auf den Schreibpuffer.
//	iSize		: Ist die Anzahl der speichernden Bytes
//	Ergibt die Anzahl der geladenen Bytes, oder einen Fehlercode kleiner 0.
int SysFileWrite(SysFile hFile,const void *pData,int iSize)
{
int 	iByte;


	if(hFile==0)return SYS_ERR_PARAM;
	if(iSize<=0)return 0;

	iByte = fwrite(pData,1,iSize,(FILE*)hFile);

	if(iByte<=0)
		{
		return SysError();
		}


return iByte;
}


//******************************************************************************
//*
//*		SysFileClose
//*
//******************************************************************************
//	Schliet eine geffnete Datei.
//	hFile		: File-Handle der Datei
//	Ergbibt 0 wenn die Datei geschlossen werden konnte, bzw einen Fehlercode.
int SysFileClose(SysFile hFile)
{

	if(!hFile)return SYS_ERR_PARAM;

	fclose((FILE*)hFile);


return 0;
}



//******************************************************************************
//*
//*		SysFileSeek
//*
//******************************************************************************
//	Verschiet die Lese/Schreibposition in einer Datei
//	hFile		: File-Handle der Datei
//	iPos		: Neue Lese/Schreibposition
//	iMode		: Bestimmt wie die Position bestimmt werden soll.
//					SYS_SEEK_SET	0	Absolute Position
//					SYS_SEEK_CUR	1	Relativ zur aktuellen Position
//					SYS_SEEK_END	2	Relativ zum Datei Ende
//	Ergibt die neue Position in der Datei.
int SysFileSeek(SysFile hFile,int iPos,int iMode)
{

	if(!hFile)return SYS_ERR_PARAM;

	fseek((FILE*)hFile,iPos,iMode);


return ftell((FILE*)hFile);
}

//*****************************************************************************
//*
//*     SysFileSeekEx
//*
//*****************************************************************************
//	Verschiet die Lese/Schreibposition in einer Datei (64-Bit Version)
//	hFile		: File-Handle der Datei
//	iPos		: Neue Lese/Schreibposition
//	iMode		: Bestimmt wie die Position bestimmt werden soll.
//					SYS_SEEK_SET	0	Absolute Position
//					SYS_SEEK_CUR	1	Relativ zur aktuellen Position
//					SYS_SEEK_END	2	Relativ zum Datei Ende
//	Ergibt die neue Position in der Datei.
SysPos SysFileSeekEx(SysFile hFile,SysPos iPos,int iMode)
{

	if(!hFile)return SYS_ERR_PARAM;

	fseeko64((FILE*)hFile,iPos,iMode);


return ftello64((FILE*)hFile);
}


//******************************************************************************
//*
//*		SysFileTell
//*
//******************************************************************************
//	Verschiet die Lese/Schreibposition in einer Datei
//	hFile		: File-Handle der Datei
//	iPos		: Neue Lese/Schreibposition
//	iMode		: Bestimmt wie die Position bestimmt werden soll.
//					SYS_SEEK_SET	0	Absolute Position
//					SYS_SEEK_CUR	1	Relativ zur aktuellen Position
//					SYS_SEEK_END	2	Relativ zum Datei Ende
//	Ergibt die neue Position in der Datei.
int SysFileTell(SysFile hFile)
{

	if(!hFile)return SYS_ERR_PARAM;

return ftell((FILE*)hFile);
}

//******************************************************************************
//*
//*		SysFileTellEx
//*
//******************************************************************************
//	Verschiet die Lese/Schreibposition in einer Datei  (64-Bit-Version)
//	hFile		: File-Handle der Datei
//	iPos		: Neue Lese/Schreibposition
//	iMode		: Bestimmt wie die Position bestimmt werden soll.
//					SYS_SEEK_SET	0	Absolute Position
//					SYS_SEEK_CUR	1	Relativ zur aktuellen Position
//					SYS_SEEK_END	2	Relativ zum Datei Ende
//	Ergibt die neue Position in der Datei.
SysPos SysFileTellEx(SysFile hFile)
{

	if(!hFile)return SYS_ERR_PARAM;

return ftello64((FILE*)hFile);
}

//******************************************************************************
//*
//*		SysFileFlush
//*
//******************************************************************************
//	Leert alle Zwischenpuffer einer Datei.
//	hFile		: File-Handle der Datei
//	Ergbibt 1 wenn die Puffer gellert werden konnten, bzw. 0 bei einem Fehler.
int SysFileFlush(SysFile hFile)
{

	if(!hFile)return 0;

	fflush((FILE*)hFile);


return 1;
}



//*****************************************************************************
//*
//*		SysSetCurrentPath
//*
//*****************************************************************************
//	Stellt fr den Prozess das aktuelle Verzeichnis ein
//	pFilename	: Ist der Zeiger auf den Verzeichnisnamen.
//				  Der Pfad kann auch '/' am Ende haben.
//	Ergibt 0 wenn ok, bzw eine Fehlernummer.
int SysSetCurrentPath(const char *pFilename)
{
char	cPath[260];
int		iLen;


	   iLen = SysFullPath(pFilename,cPath,sizeof(cPath));
	if(iLen<=0)return SysError();

	if(iLen>0 && (cPath[iLen-1]=='\\' || cPath[iLen-1]=='/'))
		{
		cPath[iLen-1]=0;
		}

	if(chdir(cPath))
		{
		return SysError();
		}


return SYS_ERR_NONE;
}


//*****************************************************************************
//*
//*		SysFileCopy
//*
//*****************************************************************************
//	Copyies a file from pSource to pTarget.
//	pSource			: Path of the source file
//	pTarget			: Path of the target file
//	iFailIfExists	: fails if the target file exists
//	Rertuns the file size, or an error number below zero.
int SysFileCopy(const char *pSource,const char *pTarget,int iFailIfExists)
{
char		cBuffer[4096];
SysFile		hSource;
SysFile		hTarget;
int			iError;
int			iPos;
int			iLen;
int			iNum;


	if(iFailIfExists && SysFileExist(pTarget)!=0)	// exists the target
		{
		return SYS_ERR_ALREADY_EXISTS;
		}

													// open files
		hSource = SysFileOpenEx(pSource,SYS_READ,&iError);
	if(!hSource)return iError;

		hTarget = SysFileOpenEx(pTarget,SYS_CREATE|SYS_WRITE,&iError);
	if(!hTarget)
		{
		SysFileClose(hSource);
		return iError;
		}


	iLen = sizeof(cBuffer);

	for(iPos=0;iLen==sizeof(cBuffer);iPos+=iLen)	// copy data
		{
		   iLen = SysFileRead(hSource,cBuffer,sizeof(cBuffer));
		if(iLen<=0){iError=iLen;break;}

		   iNum = SysFileWrite(hTarget,cBuffer,iLen);
		if(iNum<=0){iError=iNum;break;}
		if(iNum!=iLen)
			{
			iError = SYS_ERR_DISKFULL;
			break;
			}
		}


	SysFileClose(hSource);
	SysFileClose(hTarget);

	if(iError<0)									// all ok
		{
		SysDelete(pTarget);
		return iError;
		}


return iPos;
}


//*****************************************************************************
//*
//*     SysGetFileTime
//*
//*****************************************************************************
//  Gets all times of a opened file
//	hFile		: File-Handle der Datei
//	pWrite		: here the write time is stored (NULL means not storing)
//	pCreate		: here the create time is stored (NULL means not storing)
//	pAccess		: here the last access time is stored (NULL means not storing)
//	Returns 0 if ok, or an error value below zero.
int SysGetFileTime(SysFile hFile,SysDW *pWrite,SysDW *pCreate,SysDW *pAccess)
{
struct stat sStat;
int     	iFile;
int     	iOk;


	
       iFile = fileno((FILE*)hFile);
	if(iFile<0)
		{
		return SYS_ERR_PARAM;	
		}

       iOk = fstat(iFile,&sStat);
    if(iOk<0)
        {
        return SysError();
        }


    *pWrite  = (SysUtc)sStat.st_mtime;
    *pCreate = (SysUtc)sStat.st_ctime;
    *pAccess = (SysUtc)sStat.st_atime;


return 0;
}


//*****************************************************************************
//*
//*     SysSetFileTime
//*
//*****************************************************************************
//  Sets all times of an opened file. Attension, this works only if no data was written.
//	hFile		: File-Handle der Datei
//	dwWrite		: it the new write time (0 means not changing)
//	dwCreate	: it the new create time (0 means not changing)
//	dwAccess	: it the new last access time (0 means not changing)
//	Returns 0 if ok, or an error value below zero.
int SysSetFileTime(SysFile hFile,SysDW dwWrite,SysDW dwCreate,SysDW dwAccess)
{
struct timeval  sTv[2];
struct stat     sStat;
int             iFile;
int             iOk;


       iFile = fileno((FILE*)hFile);
	if(iFile<0)
		{
		return SYS_ERR_PARAM;	
		}

    sTv[0].tv_sec   = dwAccess;
    sTv[1].tv_sec   = dwWrite;

    if(!dwWrite || !dwAccess)                           // Alte Werte laden
        {
           iOk = fstat(iFile,&sStat);
        if(iOk<0)
            {
            return SysError();
            }

        if(!dwAccess)sTv[0].tv_sec  = sStat.st_atime;
        if(!dwWrite )sTv[1].tv_sec  = sStat.st_mtime;
        }


    sTv[0].tv_usec  = 0;
    sTv[1].tv_usec  = 0;


       iOk = futimes(iFile,sTv);
    if(iOk<0)
        {
        return SysError();
        }


return 0;
}



//******************************************************************************
//*
//*		SysGetFileSize
//*
//******************************************************************************
//	Gets the size of an opend file.
//	hFile		: handle of the file
//	Returns the size of the file, or an error value below zero.
int SysGetFileSize(SysFile hFile)
{
int 	iSize;
int 	iPos;


	if(!hFile)return SYS_ERR_PARAM;

	iPos  = ftell((FILE*)hFile);
	
	fseek((FILE*)hFile,0,SEEK_END);
	
	iSize = ftell((FILE*)hFile);
	
	fseek((FILE*)hFile,iPos,SEEK_SET);


return iSize;
}

//******************************************************************************
//*
//*		SysSetFileSize
//*
//******************************************************************************
//	Sets the size of an opend file.
//	hFile		: handle of the file
//	uSize		: is the new file size
//	Returns the size of the file, or an error value below zero.
int SysSetFileSize(SysFile hFile,SysDW uSize)
{
int     iFile;
int     iOk;


       iFile = fileno((FILE*)hFile);
	if(iFile<0)
		{
		return SYS_ERR_PARAM;	
		}

       iOk=ftruncate(iFile,uSize);
    if(iOk<0)
		{
		return SysError();
		}


return uSize;
}

//******************************************************************************
//*
//*		SysGetFileInfo
//*
//******************************************************************************
//  Get informations of an opened file.
//	hFile	: handle of the file
//  pFind   : pointer to file structer
//  returns 0 if a file was found or an error code below zero
int SysGetFileInfo(SysFile hFile,SysFindStruct *pFind)
{
char			cPath[256];
struct stat     sStat;
int 			iFile;
int				iAttr;
int				iMode;
int 			iPos;
int 			iOk;


       iFile = fileno((FILE*)hFile);
	if(iFile<0)
		{
		return SYS_ERR_PARAM;	
		}
	
       iOk = fstat(iFile,&sStat);
    if(iOk<0)
        {
        return SysError();
        }

	if(fcntl(iFile,F_GETPATH,cPath)<0)
		{
		pFind->cName[0] = 0;
		}	
	else{	
		iPos = strlen(cPath);
		
		for(;iPos>0;iPos--)
			{
			if(cPath[iPos-1]=='/')break;
			}
		
		strcpy(pFind->cName,cPath+iPos);
		}

	pFind->dwTimeCreate = (SysUtc)sStat.st_ctime;
	pFind->dwTimeAccess = (SysUtc)sStat.st_atime;
	pFind->dwTimeWrite  = (SysUtc)sStat.st_mtime;
	pFind->dwSize       = (SysDW )sStat.st_size;

	iAttr  = (sStat.st_size>>(32-SYS_HIGHSHIFT))&SYS_HIGHMASK;
	iAttr |= (sStat.st_mode)<<16;
	iMode  = (sStat.st_mode& S_IFMT);

	if(sStat.st_mode&S_IWUSR)iAttr|=SYS_WRITE;
	if(sStat.st_mode&S_IRUSR)iAttr|=SYS_READ;
	if(sStat.st_mode&S_IXUSR)iAttr|=SYS_EXEC;
	if(iMode       ==S_IFDIR)iAttr|=SYS_DIR;

	pFind->dwAttr = iAttr;
	


return 0;
}

