//*****************************************************************************
//*
//*
//*		Mailslots.c
//*
//*
//*****************************************************************************
//
//  This is a library for a mailslot support for WINDOWS.
//	Mailslots are thread save FIFO buffers
//	The elenent size and FIFO size could conigured individually.
//	Note: MailslotCreate use malloc, use this function only at startup
//
//	Global Macros:
//
//		GLOBAL_CONFIG		Enable the configuration via GlobalConfig.h header file (1=on 0=off)
//		MAX_MAILSLOTS		Maximum count of mailslots
//
//
//  Copyright  2018    Anton Zechner
//
//
#include	<windows.h>
#include	<malloc.h>
#include	<memory.h>
#include	"Mailslots.h"

#ifdef		_DEBUG
#ifdef		_MSC_VER
#ifndef		 NO_DEBUG_NEW
#define		_CRTDBG_MAP_ALLOC
#ifdef 		_malloca
#undef 		_malloca
#endif
#include	<stdlib.h>
#include	<crtdbg.h>
#endif
#endif
#endif


#ifndef		MAX_MAILSLOTS
#define		MAX_MAILSLOTS		256
#endif




typedef struct
	{
	HANDLE			hSemEmpty;						// semaphore for empty condion
	HANDLE			hSemFull;						// semaphore for full condion
	HANDLE			hSemEdit;						// semaphore for edit operations
	unsigned		uItemSize;						// size of one item in the buffer
	unsigned 		uSlotSize;						// size of one slot in the buffer
	unsigned		uQueueSize;						// maximum count of items in he fifo
	unsigned		uQueueCount;					// count of items in the buffer
	unsigned		uReadPos;						// current read position
	unsigned		uWritePos;						// current write position
	unsigned char 	ucData[1];						// dummy for data memory
	}MailslotData;


static	MailslotData	   *aSlotPointers[MAX_MAILSLOTS];
static	CRITICAL_SECTION	sLockSection;
static	unsigned			uInitDone= 0;
static	unsigned 			uNextPos = 0;


#define	NUM_SLOT(i)		((unsigned)((unsigned)i)-1)
#define	GET_SLOT(i)		(( NUM_SLOT(i)>=MAX_MAILSLOTS)? 0:aSlotPointers[NUM_SLOT(i)])
#define	CLR_SLOT(i)		if(NUM_SLOT(i)< MAX_MAILSLOTS)    aSlotPointers[NUM_SLOT(i)]=0



//*****************************************************************************
//*
//*		MailslotCreate
//*
//*****************************************************************************
//	This function creates a new mailslot
//	iItemSize	: size of one item
//	iQueueSize	: maximum size of the fifo buffer
//	Returns a handle to the mailslot or zero.
//	Note: MailslotCreate use malloc, use this function only at startup
Mailslot MailslotCreate(int iItemSize,int iQueueSize)
{
MailslotData	*pSlot;
unsigned 		 uSlotSize;
unsigned 		 uPos;


	if(!uInitDone)								// Create a global semaphor
		{
		uInitDone	 = 1;
		InitializeCriticalSection(&sLockSection);
		}

	if(iQueueSize<=0 || iItemSize<=0)return 0;

	EnterCriticalSection(&sLockSection);


	for(uPos=uNextPos;aSlotPointers[uPos];)		// Find empty handle
		{
		   uPos++;
		if(uPos>=MAX_MAILSLOTS)uPos=0;
		if(uPos==uNextPos)
			{
			LeaveCriticalSection(&sLockSection);
			return 0;
			}
		}



	aSlotPointers[uPos] = (MailslotData*)1;

	   uNextPos=uPos+1;
	if(uNextPos>=MAX_MAILSLOTS)uNextPos=0;

	LeaveCriticalSection(&sLockSection);


	uSlotSize = ((iItemSize+3)&~3)+sizeof(unsigned);

		pSlot = (MailslotData*)malloc(sizeof(MailslotData)+iQueueSize*uSlotSize);
	if(!pSlot)
		{
		aSlotPointers[uPos]=0;
		return 0;
		}

	aSlotPointers[uPos] = pSlot;

	pSlot->hSemEmpty	= CreateSemaphoreW(0,0         ,0x70000000,0);
	pSlot->hSemFull		= CreateSemaphoreW(0,iQueueSize,0x70000000,0);
	pSlot->hSemEdit		= CreateSemaphoreW(0,1         ,0x70000000,0);
	pSlot->uItemSize	= iItemSize;
	pSlot->uSlotSize	= uSlotSize;
	pSlot->uQueueSize	= iQueueSize;
	pSlot->uQueueCount	= 0;
	pSlot->uReadPos		= 0;
	pSlot->uWritePos	= 0;




return 	uPos+1;
}


//*****************************************************************************
//*
//*		MailslotDelete
//*
//*****************************************************************************
//	Deletes a mailslot
//	hMailslot	: handle of the mailslot
void MailslotDelete(Mailslot hMailslot)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);


	if(!pSlot)return;

	aSlotPointers[hMailslot-1] = 0;

	CloseHandle(pSlot->hSemEdit );
	CloseHandle(pSlot->hSemFull );
	CloseHandle(pSlot->hSemEmpty);

	free(pSlot);

	CLR_SLOT(hMailslot);

}


//*****************************************************************************
//*
//*		MailslotCount
//*
//*****************************************************************************
//	Calculates the count of messages in a mailslot
//	hMailslot	: handle of the mailslot
//	Returns the count of messages
int MailslotCount(Mailslot hMailslot)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);



	if(!pSlot)return 0;


return pSlot->uQueueCount;
}

//*****************************************************************************
//*
//*		MailslotGetWait
//*
//*****************************************************************************
//	Gets a message from a mailslot, if the fifo is empty
//	the function waits a time for arraiving a message.
//	hMailslot	: handle of the mailslot
//	pData		: pointer to the message memory
//	uTimeout	: maximum wait time in ms
//	Returns the size of the message, or 0 if no message was copied
int	MailslotGetWait(Mailslot hMailslot,void *pData,unsigned uTimeout)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned		 uOffset;
unsigned 		 uSize;



	if(!pSlot)
		{
		if(uTimeout)Sleep(uTimeout);
		return 0;
		}

	if(WaitForSingleObject(pSlot->hSemEmpty,uTimeout)!=WAIT_OBJECT_0)
		{
		return 0;
		}



	WaitForSingleObject(pSlot->hSemEdit,0xFFFFFFFF);

	uOffset = pSlot->uReadPos*pSlot->uSlotSize;
	uSize   = *(unsigned*)(pSlot->ucData+uOffset);

	memcpy(pData,pSlot->ucData+uOffset+sizeof(unsigned),uSize);

	   pSlot->uReadPos++;
	if(pSlot->uReadPos>=pSlot->uQueueSize)pSlot->uReadPos=0;
	   pSlot->uQueueCount--;


	ReleaseSemaphore(pSlot->hSemEdit,1,0);
	ReleaseSemaphore(pSlot->hSemFull,1,0);



return uSize;
}


//*****************************************************************************
//*
//*		MailslotPutWait
//*
//*****************************************************************************
//	Puts a message into a mailslot, if the fifo is full
//	the function waits a time for geting an empty slot.
//	hMailslot	: handle of the mailslot
//	pData		: pointer to the message memory
//	uSize		: size of the message in bytes
//	uTimeout	: maximum wait time in ms
//	Returns the inserted bytes of the message, or 0 if no message was copied
int	MailslotPutWait(Mailslot hMailslot,const void *pData,unsigned uSize,unsigned uTimeout)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned		 uOffset;


	if(!uSize)return 0;
	if(!pSlot)
		{
		if(uTimeout)Sleep(uTimeout);
		return 0;
		}

	if(WaitForSingleObject(pSlot->hSemFull,uTimeout)!=WAIT_OBJECT_0)
		{
		return 0;
		}


	WaitForSingleObject(pSlot->hSemEdit,0xFFFFFFFF);


	if(uSize>pSlot->uItemSize)uSize=pSlot->uItemSize;

	uOffset = pSlot->uWritePos*pSlot->uSlotSize;
	memcpy(pSlot->ucData+uOffset+sizeof(unsigned),pData,uSize);
	*(unsigned*)(pSlot->ucData+uOffset)=uSize;

	   pSlot->uWritePos++;
	if(pSlot->uWritePos>=pSlot->uQueueSize)pSlot->uWritePos=0;
	   pSlot->uQueueCount++;


	ReleaseSemaphore(pSlot->hSemEdit ,1,0);
	ReleaseSemaphore(pSlot->hSemEmpty,1,0);



return uSize;
}




//*****************************************************************************
//*
//*		MailslotPeek
//*
//*****************************************************************************
//	Gets a message from a mailslot, put didn't remove a message
//	if the fifo is empty the function returns zero
//	hMailslot	: handle of the mailslot
//	pData		: pointer to the message memory
//	uPosition	: is the position of the message to take
//	Returns the size of the message, or 0 if no message was copied
int	MailslotPeek(Mailslot hMailslot,void *pData,unsigned uPosition)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned 		 uSize;
unsigned		 uOffset;


	if(!pSlot)return 0;


	WaitForSingleObject(pSlot->hSemEdit,0xFFFFFFFF);


	if(uPosition>=pSlot->uQueueCount)
		{
		ReleaseSemaphore(pSlot->hSemEdit,1,0);
		return 0;
		}


	uOffset  = pSlot->uReadPos;
	uOffset += uPosition;

	if(uOffset>=pSlot->uQueueSize)uOffset -= pSlot->uQueueSize;

	uOffset *= pSlot->uSlotSize;
	uSize    = *(unsigned*)(pSlot->ucData+uOffset);

	memcpy(pData,pSlot->ucData+uOffset+sizeof(unsigned),uSize);

	ReleaseSemaphore(pSlot->hSemEdit,1,0);


return uSize;
}

//*****************************************************************************
//*
//*		MailslotClear
//*
//*****************************************************************************
//	Remove some messages from a mailslot
//	hMailslot	: handle of the mailslot
//	iCount		: count of messgae to remove
//	Returns the count of cleared messages
int MailslotClear(Mailslot hMailslot,int iCount)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned		 uRet=0;


	if(!pSlot)return 0;


	while(iCount>0)
		{
		if(WaitForSingleObject(pSlot->hSemEmpty,0)!=WAIT_OBJECT_0)
			{
			return uRet;
			}


		WaitForSingleObject(pSlot->hSemEdit,0xFFFFFFFF);

		   pSlot->uReadPos++;
		if(pSlot->uReadPos>=pSlot->uQueueSize)pSlot->uReadPos=0;
		   pSlot->uQueueCount--;

		ReleaseSemaphore(pSlot->hSemEdit,1,0);
		ReleaseSemaphore(pSlot->hSemFull,1,0);

		iCount--;
		uRet++;
		}

return uRet;
}


//*****************************************************************************
//*
//*		MailslotInfo
//*
//*****************************************************************************
//	Gets infos of a mailslot
//	hMailslot	: handle of the mailslot
//	pItemSize	: here the size of a singel slot will be stored.
//				  if the pointer is zero the nothing will be stored.
//	pSlotCount	: here the count of all slots will be stored.
//				  if the pointer is zero the nothing will be stored.
//	Returns the count of messages in the mailslot or -1 if an error occurs
int MailslotInfo(Mailslot hMailslot,int *pItemSize,int *pSlotCount)
{
MailslotData   *pSlot=GET_SLOT(hMailslot);
unsigned 		uRet;


	if(!pSlot)
		{
		if(pItemSize )*pItemSize =0;
		if(pSlotCount)*pSlotCount=0;
		return -1;
		}


	WaitForSingleObject(pSlot->hSemEdit,0xFFFFFFFF);

	if(pItemSize )*pItemSize =pSlot->uItemSize;
	if(pSlotCount)*pSlotCount=pSlot->uQueueSize;

	uRet = pSlot->uQueueCount;

	ReleaseSemaphore(pSlot->hSemEdit,1,0);


return uRet;
}
