//*****************************************************************************
//*
//*
//*		MailslotsFree.c
//*
//*
//*****************************************************************************
//
//  This is a library for a mailslot support for FREERTOS.
//	Mailslots are thread save FIFO buffers
//	The elenent size and FIFO size could conigured individually.
//	Note: MailslotCreate use malloc, use this function only at startup
//
//	Global Macros:
//
//		GLOBAL_CONFIG		Enable the configuration via GlobalConfig.h header file (1=on 0=off)
//		MAX_MAILSLOTS		Maximum count of mailslots
//
//
//  Copyright  2018    Anton Zechner
//
//
#include	<FreeRTOS.h>
#include	<string.h>
#include	"task.h"
#include	"semphr.h"
#include	"Mailslots.h"


#ifdef		portUSING_MPU_WRAPPERS
#if			portUSING_MPU_WRAPPERS
#error		"wrong FreeRtos configuration, portUSING_MPU_WRAPPERS isn't allowed"
#endif
#endif

#if			configUSE_COUNTING_SEMAPHORES==0
#error		"wrong FreeRtos configuration, configUSE_COUNTING_SEMAPHORES must be 1"
#endif

#if			INCLUDE_vTaskDelay==0
#error		"wrong FreeRtos configuration, INCLUDE_vTaskDelay must be 1"
#endif


#ifndef		MAX_MAILSLOTS
#define		MAX_MAILSLOTS		256
#endif


typedef struct
	{
	QueueHandle_t	hSemEmpty;						// semaphore for empty condion
	QueueHandle_t	hSemFull;						// semaphore for full condion
	QueueHandle_t	hSemEdit;						// semaphore for edit operations
	unsigned		uItemSize;						// size of one item in the buffer
	unsigned 		uSlotSize;						// size of one slot in the buffer
	unsigned		uQueueSize;						// maximum count of items in he fifo
	unsigned		uQueueCount;					// count of items in the buffer
	unsigned		uReadPos;						// current read position
	unsigned		uWritePos;						// current write position
	unsigned char 	ucData[1];						// dummy for data memory
	}MailslotData;

static		MailslotData   *aSlotPointers[MAX_MAILSLOTS];
static		unsigned 		uNextPos=0;
static		QueueHandle_t	hGlobalSem=0;

#if			configTICK_RATE_HZ!=1000
#define 	TICKS_TO_MS(m)	(((m)*((configTICK_RATE_HZ*1024)/1000))/1024)
#else
#define 	TICKS_TO_MS(m)	m
#endif

#define 	SemDelete(h)	vSemaphoreDelete(h)
#define 	SemCreate(c,n)	xSemaphoreCreateCounting(0x70000000,c)
#define 	SemLockEx(h,t) (xSemaphoreTake(h,TICKS_TO_MS(t))? 1:0)
#define 	SemLock(h)		xSemaphoreTake(h,portMAX_DELAY)
#define 	SemUnlock(h)	xSemaphoreGive(h)

#define		NUM_SLOT(i)		((unsigned)((unsigned)i)-1)
#define		GET_SLOT(i)		(( NUM_SLOT(i)>=MAX_MAILSLOTS)? 0:aSlotPointers[NUM_SLOT(i)])
#define		CLR_SLOT(i)		if(NUM_SLOT(i)< MAX_MAILSLOTS)    aSlotPointers[NUM_SLOT(i)]=0


//*****************************************************************************
//*
//*		MailslotCreate
//*
//*****************************************************************************
//	This function creates a new mailslot
//	iItemSize	: size of one item
//	iQueueSize	: maximum size of the fifo buffer
//	Returns a handle to the mailslot or zero.
//	Note: MailslotCreate use malloc, use this function only at startup
Mailslot MailslotCreate(int iItemSize,int iQueueSize)
{
MailslotData	*pSlot;
unsigned 		 uSlotSize;
unsigned		 uPos;

uPos=TICKS_TO_MS(33);


	if(!hGlobalSem)								// Create a global semaphor
		{
		hGlobalSem=SemCreate(1,NULL);
		}

	if(iQueueSize<=0 || iItemSize<=0)return 0;

	SemLock(hGlobalSem);


	for(uPos=uNextPos;aSlotPointers[uPos];)		// Find empty handle
		{
		   uPos++;
		if(uPos>=MAX_MAILSLOTS)uPos=0;
		if(uPos==uNextPos)
			{
			SemUnlock(hGlobalSem);
			return 0;
			}
		}

	aSlotPointers[uPos] = (MailslotData*)1;

	   uNextPos=uPos+1;
	if(uNextPos>=MAX_MAILSLOTS)uNextPos=0;

	SemUnlock(hGlobalSem);


	uSlotSize = ((iItemSize+3)&~3)+sizeof(unsigned);

		pSlot = (MailslotData*)pvPortMalloc(sizeof(MailslotData)+iQueueSize*uSlotSize);
	if(!pSlot)
		{
		aSlotPointers[uPos]=0;
		return 0;
		}

	aSlotPointers[uPos] = pSlot;


	pSlot->hSemEmpty	= SemCreate(0         ,"MailslotEmpty");
	pSlot->hSemFull		= SemCreate(iQueueSize,"MailslotFull");
	pSlot->hSemEdit		= SemCreate(1         ,"MailslotEdit");
	pSlot->uItemSize	= iItemSize;
	pSlot->uSlotSize	= uSlotSize;
	pSlot->uQueueSize	= iQueueSize;
	pSlot->uQueueCount	= 0;
	pSlot->uReadPos		= 0;
	pSlot->uWritePos	= 0;



return 	uPos+1;
}


//*****************************************************************************
//*
//*		MailslotDelete
//*
//*****************************************************************************
//	Deletes a mailslot
//	hMailslot	: handle of the mailslot
void MailslotDelete(Mailslot hMailslot)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);



	if(!pSlot)return;

	SemDelete(pSlot->hSemEdit );
	SemDelete(pSlot->hSemFull );
	SemDelete(pSlot->hSemEmpty);

	vPortFree(pSlot);

	CLR_SLOT(hMailslot);

}


//*****************************************************************************
//*
//*		MailslotCount
//*
//*****************************************************************************
//	Calculates the count of messages in a mailslot
//	hMailslot	: handle of the mailslot
//	Returns the count of messages
int MailslotCount(Mailslot hMailslot)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);



	if(!pSlot)return 0;


return pSlot->uQueueCount;
}

//*****************************************************************************
//*
//*		MailslotGetWait
//*
//*****************************************************************************
//	Gets a message from a mailslot, if the fifo is empty
//	the function waits a time for arraiving a message.
//	hMailslot	: handle of the mailslot
//	pData		: pointer to the message memory
//	uTimeout	: maximum wait time in ms
//	Returns the size of the message, or 0 if no message was copied
int	MailslotGetWait(Mailslot hMailslot,void *pData,unsigned uTimeout)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned		 uOffset;
unsigned 		 uSize;



	if(uTimeout==0xFFFFFFFF)								// Wait indefinitely
		{
		if(!pSlot)
			{
			if(uTimeout)vTaskDelay(0xFFFFFFFF);
			return 0;
			}

		SemLock(pSlot->hSemEmpty);
		}
	else{													// Timed wait
		if(!pSlot)
			{
			if(uTimeout)vTaskDelay(TICKS_TO_MS(uTimeout));
			return 0;
			}

		if(!SemLockEx(pSlot->hSemEmpty,uTimeout))
			{
			return 0;
			}
		}

	SemLock(pSlot->hSemEdit);

	uOffset = pSlot->uReadPos*pSlot->uSlotSize;
	uSize   = *(unsigned*)(pSlot->ucData+uOffset);

	memcpy(pData,pSlot->ucData+uOffset+sizeof(unsigned),uSize);

	   pSlot->uReadPos++;
	if(pSlot->uReadPos>=pSlot->uQueueSize)pSlot->uReadPos=0;
	   pSlot->uQueueCount--;

	SemUnlock(pSlot->hSemEdit);
	SemUnlock(pSlot->hSemFull);



return uSize;
}


//*****************************************************************************
//*
//*		MailslotPutWait
//*
//*****************************************************************************
//	Puts a message into a mailslot, if the fifo is full
//	the function waits a time for geting an empty slot.
//	hMailslot	: handle of the mailslot
//	pData		: pointer to the message memory
//	uSize		: size of the message in bytes
//	uTimeout	: maximum wait time in ms
//	Returns the inserted bytes of the message, or 0 if no message was copied
int	MailslotPutWait(Mailslot hMailslot,const void *pData,unsigned uSize,unsigned uTimeout)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned		 uOffset;


	if(!uSize)return 0;

	if(uTimeout==0xFFFFFFFF)								// Wait indefinitely
		{
		if(!pSlot)
			{
			if(uTimeout)vTaskDelay(0xFFFFFFFF);
			return 0;
			}

		SemLock(pSlot->hSemFull);
		}
	else{													// Timed wait
		if(!pSlot)
			{
			if(uTimeout)vTaskDelay(TICKS_TO_MS(uTimeout));
			return 0;
			}

		if(!SemLockEx(pSlot->hSemFull,uTimeout))
			{
			return 0;
			}
		}

	SemLock(pSlot->hSemEdit);

	if(uSize>pSlot->uItemSize)uSize=pSlot->uItemSize;

	uOffset = pSlot->uWritePos*pSlot->uSlotSize;
	memcpy(pSlot->ucData+uOffset+sizeof(unsigned),pData,uSize);
	*(unsigned*)(pSlot->ucData+uOffset)=uSize;

	   pSlot->uWritePos++;
	if(pSlot->uWritePos>=pSlot->uQueueSize)pSlot->uWritePos=0;
	   pSlot->uQueueCount++;

	SemUnlock(pSlot->hSemEdit);
	SemUnlock(pSlot->hSemEmpty);

return uSize;
}



//*****************************************************************************
//*
//*		MailslotPeek
//*
//*****************************************************************************
//	Gets a message from a mailslot, put didn't remove a message
//	if the fifo is empty the function returns zero
//	hMailslot	: handle of the mailslot
//	pData		: pointer to the message memory
//	uPosition	: is the position of the message to take
//	Returns the size of the message, or 0 if no message was copied
int	MailslotPeek(Mailslot hMailslot,void *pData,unsigned uPosition)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned 		 uSize;
unsigned		 uOffset;


	if(!pSlot)return 0;


	SemLock(pSlot->hSemEdit);


	if(uPosition>=pSlot->uQueueCount)
		{
		SemUnlock(pSlot->hSemEdit);
		return 0;
		}



	uOffset  = pSlot->uReadPos;
	uOffset += uPosition;

	if(uOffset>=pSlot->uQueueSize)uOffset -= pSlot->uQueueSize;

	uOffset *= pSlot->uSlotSize;
	uSize    = *(unsigned*)(pSlot->ucData+uOffset);

	memcpy(pData,pSlot->ucData+uOffset+sizeof(unsigned),uSize);


	SemUnlock(pSlot->hSemEdit);


return uSize;
}


//*****************************************************************************
//*
//*		MailslotClear
//*
//*****************************************************************************
//	Remove some messages from a mailslot
//	hMailslot	: handle of the mailslot
//	iCount		: count of messgae to remove
//	Returns the count of removed messages
int MailslotClear(Mailslot hMailslot,int iCount)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned		 uRet=0;


	if(!pSlot)return 0;


	while(iCount>0)
		{
		if(!SemLockEx(pSlot->hSemEmpty,0))
			{
			return uRet;
			}

		SemLock(pSlot->hSemEdit);

		   pSlot->uReadPos++;
		if(pSlot->uReadPos>=pSlot->uQueueSize)pSlot->uReadPos=0;
		   pSlot->uQueueCount--;

		SemUnlock(pSlot->hSemEdit);
		SemUnlock(pSlot->hSemFull);

		iCount--;
		uRet++;
		}

return uRet;
}


//*****************************************************************************
//*
//*		MailslotInfo
//*
//*****************************************************************************
//	Gets infos of a mailslot
//	hMailslot	: handle of the mailslot
//	pItemSize	: here the size of a singel slot will be stored.
//				  if the pointer is zero the nothing will be stored.
//	pSlotCount	: here the count of all slots will be stored.
//				  if the pointer is zero the nothing will be stored.
//	Returns the count of messages in the mailslot or -1 if an error occurs
int MailslotInfo(Mailslot hMailslot,int *pItemSize,int *pSlotCount)
{
MailslotData   *pSlot=GET_SLOT(hMailslot);
unsigned 		uRet;


	if(!pSlot)
		{
		if(pItemSize )*pItemSize =0;
		if(pSlotCount)*pSlotCount=0;
		return -1;
		}


	SemLock(pSlot->hSemEdit);

	if(pItemSize )*pItemSize =pSlot->uItemSize;
	if(pSlotCount)*pSlotCount=pSlot->uQueueSize;

	uRet = pSlot->uQueueCount;

	SemUnlock(pSlot->hSemEdit);


return uRet;
}
