//*****************************************************************************
//*
//*
//*		MailslotsRt.c
//*
//*
//*****************************************************************************
//
//  This is a library for a mailslot support for RTTARGET.
//	Mailslots are thread save FIFO buffers
//	The elenent size and FIFO size could conigured individually.
//	Note: MailslotCreate use malloc, use this function only at startup
//
//	Global Macros:
//
//		GLOBAL_CONFIG		Enable the configuration via GlobalConfig.h header file (1=on 0=off)
//		MAX_MAILSLOTS		Maximum count of mailslots
//
//
//  Copyright  2018    Anton Zechner
//
//
#include	<rtk32.h>
#include	<clock.h>
#include	<malloc.h>
#include	<memory.h>
#include	"Mailslots.h"


#ifndef		MAX_MAILSLOTS
#define		MAX_MAILSLOTS		256
#endif


typedef struct
	{
	RTKSemaphore	hSemEmpty;						// semaphore for empty condion
	RTKSemaphore	hSemFull;						// semaphore for full condion
	RTKSemaphore	hSemEdit;						// semaphore for edit operations
	unsigned		uItemSize;						// size of one item in the buffer
	unsigned 		uSlotSize;						// size of one slot in the buffer
	unsigned		uQueueSize;						// maximum count of items in he fifo
	unsigned		uQueueCount;					// count of items in the buffer
	unsigned		uReadPos;						// current read position
	unsigned		uWritePos;						// current write position
	unsigned char 	ucData[1];						// dummy for data memory
	}MailslotData;


static	MailslotData   *aSlotPointers[MAX_MAILSLOTS];
static	unsigned 		uNextPos=0;
static	RTKSemaphore	hGlobalSem=0;

#define	NUM_SLOT(i)		((unsigned)((unsigned)i)-1)
#define	GET_SLOT(i)		(( NUM_SLOT(i)>=MAX_MAILSLOTS)? 0:aSlotPointers[NUM_SLOT(i)])
#define	CLR_SLOT(i)		if(NUM_SLOT(i)< MAX_MAILSLOTS)    aSlotPointers[NUM_SLOT(i)]=0

//*****************************************************************************
//*
//*		MailslotCreate
//*
//*****************************************************************************
//	This function creates a new mailslot
//	iItemSize	: size of one item
//	iQueueSize	: maximum size of the fifo buffer
//	Returns a handle to the mailslot or zero.
//	Note: MailslotCreate use malloc, use this function only at startup
Mailslot MailslotCreate(int iItemSize,int iQueueSize)
{
MailslotData	*pSlot;
unsigned 		 uSlotSize;
unsigned		 uPos;



	if(!hGlobalSem)								// Create a global semaphor
		{
		hGlobalSem=RTKCreateSemaphore(ST_COUNTING,1,NULL);
		}

	if(iQueueSize<=0 || iItemSize<=0)return 0;

	RTKWait(hGlobalSem);


	for(uPos=uNextPos;aSlotPointers[uPos];)		// Find empty handle
		{
		   uPos++;
		if(uPos>=MAX_MAILSLOTS)uPos=0;
		if(uPos==uNextPos)
			{
			RTKSignal(hGlobalSem);
			return 0;
			}
		}

	aSlotPointers[uPos] = (MailslotData*)1;

	   uNextPos=uPos+1;
	if(uNextPos>=MAX_MAILSLOTS)uNextPos=0;

	RTKSignal(hGlobalSem);


	uSlotSize = ((iItemSize+3)&~3)+sizeof(unsigned);

		pSlot = (MailslotData*)malloc(sizeof(MailslotData)+iQueueSize*uSlotSize);
	if(!pSlot)
		{
		aSlotPointers[uPos]=0;
		return 0;
		}

	aSlotPointers[uPos] = pSlot;


	pSlot->hSemEmpty	= RTKCreateSemaphore(ST_COUNTING,0         ,"MailslotEmpty");
	pSlot->hSemFull		= RTKCreateSemaphore(ST_COUNTING,iQueueSize,"MailslotFull");
	pSlot->hSemEdit		= RTKCreateSemaphore(ST_COUNTING,1         ,"MailslotEdit");
	pSlot->uItemSize	= iItemSize;
	pSlot->uSlotSize	= uSlotSize;
	pSlot->uQueueSize	= iQueueSize;
	pSlot->uQueueCount	= 0;
	pSlot->uReadPos		= 0;
	pSlot->uWritePos	= 0;



return 	uPos+1;
}


//*****************************************************************************
//*
//*		MailslotDelete
//*
//*****************************************************************************
//	Deletes a mailslot
//	hMailslot	: handle of the mailslot
void MailslotDelete(Mailslot hMailslot)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);



	if(!pSlot)return;

	RTKDeleteSemaphore(&pSlot->hSemEdit );
	RTKDeleteSemaphore(&pSlot->hSemFull );
	RTKDeleteSemaphore(&pSlot->hSemEmpty);

	free(pSlot);

	CLR_SLOT(hMailslot);

}


//*****************************************************************************
//*
//*		MailslotCount
//*
//*****************************************************************************
//	Calculates the count of messages in a mailslot
//	hMailslot	: handle of the mailslot
//	Returns the count of messages
int MailslotCount(Mailslot hMailslot)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);



	if(!pSlot)return 0;


return pSlot->uQueueCount;
}

//*****************************************************************************
//*
//*		MailslotGetWait
//*
//*****************************************************************************
//	Gets a message from a mailslot, if the fifo is empty
//	the function waits a time for arraiving a message.
//	hMailslot	: handle of the mailslot
//	pData		: pointer to the message memory
//	uTimeout	: maximum wait time in ms
//	Returns the size of the message, or 0 if no message was copied
int	MailslotGetWait(Mailslot hMailslot,void *pData,unsigned uTimeout)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned		 uOffset;
unsigned 		 uSize;



	if(uTimeout==0xFFFFFFFF)								// Wait indefinitely
		{
		if(!pSlot)
			{
			if(uTimeout)RTKDelay(0xFFFFFFFF);
			return 0;
			}

		RTKWait(pSlot->hSemEmpty);
		}
	else if(uTimeout==0)									// No wait
		{
		if(!pSlot)
			{
			return 0;
			}

		if(!RTKWaitCond(pSlot->hSemEmpty))
			{
			return 0;
			}
		}
	else{													// Timed wait
		if(!pSlot)
			{
			if(uTimeout)RTKDelay(CLKMilliSecsToTicks(uTimeout));
			return 0;
			}

		if(!RTKWaitTimed(pSlot->hSemEmpty,CLKMilliSecsToTicks(uTimeout)))
			{
			return 0;
			}
		}

	RTKWait(pSlot->hSemEdit);

	uOffset = pSlot->uReadPos*pSlot->uSlotSize;
	uSize   = *(unsigned*)(pSlot->ucData+uOffset);

	memcpy(pData,pSlot->ucData+uOffset+sizeof(unsigned),uSize);

	   pSlot->uReadPos++;
	if(pSlot->uReadPos>=pSlot->uQueueSize)pSlot->uReadPos=0;
	   pSlot->uQueueCount--;

	RTKSignal(pSlot->hSemEdit);
	RTKSignal(pSlot->hSemFull);



return uSize;
}


//*****************************************************************************
//*
//*		MailslotPutWait
//*
//*****************************************************************************
//	Puts a message into a mailslot, if the fifo is full
//	the function waits a time for geting an empty slot.
//	hMailslot	: handle of the mailslot
//	pData		: pointer to the message memory
//	uSize		: size of the message in bytes
//	uTimeout	: maximum wait time in ms
//	Returns the inserted bytes of the message, or 0 if no message was copied
int	MailslotPutWait(Mailslot hMailslot,void *pData,unsigned uSize,unsigned uTimeout)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned		 uOffset;


	if(!uSize)return 0;

	if(uTimeout==0xFFFFFFFF)								// Wait indefinitely
		{
		if(!pSlot)
			{
			if(uTimeout)RTKDelay(0xFFFFFFFF);
			return 0;
			}

		RTKWait(pSlot->hSemFull);
		}
	else if(uTimeout==0)									// No wait
		{
		if(!pSlot)
			{
			return 0;
			}

		if(!RTKWaitCond(pSlot->hSemFull))
			{
			return 0;
			}
		}
	else{													// Timed wait
		if(!pSlot)
			{
			if(uTimeout)RTKDelay(CLKMilliSecsToTicks(uTimeout));
			return 0;
			}

		if(!RTKWaitTimed(pSlot->hSemFull,CLKMilliSecsToTicks(uTimeout)))
			{
			return 0;
			}
		}

	RTKWait(pSlot->hSemEdit);

	if(uSize>pSlot->uItemSize)uSize=pSlot->uItemSize;

	uOffset = pSlot->uWritePos*pSlot->uSlotSize;
	memcpy(pSlot->ucData+uOffset+sizeof(unsigned),pData,uSize);
	*(unsigned*)(pSlot->ucData+uOffset)=uSize;

	   pSlot->uWritePos++;
	if(pSlot->uWritePos>=pSlot->uQueueSize)pSlot->uWritePos=0;
	   pSlot->uQueueCount++;

	RTKSignal(pSlot->hSemEdit);
	RTKSignal(pSlot->hSemEmpty);

return uSize;
}



//*****************************************************************************
//*
//*		MailslotPeek
//*
//*****************************************************************************
//	Gets a message from a mailslot, put didn't remove a message
//	if the fifo is empty the function returns zero
//	hMailslot	: handle of the mailslot
//	pData		: pointer to the message memory
//	uPosition	: is the position of the message to take
//	Returns the size of the message, or 0 if no message was copied
int	MailslotPeek(Mailslot hMailslot,void *pData,unsigned uPosition)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned 		 uSize;
unsigned		 uOffset;


	if(!pSlot)return 0;


	RTKWait(pSlot->hSemEdit);


	if(uPosition>=pSlot->uQueueCount)
		{
		RTKSignal(pSlot->hSemEdit);
		return 0;
		}



	uOffset  = pSlot->uReadPos;
	uOffset += uPosition;

	if(uOffset>=pSlot->uQueueSize)uOffset -= pSlot->uQueueSize;

	uOffset *= pSlot->uSlotSize;
	uSize    = *(unsigned*)(pSlot->ucData+uOffset);

	memcpy(pData,pSlot->ucData+uOffset+sizeof(unsigned),uSize);


	RTKSignal(pSlot->hSemEdit);


return uSize;
}


//*****************************************************************************
//*
//*		MailslotClear
//*
//*****************************************************************************
//	Remove some messages from a mailslot
//	hMailslot	: handle of the mailslot
//	iCount		: count of messgae to remove
//	Returns the count of removed messages
int MailslotClear(Mailslot hMailslot,int iCount)
{
MailslotData	*pSlot=GET_SLOT(hMailslot);
unsigned		 uRet=0;


	if(!pSlot)return 0;


	while(iCount>0)
		{
		if(!RTKWaitTimed(pSlot->hSemEmpty,0))
			{
			return uRet;
			}

		RTKWait(pSlot->hSemEdit);

		   pSlot->uReadPos++;
		if(pSlot->uReadPos>=pSlot->uQueueSize)pSlot->uReadPos=0;
		   pSlot->uQueueCount--;

		RTKSignal(pSlot->hSemEdit);
		RTKSignal(pSlot->hSemFull);

		iCount--;
		uRet++;
		}

return uRet;
}


//*****************************************************************************
//*
//*		MailslotInfo
//*
//*****************************************************************************
//	Gets infos of a mailslot
//	hMailslot	: handle of the mailslot
//	pItemSize	: here the size of a singel slot will be stored.
//				  if the pointer is zero the nothing will be stored.
//	pSlotCount	: here the count of all slots will be stored.
//				  if the pointer is zero the nothing will be stored.
//	Returns the count of messages in the mailslot or -1 if an error occurs
int MailslotInfo(Mailslot hMailslot,int *pItemSize,int *pSlotCount)
{
MailslotData   *pSlot=GET_SLOT(hMailslot);
unsigned 		uRet;


	if(!pSlot)
		{
		if(pItemSize )*pItemSize =0;
		if(pSlotCount)*pSlotCount=0;
		return -1;
		}


	RTKWait(pSlot->hSemEdit);

	if(pItemSize )*pItemSize =pSlot->uItemSize;
	if(pSlotCount)*pSlotCount=pSlot->uQueueSize;

	uRet = pSlot->uQueueCount;

	RTKSignal(pSlot->hSemEdit);


return uRet;
}
