//*****************************************************************************
//*
//*
//*		RemoteIntClient.c
//*
//*
//*****************************************************************************
//
//	(C) Copyright Anton Zechner 2009
//
//	Mit diesem Interface knnen ber Shared-Memory und einem TCP-Prozess
//	Pakete an eine Remote-Station versendet werden.
//
//	Diese Datei ist der Client auf dem Logging-System.
//
//
//	Client-Beispiel:
//
//		RemoteConn		conn;
//		RemoteTimeStemp	stamp;
//		unsigned		id;
//		char			puffer[1024];
//		int				port=777;
//		int				len;
//
//		conn = RemoteConnect("127.0.0.1",port);
//
//		for(;;)
//			{
//			Sleep(100);
//
//			   len = RemoteGetPacket(conn,puffer,sizeof(puffer),&id,&stamp);
//			if(len<0)break;
//
//			RemotePutPacket(conn,"Hallo",5,8888);
//			}
//
//		RemoteDisconnect(conn);
//
#ifdef		GLOBAL_CONFIG
#define 	REMOTEINTCLIENT_C
#include	<GlobalConfig.h>
#undef		REMOTEINTCLIENT_C
#endif

#include	"RemoteInterface.h"
#include	"SocketUtility.h"
#include	"ThreadUtility.h"
#include	<malloc.h>
#include	<stdio.h>
#include	<time.h>


#define 	LOG_RECV_ON		0
#define 	LOG_SEND_ON		0
#define 	LOG_CONN_ON		0


#if			LOG_RECV_ON
#define 	LOG_RECV(p,l)	printf("RECV:  %s %i\n"   ,p,l);
#else
#define 	LOG_RECV(p,l)
#endif

#if			LOG_SEND_ON
#define 	LOG_SEND(p,l)	printf("SEND:  %s %i\n"   ,p,l);
#define 	LOG_SEND_X(p,l)	printf("SEND:  %s %08Xh\n",p,l);
#else
#define 	LOG_SEND(p,l)
#define 	LOG_SEND_X(p,l)
#endif

#if			LOG_CONN_ON
#define 	LOG_CONN(i,p,c)	{char cAddr[64];	IpToString(&i,cAddr); printf("CONN:  %s to %s:%u\n",c,cAddr,p);}
#define 	LOG_CONN_EX(c)	{printf("CONN:  %s",c);}
#else
#define 	LOG_CONN(i,p,c)
#define 	LOG_CONN_EX(c)
#endif



typedef struct
	{
	TcpHandle		hTcp;
	unsigned		uTime;
	unsigned		uTimeSub;
	unsigned		uLifeSend;
	unsigned		uLifeRecv;
	}RemoteClient;




//*****************************************************************************
//*
//*		RemoteConnect
//*
//*****************************************************************************
//	ffnet eine Verbindung zu einem Remote-Server.
//	pHostname		: Ist der Hostname oder die IP-Adresse vom Server.
//	uPort			: Ist das TCP-Port, bei 0 wird die Standartadresse benutzt.
//	Ergibt ein Verbindungs-Handle oder 0 bei einem Fehler.
RemoteConn RemoteConnect(const char *pHostname,unsigned uPort)
{
RemoteClient   *pHandle;
TcpHandle		hTcp;
IpAddr			iAddr;
unsigned		uTime;


	if(!IpHostname(&iAddr,pHostname))
		{
		return 0;
		}


	if(!uPort)uPort=REMOTE_PORT;


		hTcp = TcpOpen(&iAddr,(unsigned short)uPort,0,0);
	if(!hTcp)return 0;

	LOG_CONN(iAddr,uPort,"connect to");

	uTime				= ThreadTickCount();
	pHandle				= (RemoteClient*)malloc(sizeof(RemoteClient));
	pHandle->uLifeSend	= uTime;
	pHandle->uLifeRecv	= uTime;
	pHandle->uTimeSub	= uTime;
	pHandle->uTime		= time(0);
	pHandle->hTcp		= hTcp;



return	pHandle;
}

//*****************************************************************************
//*
//*		RemoteDisconnect
//*
//*****************************************************************************
//	Schliet eine Verbindung zu einem Remote-Server.
//	hHandle			: Ist das Handle der Verbindung.
//	Ergibt 0 wenn ok, oder -1 bei einem Fehler.
int RemoteDisconnect(RemoteConn hHandle)
{
RemoteClient   *pHandle=(RemoteClient*)hHandle;


	if(!hHandle)return -1;


	if(pHandle->hTcp)
		{
		LOG_CONN_EX("disconnect");
		TcpClose(pHandle->hTcp);
		pHandle->hTcp = 0;
		}

	free(pHandle);


return 0;
}

//*****************************************************************************
//*
//*		RemoteSetFilter
//*
//*****************************************************************************
//	Setzt den Sub-Filter auf der Gegenstelle
//	hHandle		: Handle der Verbindung.
//	uSubId		: Ist die 16-Bit Paket-ID.
//						Bit  0..15 = Neuer Wert
//						Bit 16..31 = Maske
//	Ergibt 0 wenn ok, bzw eine Fehlernummer kleiner 0.
//			-1 = Falsches Handle
//			-2 = Keine TCP-Verbindung
int RemoteSetFilter(RemoteConn hHandle,unsigned uSubId)
{
RemoteClient   *pHandle=(RemoteClient*)hHandle;
unsigned		aBuffer[3];
unsigned		uTime;


	if(pHandle==0      )return -1;
	if(pHandle->hTcp==0)return -2;


	uTime		= ThreadTickCount();
	aBuffer[0]	= 0xFFFFFFFE;
	aBuffer[1]	= uSubId;


	if(TcpPut(pHandle->hTcp,aBuffer,sizeof(unsigned)*2)!=sizeof(unsigned)*2)
		{
		LOG_CONN_EX("broken F");
		TcpClose(pHandle->hTcp);
		pHandle->hTcp = 0;
		return -2;
		}

	LOG_SEND_X("Filter bits =",uSubId);

	pHandle->uLifeSend = uTime;



return 0;
}

//*****************************************************************************
//*
//*		RemotePutPacket
//*
//*****************************************************************************
//	Sendet ein Paket zum Server.
//	hHandle		: Handle der Verbindung.
//	pData		: Zeiger auf die Paketdaten.
//	uSize		: Gre der Paketdaten.
//	uPacketId	: Ist die 16-Bit Paket-ID.
//				  Wenn der Zeiger 0 ist wird nichts gespeichert.
//	Ergibt 0 wenn das Paket versandt wurde, bzw. eine Fehlernummer kleiner 0.
//	         0 = Paket wurde versandt
//			-1 = Falsches Handle
//			-2 = Keine TCP-Verbindung
//			-3 = Falsche Paketgre
int RemotePutPacket(RemoteConn hHandle,const void *pData,unsigned uSize,unsigned uPacketId)
{
RemoteClient   *pHandle=(RemoteClient*)hHandle;
unsigned		aBuffer[3];
unsigned		uTime;


	if(pHandle==0      )return -1;
	if(pHandle->hTcp==0)return -2;
	if(uSize>0xFFF0    )return -3;


	uTime		= ThreadTickCount();
	aBuffer[0]	= uSize|(uPacketId<<16);
	aBuffer[1]	= pHandle->uTime;
	aBuffer[2]	= uTime-pHandle->uTimeSub;


	if(TcpPut(pHandle->hTcp,aBuffer,sizeof(unsigned)*3)!=sizeof(unsigned)*3)
		{
		LOG_CONN_EX("broken C");
		TcpClose(pHandle->hTcp);
		pHandle->hTcp = 0;
		return -2;
		}

	LOG_SEND("Header size =",uSize);

	if(TcpPut(pHandle->hTcp,pData,uSize)!=(int)uSize)
		{
		LOG_CONN_EX("broken D");
		TcpClose(pHandle->hTcp);
		pHandle->hTcp = 0;
		return -2;
		}

	LOG_SEND("Data   num  =",uSize);

	pHandle->uLifeSend = uTime;



return 0;
}

//*****************************************************************************
//*
//*		RemoteWait
//*
//*****************************************************************************
//	Wartet auf das nchste Paket
//	hHandle		: Handle der Verbindung.
//	iTimeout	: Wartetzeit in Millisekunden (Maximal 3000)
//	Ergibt 1 wenn etwas empfangen werden kann, bzw 0 wenn nicht, oder eine Fehlernummer
//		   >=1 = Daten sind vorhanden
//			 0 = Kein Empfang
//			-1 = Falsches Handle
//			-2 = Keine TCP-Verbindung
//			-3 = Verbindung abgebrochen
int RemoteWait(RemoteConn hHandle,int iTimeout)
{
RemoteClient   *pHandle=(RemoteClient*)hHandle;
int				iLen;


	if(iTimeout>3000)iTimeout=3000;
	if(iTimeout<   0)iTimeout=0;


	if(pHandle==0)
		{
		if(iTimeout)ThreadSleep((iTimeout<500)? iTimeout:500);
		return -1;
		}

	if(pHandle->hTcp==0)
		{
		if(iTimeout)ThreadSleep((iTimeout<500)? iTimeout:500);
		return -1;
		}
	
	   iLen = TcpWait(pHandle->hTcp,iTimeout);
	if(iLen<0)
		{
		if(iTimeout)ThreadSleep((iTimeout<50)? iTimeout:50);
		return -3;
		}


return iLen;
}

//*****************************************************************************
//*
//*		RemoteGetPacket
//*
//*****************************************************************************
//	Liet ein Paket vom Server, und sendet Lifeticks.
//	Diese Funktion muss mindestens alle 5 Sekunden aufgerufen werden.
//	hHandle		: Handle der Verbindung.
//	pData		: Zeiger auf den Puffer fr die Paketdaten.
//	iSize		: Gre des Puffers in Byte.
//	pPacketId	: Hier wird die 16-Bit Paket-ID gespeichert.
//				  Wenn der Zeiger 0 ist wird nichts gespeichert.
//	pTime		: Hier wird der Zeitstempel der Pakets gespeichert.
//				  Wenn der Zeiger 0 ist wird nichts gespeichert.
//	Ergibt die empfangene Paketlnge, bzw. eine Fehlernummer kleiner 0.
//			-1 = Falsches Handle
//			-2 = Keine TCP-Verbindung
//			-3 = Falsche Paketgre
int RemoteGetPacket(RemoteConn hHandle,void *pData,int iSize,unsigned *pPacketId,RemoteTimeStemp *pTime)
{
RemoteClient   *pHandle=(RemoteClient*)hHandle;
unsigned		aBuffer[1024];
unsigned		uTime;
unsigned		uDiff;
int				iByte;
int				iLen;
int				iNum;
int				iCnt;



	if(pHandle==0      )return -1;
	if(pHandle->hTcp==0)return -2;

	uTime = ThreadTickCount();
	uDiff = uTime-pHandle->uLifeSend;


	if(uDiff>REMOTE_LIFETICK)								// Lifetick senden
		{
		pHandle->uLifeSend = uTime;
		aBuffer[0] = 0xFFFFFFFF;


		LOG_SEND("Lifetick",-1);

		if(TcpPut(pHandle->hTcp,aBuffer,sizeof(unsigned))!=sizeof(unsigned))
			{
			LOG_CONN_EX("broken E");
			TcpClose(pHandle->hTcp);
			pHandle->hTcp = 0;
			return -2;
			}
		}


	iLen = TcpGetNext(pHandle->hTcp,aBuffer,sizeof(unsigned),0);

	if(iLen<0)
		{
		TcpClose(pHandle->hTcp);
		pHandle->hTcp = 0;
		return -2;
		}

	if(iLen==0)												// Wird ein Paket bertragen
		{
		uDiff = uTime-pHandle->uLifeRecv;
		if(uDiff>REMOTE_TIMEOUT)
			{
			LOG_CONN_EX("broken timeout");
			TcpClose(pHandle->hTcp);
			pHandle->hTcp = 0;
			return -2;
			}

		return 0;
		}


	pHandle->uLifeRecv = uTime;


	if(aBuffer[0]==0xFFFFFFFF)								// Lifetick empfangen
		{
		LOG_RECV("Lifetick ",-1);
		return 0;
		}

															// Zeitstempel einlesen
	if(TcpGetWait(pHandle->hTcp,aBuffer+1,sizeof(unsigned)*2,5000)!=sizeof(unsigned)*2)
		{
		LOG_CONN_EX("broken A");
		TcpClose(pHandle->hTcp);
		pHandle->hTcp = 0;
		return -2;
		}


	LOG_RECV("Header size =",aBuffer[0]&0xFFFF);


	if(pPacketId)
		{
		*pPacketId = aBuffer[0]>>16;
		}

	if(pTime)
		{
		pTime->uUtcTime  = aBuffer[1];
		pTime->uMsOffset = aBuffer[2];
		}

	iNum = aBuffer[0]&0xFFFF;

	if(iNum>iSize)											// Ist das Paket zu gro fr den Puffer
		{
		iLen = TcpGetWait(pHandle->hTcp,pData,iSize,1000);

		LOG_RECV("Data-  num  =",iLen);

		if(iLen==iSize)										// Paketrest empfangen
			{
			iCnt = iNum-iSize;

			while(iCnt>0)
				{
				iByte = (iCnt>=sizeof(aBuffer))? sizeof(aBuffer):iCnt;
				iLen  = TcpGetWait(pHandle->hTcp,aBuffer,iByte,300);

				LOG_RECV("Data+  num  =",iLen);

				if(iLen!=iByte)
					{
					iLen = -1;
					break;
					}

				iCnt -= iLen;
				}
			}
		else{
			iLen = -1;
			}

		iNum = iSize;
		}
	else{													// Paketdaten empfangen
		iLen = TcpGetWait(pHandle->hTcp,pData,iNum,1000);
		LOG_RECV("Data   num  =",iLen);
		}




	if(iLen<0 || iLen!=iNum)								// Verbindungsabbruch entdeckt
		{
		LOG_CONN_EX("broken B");
 		TcpClose(pHandle->hTcp);
		pHandle->hTcp = 0;
		return -2;
		}


return iNum;
}
