//*****************************************************************************
//*
//*
//*		RemoteInterface.c
//*
//*
//*****************************************************************************
//
//	(C) Copyright Anton Zechner 2009
//
//	Mit diesem Interface knnen ber Shared-Memory und einem TCP-Prozess
//	Pakete an eine Remote-Station versendet werden.
//
//	Diese Datei ist das Interface in der Echtzeitumgebung.
//
//
//	Beispiel:
//
//
//		RemoteTimeStemp	stamp;
//		RemoteHandle	interface;
//		unsigned		key=123;
//		unsigned		id;
//		int				len;
//
//
//		interface = RemoteInit(key,1024*32,1024*32,1024);
//
//		for(;;)
//			{
//			len = RemoteRecv(interface,puffer,sizeof(puffer),&id,&stamp);
//
//			if(len>=0)
//				{
//				...
//				}
//
//			RemoteSend(interface,"Hallo",5,id);
//			}
//
//		RemoteExit(interface);
//
#ifdef		GLOBAL_CONFIG
#define 	REMOTEINTERFACE_C
#include	<GlobalConfig.h>
#undef		REMOTEINTERFACE_C
#endif

#include	"ShmUtility.h"
#include	"RemoteInterface.h"
#include	<memory.h>
#include	<time.h>

#define		SEM_KEY		0x2964
typedef		void	   *RemoteSem;
static		RemoteSem	hRemoteSem = 0;					// Semaphore fr's Senden
static		int			iSemCount  = 0;					// Referenzanzahl der Semaphore



//*****************************************************************************
//*
//*
//*		LINUX
//*
//*
//*****************************************************************************

#ifndef		LINUX
#ifdef		linux
#define 	LINUX	1
#endif
#endif

#ifdef		LINUX

#define	   _GNU_SOURCE 

#include	<time.h>
#include 	<errno.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<semaphore.h>
#include 	<sys/ipc.h>
#include 	<sys/sem.h>
#include 	<sys/time.h>
#include	<sys/times.h>
#include 	<sys/types.h>
#include	<sys/syscall.h>
#ifdef		__i386__
#include	<linux/version.h>
#else
#ifndef		LINUX_VERSION_CODE
#define		LINUX_VERSION_CODE	0x20610
#endif
#endif

extern int  semtimedop(int,struct sembuf*,unsigned,struct timespec*);


//*****************************************************************************
//*
//*		semtimedop
//*
//*****************************************************************************
//  Perform user-defined atomical operation of array of semaphores.
#if defined(__i386__) && LINUX_VERSION_CODE < 0x20416

static inline int semtimedopx(int iSemId,struct sembuf *pSops,size_t uSopsN,const struct timespec *pTimeout)
{
int 	iRet;



	__asm__ volatile
		(
		"pushl	%%ebp			\n"
		"movl	$117 ,%%eax		\n"
		"movl	$4   ,%%ebx		\n"
		"movl	%%esi,%%ebp		\n"
		"call   *%%gs:0x10      \n"
		"popl	%%ebp			\n"
        : "=a" (iRet)
        : "c"  (iSemId), "d" (uSopsN), "S" (pTimeout), "D" (pSops)
		: "ebx"
		);

	if(iRet<0 && iRet>-125)
		{

		errno = -iRet;
		return -1;
		}

return 0;
}

#else

#define	semtimedopx(a,b,c,d)	semtimedop(a,b,c,d)

#endif


//*****************************************************************************
//*
//*		RemoteTickCount
//*
//*****************************************************************************
//	Returns the current time in ms
unsigned RemoteTickCount()
{
struct timespec 	sTime;


	if(clock_gettime(CLOCK_MONOTONIC,&sTime))
		{
		clock_gettime(CLOCK_REALTIME,&sTime);
		}


return sTime.tv_sec*1000+sTime.tv_nsec/1000000;
}

//*****************************************************************************
//*
//*		SemCreate
//*
//*****************************************************************************
//	Creates a new semaphore
//	iKey	: is the name of the semaphore, 0 is for a local semaphore
//	Returns a handle to a semaphore or NULL if an error occurs
static RemoteSem SemCreate(int iKey)
{
struct sembuf	sOp;
int				iInc;
int 			iSem;



	   iSem = semget(iKey,2,IPC_CREAT|IPC_EXCL|0666);
	if(iSem==-1)
		{
		   iSem = semget(iKey,2,0666);						// did the semaphore allready exist ?
		if(iSem==-1)return NULL;

		   iInc = semctl(iSem,1,GETVAL,0);					// if open count is zero the semaphore can be used
		if(iInc!=0)return NULL;
		}


	sOp.sem_num = 1;										// increment open count
	sOp.sem_op  = 1;
	sOp.sem_flg = SEM_UNDO;

	semop(iSem,&sOp,1);


	semctl(iSem,0,SETVAL,1);								// set starting value



return (RemoteSem)((intptr_t)iSem+1);
}

//*****************************************************************************
//*
//*		SemOpen
//*
//*****************************************************************************
//	Opens an existing semaphore
//	iKey	: is the name of the semaphore, 0 is for a local semaphore
//	Returns a handle to a semaphore or NULL if an error occurs
static RemoteSem SemOpen(int iKey)
{
struct sembuf	sOp;
int				hSem;


	   hSem = semget(iKey,2,0666);
	if(hSem==-1)return NULL;


	sOp.sem_num = 1;										// increment open count
	sOp.sem_op  = 1;
	sOp.sem_flg = SEM_UNDO;

	semop(hSem,&sOp,1);




return (RemoteSem)((intptr_t)hSem+1);
}

//*****************************************************************************
//*
//*		SemClose
//*
//*****************************************************************************
//	Closes a semaphore. If all handels are closed the semaphor will be deleted.
//	If the last handle of a semaphore was closed the semaphore will be removed.
//	hSemaphore	: ist the handle of the semaphore
//	Return 1 if the semaphore was closed or 0 if an error occurs
/*
static int SemClose(RemoteSem hSemaphore)
{
struct timespec	sTimeout;
struct sembuf	sOp;
int 			iInc;


	if(!hSemaphore)return 0;


	sTimeout.tv_sec	 =  0;									// decrement open count
	sTimeout.tv_nsec =  0;
	sOp.sem_num      =  1;
	sOp.sem_op		 = -1;
	sOp.sem_flg		 = SEM_UNDO;


	if(semtimedopx(((int)hSemaphore)-1,&sOp,1,&sTimeout)==-1)
		{
		return 0;
		}

	   iInc = semctl(((int)hSemaphore)-1,1,GETVAL,0);		// if open count is zero the semaphore can be deleded
	if(iInc>0)return 1;





	if(semctl(((int)hSemaphore)-1,0,IPC_RMID,0)==-1)		// close semaphore
		{
		return 0;
		}



return 1;
}
*/

//*****************************************************************************
//*
//*		SemLock
//*
//*****************************************************************************
//	Locks a semaphore, the semaphore value is decremented by 1. If the
//  semaphore value is 0 the function waits unti the value is above zero.
//	If the semaphore is locked longer than uWaitTime (in ms) the function
//	breakes the waiting.
//	hSemaphore	: is the handle of the semaphore
//	uWaitTime	: waite time in milliseconds
//	Returns 1 if the semophore is locked, or 0 at a timeout
static int SemLock(RemoteSem hSemaphore,unsigned uWaitTime)
{
struct sembuf	sOp;
struct timespec	sTimeout;



	if(!hSemaphore)return 0;

	if(uWaitTime==0)
		{
		sOp.sem_num = 0;
		sOp.sem_op  = -1;
		sOp.sem_flg = SEM_UNDO|IPC_NOWAIT;

		if(semop(((intptr_t)hSemaphore)-1,&sOp,1)!=-1)
			{
			return 1;
			}

		return 0;
		}


	sTimeout.tv_sec	 = uWaitTime/1000;
	sTimeout.tv_nsec =(uWaitTime%1000)*1000000;
	sOp.sem_num      = 0;
	sOp.sem_op		 = -1;
	sOp.sem_flg		 = SEM_UNDO;

	if(semtimedopx(((intptr_t)hSemaphore)-1,&sOp,1,&sTimeout)!=-1)
		{
		return 1;
		}


return 0;
}


//*****************************************************************************
//*
//*		SemUnlock
//*
//*****************************************************************************
//	Unlocks a semaphore, the semaphore value is incremented by 1.
//	hSemaphore	: is the handle of the semaphore
static void SemUnlock(RemoteSem hSemaphore)
{
struct sembuf	sOp;


	if(!hSemaphore)return;

	sOp.sem_num = 0;
	sOp.sem_op  = 1;
	sOp.sem_flg = SEM_UNDO;

	semop(((intptr_t)hSemaphore)-1,&sOp,1);
}

//******************************************************************************
//*
//*
//*		RTTARGET Code
//*
//*
//******************************************************************************
#elif defined(RTTARGET)

#include	<rttarget.h>
#include 	<rtk32.h>
#include 	<clock.h>
#include 	<stdio.h>

#define 	RemoteTickCount()		CLKTicksToMilliSecs(RTKGetTime())
#define		SemLock(h,t)			RTKWaitTimed((RTKSemaphore)h,CLKMilliSecsToTicks(t))
#define		SemUnlock(h)			RTKWait((RTKSemaphore)h)


//*****************************************************************************
//*
//*		SemOpen
//*
//*****************************************************************************
//	Opens an existing global semaphore
//	iKey	: is the name of the semaphore, 0 isn't allowed
//	Returns a handle to a semaphore or NULL if an error occurs
static RemoteSem SemOpen(int iKey)
{
char		cName[32];
RemoteSem	hSem;


	sprintf(cName,"REM_%08X",iKey);
	hSem = (RemoteSem)RTKOpenSemaphore(ST_COUNTING,0,0,cName)



return hSem;
}

//*****************************************************************************
//*
//*		SemCreate
//*
//*****************************************************************************
//	Creates a new semaphore
//	iKey	: is the name of the semaphore, 0 is for a local semaphore
//	Returns a handle to a semaphore or NULL if an error occurs
static RemoteSem SemCreate(int iKey)
{
char		cName[32];
RemoteSem	hSem;



	sprintf(cName,"REM_%08X",iKey);

	   hSem = (RemoteSem)RTKOpenSemaphore(ST_COUNTING,0,0,cName)
	if(hSem)												// the semaphore allready exists
		{
		RTKDeleteSemaphore(&hSem);
		return NULL;
		}


	hSem = (RemoteSem)RTKOpenSemaphore(ST_COUNTING,1,0x70000000,cName)


return (RemoteSem)hSem;
}

//*****************************************************************************
//*
//*		SemClose
//*
//*****************************************************************************
//	Closes a semaphore. If all handels are closed the semaphor will be deleted.
//	hSemaphore	: ist the handle of the semaphore
//	Return 1 if the semaphore was closed or 0 if an error occurs
/*
static int SemClose(RemoteSem hSemaphore)
{

return RTKDeleteSemaphore(&hSem);
}
*/



//******************************************************************************
//*
//*
//*		WINDOWS Code
//*
//*
//******************************************************************************
#else

#include	<windows.h>
#include	<stdio.h>
#pragma		warning( disable :4244 )
#pragma		warning( disable :4505 )
#pragma		warning( disable :4115 )

#ifdef	   _MSC_VER
#pragma		warning( disable :4996 )
#endif

#define		SemLock(h,t)		  ((WaitForSingleObject(h,t)==0)? 1:0)
#define		SemUnlock(h)			ReleaseSemaphore(h,1,NULL)


//*****************************************************************************
//*
//*		RemoteTickCount
//*
//*****************************************************************************
//	Returns the current time in ms
unsigned RemoteTickCount()
{
return GetTickCount();
}



//*****************************************************************************
//*
//*		SemOpen
//*
//*****************************************************************************
//	Opens an existing global semaphore
//	iKey	: is the name of the semaphore, 0 isn't allowed
//	Returns a handle to a semaphore or NULL if an error occurs
static RemoteSem SemOpen(int iKey)
{
char	cName[32];
HANDLE	hSem;


	sprintf(cName,"Global\\REM_%08X",iKey);
	hSem = OpenSemaphore(SEMAPHORE_ALL_ACCESS,FALSE,cName);


return (RemoteSem)hSem;
}

//*****************************************************************************
//*
//*		SemCreate
//*
//*****************************************************************************
//	Creates a new semaphore
//	iKey	: is the name of the semaphore, 0 is for a local semaphore
//	Returns a handle to a semaphore or NULL if an error occurs
static RemoteSem SemCreate(int iKey)
{
SECURITY_ATTRIBUTES sAttr;
char				sDesc[SECURITY_DESCRIPTOR_MIN_LENGTH];
char				cName[32];
HANDLE				hSem;



	sprintf(cName,"Global\\REM_%08X",iKey);

	   hSem = OpenSemaphore(SEMAPHORE_ALL_ACCESS,FALSE,cName);
	if(hSem)												// the semaphore allready exists
		{
		CloseHandle(hSem);
		return NULL;
		}

	sAttr.nLength				= sizeof(sAttr);			// create security descriptor
	sAttr.bInheritHandle		= FALSE;
	sAttr.lpSecurityDescriptor	= &sDesc;
	InitializeSecurityDescriptor(sAttr.lpSecurityDescriptor,SECURITY_DESCRIPTOR_REVISION);
	SetSecurityDescriptorDacl   (sAttr.lpSecurityDescriptor,TRUE,NULL,FALSE);

	hSem = CreateSemaphoreA(&sAttr,1,0x70000000,cName);		// create global semaphore


return (RemoteSem)hSem;
}


//*****************************************************************************
//*
//*		SemClose
//*
//*****************************************************************************
//	Closes a semaphore. If all handels are closed the semaphor will be deleted.
//	hSemaphore	: ist the handle of the semaphore
//	Return 1 if the semaphore was closed or 0 if an error occurs
/*
static int SemClose(RemoteSem hSemaphore)
{

return CloseHandle((HANDLE)hSemaphore);
}
*/


#endif



//******************************************************************************
//*
//*		RemoteSendEx
//*
//******************************************************************************
//	Speichert ein Paket im Senderingpuffer.
//	pRemote			: Handle des Remote-Interfaces.
//	pData			: Zeiger auf die Paketdaten.
//	uSize			: Gre des Paketes in Byte.
//	uPacketId		: User ID fr das Paket (16-Bit).
//	Ergibt 0 wenn das Paket gespeichert wurde, ansonsten -1.
int RemoteSendEx(RemoteHandle pRemote,const void *pData,unsigned uSize,unsigned uPacketId)
{
unsigned   *pOut;
unsigned   *pPtr;
unsigned 	uFree;
unsigned 	uOff;
unsigned 	uMax;




	if(pRemote->usMaxPacket<uSize)							// Ist das Paket zu gro
		{
		return -1;
		}

	if(pRemote->ucSemaphore)								// Semaphor benutzen
		{
		if(!SemLock(hRemoteSem,(pRemote->ucSemaphore-1)*16))
			{
			return -1;
			}
		}


	   uFree=pRemote->uSendWr-pRemote->uSendRd;
	if(uFree>pRemote->uSendSize)uFree+=pRemote->uSendSize;

	uFree = pRemote->uSendSize-uFree;

	if(uFree<=pRemote->usMaxPacket+32U)
		{
		if(pRemote->ucSemaphore)							// Semaphor benutzen
			{
			SemUnlock(hRemoteSem);
			}

		return -1;
		}


	pOut = pRemote->aData;
	uOff = pRemote->uSendWr  /sizeof(unsigned);
	uMax = pRemote->uSendSize/sizeof(unsigned);


	pOut[uOff] = uSize|(uPacketId<<16);					uOff++;	if(uOff>=uMax)uOff=0;
	pOut[uOff] = pRemote->uTime;						uOff++;	if(uOff>=uMax)uOff=0;
	pOut[uOff] = RemoteTickCount()-pRemote->uTimeSub;	uOff++;	if(uOff>=uMax)uOff=0;

	uSize = (uSize+sizeof(unsigned)-1)/sizeof(unsigned);
	pPtr  = (unsigned*)pData;


	while(uSize>0)											// Die Paketdaten kopieren
		{
		pOut[uOff] = *pPtr;

		uSize--;
		pPtr++;
		uOff++;

		if(uOff>=uMax)uOff=0;
		}


	pRemote->uSendWr = uOff*sizeof(unsigned);				// Neue Schreibposition fr Ringpuffer


	if(pRemote->ucSemaphore)								// Semaphor benutzen
		{
		SemUnlock(hRemoteSem);
		}



return 0;
}

//******************************************************************************
//*
//*		RemoteRecvEx
//*
//******************************************************************************
//	Liet ein Paket aus dem Empfangsringpuffer.
//	pRemote			: Handle des Remote-Interfaces
//	pData			: Zeiger auf den Puffer fr die Paketdaten.
//	iSize			: Gre des Puffers in Byte.
//	pPacketId		: Hier wird die User ID vom das Paket (16-Bit) gespeichert.
//					  Wenn der Zeiger 0 ist wird nichts gespeichert.
//	pTime			: Hier wird der Zeitstempel gespeichert.
//					  Wenn der Zeiger 0 ist wird nichts gespeichert.
//	Ergibt die Paketlnge wenn ein Paket empfangen wurde, ansonsten -1.
int RemoteRecvEx(RemoteHandle pRemote,void *pData,int iSize,unsigned *pPacketId,RemoteTimeStemp *pTime)
{
unsigned   *pInp;
unsigned   *pPtr;
unsigned 	uUsed;
unsigned 	uSize=0;
unsigned 	uOff;
unsigned 	uMax;
unsigned 	uCnt;



	   uUsed=pRemote->uRecvWr-pRemote->uRecvRd;
	if(uUsed>pRemote->uRecvSize)uUsed+=pRemote->uRecvSize;
	if(uUsed==0)return -1;


	pInp = pRemote->aData+pRemote->uSendSize/sizeof(unsigned);
	uOff = pRemote->uRecvRd                 /sizeof(unsigned);
	uMax = pRemote->uRecvSize				/sizeof(unsigned);


	if(iSize>=sizeof(unsigned))

	uSize = pInp[uOff];
	if(pPacketId)*pPacketId=uSize>>16;							// Packed ID bernehmen

	uSize &= 0xFFFF;
	uOff++;	if(uOff>=uMax)uOff=0;


	if(pTime)
		{
		pTime->uUtcTime = pInp[uOff];							// UTC-Refenzezeit wann die Aufzeichnung begonnen hat
		uOff++;	if(uOff>=uMax)uOff=0;
		pTime->uMsOffset = pInp[uOff];							// Offset in ms wann das Paket erzeugt wurde.
		uOff++;	if(uOff>=uMax)uOff=0;
		}
	else{
		uOff++;	if(uOff>=uMax)uOff=0;
		uOff++;	if(uOff>=uMax)uOff=0;
		}


	uCnt = (uSize+sizeof(unsigned)-1)/sizeof(unsigned);
	pPtr = (unsigned*)pData;

	while(uCnt>0)												// Die Paketdaten kopieren
		{
		if(iSize>=sizeof(unsigned))
			{
			*pPtr  = pInp[uOff];
			iSize -= sizeof(unsigned);
			}

		uCnt--;
		pPtr++;
		uOff++;

		if(uOff>=uMax)uOff=0;
		}


	pRemote->uRecvRd = uOff*sizeof(unsigned);					// Neue Leseposition fr Ringpuffer



return uSize;
}

//******************************************************************************
//*
//*		RemoteInit
//*
//******************************************************************************
//	Erzeugt einen Kommunikationspuffer.
//	uId				: Ist die ID fr den Puffer
//	uSendSize		: Gre des Sendepuffers in Byte.
//	uRecvSize		: Gre des Empfangspuffers in Byte.
//	uMaxPacket		: Maximale Paketgre.
//	uUseSem			: Soll beim Senden eine RemoteSem ein Semaphor benutzt
//					  werden. Der Wert ist die maximal Wartezeit in 16 ms
//					  Schritten (1=0ms 2=16ms ...)
//					  Nur wenn eine Semaphor benutzt wird knnen mehrere
//					  Prozesse sich einen Kommunikationspuffer teilen.
//	Ergibt einen Zeiger auf den Kommunikationspuffer, oder 0 bei einem Fehler.
RemoteHandle RemoteInit(unsigned uId,unsigned uSendSize,unsigned uRecvSize,unsigned uMaxPacket,unsigned uUseSem)
{
RemoteHandle	pRemote;
unsigned 		uIdKey;


	if(uId     >=0x10000 || uId      ==0x00000000)return 0;
	if(uSendSize <0x1000 || uSendSize >0x10000000)return 0;
	if(uRecvSize <0x1000 || uRecvSize >0x10000000)return 0;
	if(uMaxPacket<0x0010 || uMaxPacket>0x0000FFF0)return 0;


	uSendSize  +=  (sizeof(unsigned)-1);
	uSendSize  &= ~(sizeof(unsigned)-1);

	uRecvSize  +=  (sizeof(unsigned)-1);
	uRecvSize  &= ~(sizeof(unsigned)-1);

	uIdKey		=  uId+(~uId<<16);


		pRemote = (RemoteHandle)ShmCreate(uId,sizeof(*pRemote)+uSendSize+uRecvSize,SHM_WRITE);
	if(!pRemote)
		{
		if(!uUseSem)return 0;
																// Existiert das SHM bereits
			pRemote = (RemoteHandle)ShmOpen(uId,sizeof(*pRemote),SHM_WRITE);
		if(!pRemote)return 0;
		if(!pRemote->ucSemaphore && pRemote->uId!=uIdKey)		// Passen die Bereiche zusammen
		   	{
			ShmClose(uId);
			return 0;
			}


		uSendSize = pRemote->uSendSize;							// Gre auslesen
		uRecvSize = pRemote->uRecvSize;
		ShmClose(uId);

																// Existiert das SHM bereits
			pRemote = (RemoteHandle)ShmOpen(uId,sizeof(*pRemote)+uSendSize+uRecvSize,SHM_WRITE);
		if(!pRemote)return 0;

		if(!uUseSem)
			{
			return pRemote;
			}

		if(!hRemoteSem)											// Semaphore erzeugen
			{
				           hRemoteSem = SemCreate(SEM_KEY);
			if(!hRemoteSem)hRemoteSem = SemOpen  (SEM_KEY);
			if(!hRemoteSem)
				{
				ShmClose(uId);
				return 0;
				}
			}

		pRemote->ucAttached++;
		iSemCount++;

		return pRemote;
		}


	memset(pRemote,0,sizeof(*pRemote));


	pRemote->usMaxPacket = uMaxPacket;							// Maximale Paketgre
	pRemote->uSendSize	 = uSendSize;							// Ringpuffer frs Senden
	pRemote->uRecvSize	 = uRecvSize;							// Ringpuffer frs Empfangen
	pRemote->ucSemaphore = uUseSem;								// RemoteSem benutzen
	pRemote->uId		 = uIdKey;


	if(uUseSem)													// Wird ein Sende-Semaphor benutz
		{
		if(!hRemoteSem)
			{
				           hRemoteSem = SemCreate(SEM_KEY);
			if(!hRemoteSem)hRemoteSem = SemOpen  (SEM_KEY);
			if(!hRemoteSem)
				{
				ShmClose(uId);
				return 0;
				}
			}

		pRemote->ucAttached++;
		iSemCount++;
		}



return pRemote;
}

//******************************************************************************
//*
//*		RemoteExit
//*
//******************************************************************************
//	Lscht eine Kommunikationspuffer
//	pRemote			: Handle des Remote-Interfaces.
//	Ergibt 0 wenn ok, ansonsten -1.
int RemoteExit(RemoteHandle pRemote)
{
unsigned	uId;


	if(!pRemote)return -1;

	if(pRemote->ucSemaphore && iSemCount>0)						// Wird ein Sende-Semaphor benutzt
		{
		if(iSemCount>0)iSemCount--;
//		if(iSemCount==0)
//			{
//			SemClose(hRemoteSem);
//			hRemoteSem = 0;
//			}

		if(pRemote->ucAttached)
			{
			pRemote->ucAttached--;
			}
		}
	else{
		pRemote->ucAttached = 0;
		}

	uId = pRemote->uId&0xFFFF;

	if(pRemote->ucAttached==0)
		{
		memset(pRemote,0,sizeof(*pRemote));
		}


	if(!ShmClose(uId))return -1;


return 0;
}


//******************************************************************************
//*
//*		RemoteEnable
//*
//******************************************************************************
//	Ein- und ausschalten des Remote-Interfaces. Damit kann der Puffer unabhngig
//	von einer Remoteverbindung eingeschalten werden.
//	pRemote			: Handle des Remote-Interfaces.
//	uEnable			: Neuer Zustand	(0=aus 1=ein 2=nicht ndern)
//	uSubIdMask		: Neuer Zustand fr SubID
//						Bit  0..15 = Neuer Wert
//						Bit 16..31 = Maske
void RemoteEnable(RemoteHandle pRemote,unsigned uEnable,unsigned uSubIdMask)
{
unsigned 	uMask;
unsigned 	uBits;



	if(!pRemote)return;

	uMask = uSubIdMask>>16;


	if(uMask)													// Sub-Filter setzen
		{
		uBits  = uSubIdMask&0xFFFF;
		uBits &= uMask;
		uBits |= pRemote->usSubEnable&~uMask;

		pRemote->usSubEnable = uBits;
		}

	if(uEnable==2)												// Nicht ndern
		{
		return;
		}

	if(uEnable==0)												// Ausschalten
		{
		pRemote->ucEnabled &= ~2;
		return;
		}

	if(pRemote->ucEnabled&2)									// Schon eingeschalten
		{
		return;
		}


	pRemote->uTime	    =  (unsigned)time(0);					// Einschaltezeit speichern
	pRemote->uTimeSub   =  RemoteTickCount();
	pRemote->ucEnabled |=  2;

}
