//*****************************************************************************
//*
//*
//*		ThreadUtility.c
//*
//*
//*****************************************************************************
//
//	Thread Module for WIN32,LINUX,FREERTOS,CIMSIS and RTTARGET
//
//	Global Macros:
//
//		GLOBAL_CONFIG		Enable the configuration via GlobalConfig.h header file (1=on 0=off)
//		THREAD_CHECK		Store ckeck points with ThreadCheck()/ThreadCheckSub() (1=on 0=off)
//		THREAD_TLS_SIZE		Size of TLS data for the threads (0=none, 1...n=size)
//		THREAD_NAMES		Size of thread names (0=no names, 1...n=size, not for CIMSIS)
//		FREERTOS			Operating system is FreeRTOS
//		RTTARGET			Operating system is RtTarget
//		RTL					Operating system is Cimsis (Keil RTX)
//
//
//	Copyright Anton Zechner (c) 2018
//
//
#ifdef		GLOBAL_CONFIG
#define 	THREAD_C
#include	<GlobalConfig.h>
#undef		THREAD_C
#endif

#define 	__THREAD_C__
#include	"ThreadUtility.h"


typedef struct
	{
	volatile const char    *pFile;					// filename of last check    point
	volatile const char    *pSubFile;				// filename of last subcheck point
	volatile int	 		iLine;					// line 	of last check    point
	volatile unsigned		uTick;					// time 	of last check    point
	volatile int	 		iSubLine;				// line 	of last subcheck point
	volatile unsigned		uSubTick;				// time 	of last subcheck point
	volatile unsigned		uSubParam;				// param 	of last subcheck point
	}CheckData;


#ifndef	THREAD_TLS_SIZE
#define THREAD_TLS_SIZE		0
#endif

#if		defined(RTTARGET) || defined(FREERTOS)
#ifdef	THREAD_NAMES
#undef	THREAD_NAMES
#endif
#define THREAD_NAMES		0
#endif

#ifndef	THREAD_NAMES
#define THREAD_NAMES		0
#endif

#ifndef	THREADS_MAX
#define THREADS_MAX			128
#endif


#if		THREAD_CHECK || THREAD_NAMES || (THREAD_TLS_SIZE>0)
static  void ThreadTlsAdd(char*);
static  void ThreadTlsInit();
#define USE_TLS				1
#else
#define USE_TLS				0
#endif

#define	THREAD_TLS_BYTE		((THREAD_TLS_SIZE+sizeof(char*)-1)&~(sizeof(char*)-1))
#define THREAD_NAME_OFF		(TLS_SIZE-sizeof(char*))

#if		THREAD_CHECK
#define TLS_SIZE			(THREAD_TLS_BYTE+sizeof(char*)*THREAD_NAMES+sizeof(CheckData))
#else
#define TLS_SIZE			(THREAD_TLS_BYTE+sizeof(char*)*THREAD_NAMES+4*!THREAD_NAMES)
#endif

#if 	USE_TLS
static	char				aMainData[TLS_SIZE]={0};// is the tls data for the main thread
static	int					iThreadCheckInit   = 0;	// 0=not set 1=off 2=on
#endif

//*****************************************************************************
//*
//*
//*		WINDOWS-Code
//*
//*
//*****************************************************************************

#if defined(_WIN32)	&& !defined(RTTARGET) && !defined(FREERTOS)

#include	<windows.h>
#include	<winbase.h>
#include	<malloc.h>


#ifdef		_MSC_VER
#pragma		warning( disable :4990)
#pragma		warning( disable :4311)
#pragma		warning( disable :4312)
#endif


typedef struct
   {
   void (_cdecl *pProc)(void*);
   void			*hThread;
   void         *pParam;
   const char   *pName;
   }ThreadData;


#define		ThreadTick()			GetTickCount()
#define 	ThreadTickToMs(t)		(t)
#define 	ThreadTlsAlloc()		TlsAlloc()
#define 	ThreadTlsSet(n,p)		TlsSetValue(n,p);
#define 	ThreadTlsGet(n)			(char*)(TlsGetValue(n))

			int			iThreadTlsNum = -1;			// is the used TLS section
static		const int	aPrio[] = { THREAD_PRIORITY_TIME_CRITICAL,
                                    THREAD_PRIORITY_HIGHEST,
                                    THREAD_PRIORITY_ABOVE_NORMAL,
                                    THREAD_PRIORITY_NORMAL,
                                    THREAD_PRIORITY_BELOW_NORMAL,
                                    THREAD_PRIORITY_LOWEST,
                                    THREAD_PRIORITY_IDLE
                                  };

#if			THREAD_CHECK

typedef struct
   {
   void    *hThread;
   char    *pTlsData;
   }TlsMap;


static		TlsMap		aTlsMap[THREADS_MAX];
static		unsigned 	uTlsMapPos = 0;


//*****************************************************************************
//*
//*		ThreadTlsGetOf
//*
//*****************************************************************************
//	Gets the info data field to a thread handle
//	hThread		: is the thread handle
//	iNum		: ist the tls slot number
//	Returns the pointer to the info data field, or NULL if an error occurs.
static char *ThreadTlsGetOf(void *hThread,int iNum)
{
unsigned	uPos;


	for(uPos=0;uPos<THREADS_MAX;uPos++)
		{
		if(aTlsMap[uPos].hThread!=hThread)continue;
		return aTlsMap[uPos].pTlsData;
		}


return 0;
}

//*****************************************************************************
//*
//*		ThreadTlsRegister
//*
//*****************************************************************************
//	Insert a thread handle in the global thread list
//	hThread		: is the thread handle
//	pData		: pointer to the info data field
static void ThreadTlsRegister(void *hThread,char *pData)
{
unsigned	uPos;


	for(uPos=0;uPos<THREADS_MAX;uPos++)
		{
		if(InterlockedCompareExchange(&aTlsMap[uPos].hThread,hThread,0))continue;
		aTlsMap[uPos].pTlsData = pData;
		break;
		}

}

//*****************************************************************************
//*
//*		ThreadTlsUnregister
//*
//*****************************************************************************
//	Removes a thread handle from the global thread list
//	hThread		: is the thread handle
static void ThreadTlsUnregister(void *hThread)
{
unsigned	uPos;


	for(uPos=0;uPos<THREADS_MAX;uPos++)
		{
		if(aTlsMap[uPos].hThread!=hThread)continue;
		aTlsMap[uPos].hThread  = 0;
		break;
		}

}

#endif

//*****************************************************************************
//*
//*		ThreadHandleError
//*
//*****************************************************************************
//	Shows an error if a handle is zero
#ifdef	_DEBUG
#include	<stdio.h>
int ThreadHandleError(const char *pFile,int iLine)
{
int		iLen;

	iLen = (int)strlen(pFile);
	while(iLen>0 && pFile[iLen-1]!='/' && pFile[iLen-1]!='\\')iLen--;
	printf("Handle-Error at %s:(%i)\n",pFile+iLen,iLine);

return 0;
}
#endif

//*****************************************************************************
//*
//*		ThreadMutexInitEx
//*
//*****************************************************************************
//	Allocates a new mutex
//	iLock		: locks the mutex after creation.
//	Returns the mutex handle.
void *ThreadMutexInitEx(int iLock)
{
LPCRITICAL_SECTION	pMutex;

	pMutex = (LPCRITICAL_SECTION)malloc(sizeof(CRITICAL_SECTION));
	InitializeCriticalSection(pMutex);
	if(iLock)EnterCriticalSection(pMutex);

return pMutex;
}

//*****************************************************************************
//*
//*		ThreadMutexExit
//*
//*****************************************************************************
//	Deletes a mutex
//	hMutex		: is the handle of the mutex
void ThreadMutexExit(ThreadMutex hMutex)
{
LPCRITICAL_SECTION	pMutex=(LPCRITICAL_SECTION)hMutex;

	if(!pMutex)return;
	DeleteCriticalSection(pMutex);
	free(pMutex);

}

//*****************************************************************************
//*
//*		ThreadBegin
//*
//*****************************************************************************
DWORD WINAPI ThreadBegin(LPVOID pParam)
{
ThreadData	   *pData=(ThreadData*)pParam;
void( _cdecl   *pStartProc)(void*);
void           *pStartParam;
const char 	   *pStartName;
void		   *hHandle;
#if				USE_TLS
char			aTlsData[TLS_SIZE];
#endif


	pStartParam	= pData->pParam;
	pStartProc	= pData->pProc;
	pStartName	= pData->pName;
	hHandle		= pData->hThread;

	free(pParam);

	#if USE_TLS
	memset(aTlsData,0,TLS_SIZE);
	#if THREAD_NAMES
	*(const char**)(aTlsData+THREAD_NAME_OFF)=pStartName;
	#endif
	ThreadTlsAdd(aTlsData);
	#if THREAD_CHECK
	ThreadTlsRegister((void*)GetCurrentThreadId(),aTlsData);
	#endif
	#endif

	pStartProc(pStartParam);

	#if THREAD_CHECK
	ThreadTlsUnregister((void*)GetCurrentThreadId());
	#endif

	CloseHandle(hHandle);


return 0;
}

//*****************************************************************************
//*
//*		ThreadStartEx
//*
//*****************************************************************************
//	pStartAddress	: void Proc(void *parameter)
//	uStackSize		: Size of Stack
//	parameter		: Is send to Proc
//	iPriority		: -256 to +256  ( 0 = normaly )
//	pName			: Is the name of the thread
void *ThreadStartEx(void( _cdecl  *pStartAddress )(void*),
					unsigned       uStackSize,
					void          *pParameter,
					int            iPriority,
					const char    *pName)
{
ThreadData *pStart;
HANDLE		hHandle;
DWORD		uIdNum;
int			iPrio;
int			i;



	if(uStackSize==0)
		{
		#ifdef _DEBUG
		printf("Wrong stack size in funktion ThreadStartEx(...)\n");
		#endif
		return 0;
		}


    if((iPriority&0xFFFF0000)==THREAD_GLOBAL)
        {
        iPriority &= 0xFFFF;
        }
    else{
        iPrio  = GetThreadPriority(GetCurrentThread());

        for(i=0;i<6;i++)
            {
            if(iPrio>=aPrio[i])break;
            }

        iPriority += (3-i)*73;
        }


	pStart			= (ThreadData*)malloc(sizeof(ThreadData));
	pStart->pProc	= pStartAddress;
	pStart->pParam	= pParameter;
	pStart->pName	= pName;


	#if USE_TLS
	if(iThreadTlsNum<0)ThreadTlsInit();
	ThreadTlsRegister((void*)GetCurrentThreadId(),aMainData);
	#endif


	    hHandle = CreateThread(0,uStackSize+TLS_SIZE,ThreadBegin,pStart,CREATE_SUSPENDED,&uIdNum);
	if(!hHandle)return 0;

	pStart->hThread = hHandle;

	#ifdef  THREAD_ALLOW_REALTIME

	     if(iPriority> 183)iPrio=THREAD_PRIORITY_TIME_CRITICAL;
	else if(iPriority> 110)iPrio=THREAD_PRIORITY_HIGHEST;
	else

	#endif

	     if(iPriority>  37)iPrio=THREAD_PRIORITY_ABOVE_NORMAL;
	else if(iPriority> -37)iPrio=THREAD_PRIORITY_NORMAL;
	else if(iPriority>-110)iPrio=THREAD_PRIORITY_BELOW_NORMAL;
	else if(iPriority>-183)iPrio=THREAD_PRIORITY_LOWEST;
	else                   iPrio=THREAD_PRIORITY_IDLE;

	//SetPriorityClass(hHandle,uClassNum);
	SetThreadPriority(hHandle,iPrio);

	if(ResumeThread(hHandle)==0xFFFFFFFF)
		{
		CloseHandle(hHandle);
		return 0;
		}



return (void*)uIdNum;
}

//*****************************************************************************
//*
//*		ThreadSemaphoreCount
//*
//*****************************************************************************
//	Gets the value of a semaphore
//	hSemaphore		: handle of the semaphore
//	Returns the semophore value between 0 and 0x70000000
int ThreadSemaphoreCount(ThreadSemaphore hSemaphore)
{
typedef struct  {UINT		uCount;	UINT uLimit;}SemaphoreInfo;
typedef LONG    (__stdcall *NtQuerySemaphoreType)(HANDLE hHandle,UINT uInfoClass,SemaphoreInfo *pInfo,UINT uInfoSize,PUINT pRetLen);
static NtQuerySemaphoreType	pNtQuerySemaphore=0;
static HMODULE				hNtDll=0;
static int					iDone=0;
SemaphoreInfo				sInfo;
UINT						uByte;
LONG						lRet;


	if(pNtQuerySemaphore==0)
		{
		if(hNtDll==0 && iDone<=0)
			{
			hNtDll = GetModuleHandle("NTDLL");
			iDone  = 1;
			}

		if(hNtDll && pNtQuerySemaphore==0 && iDone<=1)
			{
			pNtQuerySemaphore = (NtQuerySemaphoreType)GetProcAddress(hNtDll,"NtQuerySemaphore");
			iDone  = 2;
			}

		if(pNtQuerySemaphore==0)
			{
			return 0;
			}
		}

	   lRet = pNtQuerySemaphore(hSemaphore,0,&sInfo,sizeof(sInfo),&uByte);
    if(lRet<0)
		{
		return 0;
		}


return sInfo.uCount;
}



#endif



//*****************************************************************************
//*
//*
//*		LINUX-Code
//*
//*
//*****************************************************************************

#ifdef	linux

// add "/usr/lib/libpthread.so" and "/usr/lib/librt.so" to the project

#include	<unistd.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<time.h>
#include	<sched.h>
#include	<setjmp.h>
#include	<memory.h>
#include	<pthread.h>
#include	<semaphore.h>
#include	<sys/time.h>
#include	<sys/times.h>
#include	<sys/errno.h>
#include	<sys/resource.h>

static		int						iSchedMin=0;
static		int						iSchedMax=0;
static		clockid_t				iClockId=-1;

static		pthread_mutex_t 		hThreadMutex;	// is the handle mutex for all handles
#define		THREAD_INIT()			pthread_mutex_init  (&hThreadMutex,0);
#define		THREAD_LOCK()			pthread_mutex_lock  (&hThreadMutex);
#define		THREAD_UNLOCK()			pthread_mutex_unlock(&hThreadMutex);
extern		int 					pthread_getattr_np(pthread_t,pthread_attr_t*);

#ifndef		CLOCK_MONOTONIC
#define 	CLOCK_MONOTONIC     	1
#endif

typedef struct
	{
	void      (*pStartProc)(void*);
	void       *pStartArg;
	const char *pStartName;
	}ThreadData;

typedef struct
	{
	int 				iId;
	int 				iValue;
	int 				iNumber;
	pthread_mutex_t 	hAccess;
	pthread_cond_t  	hWait;
	}SemData;

typedef struct
	{
	int 				iId;
	pthread_mutex_t 	hAccess;
	}MtxData;

static		pthread_key_t			iThreadTlsNum = -1;

#define 	ThreadTick()			ThreadTickCount()
#define 	ThreadTickToMs(t)		(t)
#define 	ThreadTlsSet(n,p)		pthread_setspecific(n,p)
#define 	ThreadTlsGet(n)			(char*)(pthread_getspecific(n))

#if			THREAD_CHECK

typedef struct
   {
   void    *hThread;
   char    *pTlsData;
   }TlsMap;


static		TlsMap			aTlsMap[THREADS_MAX];
static		unsigned 		uTlsMapPos = 0;

//*****************************************************************************
//*
//*		ThreadTlsGetOf
//*
//*****************************************************************************
//	Gets the info data field to a thread handle
//	hThread		: is the thread handle
//	iNum		: ist the tls slot number
//	Returns the pointer to the info data field, or NULL if an error occurs.
static char *ThreadTlsGetOf(void *hThread,int iNum)
{
unsigned	uPos;


	for(uPos=0;uPos<THREADS_MAX;uPos++)
		{
		if(aTlsMap[uPos].hThread!=hThread)continue;
		return aTlsMap[uPos].pTlsData;
		}


return 0;
}

//*****************************************************************************
//*
//*		ThreadTlsRegister
//*
//*****************************************************************************
//	Insert a thread handle in the global thread list
//	hThread		: is the thread handle
//	pData		: pointer to the info data field
static void ThreadTlsRegister(void *hThread,char *pData)
{
unsigned	uPos;


	for(uPos=0;uPos<THREADS_MAX;uPos++)
		{
		if(aTlsMap[uPos].hThread)continue;
		aTlsMap[uPos].hThread  = hThread;
		aTlsMap[uPos].pTlsData = pData;
		break;
		}

}

//*****************************************************************************
//*
//*		ThreadTlsUnregister
//*
//*****************************************************************************
//	Removes a thread handle from the global thread list
//	hThread		: is the thread handle
void ThreadTlsUnregister(void *hThread)
{
unsigned	uPos;


	for(uPos=0;uPos<THREADS_MAX;uPos++)
		{
		if(aTlsMap[uPos].hThread!=hThread)continue;
		aTlsMap[uPos].hThread  = 0;
		break;
		}

}

#endif

//*****************************************************************************
//*
//*		ThreadTlsAlloc
//*
//*****************************************************************************
//	Allocates a new thread local storage
#if USE_TLS
static int ThreadTlsAlloc()
{
pthread_key_t	iKey;

	if(!pthread_key_create(&iKey,0))
		{
		return iKey;
		}

return -1;
}
#endif

//*****************************************************************************
//*
//*		ThreadTlsPointer
//*
//*****************************************************************************
//	Gets a pointer to the tls memory
char *ThreadTlsPointer()
{
return ThreadTlsGet(iThreadTlsNum);
}

//*****************************************************************************
//*
//*		ThreadTickCount
//*
//*****************************************************************************
//	Returns the current time in ms
unsigned ThreadTickCount()
{
struct timespec 	sTime;
struct timeval		sTval;



	if(iClockId>=0)
		{
		clock_gettime(iClockId,&sTime);
		return sTime.tv_sec*1000+sTime.tv_nsec/1000000;
		}

	if(iClockId==-1)
		{
		if(!clock_gettime(CLOCK_MONOTONIC,&sTime))
			{
			iClockId = CLOCK_MONOTONIC;
			return sTime.tv_sec*1000+sTime.tv_nsec/1000000;
			}

		if(!clock_gettime(CLOCK_REALTIME,&sTime))
			{
			iClockId = CLOCK_REALTIME;
			return sTime.tv_sec*1000+sTime.tv_nsec/1000000;
			}

		iClockId = -2;
		}

	gettimeofday(&sTval,0);


return sTval.tv_sec*1000+sTval.tv_usec/1000;
}

//*****************************************************************************
//*
//*		ThreadSleep
//*
//*****************************************************************************
// Stops the current thread for some milliseconds
void ThreadSleep(unsigned uMilliSec)
{
pthread_condattr_t	sAttr;
pthread_mutex_t 	hMtx;
pthread_cond_t  	hCnd;
struct timespec 	sTi;
struct timeval		sTv;



	if(iClockId==-1)
		{
		ThreadTickCount();
		}

    pthread_mutex_init(&hMtx,0);
    pthread_cond_init (&hCnd,0);

	if(iClockId>=0 && !clock_gettime(iClockId,&sTi))	// use monotonic time
		{
		pthread_condattr_init(&sAttr);
		pthread_condattr_setclock(&sAttr,iClockId);
		pthread_cond_init(&hCnd,&sAttr);

	    sTi.tv_nsec += (uMilliSec%1000)*1000000;
		sTi.tv_sec  += (sTi.tv_nsec   )/1000000000;
	    sTi.tv_sec  += (uMilliSec/1000);
	    sTi.tv_nsec %= 1000000000;
		}
	else{												// calculate wait point
		gettimeofday(&sTv, 0);

	    sTi.tv_nsec  =(sTv.tv_usec + (uMilliSec % 1000) * 1000) * 1000;
	    sTi.tv_sec   = sTv.tv_sec  + (uMilliSec / 1000) + (sTi.tv_nsec / 1000000000);
	    sTi.tv_nsec %= 1000000000;
		}


    pthread_mutex_lock    (&hMtx);
    pthread_cond_timedwait(&hCnd,&hMtx,&sTi);
    pthread_mutex_unlock  (&hMtx);

	pthread_cond_destroy  (&hCnd);
    pthread_mutex_destroy (&hMtx);

}



//*****************************************************************************
//*
//*		ThreadBegin
//*
//*****************************************************************************
//	local dummy function for thread starting
static void *ThreadBegin(void *pParam)
{
ThreadData		sThread;
#if				USE_TLS
char			aTlsData[TLS_SIZE];
#endif


	memcpy(&sThread,pParam,sizeof(sThread));
	free(pParam);

	#if USE_TLS
	memset(aTlsData,0,TLS_SIZE);
	#if THREAD_NAMES
	*(char**)(aTlsData+THREAD_NAME_OFF)=pStartName;
	#endif
	ThreadTlsAdd(aTlsData);
	#if THREAD_CHECK
	THREAD_LOCK();
	ThreadTlsRegister((void*)(pthread_self()+1),aTlsData);
	THREAD_UNLOCK();
	#endif
	#endif

	sThread.pStartProc(sThread.pStartArg);

	#if THREAD_CHECK
	ThreadTlsUnregister((void*)(pthread_self()+1));
	#endif



return NULL;
}

//*****************************************************************************
//*
//*		ThreadStartEx
//*
//*****************************************************************************
//	Start a new thread
//	pStartAddress	: void Proc(void *pParam)
//	uStackSize		: Size of Stack
//	pParam			: Is send to Proc
//	iPriority		: -256 to +256  ( 0 = normaly )
//	pName			: Is the name of the thread
//	Returns the thread ID or 0 if an error occurs.
void *ThreadStartEx(void( _cdecl  *pStartAddress)(void*),
					unsigned       uStackSize,
					void          *pParam,
					int            iPriority,
					const char    *pName)
{
static int 			iIsInit=0;
ThreadData		   *pThread;
struct sched_param	sParam;
pthread_t			iThreadId;
pthread_attr_t		sCurrent;
pthread_attr_t		sAttrib;
int					iPrior;



	if(!uStackSize)
		{
		return 0;
		}


		pThread = (ThreadData*)malloc(sizeof(ThreadData));
	if(!pThread)return 0;

	if(!iIsInit)
		{
		iIsInit=1;
		THREAD_INIT();

		#if USE_TLS

		if((int)iThreadTlsNum==-1)
			{
			ThreadTlsInit();
			}

		#endif

		iSchedMin = sched_get_priority_max(SCHED_FIFO);
		iSchedMax = sched_get_priority_min(SCHED_FIFO);

		if(iSchedMax<=iSchedMin)
			{
			iSchedMax = iSchedMin+1;
			}
		}


	memset(&sAttrib,0,sizeof(sAttrib));
	memset(&sParam ,0,sizeof(sParam ));

	pthread_attr_init(&sAttrib);

	if((iPriority&0xFFFF0000)==THREAD_GLOBAL)
		{
		iPrior		= iPriority&0xFFFF;
		iPrior	   *= iSchedMax-iSchedMin;
		iPrior     /= 256;
		iPrior     += iSchedMin;
		}
	else{
		pthread_getattr_np(pthread_self(),&sCurrent);
		pthread_attr_getschedparam(&sCurrent,&sParam);

		iPrior      = iPriority;
		iPrior	   *= iSchedMax-iSchedMin;
		iPrior     /= 256;
		iPrior     += iSchedMin;
		iPrior	   += sParam.sched_priority;
		}

	if(iPrior<iSchedMin)iPrior=iSchedMin;
	if(iPrior>iSchedMax)iPrior=iSchedMax;

	sParam.sched_priority = iPrior;
	pThread->pStartProc	  = pStartAddress;
	pThread->pStartArg	  = pParam;
	pThread->pStartName	  = pName;


		pthread_attr_setschedparam(&sAttrib,&sParam);
		pthread_attr_setstacksize (&sAttrib,uStackSize+TLS_SIZE);
	if( pthread_create(&iThreadId ,&sAttrib,ThreadBegin,pThread))
		{
		free(pThread);
		return 0;
		}


	pthread_detach(iThreadId);



return (void*)(iThreadId+1);
}

//*****************************************************************************
//*
//*		ThreadMutexInitEx
//*
//*****************************************************************************
//	Create a new mutex
//	iLock	: shold be mutex locked after creation
//	Returns a handle to the mutex
ThreadMutex ThreadMutexInitEx(int iLock)
{
pthread_mutexattr_t  sAttrib;
MtxData				*pMtx;


	pMtx=(MtxData*)malloc(sizeof(MtxData));

	pthread_mutexattr_init(&sAttrib);
	pthread_mutexattr_settype(&sAttrib,PTHREAD_MUTEX_RECURSIVE);
	pthread_mutex_init(&pMtx->hAccess,&sAttrib);

	pMtx->iId     = 0x1255AA3C;

	if(!iLock)pthread_mutex_unlock(&pMtx->hAccess);


return (ThreadMutex)pMtx;
}

//*****************************************************************************
//*
//*		ThreadMutexExit
//*
//*****************************************************************************
//	Deletes a mutex
//	hMtxHandle		: handle of the mutex
void ThreadMutexExit(ThreadMutex hMtxHandle)
{
MtxData		*pMtx;


	if(!hMtxHandle)return;

	   pMtx=(MtxData*)hMtxHandle;
	if(pMtx->iId!=0x1255AA3C)return;

	pMtx->iId     = 0;

	pthread_mutex_unlock (&pMtx->hAccess);
    pthread_mutex_destroy(&pMtx->hAccess);

	free(pMtx);

}

//*****************************************************************************
//*
//*		ThreadMutexTake
//*
//*****************************************************************************
//	Locks a mutex only if it is unlocked
//	hMtxHandle		: handle of the mutex
//	Returns 1 if successful, or 0 if not
int ThreadMutexTake(ThreadMutex hMtxHandle)
{
MtxData		*pMtx;



	if(!hMtxHandle)return 0;

	   pMtx=(MtxData*)hMtxHandle;
	if(pMtx->iId!=0x1255AA3C)return 0;						// check mutex

	// REM AZ sollte noch getetset werden

return (pthread_mutex_trylock(&pMtx->hAccess))? 0:1;
}

//*****************************************************************************
//*
//*		ThreadMutexWait
//*
//*****************************************************************************
//	Waits until a mutex is signaled and locks it
//	hMtxHandle		: handle of the mutex
void ThreadMutexWait(ThreadMutex hMtxHandle)
{
MtxData		*pMtx;



	if(!hMtxHandle)return;

	   pMtx=(MtxData*)hMtxHandle;
	if(pMtx->iId!=0x1255AA3C)return;						// check mutex

	pthread_mutex_lock(&pMtx->hAccess);

}

//*****************************************************************************
//*
//*		ThreadMutexRelease
//*
//*****************************************************************************
//	Releases a mutex
//	hMtxHandle		: handle of the mutex
void ThreadMutexRelease(ThreadMutex hMtxHandle)
{
MtxData    *pMtx;


	if(!hMtxHandle)return;

	   pMtx=(MtxData*)hMtxHandle;
	if(pMtx->iId!=0x1255AA3C)return;

	pthread_mutex_unlock(&pMtx->hAccess);

}


//*****************************************************************************
//*
//*		ThreadSemaphoreInitEx
//*
//*****************************************************************************
//	Creates a semaphore
ThreadSemaphore ThreadSemaphoreInitEx(int iInitCount)
{
pthread_condattr_t	sAttr;
SemData			   *pSem;



	pSem	= (SemData*)malloc(sizeof(SemData));

	pthread_mutex_init (&pSem->hAccess, 0);
	pthread_cond_init  (&pSem->hWait  , 0);


	if(iClockId==-1)
		{
		ThreadTickCount();
		}

	if(iClockId>=0)
		{
		pthread_condattr_init    (&sAttr);
		pthread_condattr_setclock(&sAttr,iClockId);
		pthread_cond_init(&pSem->hWait,&sAttr);
		}


	pSem->iId		= 0x1255AA78;
	pSem->iValue	= iInitCount;
	pSem->iNumber	= 0;


return (ThreadSemaphore)pSem;
}

//*****************************************************************************
//*
//*		ThreadSemaphoreExit
//*
//*****************************************************************************
//	Deletes a semaphore
void ThreadSemaphoreExit(ThreadSemaphore hSemHandle)
{
SemData		*pSem;


	if(!hSemHandle)return;

	   pSem=(SemData*)hSemHandle;
	if(pSem->iId!=0x1255AA78)return;

	pSem->iId     = 0;
	pSem->iValue  = 0x40000000;
	pSem->iNumber = 0;

	pthread_cond_signal  (&pSem->hWait);
	pthread_mutex_unlock (&pSem->hAccess);
	pthread_cond_destroy (&pSem->hWait);
    pthread_mutex_destroy(&pSem->hAccess);

	free(pSem);

}

//*****************************************************************************
//*
//*		ThreadSemaphoreWait
//*
//*****************************************************************************
//	Waits until a semophore is signaled and locks it
void ThreadSemaphoreWait(ThreadSemaphore hSemHandle)
{
SemData		*pSem;



	if(!hSemHandle)return;

	   pSem=(SemData*)hSemHandle;
	if(pSem->iId!=0x1255AA78)return;						// check semophore

	pthread_mutex_lock(&pSem->hAccess);

	   pSem->iValue--;
	if(pSem->iValue<0)										// is the semophore locked ?
		{
		pSem->iNumber++;

		pthread_cond_wait(&pSem->hWait,&pSem->hAccess);		// UNLOCK(hAccess) --> LOCK(hAccess)

		if(pSem->iId!=0x1255AA78)return;					// check semophore

		pSem->iNumber--;
		}

	pthread_mutex_unlock(&pSem->hAccess);


}

//*****************************************************************************
//*
//*		ThreadSemaphoreTimed
//*
//*****************************************************************************
//	Waits until a semophore is signaled and locks it
//	hSemaphore	: hHandle of the semaphore
//	uMillisec	: wait time in milliseconds
//	returns 1 if the semophore is locked, or 0 at a timeout
int ThreadSemaphoreTimed(ThreadSemaphore hSemaphore,unsigned uMillisec)
{
SemData			   *pSem;
struct timeval		sTv;
struct timespec 	sTi;
int					iRet;



	if(!hSemaphore)return 1;

	   pSem=(SemData*)hSemaphore;
	if(pSem->iId!=0x1255AA78)return 1;						// check semophore

	pthread_mutex_lock(&pSem->hAccess);

	   pSem->iValue--;
	if(pSem->iValue<0)										// is the semophore locked ?
		{
		pSem->iNumber++;

		if(iClockId>=0)										// use monotonic time
			{
			clock_gettime(iClockId,&sTi);

		    sTi.tv_nsec += (uMillisec%1000)*1000000;
			sTi.tv_sec  += (sTi.tv_nsec   )/1000000000;
		    sTi.tv_sec  += (uMillisec/1000);
		    sTi.tv_nsec %= 1000000000;
			}
		else{
			gettimeofday(&sTv,0);							// calculate timeout point

		    sTi.tv_nsec  =(sTv.tv_usec + (uMillisec%1000) * 1000) * 1000;
		    sTi.tv_sec   = sTv.tv_sec  + (uMillisec/1000) + (sTi.tv_nsec/1000000000);
		    sTi.tv_nsec %= 1000000000;
			}

															// UNLOCK(hAccess) --> LOCK(hAccess)
		iRet = pthread_cond_timedwait(&pSem->hWait,&pSem->hAccess,&sTi);

		if(pSem->iId!=0x1255AA78)return 0;					// check semophore

		pSem->iNumber--;

		if(iRet==ETIMEDOUT)									// timeout happend
			{
			pSem->iValue++;
			pthread_mutex_unlock(&pSem->hAccess);
			return 0;
			}
		}

	pthread_mutex_unlock(&pSem->hAccess);


return 1;
}

//*****************************************************************************
//*
//*		ThreadSemaphoreReleaseEx
//*
//*****************************************************************************
//	Releases a semaphore several times
//	hSemaphore		: handle of the semaphore
//	iCount			: is the count of releases (must be greater 0)
void ThreadSemaphoreReleaseEx(ThreadSemaphore hSemHandle,int iCount)
{
SemData    *pSem;
int			iNum;
int			iVal;


	if(!hSemHandle || iCount<=0)return;

	   pSem=(SemData*)hSemHandle;
	if(pSem->iId!=0x1255AA78)return;

	pthread_mutex_lock(&pSem->hAccess);

	if(iCount>0x70000000)
		{
		iCount = 0x70000000;
		}

	if(pSem->iValue>0)										// check count value
		{
		   iVal = pSem->iValue+iCount;
		if(iVal<=0 || iVal>0x70000000)
			{
			iCount = 0x70000000-pSem->iValue;
			}
		}

	   iNum = pSem->iNumber;
	if(iNum>iCount)iNum=iCount;

	pSem->iValue += iCount;

	while(iNum>0)											// waits a thread on this semaphore
		{
		pthread_cond_signal(&pSem->hWait);
		iNum--;
		}

	pthread_mutex_unlock(&pSem->hAccess);

}


//*****************************************************************************
//*
//*		ThreadSemaphoreCount
//*
//*****************************************************************************
//	Gets the value of a semaphore
//	hSemaphore		: handle of the semaphore
//	Returns the semophore value between 0 and 0x70000000
int ThreadSemaphoreCount(ThreadSemaphore hSemaphore)
{
SemData	   *pSem;
int			iRet;


	if(!hSemaphore)return 0;

	   pSem=(SemData*)hSemaphore;
	if(pSem->iId!=0x1255AA78)return 0;						// check semophore

	   iRet = pSem->iValue;
	if(iRet<0)return 0;


return iRet;
}


#endif


//*****************************************************************************
//*
//*
//*		RTTARGET-Code
//*
//*
//*****************************************************************************

#ifdef RTTARGET

#include	<rttarget.h>
#include	<rtfiles.h>
#include	<string.h>
#include	<rtk32.h>
#include	<clock.h>
#include	<finetime.h>



typedef struct
	{
	volatile void	   (_cdecl *pProc)(void*);				// function for start
	volatile void			   *pParam;						// parameter for function
	}ThreadData;


#ifdef		_MSC_VER
#pragma		warning( disable :4090)
#pragma		warning( disable :4990)
#pragma		warning( disable :4311)
#pragma		warning( disable :4312)
#endif

#define 	ThreadTick()		RTKGetTime()
#define 	ThreadTickToMs(t)	CLKTicksToMilliSecs(t)
#define 	ThreadTlsAlloc()	RTKAllocUserData()
#define 	ThreadTlsSet(n,p)	RTKSetUserData(RTKCurrentTaskHandle(),n,p)
#define 	ThreadTlsGet(n)		((char*)RTKGetUserData(RTKCurrentTaskHandle(),n))
#define 	ThreadTlsGetOf(t,n)	((char*)RTKGetUserData(t,n))


extern		UINT RTKAPI        _rtkCLKCurrentTimerInterval(void);
extern		DWORD				FTUnitsPerSecond;
extern		int					FTDivisor;
static		RTKSemaphore		hStartSem		 =  0;		// semaphore for Starting threads
			int					iThreadTlsNum	 = -1;		// is the used TLS section


//*****************************************************************************
//*
//*		ThreadWithTls
//*
//*****************************************************************************
#if USE_TLS

static void _cdecl ThreadWithTls(void *pParam)
{
ThreadData	   *pData=(ThreadData*)pParam;
void( _cdecl   *pStartProc)(void*);
char			aTlsData[TLS_SIZE];


	pParam		= pData->pParam;
	pStartProc	= pData->pProc;

	memset(aTlsData,0,TLS_SIZE);
	ThreadTlsAdd(aTlsData);

	RTKSignal(hStartSem);

	pStartProc(pParam);

}

#endif

//*****************************************************************************
//*
//*		ThreadStartEx
//*
//*****************************************************************************
//	pStartProc		: void Proc(void *parameter)
//	dwStackSize		: Size of Stack
//	pParam			: Is send to Proc
//	iPriority		: -256 to +256  ( 0 = normaly )
//	pName			: Is the name of the thread
void *ThreadStartEx(void( _cdecl  *pStartProc)(void*),
					unsigned       dwStackSize,
					void          *pParam,
					int            iPriority,
					const char    *pName)
{
RTKTaskHandle	uId;
int				iPrior;


	if(!dwStackSize)
		{
		return 0;
		}

    if((iPriority&0xFFFF0000)==THREAD_GLOBAL)
        {
        iPrior = iPriority&0xFFFF;

		#if RTT32_VER<600
		iPrior /= 4;
		#endif
        }
    else{
        iPrior = RTKGetTaskPrio(RTKCurrentTaskHandle());
        iPrior = iPriority;

		#if RTT32_VER<600
		iPrior /= 4;
		#endif
        }

	#if RTT32_VER>=600

	if(iPrior<  1)iPrior=1;
	if(iPrior>256)iPrior=256;

	#else

	if(iPrior< 1)iPrior=1;
	if(iPrior>64)iPrior=64;

	#endif

	#if USE_TLS
	{
	static ThreadData	sData;

	if(iThreadTlsNum==-1)
		{
		ThreadTlsInit();
		}

	if(!hStartSem)hStartSem=RTKOpenSemaphore(ST_COUNTING,1,0,0)	;

	RTKWait(hStartSem);

	sData.pParam = pParam;
	sData.pProc	 = pStartProc;


		uId=RTKCreateThread(ThreadWithTls,iPrior,dwStackSize+TLS_SIZE,0,&sData,pName);
	if(!uId)RTKSignal(hStartSem);
	}
	#else

	uId = RTKCreateThread(pStartProc,iPrior,dwStackSize,0,pParam,pName);

	#endif


return (void*)uId;
}

//*****************************************************************************
//*
//*		ThreadTickCount
//*
//*****************************************************************************
unsigned ThreadTickCount()
{
const unsigned 	uClkUnitsPerSecond = 14318180/4/1000;
const unsigned 	uClkDivisor        = 12/4;
RTKFineTime		sTime;
unsigned		uTick;
unsigned		uNext;
unsigned		uUpper;
static unsigned uHigh[8]={0,0,0,0,0,0,0,0};
static unsigned	uLast=0;



	uTick	= RTKGetTime();
	uUpper	= uTick>>29;

	if(uUpper!=uLast)
		{
		uLast  =  uUpper;
		uNext  = (uUpper-4)&7;

		if(uUpper>=4)
			{
			uHigh[uNext] = uHigh[uUpper]+1;
			}
		else{
			uHigh[uNext] = uHigh[uUpper];
			}
		}


	sTime.TimeHigh = uHigh[uUpper];
	sTime.TimeLow  = uTick;

	FTMultiply(&sTime,&sTime,_rtkCLKCurrentTimerInterval());


return FTMultDivDouble(&sTime,&sTime,uClkDivisor,uClkUnitsPerSecond);
}

//*****************************************************************************
//*
//*		ThreadMutexInitEx
//*
//*****************************************************************************
//	Creates a mutex
ThreadSemaphore ThreadMutexInitEx(int iLock)
{
return RTKOpenSemaphore(ST_MUTEX,iLock,0,0);
}

//*****************************************************************************
//*
//*		ThreadMutexExit
//*
//*****************************************************************************
//	Deletes a mutex
//	hMutex		: is the handle of the mutex
void ThreadMutexExit(ThreadMutex hMutex)
{
	RTKDeleteSemaphore((RTKSemaphore*)&hMutex);
}

//*****************************************************************************
//*
//*		ThreadSemaphoreInitEx
//*
//*****************************************************************************
//	Creates a semaphore
ThreadSemaphore ThreadSemaphoreInitEx(int iInitCount)
{
return RTKOpenSemaphore(ST_COUNTING,iInitCount,0,0);
}

//*****************************************************************************
//*
//*		ThreadSemaphoreExit
//*
//*****************************************************************************
//	Deletes a semaphore
void ThreadSemaphoreExit(ThreadSemaphore hSemHandle)
{
	RTKDeleteSemaphore(&(RTKSemaphore)hSemHandle);
}

//*****************************************************************************
//*
//*		ThreadSemaphoreReleaseEx
//*
//*****************************************************************************
void ThreadSemaphoreReleaseEx(void *hSemaphore,int iCount)
{

	for(;iCount>0;iCount--)RTKSignal(hSemaphore);

}

#endif

//*****************************************************************************
//*
//*
//*		FreeRtos-Code
//*
//*
//*****************************************************************************

#ifdef FREERTOS

#include	<FreeRTOS.h>
#include	<string.h>
#include	"task.h"
#include	"queue.h"
#include	"semphr.h"

#ifdef		portUSING_MPU_WRAPPERS
#if			portUSING_MPU_WRAPPERS
#error		"wrong FreeRtos configuration, portUSING_MPU_WRAPPERS isn't allowed"
#endif
#endif

#ifndef		configUSE_APPLICATION_TASK_TAG
#error		"wrong FreeRtos configuration, configUSE_APPLICATION_TASK_TAG must be 1"
#endif

#if			configUSE_APPLICATION_TASK_TAG==0
#error		"wrong FreeRtos configuration, configUSE_APPLICATION_TASK_TAG must be 1"
#endif

#if			configUSE_COUNTING_SEMAPHORES==0
#error		"wrong FreeRtos configuration, configUSE_COUNTING_SEMAPHORES must be 1"
#endif

#if			configUSE_MUTEXES==0
#error		"wrong FreeRtos configuration, configUSE_MUTEXES must be 1"
#endif

#if			configUSE_RECURSIVE_MUTEXES==0
#error		"wrong FreeRtos configuration, configUSE_RECURSIVE_MUTEXES must be 1"
#endif

#if			configUSE_16_BIT_TICKS!=0
#error		"wrong FreeRtos configuration, configUSE_16_BIT_TICKS must be 0"
#endif

#if			INCLUDE_vTaskDelay==0
#error		"wrong FreeRtos configuration, INCLUDE_vTaskDelay must be 1"
#endif

#if			INCLUDE_xTaskGetCurrentTaskHandle==0
#error		"wrong FreeRtos configuration, INCLUDE_xTaskGetCurrentTaskHandle must be 1"
#endif

#ifndef		SYS_TICK_RATE_HZ
#define 	SYS_TICK_RATE_HZ	configTICK_RATE_HZ
#endif

#if			SYS_TICK_RATE_HZ != configTICK_RATE_HZ
#error		"SYS_TICK_RATE_HZ differs to configTICK_RATE_HZ of FreeRtos"
#endif

#ifdef	   _MSC_VER
#define 	uint64		unsigned __int64
#else
#define 	uint64		unsigned long long
#endif

#ifndef 	MAIN_TASK_STACK
#define 	MAIN_TASK_STACK		0x8000
#endif
#ifndef		MAIN_PRIORITY
#define		MAIN_PRIORITY		((4*configMAX_PRIORITIES)/256)
#endif


typedef struct
	{
	void (_cdecl *pProc)(void*);
	void         *pParam;
	const char   *pName;
	}ThreadData;


extern void SysMain(int iArgc,char **pArgv);				// entry point pre SysInit()
extern int	FreeRtosMainTask(int iArgc,char **pArgv);		// entry point for main task
static		char			  **pMainArgv;					// arguments pointers in main()
static		int					iMainArgcRet;				// count of arguments in main()
static		SemaphoreHandle_t	hStartSem		 =  0;		// semaphore for Starting threads
			int					iThreadTlsNum	 = -1;		// is the used TLS section


#define 	ThreadTick()		xTaskGetTickCount()
#define 	ThreadTickToMs(t)	ThreadSysToTicks(t)

#define 	ThreadTlsAlloc()	0
#define 	ThreadTlsSet(n,p)	vTaskSetApplicationTaskTag(xTaskGetCurrentTaskHandle(),(TaskHookFunction_t)p)
#define 	ThreadTlsGet(n)		((char*)xTaskGetApplicationTaskTag(xTaskGetCurrentTaskHandle()))
#define 	ThreadTlsGetOf(t,n)	((char*)xTaskGetApplicationTaskTag(t))

//*****************************************************************************
//*
//*		ThreadMutexInitEx
//*
//*****************************************************************************
//	Allocates a new mutex
//	iLock		: locks the mutex after creation.
//	Returns the mutex handle.
extern void *ThreadMutexInitEx(int iLock)
{
ThreadMutex		hMutex;


	hMutex = ThreadMutexInit();
	if(iLock)ThreadMutexWait(hMutex);

return hMutex;
}

//*****************************************************************************
//*
//*		ThreadTickCountInternal
//*
//*****************************************************************************
//	GetTictCount for FreeRtos
unsigned ThreadTickCountInternal()
{
static unsigned uHigh[8]={0,0,0,0,0,0,0,0};
static unsigned	uLast=0;
uint64			uTime;
unsigned		uTick;
unsigned		uNext;
unsigned		uUpper;



	uTick	= xTaskGetTickCount();
	uUpper	= uTick>>29;

	if(uUpper!=uLast)
		{
		uLast  =  uUpper;
		uNext  = (uUpper-4)&7;

		if(uUpper>=4)
			{
			uHigh[uNext] = uHigh[uUpper]+1;
			}
		else{
			uHigh[uNext] = uHigh[uUpper];
			}
		}


	uTime		= uHigh[uUpper];
	uTime	  <<= 32;
	uTime	   |= uTick;
	uTime	   *= (0x10000*1000)/SYS_TICK_RATE_HZ;
	uTime	   /=  0x10000;


return (unsigned)uTime;
}

//*****************************************************************************
//*
//*		ThreadSysToTicks
//*
//*****************************************************************************
//	Convert system ticks to milliseconds
int ThreadSysToTicks(unsigned uMillSec)
{
	#if SYS_TICK_RATE_HZ!=1000

	uMillSec *= (SYS_TICK_RATE_HZ*1024)/1000;
	uMillSec /= 1024;

	#endif

return uMillSec;
}


//*****************************************************************************
//*
//*		ThreadSleepInternal
//*
//*****************************************************************************
//	Sleep funktion for internal use
void ThreadSleepInternal(unsigned uMillSec)
{

	#if SYS_TICK_RATE_HZ!=1000

	uMillSec *= (SYS_TICK_RATE_HZ*1024)/1000;
	uMillSec /= 1024;

	#endif

	vTaskDelay(uMillSec);

}

//*****************************************************************************
//*
//*		ThreadSemaphoreReleaseIsr
//*
//*****************************************************************************
//	ThreadSemaphoreRelease for using in interrrupts
void ThreadSemaphoreReleaseIsr(ThreadSemaphore hSemaphore)
{
BaseType_t	iTaskWoken = pdFALSE;


	xSemaphoreGiveFromISR(hSemaphore,&iTaskWoken);
    portYIELD_FROM_ISR(iTaskWoken);

}

//*****************************************************************************
//*
//*		ThreadWithTls
//*
//*****************************************************************************
#if USE_TLS

static void _cdecl ThreadWithTls(void *pParam)
{
ThreadData	   *pData=(ThreadData*)pParam;
void( _cdecl   *pStartProc)(void*);
char			aTlsData[TLS_SIZE];


	pParam		= pData->pParam;
	pStartProc	= pData->pProc;

	memset(aTlsData,0,TLS_SIZE);
	ThreadTlsAdd(aTlsData);

	xSemaphoreGive(hStartSem);

	pStartProc(pParam);

	vTaskDelete(NULL);

}

#else

static void _cdecl ThreadStartProc(void *pParam)
{
ThreadData	   *pData=(ThreadData*)pParam;
void( _cdecl   *pStartProc)(void*);


	pParam		= pData->pParam;
	pStartProc	= pData->pProc;

	xSemaphoreGive(hStartSem);

	pStartProc(pParam);

	vTaskDelete(NULL);

}

#endif

//*****************************************************************************
//*
//*		ThreadStartEx
//*
//*****************************************************************************
//	pStartProc		: void Proc(void *parameter)
//	dwStackSize		: Size of Stack
//	pParam			: Is send to Proc
//	iPriority		: -256 to +256  ( 0 = normaly )
//	pName			: Is the name of the thread
void *ThreadStartEx(void( _cdecl  *pStartProc)(void*),
					unsigned       dwStackSize,
					void          *pParam,
					int            iPriority,
					const char    *pName)
{
static ThreadData	sData;
int					iStack;
int					iPrior;
TaskHandle_t       	hTask;


	if(!dwStackSize)
		{
		return 0;
		}

    if((iPriority&0xFFFF0000)==THREAD_GLOBAL)
        {
        iPrior = iPriority&0xFFFF;
        iPrior>>=2;
        }
    else{
        iPrior  = uxTaskPriorityGet(xTaskGetCurrentTaskHandle());

		#if configMAX_PRIORITIES!=256
		iPrior *= 256;
		iPrior /= configMAX_PRIORITIES;
		#endif

        iPrior += iPriority;
        }


	#if configMAX_PRIORITIES!=256

	iPrior *= configMAX_PRIORITIES;
	iPrior /= 256;

	#endif

	if(iPrior<                    1)iPrior=1;
	if(iPrior>=configMAX_PRIORITIES)iPrior=configMAX_PRIORITIES-1;



	if(!hStartSem)hStartSem=xSemaphoreCreateCounting(0x70000000,1);

	xSemaphoreTake(hStartSem,portMAX_DELAY);

	sData.pParam = pParam;
	sData.pProc	 = pStartProc;


	#if USE_TLS

	if(iThreadTlsNum==-1)
		{
		ThreadTlsInit();
		}

	dwStackSize += TLS_SIZE;
	pStartProc   = ThreadWithTls;

	#else

	pStartProc   = ThreadStartProc;

	#endif


	   iStack = dwStackSize/sizeof(long);
	if(iStack>0xFFFF)iStack=0xFFFF;

	if(xTaskCreate(pStartProc,pName,(unsigned short)iStack,&sData,iPrior,&hTask)!=pdPASS)
		{
		return 0;
		}


return (void*)hTask;
}

//*****************************************************************************
//*
//*		MainStarter
//*
//*****************************************************************************
//	Entry point for main task
static void _cdecl MainStarter(void *pParam)
{

	iMainArgcRet = FreeRtosMainTask(iMainArgcRet,pMainArgv);
	vTaskDelete(NULL);

}

//*****************************************************************************
//*
//*		main
//*
//*****************************************************************************
//	The function main() is defined in this library becase in FreeRtos the
//	call vTaskStartScheduler() don't return to the caller. So an other main
//	task will be started, and the original main is renamed to FreeRtosMainTask.
int main(int iArgc,char **pArgv)
{
unsigned	uStack;
unsigned	uPrio;


	pMainArgv		= pArgv;					// arguments pointers in main()
	iMainArgcRet	= iArgc;					// count of arguments in main()

	SysMain(iArgc,pArgv);						// pre call for SysInit()

	   uStack=MAIN_TASK_STACK/sizeof(StackType_t);
	if(uStack>0xFFFF)uStack=0xFFFF;

	   uPrio = MAIN_PRIORITY;
	if(uPrio<=tskIDLE_PRIORITY)uPrio=tskIDLE_PRIORITY+1;

	xTaskCreate(MainStarter,"MainTask",(unsigned short)uStack,NULL,uPrio,0);

	vTaskStartScheduler();



return iMainArgcRet;
}


#endif

//*****************************************************************************
//*
//*
//*		GLOBAL Functions
//*
//*
//*****************************************************************************


//*****************************************************************************
//*
//*		ThreadSecCount
//*
//*****************************************************************************
//	returns a monoton rising secound counter
unsigned ThreadSecCount()
{
unsigned		uTick;
unsigned		uNext;
unsigned		uUpper;
static unsigned uHigh[8]={0,0,0,0,0,0,0,0};
static unsigned	uLast=0;



	uTick	= ThreadTickCount();
	uUpper	= uTick>>29;

	if(uUpper!=uLast)
		{
		uLast  =  uUpper;
		uNext  = (uUpper-4)&7;

		if(uUpper>=4)
			{
			uHigh[uNext] = uHigh[uUpper]+1;
			}
		else{
			uHigh[uNext] = uHigh[uUpper];
			}
		}


return (uHigh[uUpper]<<22)|(uTick/1024);
}

//*****************************************************************************
//*
//*
//*		TLS - Code
//*
//*
//*****************************************************************************

#if USE_TLS

//*****************************************************************************
//*
//*		ThreadTlsInit
//*
//*****************************************************************************
//	Initalises the tls area for the main thread
static void ThreadTlsInit()
{

	if(iThreadTlsNum==-1)
		{
		   iThreadTlsNum = ThreadTlsAlloc();
		if(iThreadTlsNum==-1)return;
		}

	#if THREAD_NAMES
	*(char**)(aMainData+THREAD_NAME_OFF)="MainThread";
	#endif

	#if THREAD_CHECK

	if(iThreadCheckInit!=0)return;

	iThreadCheckInit = 1;
	{
	CheckData   *pChkData=(CheckData*)(aMainData+THREAD_TLS_BYTE);

	pChkData->uTick		= ThreadTick();
	pChkData->uSubTick	= pChkData->uTick;
	pChkData->pFile		= "No info";
	pChkData->pSubFile	= 0;
	pChkData->iLine		= 0;
	pChkData->iSubLine	= 0;
	pChkData->uSubParam	= 0;
	}

	ThreadTlsSet(iThreadTlsNum,aMainData);

	#else

	ThreadTlsSet(iThreadTlsNum,aMainData);

	#endif

	iThreadCheckInit = 2;

}

//*****************************************************************************
//*
//*		ThreadTlsAdd
//*
//*****************************************************************************
//	Adds the tls data to an thread for checking
//	pData	: is the poiter to the check data
static void ThreadTlsAdd(char *pData)
{

	if((int)iThreadTlsNum<0)
		{
		return;
		}

	#if THREAD_CHECK
	{
	CheckData   *pChkData=(CheckData*)(pData+THREAD_TLS_BYTE);

	pChkData->uTick		= ThreadTick();
	pChkData->uSubTick	= pChkData->uTick;
	pChkData->pFile		= "No info";
	pChkData->pSubFile	= 0;
	pChkData->iLine		= 0;
	pChkData->iSubLine	= 0;
	pChkData->uSubParam	= 0;
	}
	#endif

	ThreadTlsSet(iThreadTlsNum,pData);

}

#endif

//*****************************************************************************
//*
//*
//*		Thread-Check
//*
//*
//*****************************************************************************
#if		THREAD_CHECK

#include	<string.h>
#include	<stdio.h>

#ifdef	_MSC_VER
#pragma warning( disable :4090)
#endif



//*****************************************************************************
//*
//*		ThreadCheckSet
//*
//*****************************************************************************
//	Set the thread information
//	pFile	: is the current file name of the threads code position
//	iLine	: is the current file line of the threads code position
void ThreadCheckSet(const char *pFile,int iLine)
{
CheckData	*pData;


	if(iThreadCheckInit!=2)return;

		pData = (CheckData*)(ThreadTlsGet(iThreadTlsNum)+THREAD_TLS_BYTE);
	if(!pData)return;

	pData->pFile = NULL;
	pData->iLine = iLine;
	pData->pFile = pFile;
	pData->uTick = ThreadTick();

}

//*****************************************************************************
//*
//*		ThreadCheckSetSub
//*
//*****************************************************************************
//	Set the thread ip information
//	pFile	: is the current file name of the threads code position
//	iLine	: is the current file line of the threads code position
void ThreadCheckSetSub(const char *pFile,int iLine,unsigned uHandle)
{
CheckData	*pData;


	if(iThreadCheckInit!=2)return;

		pData = (CheckData*)(ThreadTlsGet(iThreadTlsNum)+THREAD_TLS_BYTE);
	if(!pData)return;

	pData->pSubFile  = NULL;
	pData->iSubLine  = iLine;
	pData->uSubParam = uHandle;
	pData->pSubFile  = pFile;
	pData->uSubTick  = ThreadTick();

}

//*****************************************************************************
//*
//*		ThreadCheckGet
//*
//*****************************************************************************
//	Reports the information of an thread
//	hTread		: handle of the thread
//	pBuffer		: pointer where the information will be written
//	iSize		: size of the buffer in bytes
//	iAddPath	: insert full path
int ThreadCheckGet(void *hThread,char *pBuffer,int iSize,int iAddPath)
{
CheckData	*pData;
const char 	*pFile;
unsigned	 uTime;
int			 iLine;
int		 	 iLen;
int			 iPos;
int			 iSub;


	if(iSize<=0)return 0;
	if(iSize<8)
		{
		pBuffer[0]=0;
		return 0;
		}

	if(iThreadCheckInit!=2)
		{
		pBuffer[0]=0;
		return 0;
		}


		pData = (CheckData*)(ThreadTlsGetOf(hThread,iThreadTlsNum)+THREAD_TLS_BYTE);
	if(!pData)
		{
		pBuffer[0]=0;
		return 0;
		}

	iLine =				 pData->iLine;
	pFile = (const char*)pData->pFile;

	if(!pFile)
		{
		iLen = 3;
		iLine=-1;
		pFile="???";
		}
	else if(!iAddPath)
		{
		iLen = (int)strlen(pFile);
		iSub = 0;

		for(iPos=iLen;iPos>0;iPos--)
			{
			if(pFile[iPos]=='.' ){iSub=iPos;}
			if(pFile[iPos]=='/' ){iPos++;break;}
			if(pFile[iPos]=='\\'){iPos++;break;}
			}

		pFile += iPos;
		iLen  -= iPos;
		iSub  -= iPos;

		if(iSub>0)iLen=iSub;
		}

	if(iLine<0    )iLine=-1;
	if(iLine>99999)iLine=99999;

	if(iLen>iSize-8)iLen=iSize-8;

	uTime = ThreadTick()-pData->uTick;
	uTime = ThreadTickToMs(uTime);

	memcpy(pBuffer,pFile,iLen);
	iLen+=sprintf(pBuffer+iLen,"(%i)\t%u ms",iLine,uTime);



return iLen;
}

//*****************************************************************************
//*
//*		ThreadCheckGetSub
//*
//*****************************************************************************
//	Reports the information of an thread ip call
//	hTread		: handle of the thread
//	pBuffer		: pointer where the information will be written
//	iSize		: size of the buffer in bytes
//	iAddPath	: insert full path
int ThreadCheckGetSub(void *hThread,char *pBuffer,int iSize,int iAddPath)
{
CheckData	*pData;
const char 	*pFile;
unsigned	 uSema;
unsigned	 uTime;
int			 iLine;
int		 	 iLen;
int			 iPos;
int			 iSub;


	if(iSize<=0)return 0;
	if(iSize<8)
		{
		pBuffer[0]=0;
		return 0;
		}

	if(iThreadCheckInit!=2)
		{
		pBuffer[0]=0;
		return 0;
		}


		pData = (CheckData*)(ThreadTlsGetOf(hThread,iThreadTlsNum)+THREAD_TLS_BYTE);
	if(!pData)
		{
		pBuffer[0]=0;
		return 0;
		}

	iLine =				 pData->iSubLine;
	pFile = (const char*)pData->pSubFile;
	uSema =              pData->uSubParam;

	if(!pFile)
		{
		return 0;
		}

	if(!iAddPath)
		{
		iLen = (int)strlen(pFile);
		iSub = 0;

		for(iPos=iLen;iPos>0;iPos--)
			{
			if(pFile[iPos]=='.' ){iSub=iPos;}
			if(pFile[iPos]=='/' ){iPos++;break;}
			if(pFile[iPos]=='\\'){iPos++;break;}
			}

		pFile += iPos;
		iLen  -= iPos;
		iSub  -= iPos;

		if(iSub>0)iLen=iSub;
		}

	if(iLine<0    )iLine=-1;
	if(iLine>99999)iLine=99999;

	if(iLen>iSize-20)iLen=iSize-20;

	uTime = ThreadTick()-pData->uTick;
	uTime = ThreadTickToMs(uTime);

	memcpy(pBuffer,pFile,iLen);
	iLen+=sprintf(pBuffer+iLen,"(%i)\ts:%08X\t%u ms",iLine,uSema,uTime);



return iLen;
}

#endif

//******************************************************************************
//*
//*		ThreadSetName
//*
//******************************************************************************
//	Sets the name of the current thread
//	pName	: Is the pointer to the new name of the thread.
//			  The pointer must point on a static memory.
void ThreadSetName(const char *pName)
{

	#ifdef  RTTARGET

	RTKTaskHandle	hTask;
	hTask=RTKCurrentTaskHandle();
	hTask->Name = pName;

	#elif defined(FREERTOS)


	TaskHandle_t	hTask;
	char		   *pMem;
	int				iLen;

	hTask=xTaskGetCurrentTaskHandle();
	pMem =pcTaskGetTaskName(hTask);

	if(pName)
		{
		for(iLen=0;iLen<configMAX_TASK_NAME_LEN-1;iLen++)
			{
			if(!pName[iLen])break;
			pMem[iLen] = pName[iLen];
			}

		pMem[iLen] = 0;
		}

	#elif	THREAD_NAMES

	char *pPtr;

		pPtr = ThreadTlsPointer();
	if(!pPtr)return;

	*(const char**)(pPtr+THREAD_NAME_OFF)=pName;

	#endif

}


//******************************************************************************
//*
//*		ThreadGetName
//*
//******************************************************************************
//	Gets the name of the current thread
const char *ThreadGetName()
{

	#ifdef  RTTARGET

	RTKTaskHandle	hTask;
	hTask=RTKCurrentTaskHandle();
	return hTask->Name;

	#elif defined(FREERTOS)


	TaskHandle_t	hTask;
	hTask = xTaskGetCurrentTaskHandle();
	return pcTaskGetTaskName(hTask);

	#elif	THREAD_NAMES

	char *pPtr;

		pPtr = ThreadTlsPointer();
	if(!pPtr)return 0;

	return *(char**)(pPtr+THREAD_NAME_OFF);

	#else

	return 0;

	#endif

}

