//*****************************************************************************
//*
//*
//*		TimeUtilityLinux.c
//*
//*
//*****************************************************************************
//
//  This is a library with time support for LINUX
//
//	Global Macros:
//	
//		GLOBAL_CONFIG		Enable the configuration via GlobalConfig.h header file (1=on 0=off)
//		SYS_TIME_HAS_FT		use the windows FILETIME structure as SysTime (1=on 0=off)
//		SYS_TIME_HAS_NS		add a nanosecont support to the library (0=seconds 1=ns)
//		SYS_BASE_YEAR		is the base year for SysTime
//
//	(C) Copyright Anton Zechner 2018
//
//
#ifdef		GLOBAL_CONFIG
#define 	TIMEUTILITYLINUX_C
#include	<GlobalConfig.h>
#undef		TIMEUTILITYLINUX_C
#endif

#include	<time.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<memory.h>
#include	<sys/time.h>
#include	<sys/types.h>
#include	"TimeUtility.h"


#ifndef			DST_NONE
#define 		DST_NONE 		0 							// not on dst
#endif
#ifndef			DST_USA
#define 		DST_USA  		1 							// USA style dst
#endif
#ifndef			DST_AUST
#define 		DST_AUST 		2 							// Australian style dst
#endif
#ifndef			DST_WET
#define 		DST_WET  		3 							// Western European dst
#endif
#ifndef			DST_MET
#define 		DST_MET  		4 							// Middle European dst
#endif
#ifndef			DST_EET
#define 		DST_EET  		5 							// Eastern European dst
#endif
#ifndef			DST_CAN
#define 		DST_CAN  		6 							// Canada
#endif
#ifndef			DST_GB
#define 		DST_GB  		7 							// Great Britain and Eire
#endif
#ifndef			DST_GB
#define 		DST_GB  		8 							// Rumania
#endif
#ifndef			DST_TUR
#define 		DST_TUR 		9 							// Turkey
#endif
#ifndef			DST_AUSTALT
#define 		DST_AUSTALT 	10 							// Australian style with shift in 1986
#endif


#ifndef			HAS_TIME_SET
#define 		HAS_TIME_SET	0							// allow time changing
#endif

#ifndef			HAS_TIME_OFFSET
#define 		HAS_TIME_OFFSET	1							// set time via time offset
#endif


#if 			SYS_TIME_HAS_FT
extern SysTime	iTimeZoneFt;								// current time zone (difference in 100 nanosecs)
#elif			SYS_TIME_HAS_NS
extern SysTime	iTimeZoneNs;								// current time zone (difference in nano secounds)
#endif

extern int		iInitTimeZone;								// is time zone initalized
extern int 		iTimeZone;									// current time zone (difference in secounds)
extern int 		iHasDst;									// hase daylight saving
static clockid_t iClockId=-1;								// is the current if for the realtime clock

#if				HAS_TIME_OFFSET
#if 			SYS_TIME_HAS_NS
static long long iTimeOffset;
#else
static int 		iTimeOffset;
#endif
#endif

#define 		YEAR_DAYS			(365)					// days in   1 year
#define			YEAR_DAYS_4			(YEAR_DAYS*4+1)			// days in   4 years
#define			YEAR_DAYS_100		(YEAR_DAYS_4*25-1)		// days in 100 years
#define			YEAR_DAYS_400		(YEAR_DAYS_100*4+1)		// days in 400 years
#define 		IS_LEAP_YEAR(y)		(((y)%4)? 0:(((y)%100)? 1:(((y)%400)? 0:1)))
#define 		DIF_OFFSET(y)		((y)*365+(y)/4-(y)/100+(y)/400)
#define 		DAY_OFFSET(y)		(DIF_OFFSET(y)+1-IS_LEAP_YEAR(y))
#define 		TIME_OFFSET			(0x1ea7e79c20d10000LL+(long long)(DAY_OFFSET(SYS_BASE_YEAR)-DAY_OFFSET(1900))*24LL*3600LL*1000000000LL)
#define 		TIME_OFFSET_UTC		((DAY_OFFSET(1970)-DAY_OFFSET(SYS_BASE_YEAR))*24L*3600L)




//*****************************************************************************
//*
//*		SysInitTimeZone
//*
//*****************************************************************************
//	Sets the timezone values
void SysInitTimeZone()
{

	tzset();

	iTimeZone		= timezone;
	iHasDst			= daylight;

	#if   SYS_TIME_HAS_FT
	iTimeZoneFt		= (SysTime)iTimeZone*(SysTime)10000000;
	#elif SYS_TIME_HAS_NS
	iTimeZoneNs   	= (SysTime)iTimeZone*(SysTime)1000000000;
	#endif

	iInitTimeZone   = 1;


}

//*****************************************************************************
//*
//*		SysTimeTick
//*
//*****************************************************************************
//	Returns a nanoseconds counter
//	The time counter is always incremented.
SysTick SysTimeTick()
{
struct timespec 	sTime;
struct timeval		sTval;



	if(iClockId>=0)
		{
		clock_gettime(iClockId,&sTime);
		return sTime.tv_sec*1000+sTime.tv_nsec/1000;
		}

	if(iClockId==-1)
		{
		if(!clock_gettime(CLOCK_MONOTONIC,&sTime))
			{
			iClockId = CLOCK_MONOTONIC;
			return (SysTick)(sTime.tv_sec*1000+sTime.tv_nsec/1000);
			}

		if(!clock_gettime(CLOCK_REALTIME,&sTime))
			{
			iClockId = CLOCK_REALTIME;
			return (SysTick)(sTime.tv_sec*1000+sTime.tv_nsec/1000);
			}

		iClockId = -2;
		}

	gettimeofday(&sTval,0);


return (SysTick)(sTval.tv_sec*1000+sTval.tv_usec);
}

//*****************************************************************************
//*
//*		SysTimeGet
//*
//*****************************************************************************
//	Returns the current time in [nano]seconds from 1.Jan.1900
//	The time counter is always incremented, also if the system time is changed
SysTime SysTimeGet()
{
struct timeval	sTime;
SysTime			iRet;


	gettimeofday(&sTime,NULL);


	#if   SYS_TIME_HAS_FT

	iRet  = TIME_OFFSET;
	iRet +=	sTime.tv_usec*10;
	iRet +=	((SysTime)sTime.tv_sec)*10000000LL;

	#elif SYS_TIME_HAS_NS

	iRet  = TIME_OFFSET;
	iRet +=	sTime.tv_usec*1000;
	iRet +=	((SysTime)sTime.tv_sec)*1000000000LL;

	#else

	iRet  = TIME_OFFSET_UTC;
	iRet +=	(SysTime)sTime.tv_sec;

	#endif

	#if HAS_TIME_OFFSET

	iRet += iTimeOffset;

	#endif


return iRet;
}

//*****************************************************************************
//*
//*		SysTimeGetMs
//*
//*****************************************************************************
//	Returns the current time in [nano]seconds from 1.Jan.1900
//	The time counter is always incremented, also if the system time is changed
//	pMilli		: Here the milli seconds will be stored (NULL=none ms)
SysTime SysTimeGetMs(unsigned *pMilli)
{
struct timeval	sTime;
SysTime			iRet;


	gettimeofday(&sTime,NULL);



	#if   SYS_TIME_HAS_FT

	iRet  = TIME_OFFSET;
	iRet +=	sTime.tv_usec*10;
	iRet +=	((SysTime)sTime.tv_sec)*10000000LL;

	#elif SYS_TIME_HAS_NS

	iRet  = TIME_OFFSET;
	iRet +=	sTime.tv_usec*1000;
	iRet +=	((SysTime)sTime.tv_sec)*1000000000LL;

	#else

	iRet  = TIME_OFFSET_UTC;
	iRet +=	(SysTime)sTime.tv_sec;

	#endif

	if(pMilli)
		{
		*pMilli = (sTime.tv_usec/1000)%1000;
		}

	#if HAS_TIME_OFFSET

	iRet += iTimeOffset;

	#endif


return iRet;
}

//*****************************************************************************
//*
//*		SysTimeZone
//*
//*****************************************************************************
//	Sets the current time zone
//	iZone			: Seconds between UTC and local time (-3600 for middle erope)
//	iDaylightSaving	: Enables daylight saving
void SysTimeZone(int iZone,int iDaylightSaving)
{


	if(!iInitTimeZone)SysInitTimeZone();

	iHasDst				= iDaylightSaving;	// hase daylight saving

	#if SYS_TIME_HAS_FT
	iTimeZoneFt			= (SysTime)iZone;	// timezone in seconds     since Greenwich
	iTimeZoneFt		   *= 10000000;			// timezone in 100 nanosec since Greenwich
	#elif SYS_TIME_HAS_NS
	iTimeZoneNs			= (SysTime)iZone;	// timezone in seconds     since Greenwich
	iTimeZoneNs		   *= 1000000000;		// timezone in nanoseconds since Greenwich
	#endif

	iTimeZone			= iZone;			// timezone in seconds since Greenwich
	iInitTimeZone		= 1;				// are the timezone values loaded

}


//*****************************************************************************
//*
//*		SysTimeDataSet
//*
//*****************************************************************************
//	Sets the current time
//	pData		: is the new time
//	iLocal		: is time data a local time
//	Returns the new system time or 0 if an error occurs
#if HAS_TIME_SET
SysTime SysTimeDataSet(const SysTimeStruct *pData,int iLocal)
{
SysTime			iTimeNew;
struct timeval	sTime;



															// check new date
	if(pData->usYear<SYS_BASE_YEAR || pData->usYear>SYS_LAST_YEAR)
		{
		return 0;
		}

	if(iLocal)												// new time is local time
		{
		iTimeNew = SysTimeFromLocal(pData);
		}
	else{
		iTimeNew = SysTimeFromGlobal(pData);
		}



	#if HAS_TIME_OFFSET
		{													// set time via offset
		SysTime	iTimeNow;

		gettimeofday(&sTime,NULL);

		#if   SYS_TIME_HAS_FT

		iTimeNow	= TIME_OFFSET;
		iTimeNow   += sTime.tv_usec*10;
		iTimeNow   += ((SysTime)sTime.tv_sec)*10000000LL;
		iTimeOffset = iTimeNew-iTimeNow;

		#elif SYS_TIME_HAS_NS

		iTimeNow	= TIME_OFFSET;
		iTimeNow   += sTime.tv_usec*1000;
		iTimeNow   += ((SysTime)sTime.tv_sec)*1000000000LL;
		iTimeOffset = iTimeNew-iTimeNow;

		#else

		iTimeNow	= TIME_OFFSET_UTC;
		iTimeNow   += (SysTime)sTime.tv_sec;
		iTimeOffset = iTimeNew-iTimeNow;

		#endif
		}
	#else
		{													// set time directly
		#if   SYS_TIME_HAS_FT

		sTime.tv_usec	= (int   )(iTimeNew%10000000)/10;
		sTime.tv_sec	= (time_t)(iTimeNew/10000000)-TIME_OFFSET;

		#elif SYS_TIME_HAS_NS

		sTime.tv_usec	= (int   )(iTimeNew%1000000000)/1000;
		sTime.tv_sec	= (time_t)(iTimeNew/1000000000)-TIME_OFFSET;

		#else

		sTime.tv_usec	= 0;
		sTime.tv_sec	= (time_t)(iTimeNew)-TIME_OFFSET_UTC;

		#endif

		if(settimeofday(&sTime,0))
			{
			return 0;
			}
		}
	#endif



return iTimeNew;
}

#endif
