//*****************************************************************************
//*
//*
//*		TimeUtilityRt.c
//*
//*
//*****************************************************************************
//
//  This is a library with time support for RTTARGET
//
//	Global Macros:
//	
//		GLOBAL_CONFIG		Enable the configuration via GlobalConfig.h header file (1=on 0=off)
//		SYS_TIME_HAS_FT		use the windows FILETIME structure as SysTime (1=on 0=off)
//		SYS_TIME_HAS_NS		add a nanosecont support to the library (0=seconds 1=ns)
//		SYS_BASE_YEAR		is the base year for SysTime
//
//	(C) Copyright Anton Zechner 2018
//
//
#ifdef		GLOBAL_CONFIG
#define 	TIMEUTILITYRT_C
#include	<GlobalConfig.h>
#undef		TIMEUTILITYRT_C
#endif

#include	"TimeUtility.h"
#include	<finetime.h>
#include	<rttarget.h>
#include	<rtfiles.h>
#include	<windows.h>
#include	<rtkclk.h>
#include	<rtk32.h>
#include	<clock.h>
#include	<time.h>


// REM AZ
extern unsigned long	CLKUnitsPerSecond;				// APIC: Clock*MHZ	Common: 14.318 MHz/4
extern int				CLKDivisor;						// APIC: 1			Common:	12        /4

#define 	YEAR_DAYS			(365)					// days in   1 year
#define		YEAR_DAYS_4			(YEAR_DAYS*4+1)			// days in   4 years
#define		YEAR_DAYS_100		(YEAR_DAYS_4*25-1)		// days in 100 years
#define		YEAR_DAYS_400		(YEAR_DAYS_100*4+1)		// days in 400 years
#define 	IS_LEAP_YEAR(y)		(((y)%4)? 0:(((y)%100)? 1:(((y)%400)? 0:1)))
#define 	DIF_OFFSET(y)		((y)*365+(y)/4-(y)/100+(y)/400)
#define 	DAY_OFFSET(y)		(DIF_OFFSET(y)+1-IS_LEAP_YEAR(y))
#define 	FILETIME_OFFSET		(0x014F373BFDE04000L+(__int64)(DAY_OFFSET(SYS_BASE_YEAR)-DAY_OFFSET(1900))*(__int64)24L*(__int64)3600L*(__int64)10000000)


#if    SYS_TIME_HAS_FT
extern SysTime			iTimeZoneFt;					// current time zone (difference in 100 nanosecs)
#elif  SYS_TIME_HAS_NS
extern SysTime			iTimeZoneNs;					// current time zone (difference in nano secounds)
#endif


extern int				iInitTimeZone;					// is time zone initalized
extern int 				iTimeZone;						// current time zone (difference in secounds)
extern int 				iHasDst;						// hase daylight saving


//*****************************************************************************
//*
//*		SysTimeConvertDU	   DOSTIME --> SysTime
//*
//*****************************************************************************
//	converts a DOSTIME to an SysTime (used in FileUtilityRt.c)
//	pFileTime	: pointer to a windows file time item
//	pSysTime	: pointer to a system time item
//	Returns 1 if the conversion was ok or 0 if an error occurs
int SysTimeConvertDU(SysDW uDosTime ,SysTime *pTime)
{
unsigned long	uVal,uYear,uDays;
unsigned short	wTime,wDate;



	wTime=(unsigned short) uDosTime;
	wDate=(unsigned short)(uDosTime>>16);

	   uYear = (wDate>>9)+80;
	if(uYear>2038)
		{
		*pTime=0;
		return FALSE;
		}

	if(uYear&3)uDays  = aTpDays  [((wDate>>5)-1)&15];
	else	   uDays  = aTpPDays [((wDate>>5)-1)&15];
			   uDays += ((wDate)&0x1F)-1;
			   uDays += (uYear-70  )*365;
			   uDays += (uYear-70+1)>>2;

			   uVal   = uDays*24;
			   uVal  += (wTime>>11);
			   uVal  *= 60;
			   uVal  += (wTime>> 5)&0x3F;
			   uVal  *= 60;
			   uVal  += (wTime<< 1)&0x3F;


	*pTime = SysTimeFromUtc(uVal);


return 1;
}


//*****************************************************************************
//*
//*		SysTimeZone
//*
//*****************************************************************************
//	Sets the current time zone
//	iZone			: Seconds between UTC and local time (-3600 for middle erope)
//	iDaylightSaving	: Enables daylight saving
void SysTimeZone(int iZone,int iDaylightSaving)
{

	// This is a dummy function
	// The time zone from the pc is used

}



//*****************************************************************************
//*
//*		SysTimeGet
//*
//*****************************************************************************
//	Returns the current time in nanoseconds from 1.Jan.SYS_BASE_YEAR
SysTime SysTimeGet()
{
SysTime				iTime;



	GetSystemTimeAsFileTime((FILETIME*)&iTime);


	#if   SYS_TIME_HAS_FT
	//iTime-= 0L;
	//iTime*= 1L;
	#elif SYS_TIME_HAS_NS
	iTime  -= FILETIME_OFFSET;
	iTime  *= 100L;
	#else
	iTime  -= FILETIME_OFFSET;
	iTime  /= 10000000L;
	#endif


return iTime;
}


//*****************************************************************************
//*
//*		SysTimeGetMs
//*
//*****************************************************************************
//	Returns the current time in nanoseconds from 1.Jan.SYS_BASE_YEAR
//	pMilli		: Here the milli seconds will be stored (NULL=none ms)
SysTime	SysTimeGetMs(unsigned *pMilli)
{
__int64				iMsVal;
__int64				iTime;



	GetSystemTimeAsFileTime((FILETIME*)&iTime);


	if(pMilli)
		{
		iMsVal	= iTime;
		iMsVal /= 10000L;
		iMsVal %= 1000L;
		*pMilli = (unsigned)iMsVal;
		}

	#if   SYS_TIME_HAS_FT
	//iTime-= 0L;
	//iTime*= 1L;
	#elif SYS_TIME_HAS_NS
	iTime  -= FILETIME_OFFSET;
	iTime  *= 100L;
	#else
	iTime  -= FILETIME_OFFSET;
	iTime  /= 10000000L;
	#endif


return iTime;
}


//*****************************************************************************
//*
//*		SysTimeTick
//*
//*****************************************************************************
//	Returns a increasing time in nanoseconds
SysTick SysTimeTick()
{
unsigned 					uAdd;
unsigned 					uNext;
unsigned 					uShift;
unsigned 					uTicker;
unsigned 					uFromShift;
unsigned __int64			uFullTick;
static unsigned volatile	uOldShift=0;
static unsigned __int64		uAddTick[8];
static unsigned 			uTickMul;
static unsigned 			uTickDiv;
SysTime						iTime;



	uTicker = RTKGetTime();
	uShift	= uTicker>>29;


	if(!iInit)								// calculate the time value at the first call
		{
		uAddTick[(uShift+0)&7] = 0x20000000*4;
		uAddTick[(uShift+1)&7] = 0x20000000*5;
		uAddTick[(uShift+2)&7] = 0x20000000*6;
		uAddTick[(uShift+3)&7] = 0x20000000*7;
		uAddTick[(uShift+4)&7] = 0x20000000*0;
		uAddTick[(uShift+5)&7] = 0x20000000*1;
		uAddTick[(uShift+6)&7] = 0x20000000*2;
		uAddTick[(uShift+7)&7] = 0x20000000*3;


		uTickDiv   = CLKUnitsPerSecond;
		uFullTick  = CLKDivisor;
		uFullTick *=_rtkCLKCurrentTimerInterval();
		uFullTick *  1000000000;

		while(uFullTick>0x100000000)
			{
			uFullTick>>=1;
			uTickDiv >>=1;
			}

		if(!uTickDiv)uTickDiv=0;

		uTickMul   = (unsigned)uFullTick;
		uOldShift  = uShift;
		iInit	   = 1;
		}


	uFromShift = uOldShift;

	if(uShift!=uFromShift)					// update next add values
		{
		do	{
			uNext           = (uFromShift+4)&7;
			uAddTime[uNext] =  uAddTick[uFromShift]+(unsigned __int64)0x80000000;
			uFromShift      = (uFromShift+1)&7;
			}while(uShift!=uFromShift);

		uOldShift = uFromShift;
		}

	uFullTick  = uTicker&0x1FFFFFFF;
	uFullTick += uAddTick[uShift];


	#ifdef RTK_INT_64

	return (SysTick)FTMultDivDouble((RTKFineTime*)&iTime,(RTKFineTime*)&uFullTick,uTickDivisor,uTickDiv);

	#else

	FTMultDivDouble((RTKFineTime*)&iTime,(RTKFineTime*)&uFullTick,uTickDivisor,uTickDiv);
	return iTime;

	#endif
}

