//*****************************************************************************
//*
//*
//*		TimeUtilityWin.c
//*
//*
//*****************************************************************************
//
//  This is a library with time support for WINDOWS
//
//	Global Macros:
//	
//		GLOBAL_CONFIG		Enable the configuration via GlobalConfig.h header file (1=on 0=off)
//		SYS_TIME_HAS_FT		use the windows FILETIME structure as SysTime (1=on 0=off)
//		SYS_TIME_HAS_NS		add a nanosecont support to the library (0=seconds 1=ns)
//		SYS_BASE_YEAR		is the base year for SysTime
//
//	(C) Copyright Anton Zechner 2018
//
//
#ifdef		GLOBAL_CONFIG
#define 	TIMEUTILITYWIN_C
#include	<GlobalConfig.h>
#undef		TIMEUTILITYWIN_C
#endif

#include	"TimeUtility.h"
#include	<windows.h>
#include	<time.h>

#ifndef		HAS_TIME_SET
#define 	HAS_TIME_SET		0						// allow time changing
#endif

#ifndef		HAS_TIME_OFFSET
#define 	HAS_TIME_OFFSET		1						// set time via time offset
#endif

#define 	YEAR_DAYS			(365)					// days in   1 year
#define		YEAR_DAYS_4			(YEAR_DAYS*4+1)			// days in   4 years
#define		YEAR_DAYS_100		(YEAR_DAYS_4*25-1)		// days in 100 years
#define		YEAR_DAYS_400		(YEAR_DAYS_100*4+1)		// days in 400 years
#define 	IS_LEAP_YEAR(y)		(((y)%4)? 0:(((y)%100)? 1:(((y)%400)? 0:1)))
#define 	DIF_OFFSET(y)		((y)*365+(y)/4-(y)/100+(y)/400)
#define 	DAY_OFFSET(y)		(DIF_OFFSET(y)+1-IS_LEAP_YEAR(y))
#define 	FILETIME_OFFSET		(0x014F373BFDE04000L+(__int64)(DAY_OFFSET(SYS_BASE_YEAR)-DAY_OFFSET(1900))*(__int64)24L*(__int64)3600L*(__int64)10000000)


#if defined(_MSC_VER) && _MSC_VER>=1300

	static __inline	long GetTimeZone()
	{
	long	lZone=-3600;
	_get_timezone(&lZone);
	return lZone;
	}

#else

	#define	GetTimeZone()	_timezone

#endif

#ifdef		_MSC_VER
#pragma		warning( disable :4996)
#endif


extern int				iInitTimeZone;					// is time zone initalized
extern SysTime			iTimeZoneNs;					// current time zone (difference in nano secounds)
extern int 				iTimeZone;						// current time zone (difference in secounds)
extern int 				iHasDst;						// hase daylight saving
static LARGE_INTEGER 	uPerfomanceFrequncy;			// current Perfomance-Counter Frequenzy
#if	   HAS_TIME_OFFSET
static __int64			iTimeOffset;					// timeoffest in 100n steps for SysTimeDataSet
#endif


//*****************************************************************************
//*
//*		SysTimeConvertFS	   FILETIME --> SysTime
//*
//*****************************************************************************/
//	converts a FILETIME to an SysTime (used in FileUtility.c)
//	pFileTime	: pointer to a windows file time item
//	pSysTime	: pointer to a system time item
//	Returns 1 if the conversion was ok or 0 if an error occurs
int SysTimeConvertFS(FILETIME *pFileTime,SysTime *pSysTime)
{
union	{
		__int64	i64;
		DWORD	dw[2];
		}sTime;


	sTime.i64 = *(__int64*)pFileTime;


	if(sTime.i64<FILETIME_OFFSET)							// Zeit vor 1.Jan.SYS_BASE_YEAR
		{
		*pSysTime=0;
		return 0;
		}

	sTime.i64  = *(__int64*)pFileTime;
	sTime.i64 -= FILETIME_OFFSET;


	#if SYS_TIME_HAS_NS
	if(sTime.dw[1]>=0x7FFFFFE/100)							// Overrun
		{
		*pSysTime=0x7FFFFFFFFFFFFFFFL;
		return 0;
		}
	#elif SYS_BASE_YEAR>=1970
	if(sTime.dw[1]>=0x0098967F)								// Overrun
		{
		*pSysTime=0xFFFFFFFE;
		return 0;
		}
	#endif

	#if SYS_TIME_HAS_NS
	*pSysTime = sTime.i64*100;								// 100ns Schritte auf ns umrechen
	#else
	*pSysTime = (SysTime)(sTime.i64/(__int64)10000000);		// 100ns Schritte auf s umrechen
	#endif


return 1;
}



//*****************************************************************************
//*
//*		SysInitTimeZone
//*
//*****************************************************************************
//	Initializes all timezone settings
void SysInitTimeZone()
{
time_t	iTime;



	if(iInitTimeZone)return;

	iTime = time(0);
	localtime(&iTime);

	#if SYS_TIME_HAS_NS
	iTimeZoneNs		= (SysTime)GetTimeZone()*(SysTime)1000000000;
	#endif

	iTimeZone		=  GetTimeZone();
	iInitTimeZone	= 1;


}

//*****************************************************************************
//*
//*		SysTimeZone
//*
//*****************************************************************************
//	Sets the current time zone
//	iZone			: Seconds between UTC and local time (-3600 for middle erope)
//	iDaylightSaving	: Enables daylight saving
void SysTimeZone(int iZone,int iDaylightSaving)
{

	// This is a dummy function
	// The time zone from the pc is used

}



//*****************************************************************************
//*
//*		SysTimeGet
//*
//*****************************************************************************
//	Returns the current time in nanoseconds from 1.Jan.SYS_BASE_YEAR
SysTime SysTimeGet()
{
__int64				iTime;


	GetSystemTimeAsFileTime((FILETIME*)&iTime);

	iTime  -= FILETIME_OFFSET;

	#if	   HAS_TIME_OFFSET
	iTime  += iTimeOffset;					// timeoffest in 100n steps for SysTimeDataSet
	#endif


	#if SYS_TIME_HAS_NS
	iTime  *= (__int64)100;
	#else
	iTime  /= (__int64)10000000;
	#endif


return (SysTime)iTime;
}


//*****************************************************************************
//*
//*		SysTimeGetMs
//*
//*****************************************************************************
//	Returns the current time in nanoseconds from 1.Jan.SYS_BASE_YEAR
//	pMilli		: Here the milli seconds will be stored (NULL=none ms)
SysTime	SysTimeGetMs(unsigned *pMilli)
{
__int64				iMsVal;
__int64				iTime;


	GetSystemTimeAsFileTime((FILETIME*)&iTime);

	iTime  -= FILETIME_OFFSET;

	#if	   HAS_TIME_OFFSET
	iTime  += iTimeOffset;					// timeoffest in 100n steps for SysTimeDataSet
	#endif

	iMsVal	= iTime;
	iMsVal /= 10000L;
	iMsVal %= 1000L;
	*pMilli = (unsigned)iMsVal;


	#if SYS_TIME_HAS_NS
	iTime  *= 100L;
	#else
	iTime  /= 10000000L;
	#endif


return (SysTime)iTime;
}

//*****************************************************************************
//*
//*		SysTimeTick
//*
//*****************************************************************************
//	Returns a increasing time in nanoseconds
SysTick SysTimeTick()
{
LARGE_INTEGER 		uCounter;
static int			iInit=0;
unsigned __int64	uUpper;
unsigned __int64	uTime;



	QueryPerformanceCounter(&uCounter);

	if(!iInit)								// calculate the time value at the first call
		{
		QueryPerformanceFrequency(&uPerfomanceFrequncy);
		if(!uPerfomanceFrequncy.QuadPart)uPerfomanceFrequncy.QuadPart=1000000000;
		iInit	 = 1;
		}


	if(uCounter.HighPart)
		{
		uUpper   = (unsigned __int64)uCounter.HighPart*(unsigned __int64)1000000000;
		uUpper  /= uPerfomanceFrequncy.QuadPart;
		uUpper <<= 32;
		uTime    = (unsigned __int64)uCounter.LowPart*(unsigned __int64)1000000000;
		uTime   /= uPerfomanceFrequncy.QuadPart;
		uTime   += uUpper;
		}
	else{
		uTime    = (unsigned __int64)uCounter.LowPart*(unsigned __int64)1000000000;
		uTime   /= uPerfomanceFrequncy.QuadPart;
		}


return uTime;
}

//*****************************************************************************
//*
//*		SysTimeDataSet
//*
//*****************************************************************************
//	Gets the system time of a time structure.
//	pData		: is the new time
//	iLocal		: is time data a local time
//	Returns the new sytem time or 0 if an error occurs
#if HAS_TIME_SET
SysTime SysTimeDataSet(const SysTimeStruct *pData,int iLocal)
{
SysTime			iTimeNew;
__int64			iTimeNow;



															// check new date
	if(pData->usYear<SYS_BASE_YEAR || pData->usYear>SYS_LAST_YEAR)
		{
		return 0;
		}

	if(iLocal)												// new time is local time
		{
		iTimeNew = SysTimeFromLocal(pData);
		}
	else{
		iTimeNew = SysTimeFromGlobal(pData);
		}



	#if	HAS_TIME_OFFSET
		{													// set time via offset
		__int64		iDiffTime;

		GetSystemTimeAsFileTime((FILETIME*)&iTimeNow);

		iDiffTime	= iTimeNew;

		#if SYS_TIME_HAS_NS
		iDiffTime  /= 100L;
		#else
		iDiffTime  *= 10000000L;
		#endif

		iDiffTime  += FILETIME_OFFSET;
		iTimeOffset = iDiffTime-iTimeNow;
		}
	#else
		{													// set time directly
		SYSTEMTIME	sSysTime;

		iTimeNow	= iTimeNew;

		#if SYS_TIME_HAS_NS
		iTimeNow   /= 100L;
		#else
		iTimeNow   *= 10000000L;
		#endif

		iTimeNow   += FILETIME_OFFSET;

		if(!FileTimeToSystemTime((FILETIME*)&iTimeNow,&sSysTime))
			{
			return 0;
			}

		if(!SetSystemTime(&sSysTime))
			{
			return 0;
			}
		}
	#endif



return iTimeNew;
}

#endif
