//*****************************************************************************
//*
//*
//*		BitFunctions.cpp
//*
//*
//*****************************************************************************
//
//	Copyright Anton Zechner (C) 2003
//
//	Funktionen:
//
//   int  bit_set				: Bit in Speicherbereich setzen
//   int  bit_clear				: Bit in Speicherbereich lschen
//   int  bit_change			: Bit in Speicherbereich invertiren
//   int  bit_read				: Bit in Speicherbereich lesen
//   int  bit_write				: Bit in Speicherbereich schreiben
//   int  bit_rol				: Bits durch rollen	links herum
//   int  bit_ror				: Bits durch rollen rechts herum
//   int  bit_find_0			: Erstes  gelschtes Bit suchen
//   int  bit_find_1			: Erstes  gesetzes   Bit suchen
//   int  bit_swap				: Tauscht Bits von unten nach oben aus
//   int  bit_swap_qw			: Tauscht Bits von unten nach oben aus (64 Bit)
//   int  bit_shift				: Bits nach frei   verschieben
//   int  bit_shift_l			: Bits nach links  verschieben mit berlaufprfung
//   int  bit_shift_r			: Bits nach rechts verschieben mit berlaufprfung
//   int  bit_shift_a			: Bits nach rechts verschieben mit berlaufprfung (arimetisch)
//   int  btol_qw				: Ergibt Big-Endial zu Little-Endian fr QWORD
//   int  btol_dw				: Ergibt Big-Endial zu Little-Endian fr DWORD
//   int  btol_w				: Ergibt Big-Endial zu Little-Endian fr WORD
//   int  btol					: Ergibt Big-Endial zu Little-Endian
//
//	Makros:
//
//		[Cpu+Compiler]		[Optinale Makros]
//
//		BIT_MSC_X64			BIT_MSC_INSTR
//		BIT_MSC_X86			BIT_MSC_INSTR oder BIT_MSC_VC60
//		BIT_MSC_ARM64
//		BIT_MSC_ARM			BIT_MSC_THUMB
//
//		BIT_GNU_X64
//		BIT_GNU_X86
//		BIT_GNU_ARM64
//		BIT_GNU_ARM			BIT_GNU_THUMB
//
//		BIT_IAR_ARM			BIT_IAR_THUMB
//
//		BIT_KEIL_ARM		BIT_KEIL_THUMB
//
#ifdef		GLOBAL_CONFIG
#define		BITFUNCTIONS_CPP
#include	<GlobalConfig.h>
#undef		BITFUNCTIONS_CPP
#endif

#include	"BitFunctions.h"


#ifdef		_MSC_VER
#pragma 	warning( disable :4302 )
#pragma 	warning( disable :4311 )
#pragma 	warning( disable :4312 )
#endif


#ifdef		BIT_MSC_VC60
#pragma		warning(disable:4035)
#endif

#if			BIT_USE_BIG_ENDIAN
#define		B_ORDER_DW(d)			(btol_dw(d))
#define		B_XSWAP_DW(d)			d
#else
#define		B_ORDER_DW(d)			d
#define		B_XSWAP_DW(d)			(btol_dw(d))
#endif


//*****************************************************************************
//*
//*		Lock-Macros fr NO_BIT_ASM bit_change
//*
//*****************************************************************************

#if defined(NO_BIT_ASM)

	#ifdef __CC_ARM

		#define BITLOCK()			__disable_irq();
		#define BITUNLOCK()			__enable_irq ();

	#elif defined(__ICCARM__)

        #include	<intrinsics.h>
        #define BITLOCK()			__disable_interrupt();
		#define BITUNLOCK()			__enable_interrupt();

	#elif linux

		#include <pthread.h>

		static  pthread_mutex_t		sBitMutext;
		static  int					iBitSemInit=0;
		#define BITLOCK()			if(!iBitSemInit){pthread_mutex_init(&sBitMutext,0);iBitSemInit=1;}pthread_mutex_lock(&sBitMutext);
		#define BITUNLOCK()			pthread_mutex_unlock(&sBitMutext);

	#elif defined(_WIN32) || defined(_WIN64)

		#include <windows.h>

		static  CRITICAL_SECTION 	sBitMutext;
		static  int					iBitSemInit=0;
		#define BITLOCK()			if(!iBitSemInit){InitializeCriticalSection(&sBitMutext);iBitSemInit=1;}EnterCriticalSection(&sBitMutext);
		#define BITUNLOCK()			LeaveCriticalSection(&sBitMutext);

	#elif !defined(BITLOCK) || !defined(BITUNLOCK)

		#error	no locking defined for this operating system

	#endif


static	unsigned char 	ucBitOffset[256] =
	{
	8,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	7,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
	};

static	unsigned char 	ucBitOffsetX[256] =
	{
	8,7,6,6,5,5,5,5,4,4,4,4,4,4,4,4,
	3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
	2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
	2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	};

#endif



//*****************************************************************************
//*
//*		bit_set
//*
//*****************************************************************************
//	Setzt ein Bit in einem Speicherbereich auf 1. Die Funktion ist atomar
//	pMem		: Zeiger auf den Speicherbereich
//	iOffset		: Ist die Nummer des Bits das gesetzt werden soll.
//	Ergibt einen Wert ungleich 0 wenn das alte Bit gesetzt war sonst 0
#if defined(NO_BIT_ASM)

	int _cdecl bit_set(void *pMem,int iOffset)
	{
	char   *pPtr=(char*)pMem;
	char	cBit;
	int		iRet;


	pPtr += iOffset>>3;
	cBit  = 1<<(iOffset&7);

	BITLOCK();

	if(*pPtr&cBit)
		{
		iRet   = -1;
		}
	else{
		*pPtr |= cBit;
		iRet   = 0;
		}

	BITUNLOCK();

	return iRet;
	}

#elif !defined(_DEBUG)

	// function defined in inline in "BitFunctions.h"

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	int _cdecl bit_set(void *pMem,int iOffset)
	{
	int iRet;

	asm volatile (
		"lock; bts  %2,%1  \n"
		"sbb        %0,%0  \n"
		:"=r" (iRet), BIT_ADDR(pMem)
		:"Ir" (iOffset)
		:"memory", "cc"
		);

	return  iRet;
	}

#elif defined(BIT_MSC_X64)

	int _cdecl bit_set(void *pMem,int iOffset)
	{
	return _interlockedbittestandset64((__int64*)pMem,iOffset);
	}

#elif defined(BIT_MSC_INSTR)

	int _cdecl bit_set(void *pMem,int iOffset)
	{
	return _interlockedbittestandset((long*)pMem,iOffset);
	}

#elif defined(BIT_MSC_ARM64)

	int _cdecl bit_set(void *pMem,int iOffset)
	{
	return _interlockedbittestandset64_HLEAcquire((long*)pMem,iOffset);
	}

#elif defined(BIT_MSC_ARM)

	int _cdecl bit_set(void *pMem,int iOffset)
	{
	return _interlockedbittestandset_acq((long*)pMem,iOffset);
	}

#elif defined(BIT_MSC_VC60)

	int _cdecl bit_set(void *pMem,int iOffset)
	{
	_asm{
		mov	 	eax,pMem
		mov  	ebx,iOffset
		lock 	bts [eax],ebx
		sbb  	eax,eax
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM) || defined(BIT_KEIL_ARM)

	int _cdecl bit_set(void *pMem,int iOffset)
	{
	unsigned char  ucMask;
	unsigned char  ucBits;
	unsigned char *pPtr;


	pPtr   = (unsigned char*)pMem+iOffset/8;
	ucMask = 1<<(iOffset&7);

	#ifdef BIT_USE_IRQ

	{
	int iIrq;

	   iIrq=B_READ_IRQ();
	if(iIrq)B_DISABLE_IRQ();

	ucBits = *pPtr;
	*pPtr  = ucBits|ucMask;

	if(iIrq)B_ENABLE_IRQ();
	}

	#else

	for(;;)
		{
		ucBits          = B_LDREXB(pPtr);
		if(!B_STREXB(ucBits|ucMask,pPtr))
			{
			break;
			}
		}

	#endif

	return ucBits&ucMask;
	}

#endif

//*****************************************************************************
//*
//*		bit_set
//*
//*****************************************************************************
//	Setzt einige Bits in einem Speicherbereich auf 1.
//	pMem		: Zeiger auf den Speicherbereich
//	iOffset		: Ist die Nummer des Bits das gesetzt werden soll.
//	iCount		: Ist die Anzahl der Bits. Wenn der Wert kleiner
//				  gleich 0 ist,wird nicht gendert
void _cdecl bit_set(void *pMem,int iOffset,int iCount)
{
unsigned char  *pData=(unsigned char*)pMem;
int				iMask;
int				iOff;
int				iNum;


	if(iCount<=0)return;

	iOff = iOffset&7;

	if(iOff)
		{
		iNum	= 8-iOff;

		if(iNum>iCount)
			{
			iMask = ~(0xFFFFFFFF<<iNum);
			}
		else{
			iMask = 0xFF;
			}

		pData[iOffset/8] |= iMask<<iOff;

		iCount -= iNum;
		iOffset   += iNum;

		if(iCount<=0)return;
		}


	memset(pData+iOffset/8,0xFF,iCount/8);


	if(iCount&7)
		{
		iOffset /= 8;
		iOffset += iCount/8;
		iCount  &= 7;
		iMask	 = ~(0xFFFFFFFF<<iCount);
		pData[iOffset] |= iMask;
		}

}

//*****************************************************************************
//*
//*		bit_clear
//*
//*****************************************************************************
//	Lscht ein Bit in einem Speicherbereich auf. Die Funktion ist atomar
//	pMem		: Zeiger auf den Speicherbereich
//	iOffset		: Ist die Nummer des Bits das gelscht werden soll.
//	Ergibt einen Wert ungleich 0 wenn das alte Bit gesetzt war sonst 0
#if defined(NO_BIT_ASM)

	int _cdecl bit_clear(void *pMem,int iOffset)
	{
	char   *pPtr=(char*)pMem;
	char	cBit;
	int		iRet;


	pPtr += iOffset>>3;
	cBit  = 1<<(iOffset&7);

	BITLOCK();

	if(*pPtr&cBit)
		{
		*pPtr &= ~cBit;
		iRet   = -1;
		}
	else{
		iRet   = 0;
		}

	BITUNLOCK();

	return iRet;
	}

#elif !defined(_DEBUG)

	// function defined in inline in "BitFunctions.h"

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	int _cdecl bit_clear(void *pMem,int iOffset)
	{
	int iRet;

	asm volatile (
		"lock; btr  %2,%1  \n"
		"sbb        %0,%0  \n"
		:"=r" (iRet), BIT_ADDR(pMem)
		:"Ir" (iOffset)
		:"memory", "cc"
		);

	return  iRet;
	}

#elif defined(BIT_MSC_X64)

	int _cdecl bit_clear(void *pMem,int iOffset)
	{
	return _interlockedbittestandreset64((__int64*)pMem,iOffset);
	}

#elif defined(BIT_MSC_INSTR)

	int _cdecl bit_clear(void *pMem,int iOffset)
	{
	return _interlockedbittestandreset((long*)pMem,iOffset);
	}

#elif defined(BIT_MSC_ARM64)

	int _cdecl bit_clear(void *pMem,int iOffset)
	{
	return _interlockedbittestandreset64_HLEAcquire((__int64*)pMem,iOffset);
	}

#elif defined(BIT_MSC_ARM)

	int _cdecl bit_clear(void *pMem,int iOffset)
	{
	return _interlockedbittestandreset_acq((long*)pMem,iOffset);
	}

#elif defined(BIT_MSC_VC60)

	int _cdecl bit_clear(void *pMem,int iOffset)
	{
	_asm{
		mov	 	eax,pMem
		mov  	ebx,iOffset
		lock 	btr [eax],ebx
		sbb  	eax,eax
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM) || defined(BIT_KEIL_ARM)

	int _cdecl bit_clear(void *pMem,int iOffset)
	{
	unsigned char  ucMask;
	unsigned char  ucBits;
	unsigned char *pPtr;


	pPtr   = (unsigned char*)pMem+iOffset/8;
	ucMask = 1<<(iOffset&7);

	#ifdef BIT_USE_IRQ

	{
	int iIrq;

	   iIrq=B_READ_IRQ();
	if(iIrq)B_DISABLE_IRQ();

	ucBits = *pPtr;
	*pPtr  = ucBits&~ucMask;

	if(iIrq)B_ENABLE_IRQ();
	}

	#else

	for(;;)
		{
		ucBits           = B_LDREXB(pPtr);
		if(!B_STREXB(ucBits&~ucMask,pPtr))
			{
			break;
			}
		}

	#endif

	return ucBits&ucMask;
	}

#endif

//*****************************************************************************
//*
//*		bit_clear
//*
//*****************************************************************************
//	Lscht Bits in einem Speicherbereich.
//	pMem		: Zeiger auf den Speicherbereich
//	iOffset		: Ist die Nummer des Bits das gesetzt werden soll.
//	iCount		: Ist die Anzahl der Bits. Wenn der Wert kleiner
//				  gleich 0 ist,wird nicht gendert
void _cdecl bit_clear(void *pMem,int iOffset,int iCount)
{
unsigned char  *pData=(unsigned char*)pMem;
int				iMask;
int				iOff;
int				iNum;


	if(iCount<=0)return;

	iOff = iOffset&7;

	if(iOff)
		{
		iNum	= 8-iOff;

		if(iNum>iCount)
			{
			iMask = ~(0xFFFFFFFF<<iNum);
			}
		else{
			iMask = 0xFF;
			}

		pData[iOffset/8] &= ~(iMask<<iOff);

		iCount -= iNum;
		iOffset   += iNum;

		if(iCount<=0)return;
		}


	memset(pData+iOffset/8,0x00,iCount/8);


	if(iCount&7)
		{
		iOffset /= 8;
		iOffset += iCount/8;
		iCount  &= 7;
		iMask	 = 0xFFFFFFFF<<iCount;
		pData[iOffset] &= iMask;
		}

}


//*****************************************************************************
//*
//*		bit_change
//*
//*****************************************************************************
//	Invertiert ein Bit in einem Speicherbereich. Die Funktion ist atomar
//	pMem		: Zeiger auf den Speicherbereich
//	iOffset		: Ist die Nummer des Bits das ivertiert werden soll.
//	Ergibt einen Wert ungleich 0 wenn das alte Bit gesetzt war sonst 0
#if defined(NO_BIT_ASM)

	int _cdecl bit_change(void *pMem,int iOffset)
	{
	char   *pPtr=(char*)pMem;
	char	cBit;
	int		iRet;

	pPtr += iOffset>>3;
	cBit  = 1<<(iOffset&7);

	BITLOCK();

	if(*pPtr&cBit)
		{
		iRet = -1;
		}
	else{
		iRet =  0;
		}

	*pPtr ^= cBit;

	BITUNLOCK();

	return iRet;
	}

#elif defined(BIT_MSC_ARM64)

	int _cdecl bit_change(void *pMem,int iOffset)
	{
	char   *pPtr=(char*)pMem;
	char	cBit;
	int		iRet;

	pPtr += iOffset>>3;
	cBit  = 1<<(iOffset&7);

	#warning no atomar intrinsic function _bittestandcomplement64
	// return _bittestandcomplement64((__int64*)pMem,iOffset);

	if(*pPtr&cBit)
		{
		iRet = -1;
		}
	else{
		iRet =  0;
		}

	*pPtr ^= cBit;

	return iRet;
	}

#elif defined(BIT_MSC_THUMB)

	#pragma code_seg(push,".text")

	__declspec(allocate(".text")) unsigned short aBitChangeThumb[] =
		{
		0x108A,					// ASRS      R2, R1, #2
		0xEB01,0x7252,			// ADD.W     R2, R1, R2, LSR #29
		0xEB00,0x00E2,			// ADD.W     R0, R0, R2, ASR #3
		0xF05F,0x0201,			// MOVS.W    R2, #1
		0xF001,0x0107,			// AND.W     R1, R1, #7
		0xFA02,0xF101,			// LSL.W     R1, R2, R1
		0xB2C9,					// UXTB      R1, R1
		0xE8D0,0x2F4F,			// LDREXB    R2,[R0]
		0xEA81,0x0302,			// EOR.W     R3, R1, R2
		0xE8C0,0x3F4C,			// STREXB    R12,R3,[R0]
		0xF1BC,0x0F00,			// CMP.W     R12, #0
		0xD1F6,					// BNE.N     --> LDREXB ...
		0xEA01,0x0002,			// AND.W     R0, R1, R2
		0x4770,					// BX        LR
		};

	#pragma code_seg(pop)

	#ifdef _DEBUG

	int _cdecl bit_change(void *pMem,int iOffset)
		{
		return ((bit_change_proc)&aBitChangeThumb)(pMem,iOffset);
		}

	#endif

#elif defined(BIT_MSC_ARM)

	#pragma code_seg(push,".text")

	__declspec(allocate(".text")) unsigned int  aBitChangeArm[] =
		{
		0xE1A02141,				// ASR       R2, R1, #2
		0xE0812EA2,				// ADD       R2, R1, R2, LSR #29
		0xE2011007,				// AND       R1, R1, #7
		0xE08001C2,				// ADD       R0, R0, R2, ASR #3
		0xE3A02001,				// MOV       R2, #1
		0xE1A01112,				// LSL       R1, R2, R1
		0xE6EF1071,				// UXTB      R1, R1
		0xE1D02F9F,				// LDREXB    R2,[R0]
		0xE0213002,				// EOR       R3, R1, R2
		0xE1C0CF93,				// STREXB    R12,R3,[R0]
		0xE35C0000,				// CMP       R12,#0
		0x1AFFFFFA,				// BNE       --->LDREXB ...
		0xE0010002,				// AND       R0, R1, R2
		0xE12FFF1E,				// BX        LR
		};

	#pragma code_seg(pop)

	#ifdef _DEBUG

	int _cdecl bit_change(void *pMem,int iOffset)
		{
		return ((bit_change_proc)&aBitChangeThumb)(pMem,iOffset);
		}

	#endif

#elif defined(BIT_MSC_X64)

	#pragma code_seg(push,".text")

	__declspec(allocate(".text")) unsigned char	aBitChangeX64[] =
		{
		0xF0,0x0F,0xBB,0x11,		//	lock btc   [rcx],edx
		0x1B,0xC0,					//	sbb			eax ,eax
		0xC3						//	retn
		};

	#pragma code_seg(pop)

	#ifdef _DEBUG

	int _cdecl bit_change(void *pMem,int iOffset)
		{
		return ((bit_change_proc)&aBitChangeX64)(pMem,iOffset);
		}

	#endif

#elif !defined(_DEBUG)

	// function defined in inline in "BitFunctions.h"

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	int _cdecl bit_change(void * pMem,int iOffset)
	{
	int iRet;

	asm volatile (
		"lock; btc  %2,%1  \n"
		"sbb        %0,%0  \n"
		:"=r" (iRet), BIT_ADDR(pMem)
		:"Ir" (iOffset)
		:"memory", "cc"
		);

	return  iRet;
	}

#elif defined(BIT_MSC_X86)

	int _cdecl bit_change(void * pMem,int iOffset)
	{
	_asm{
		mov	 	eax,pMem
		mov  	ebx,iOffset
		lock 	btc [eax],ebx
		sbb  	eax,eax
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM) || defined(BIT_KEIL_ARM)

	int _cdecl bit_change(void * pMem,int iOffset)
	{
	unsigned char  ucMask;
	unsigned char  ucBits;
	unsigned char *pPtr;


	pPtr   = (unsigned char*)pMem+iOffset/8;
	ucMask = 1<<(iOffset&7);

	#ifdef BIT_USE_IRQ

	{
	int iIrq;

	   iIrq=B_READ_IRQ();
	if(iIrq)B_DISABLE_IRQ();

	ucBits = *pPtr;
	*pPtr  = ucBits^ucMask;

	if(iIrq)B_ENABLE_IRQ();
	}

	#else

	for(;;)
		{
		ucBits          = B_LDREXB(pPtr);
		if(!B_STREXB(ucBits^ucMask,pPtr))
			{
			break;
			}
		}

	#endif

	return ucBits&ucMask;
	}

#endif

//*****************************************************************************
//*
//*		bit_change
//*
//*****************************************************************************
//	Invertiert einige Bits in einem Speicherbereich auf 1.
//	pMem		: Zeiger auf den Speicherbereich
//	iOffset		: Ist die Nummer des Bits das gesetzt werden soll.
//	iCount		: Ist die Anzahl der Bits. Wenn der Wert kleiner
//				  gleich 0 ist,wird nicht gendert
void _cdecl bit_change(void *pMem,int iOffset,int iCount)
{
unsigned char  *pData=(unsigned char*)pMem;
int				iMask;
int				iOff;
int				iNum;


	if(iCount<=0)return;

	iOff = iOffset&7;

	if(iOff)
		{
		iNum	= 8-iOff;

		if(iNum>iCount)
			{
			iMask = ~(0xFFFFFFFF<<iNum);
			}
		else{
			iMask = 0xFF;
			}

		pData[iOffset/8] ^= iMask<<iOff;

		iCount   -= iNum;
		iOffset  += iNum;

		if(iCount<=0)return;
		}


	iOffset /= 8;


	for(iNum=iCount/8;iNum>0;iNum--)
		{
		pData[iOffset] ^= 0xFF;
		iOffset++;
		}


	if(iCount&7)
		{
		iCount  &= 7;
		iMask	 = ~(0xFFFFFFFF<<iCount);
		pData[iOffset] ^= iMask;
		}

}

//*****************************************************************************
//*
//*		bit_swap_qw
//*
//*****************************************************************************
//	Tauscht die Bits in einer Zahl. Dabei wird das niederwetigste Bit an die
//	hchste Position verschoben und umgekehrt. (fr 64 Bit Zahlen)
//	uValue		: Ist die Zahl in der die Bits getauscht werden sollen.
//	iCount		: Bestimmt wieviele Bits ab dem niederwetigsten getauscht werden.
//	Ergibt die vernderte Zahl
unsigned ibt64 _cdecl bit_swap_qw(unsigned ibt64 uValue,int iCount)
{
unsigned 		uLower;
unsigned 		uUpper;
unsigned ibt64	uRet;


	if(iCount<=32)
		{
 		return bit_swap((unsigned)uValue,iCount);
		}

	if(iCount>64)
		{
		return 0;
		}

	uLower	 = (unsigned)uValue;
	uValue >>= 32;
	uUpper	 =  (unsigned)uValue;

	uLower	 = bit_swap(uLower,32);
	uValue <<= 32;
	uUpper	=  bit_swap(uUpper,32);

	uRet	 = uLower;
	uRet   <<= 32;
	uRet	|= uUpper;
	uRet   >>= 64-iCount;


return uRet;
}

//*****************************************************************************
//*
//*		bit_read
//*
//*****************************************************************************
//	Liet ein Bit in einem Speicherbereich aus. Die Funktion ist atomar
//	pMem		: Zeiger auf den Speicherbereich
//	iOffset		: Ist die Nummer des Bits das gelesen werden soll.
//	Ergibt -1 wenn das Bit gesetzt war sonst 0

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	int _cdecl bit_read(const void *pMem,int iOffset)
	{
	return (((unsigned char*)pMem)[iOffset>>3]>>(iOffset&7))&1;
	}

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	int _cdecl bit_read(const void *pMem,int iOffset)
	{
	int iRet;

	asm volatile (
		"bt			%2,%1  \n"
		"sbb        %0,%0  \n"
		:"=r" (iRet), BIT_ADDR(pMem)
		:"Ir" (iOffset)
		:"memory", "cc"
		);

	return  iRet;
	}

#elif defined(BIT_GNU_X64)

	int _cdecl bit_read(const void *pMem,int iOffset)
	{
	int iRet;

    asm volatile (
        "movl      %1   ,%%rax  \n"
        "bt        %2,  (%%rax) \n"
        "sbb       %%eax,%%eax  \n"
        : "=a" (iRet)
        : "m"  (pMem) , "c" (iOffset)
        );

	return  iRet;
	}

#elif defined(BIT_MSC_X64) || defined(BIT_MSC_ARM64)

	int _cdecl bit_read(const void *pMem,int iOffset)
	{
	return _bittest64((__int64*)pMem,iOffset);
	}

#elif defined(BIT_MSC_INSTR)

	int _cdecl bit_read(const void *pMem,int iOffset)
	{
	return _bittest((long*)pMem,iOffset);
	}

#elif defined(BIT_MSC_VC60)

	int _cdecl bit_read(const void *pMem,int iOffset)
	{
	_asm{
		mov		eax,pMem
		mov		ebx,iOffset
		bt		[eax],ebx
		sbb		eax,eax
		}
	}

#else

	int _cdecl bit_read(const void *pMem,int iOffset)
	{
	return (((unsigned char*)pMem)[iOffset>>3]>>(iOffset&7))&1;
	}

#endif

//*****************************************************************************
//*
//*		bit_write
//*
//*****************************************************************************
//	Schreibt ein Bit in einen Speicherbereich.
//	pMem		: Zeiger auf den Speicherbereich
//	iOffset		: Ist die Nummer des Bits das geschrieben werden soll.
//	iBit		: Wenn ungleich 0 dann wird das Bit gesetzt.
//	Ergibt einen Wert ungleich 0 wenn das alte Bit gesetzt war sonst 0

#ifdef	_DEBUG

	int _cdecl bit_write(void *pMem,int iOffset,int iBit)
	{
	if(iBit)return bit_set  (pMem,iOffset);
			return bit_clear(pMem,iOffset);
	}

#endif

//*****************************************************************************
//*
//*		bit_swap
//*
//*****************************************************************************
//	Tauscht die Bits in einer Zahl. Dabei wird das niederwetigste Bit an die
//	hchste Position verschoben und umgekehrt.
//	uValue		: Ist die Zahl in der die Bits getauscht werden sollen.
//	iCount		: Bestimmt wieviele Bits ab dem niederwetigsten getauscht werden.
//	Ergibt die vernderte Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned int _cdecl bit_swap(unsigned int uValue,int iCount)
	{
	unsigned	uRet=0;

	for(;iCount>0;iCount--)
		{
		uRet   <<= 1;
		if(uValue&1)uRet|=1;
		uValue >>= 1;
		}

	return uRet;
	}

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	unsigned int _cdecl bit_swap(unsigned int uValue,int iCount)
	{
	unsigned	uRet;

    asm volatile(
        "xorl   %%eax,%%eax     \n"
		"jecxz  2f              \n"
        "1:                     \n"
        "shr    $1,%%edx        \n"
		"rcl    $1,%%eax        \n"
        "dec    %%ecx           \n"
        "jnz    1b              \n"
        "2:                     \n"
        :"=a" (uRet)
        :"d"  (uValue), "c" (iCount)
		:"cc"
        );

	return  uRet;
	}

#elif defined(BIT_MSC_X86)

	unsigned int _cdecl bit_swap(unsigned int uValue,int iCount)
	{
	_asm{
		mov	 	edx,uValue
		mov  	ecx,iCount
		xor  	eax,eax
		jecxz	Exit
	Schl:
		shr	 	edx,1
		rcl  	eax,1
		dec	 	ecx
		jnz	 	Schl
	Exit:
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM)

	unsigned int _cdecl bit_swap(unsigned int uValue,int iCount)
	{
	unsigned 	uRet;

    asm volatile( "RBIT %0,%1"
                  : "=r"(uRet)
                  : "r" (uValue));

	return uRet>>(32-iCount);
	}

#elif defined(BIT_KEIL_ARM)

	unsigned int _cdecl bit_swap(unsigned int uValue,int iCount)
	{
	unsigned 	uRet;

	if(iCount<=0)return 0;

    __asm
		{
		RBIT 	uRet,uValue
		}

	return uRet>>(32-iCount);
	}

#else

	unsigned int _cdecl bit_swap(unsigned int uValue,int iCount)
	{
	unsigned	uRet=0;

	for(;iCount>0;iCount--)
		{
		uRet   <<= 1;
		if(uValue&1)uRet|=1;
		uValue >>= 1;
		}

	return uRet;
	}

#endif

//*****************************************************************************
//*
//*		bit_rol_qw
//*
//*****************************************************************************
//	Rollt einen 64-Bitwert Bits um eine Anzahl an Bits nach links oder rechts.
//	uBits		: Ist der Wert der gerollt werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert gerollt werden soll.
//					>0 es wird nach links  gerollt
//					<0 es wird nach rechts gerollt
//	Ergibt die gerollte Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned ibt64 _cdecl bit_rol_qw(unsigned ibt64 uBits,int iCount)
	{
	unsigned ibt64	uValue;

	iCount &= 0x3F;
	uValue	= uBits <<  iCount;
	uValue |= uBits >> (64-iCount);

	return uValue;
	}

#elif defined(BIT_GNU_X64)

	unsigned ibt64 _cdecl bit_rol_qw(unsigned ibt64 uBits,int iCount)
	{
	ibt64  iRet;

    asm volatile(
        "rol  %%cl,%%rax  \n"
        :"=a" (iRet)
        :"a"  (uBits), "c" (iCount)
		:"cc"
        );

	return iRet;
	}

#elif defined(BIT_MSC_INSTR)

	unsigned ibt64 _cdecl bit_rol_qw(unsigned ibt64 uBits,int iCount)
	{
	return _rotl64(uBits,iCount);
	}

#elif defined(BIT_MSC_VC60)

	unsigned ibt64 _cdecl bit_rol_qw(unsigned ibt64 uBits,int iCount)
	{
	_asm{
		mov	  	esi,dword ptr uBits+0
		mov	  	edi,dword ptr uBits+4
		mov   	ecx,iCount
		and   	ecx,63
		test  	ecx,32
		jz    	No64
		and   	ecx,31
		xchg  	esi,edi
	No64:
		jecxz 	NoRol
		shld  	edx,esi,cl
		shld  	eax,edi,cl
		neg   	cl
		shl   	edx,cl
		shl   	eax,cl
		shrd  	edx,edi,cl
		shrd  	eax,esi,cl
		jmp   	RolExit
	NoRol:
		mov   	eax,esi
		mov   	edx,edi
	RolExit:
		}
	}

#else

	unsigned ibt64 _cdecl bit_rol_qw(unsigned ibt64 uBits,int iCount)
	{
	unsigned ibt64	uValue;

	iCount &= 0x3F;
	uValue	= uBits <<  iCount;
	uValue |= uBits >> (64-iCount);

	return uValue;
	}

#endif

//*****************************************************************************
//*
//*		bit_rol_dw
//*
//*****************************************************************************
//	Rollt einen 32-Bitwert Bits um eine Anzahl an Bits nach links oder rechts.
//	uBits		: Ist der Wert der gerollt werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert gerollt werden soll.
//					>0 es wird nach links  gerollt
//					<0 es wird nach rechts gerollt
//	Ergibt die gerollte Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned int _cdecl bit_rol_dw(unsigned int uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x1F;
	uValue	= uBits <<  iCount;
	uValue |= uBits >> (32-iCount);

	return uValue;
	}

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	unsigned int _cdecl bit_rol_dw(unsigned int uBits,int iCount)
	{
	int  iRet;

    asm volatile(
        "rol  %%cl,%%eax  \n"
        :"=a" (iRet)
        :"a"  (uBits), "c" (iCount)
        :"cc"
        );

	return iRet;
	}

#elif defined(BIT_MSC_INSTR)

	unsigned int _cdecl bit_rol_dw(unsigned int uBits,int iCount)
	{
	return _rotl(uBits,iCount);
	}

#elif defined(BIT_MSC_VC60)

	unsigned int _cdecl bit_rol_dw(unsigned int uBits,int iCount)
	{
	_asm{
		mov		eax,uBits
		mov		ecx,iCount
		rol		eax,cl
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM)

	unsigned int _cdecl bit_rol_dw(unsigned int uBits,int iCount)
	{
	int  iRet;

	iCount = -iCount;

    asm volatile( "ROR %0,%1,%2"
                  : "=r"(iRet)
                  : "r" (uBits),"r" (iCount)
                  : "cc" );

	return iRet;
	}

#elif defined(BIT_KEIL_ARM)

	unsigned int _cdecl bit_rol_dw(unsigned int uBits,int iCount)
	{
	int		iRet;

	iCount = -iCount;

    __asm
		{
		ROR 	iRet,uBits,iCount
		}

	return iRet;
	}

#else

	unsigned int _cdecl bit_rol_dw(unsigned int uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x1F;
	uValue	= uBits <<  iCount;
	uValue |= uBits >> (32-iCount);

	return uValue;
	}

#endif

//*****************************************************************************
//*
//*		bit_rol_w
//*
//*****************************************************************************
//	Rollt einen 16-Bitwert Bits um eine Anzahl an Bits nach links oder rechts.
//	uBits		: Ist der Wert der gerollt werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert gerollt werden soll.
//					>0 es wird nach links  gerollt
//					<0 es wird nach rechts gerollt
//	Ergibt die gerollte Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned short _cdecl bit_rol_w(unsigned short uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x0F;
	uValue	= uBits <<  iCount;
	uValue |= uBits >> (16-iCount);

	return (unsigned short)uValue;
	}

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	unsigned short _cdecl bit_rol_w(unsigned short uBits,int iCount)
	{
	int  iRet;

    asm volatile(
        "rol  %%cl,%%ax  \n"
        :"=a" (iRet)
        :"a"  (uBits), "c" (iCount)
        :"cc"
        );

	return iRet;
	}

#elif defined(BIT_MSC_INSTR)

	unsigned short _cdecl bit_rol_w(unsigned short uBits,int iCount)
	{
	return _rotl16(uBits,iCount);
	}

#elif defined(BIT_MSC_VC60)

	unsigned short _cdecl bit_rol_w(unsigned short uBits,int iCount)
	{
	_asm{
		mov		ax ,uBits
		mov		ecx,iCount
		rol		ax ,cl
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM)

	unsigned short _cdecl bit_rol_w(unsigned short uBits,int iCount)
	{
	int		iRet;
	int		iVal;

    iVal   =  uBits;
	iCount = -iCount;

    asm volatile( "ORR %1,%1,%1,ROR #16 \n"
				  "ROR %0,%1,%2         \n"
                  : "=r"(iRet)
                  : "r" (iVal),"r" (iCount)
                  : "cc" );

	return iRet;
	}

#elif defined(BIT_KEIL_ARM)

	unsigned short _cdecl bit_rol_w(unsigned short uBits,int iCount)
	{
	int		iRet;
	int		iVal;

    iVal   =  uBits;
	iCount = -iCount;

    __asm
		{
		ORR 	iVal,iVal,iVal,ROR #16
		ROR 	iRet,iVal,iCount
		}

	return iRet;
	}

#else

	unsigned short _cdecl bit_rol_w(unsigned short uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x0F;
	uValue	= uBits <<  iCount;
	uValue |= uBits >> (16-iCount);

	return (unsigned short)uValue;
	}

#endif

//*****************************************************************************
//*
//*		bit_rol_b
//*
//*****************************************************************************
//	Rollt einen 8-Bitwert Bits um eine Anzahl an Bits nach links oder rechts.
//	uBits		: Ist der Wert der gerollt werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert gerollt werden soll.
//					>0 es wird nach links  gerollt
//					<0 es wird nach rechts gerollt
//	Ergibt die gerollte Zahl
//	Ergibt die gerollte Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned char _cdecl bit_rol_b(unsigned char uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x07;
	uValue	= uBits <<  iCount;
	uValue |= uBits >> (8-iCount);

	return (unsigned char)uValue;
	}

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	unsigned char _cdecl bit_rol_b(unsigned char uBits,int iCount)
	{
	int  iRet;

    asm volatile(
        "rol  %%cl,%%al  \n"
        :"=a" (iRet)
        :"a"  (uBits), "c" (iCount)
        :"cc"
        );

	return iRet;
	}

#elif defined(BIT_MSC_INSTR)

	unsigned char _cdecl bit_rol_b(unsigned char uBits,int iCount)
	{
	return _rotl8(uBits,iCount);
	}

#elif defined(BIT_MSC_VC60)

	unsigned char _cdecl bit_rol_b(unsigned char uBits,int iCount)
	{
	_asm{
		mov		al ,uBits
		mov		ecx,iCount
		rol		al ,cl
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM)

	unsigned char _cdecl bit_rol_b(unsigned char uBits,int iCount)
	{
	int		iRet;
	int		iVal;

    iVal   =  uBits;
	iCount = -iCount;

    asm volatile( "ORR %1,%1,%1,ROR #8  \n"
				  "ORR %1,%1,%1,ROR #16 \n"
				  "ROR %0,%1,%2         \n"
                  : "=r"(iRet)
                  : "r" (iVal),"r" (iCount)
                  : "cc" );

	return iRet;
	}

#elif defined(BIT_KEIL_ARM)

	unsigned char _cdecl bit_rol_b(unsigned char uBits,int iCount)
	{
	int		iRet;
	int		iVal;

    iVal   =  uBits;
	iCount = -iCount;

    __asm
		{
		ORR 	iVal,iVal,iVal,ROR #8
		ORR 	iVal,iVal,iVal,ROR #16
		ROR 	iRet,iVal,iCount
		}

	return iRet;
	}

#else

	unsigned char _cdecl bit_rol_b(unsigned char uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x07;
	uValue	= uBits <<  iCount;
	uValue |= uBits >> (8-iCount);

	return (unsigned char)uValue;
	}

#endif

//*****************************************************************************
//*
//*		bit_ror_qw
//*
//*****************************************************************************
//	Rollt einen 64-Bitwert Bits um eine Anzahl an Bits nach rechts oder rechts.
//	uBits		: Ist der Wert der gerollt werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert gerollt werden soll.
//					>0 es wird nach rechts  gerollt
//					<0 es wird nach rechts gerollt
//	Ergibt die gerollte Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned ibt64 _cdecl bit_ror_qw(unsigned ibt64 uBits,int iCount)
	{
	unsigned ibt64	uValue;

	iCount &= 0x3F;
	uValue	= uBits >>  iCount;
	uValue |= uBits << (64-iCount);

	return uValue;
	}

#elif defined(BIT_GNU_X64)

	unsigned ibt64 _cdecl bit_ror_qw(unsigned ibt64 uBits,int iCount)
	{
	ibt64  iRet;

    asm volatile(
        "ror  %%cl,%%rax  \n"
        :"=a" (iRet)
        :"a"  (uBits), "c" (iCount)
        :"cc"
        );

	return iRet;
	}

#elif defined(BIT_MSC_INSTR)

	unsigned ibt64 _cdecl bit_ror_qw(unsigned ibt64 uBits,int iCount)
	{
	return _rotr64(uBits,iCount);
	}

#elif defined(BIT_MSC_VC60)

	unsigned ibt64 _cdecl bit_ror_qw(unsigned ibt64 uBits,int iCount)
	{
	_asm{
		mov	  	esi,dword ptr uBits+0
		mov	  	edi,dword ptr uBits+4
		mov   	ecx,iCount
		and   	ecx,63
		test  	ecx,32
		jz    	No64
		and   	ecx,31
		xchg  	esi,edi
	No64:
		jecxz 	NoRol
		shrd  	edx,esi,cl
		shrd  	eax,edi,cl
		neg   	cl
		shr   	edx,cl
		shr   	eax,cl
		shld  	edx,edi,cl
		shld  	eax,esi,cl
		jmp   	RolExit
	NoRol:
		mov   	eax,esi
		mov   	edx,edi
	RolExit:
		}
	}

#else

	unsigned ibt64 _cdecl bit_ror_qw(unsigned ibt64 uBits,int iCount)
	{
	unsigned ibt64	uValue;

	iCount &= 0x3F;
	uValue	= uBits >>  iCount;
	uValue |= uBits << (64-iCount);

	return uValue;
	}

#endif

//*****************************************************************************
//*
//*		bit_ror_dw
//*
//*****************************************************************************
//	Rollt einen 32-Bitwert Bits um eine Anzahl an Bits nach rechts oder rechts.
//	uBits		: Ist der Wert der gerollt werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert gerollt werden soll.
//					>0 es wird nach rechts  gerollt
//					<0 es wird nach rechts gerollt
//	Ergibt die gerollte Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned int _cdecl bit_ror_dw(unsigned int uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x1F;
	uValue	= uBits >>  iCount;
	uValue |= uBits << (32-iCount);

	return uValue;
	}

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	unsigned int _cdecl bit_ror_dw(unsigned int uBits,int iCount)
	{
	int  iRet;

    asm volatile(
        "ror  %%cl,%%eax   \n"
        :"=a" (iRet)
        :"a"  (uBits), "c" (iCount)
        );

	return iRet;
	}

#elif defined(BIT_MSC_INSTR)

	unsigned int _cdecl bit_ror_dw(unsigned int uBits,int iCount)
	{
	return _rotr(uBits,iCount);
	}

#elif defined(BIT_MSC_VC60)

	unsigned int _cdecl bit_ror_dw(unsigned int uBits,int iCount)
	{
	_asm{
		mov		eax,uBits
		mov		ecx,iCount
		ror		eax,cl
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM)

	unsigned int _cdecl bit_ror_dw(unsigned int uBits,int iCount)
	{
	int  iRet;

    asm volatile( "ROR %0,%1,%2"
                  : "=r"(iRet)
                  : "r" (uBits),"r" (iCount)
                  : "cc" );

	return iRet;
	}

#elif defined(BIT_KEIL_ARM)

	unsigned int _cdecl bit_ror_dw(unsigned int uBits,int iCount)
	{
	int		iRet;
	int		iVal;

    iVal   =  uBits;

    __asm
		{
		ROR 	iRet,iVal,iCount
		}

	return iRet;
	}

#else

	unsigned int _cdecl bit_ror_dw(unsigned int uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x1F;
	uValue	= uBits >>  iCount;
	uValue |= uBits << (32-iCount);

	return uValue;
	}

#endif

//*****************************************************************************
//*
//*		bit_ror_w
//*
//*****************************************************************************
//	Rollt einen 16-Bitwert Bits um eine Anzahl an Bits nach rechts oder rechts.
//	uBits		: Ist der Wert der gerollt werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert gerollt werden soll.
//					>0 es wird nach rechts  gerollt
//					<0 es wird nach rechts gerollt
//	Ergibt die gerollte Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned short _cdecl bit_ror_w(unsigned short uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x0F;
	uValue	= uBits >>  iCount;
	uValue |= uBits << (16-iCount);

	return (unsigned short)uValue;
	}

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	unsigned short _cdecl bit_ror_w(unsigned short uBits,int iCount)
	{
	int  iRet;

    asm volatile(
        "ror  %%cl,%%ax  \n"
        :"=a" (iRet)
        :"a"  (uBits), "c" (iCount)
        :"cc"
        );

	return iRet;
	}

#elif defined(BIT_MSC_INSTR)

	unsigned short _cdecl bit_ror_w(unsigned short uBits,int iCount)
	{
	return _rotr16(uBits,iCount);
	}

#elif defined(BIT_MSC_VC60)

	unsigned short _cdecl bit_ror_w(unsigned short uBits,int iCount)
	{
	_asm{
		mov		ax ,uBits
		mov		ecx,iCount
		ror		ax ,cl
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM)

	unsigned short _cdecl bit_ror_w(unsigned short uBits,int iCount)
	{
	int		iRet;
	int		iVal;

    iVal   =  uBits;

    asm volatile( "ORR %1,%1,%1,ROR #16 \n"
				  "ROR %0,%1,%2         \n"
                  : "=r"(iRet)
                  : "r" (iVal),"r" (iCount)
                  : "cc" );

	return iRet;
	}

#elif defined(BIT_KEIL_ARM)

	unsigned short _cdecl bit_ror_w(unsigned short uBits,int iCount)
	{
	int		iRet;
	int		iVal;

    iVal   =  uBits;

    __asm
		{
		ORR 	iVal,iVal,iVal,ROR #16
		ROR 	iRet,iVal,iCount
		}

	return iRet;
	}

#else

	unsigned short _cdecl bit_ror_w(unsigned short uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x0F;
	uValue	= uBits >>  iCount;
	uValue |= uBits << (16-iCount);

	return (unsigned short)uValue;
	}

#endif

//*****************************************************************************
//*
//*		bit_ror_b
//*
//*****************************************************************************
//	Rollt einen 8-Bitwert Bits um eine Anzahl an Bits nach rechts oder rechts.
//	uBits		: Ist der Wert der gerollt werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert gerollt werden soll.
//					>0 es wird nach rechts  gerollt
//					<0 es wird nach rechts gerollt
//	Ergibt die gerollte Zahl
//	Ergibt die gerollte Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned char _cdecl bit_ror_b(unsigned char uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x07;
	uValue	= uBits >>  iCount;
	uValue |= uBits << (8-iCount);

	return (unsigned char)uValue;
	}

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	unsigned char _cdecl bit_ror_b(unsigned char uBits,int iCount)
	{
	int  iRet;

    asm volatile(
        "ror  %%cl,%%al  \n"
        :"=a" (iRet)
        :"a"  (uBits), "c" (iCount)
        :"cc"
        );

	return iRet;
	}

#elif defined(BIT_MSC_INSTR)

	unsigned char _cdecl bit_ror_b(unsigned char uBits,int iCount)
	{
	return _rotr8(uBits,iCount);
	}

#elif defined(BIT_MSC_VC60)

	unsigned char _cdecl bit_ror_b(unsigned char uBits,int iCount)
	{
	_asm{
		mov		al ,uBits
		mov		ecx,iCount
		ror		al ,cl
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM)

	unsigned char _cdecl bit_ror_b(unsigned char uBits,int iCount)
	{
	int		iRet;
	int		iVal;

    iVal   =  uBits;

    asm volatile( "ORR %1,%1,%1,ROR #8  \n"
				  "ORR %1,%1,%1,ROR #16 \n"
				  "ROR %0,%1,%2         \n"
                  : "=r"(iRet)
                  : "r" (iVal),"r" (iCount)
                  : "cc" );

	return iRet;
	}

#elif defined(BIT_KEIL_ARM)

	unsigned char _cdecl bit_ror_b(unsigned char uBits,int iCount)
	{
	int		iRet;
	int		iVal;

    iVal   =  uBits;

    __asm
		{
		ORR 	iVal,iVal,iVal,ROR #8
		ORR 	iVal,iVal,iVal,ROR #16
		ROR 	iRet,iVal,iCount
		}

	return iRet;
	}

#else

	unsigned char _cdecl bit_ror_b(unsigned char uBits,int iCount)
	{
	unsigned  	uValue;

	iCount &= 0x07;
	uValue	= uBits >>  iCount;
	uValue |= uBits << (8-iCount);

	return (unsigned char)uValue;
	}

#endif

//*****************************************************************************
//*
//*		bit_shift
//*
//*****************************************************************************
//	Verschibt die Bits eines Wertes mit berlaufprfung. (nicht arithmetisch)
//	uBits		: Ist der Wert verschoben werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert verschoben werden soll.
//					>0 es wird nach links  verschoben
//					<0 es wird nach rechts verschoben
//	Ergibt die verschobene Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	int _cdecl bit_shift(unsigned int uBits,int iCount)
	{
	if(iCount<0)
		{
		if(iCount<=-32)return 0;
		return uBits>>-iCount;
		}

	if(iCount>=32)return 0;

	return uBits<<iCount;
	}

#elif defined(BIT_MSC_X86)

	int _cdecl bit_shift(unsigned int uBits,int iCount)
	{
	_asm{
		mov		ecx,iCount
		test 	ecx,80000000h
		jz	 	left

		neg  	ecx
		test 	ecx,NOT 1Fh
		jnz	 	null
		mov	 	eax,uBits
		shr	 	eax,cl
		jmp  	ende
	left:
		test 	ecx,NOT 1Fh
		jz	 	shiftl
	null:
		xor  	eax,eax
		jmp  	ende
	shiftl:
		mov	 	eax,uBits
		shl	 	eax,cl
	ende:
		}
	}

#else

	int _cdecl bit_shift(unsigned int uBits,int iCount)
	{
	if(iCount<0)
		{
		if(iCount<=-32)return 0;
		return uBits>>-iCount;
		}

	if(iCount>=32)return 0;

	return uBits<<iCount;
	}

#endif




//*****************************************************************************
//*
//*		bit_shift_l
//*
//*****************************************************************************
//	Verschibt die Bits eines Wertes mit berlaufprfung nach links.
//	iBits		: Ist der Wert verschoben werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert verschoben werden soll.
//	Ergibt die verschobene Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	int _cdecl bit_shift_l(int iBits,int iCount)
	{
	if(iCount&~0x1F)return 0;
	return iBits<<iCount;
	}

#elif defined(BIT_MSC_X86)

	int _cdecl bit_shift_l(int iBits,int iCount)
	{
	_asm{
		mov		ecx,iCount
		test 	ecx,NOT 1Fh
		jz	 	shift
		xor  	eax,eax
		jmp  	ende
	shift:
		mov	 	eax,iBits
		shl	 	eax,cl
	ende:
		}
	}

#else

	int _cdecl bit_shift_l(int iBits,int iCount)
	{
	if(iCount&~0x1F)return 0;
	return iBits<<iCount;
	}

#endif

//*****************************************************************************
//*
//*		bit_shift_r
//*
//*****************************************************************************
//	Verschibt die Bits eines Wertes mit berlaufprfung nach rechts. (nicht arithmetisch)
//	iBits		: Ist der Wert verschoben werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert verschoben werden soll.
//	Ergibt die verschobene Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	int _cdecl bit_shift_r(int iBits,int iCount)
	{
	if(iCount&~0x1F)return 0;
	unsigned uBits=iBits;
	return   uBits>>iCount;
	}

#elif defined(BIT_MSC_X86)

	int _cdecl bit_shift_r(int iBits,int iCount)
	{
	_asm{
		mov  	ecx,iCount
		test 	ecx,NOT 1Fh
		jz	 	shift
		xor  	eax,eax
		jmp  	ende
	shift:
		mov	 	eax,iBits
		shr	 	eax,cl
	ende:
		}
	}

#else

	int _cdecl bit_shift_r(int iBits,int iCount)
	{
	if(iCount&~0x1F)return 0;
	unsigned uBits=iBits;
	return   uBits>>iCount;
	}

#endif

//*****************************************************************************
//*
//*		bit_shift_a
//*
//*****************************************************************************
//	Verschibt die Bits eines Wertes mit berlaufprfung arithmetisch nach rechts.
//	iBits		: Ist der Wert verschoben werden soll.
//	iCount		: Ist die Anzahl der Bits um die der Wert verschoben werden soll.
//	Ergibt die verschobene Zahl

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	int _cdecl bit_shift_a(int iBits,int iCount)
	{
	if(iCount&~0x1F)iCount=31;
	return   iBits>>iCount;
	}

#elif defined(BIT_MSC_X86)

	int _cdecl bit_shift_a(int iBits,int iCount)
	{
	_asm{
		mov  	ecx,iCount
		mov	 	eax,iBits
		test 	ecx,NOT 1Fh
		jz	 	shift
		mov  	cl ,31
	shift:
		sar	 	eax,cl
		}
	}

#else

	int _cdecl bit_shift_a(int iBits,int iCount)
	{
	if(iCount&~0x1F)iCount=31;
	return   iBits>>iCount;
	}

#endif



//*****************************************************************************
//*
//*		btol_w
//*
//*****************************************************************************

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned short _cdecl btol_w(unsigned short uValue)
	{
	return (unsigned short)((uValue>>8)|(uValue<<8));
	}

#elif defined(BIT_GNU_4_3)

	unsigned short _cdecl btol_w(unsigned short uValue)
	{
	return __builtin_bswap16(uValue);
	}

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	unsigned short _cdecl btol_w(unsigned short uValue)
	{
	int iRet;

	asm volatile (
        "xchg   %%al,%%ah  \n"
        : "=a" (iRet)
        : "a"  (uValue)
        );

	return iRet;
	}

#elif defined(BIT_MSC_INSTR) || defined(BIT_MSC_ARM) || defined(BIT_MSC_ARM64)

	unsigned short _cdecl btol_w(unsigned short uValue)
	{
	return _byteswap_ushort(uValue);
	}

#elif defined(BIT_MSC_VC60)

	unsigned short _cdecl btol_w(unsigned short uValue)
	{
	_asm{
		mov		ax,uValue
		xchg	al,ah;				;// Tausche AL mit AH
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM)

	unsigned short _cdecl btol_w(unsigned short uValue)
	{
	unsigned 	uRet;

    asm volatile( "REVSH %0,%1"
                  : "=r"(uRet)
                  : "r" (uValue));

	return uRet;
	}

#elif defined(BIT_KEIL_ARM)

	unsigned short _cdecl btol_w(unsigned short uValue)
	{
	unsigned 	uRet;

    __asm
		{
		REVSH 	uRet,uValue
		}

	return uRet;
	}

#else

	unsigned short _cdecl btol_w(unsigned short uValue)
	{
	return (unsigned short)((uValue>>8)|(uValue<<8));
	}

#endif

//*****************************************************************************
//*
//*		btol_dw
//*
//*****************************************************************************

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned int _cdecl btol_dw(unsigned int uValue)
	{
	unsigned int  uVal;

	uVal  =            btol_w((unsigned short)(uValue>>16));
	uVal |= ((unsigned)btol_w((unsigned short)(uValue)))<<16;

	return uVal;
	}

#elif defined(BIT_GNU_4_3)

	unsigned int _cdecl btol_dw(unsigned int uValue)
	{
	return __builtin_bswap32(uValue);
	}

#elif defined(BIT_GNU_X86) || defined(BIT_GNU_X64)

	unsigned int _cdecl btol_dw(unsigned int uValue)
	{
    int iRet;

	asm volatile (
        "bswap     %%eax  \n"
        : "=a" (iRet)
        : "a"  (uValue)
        );

	return iRet;
	}

#elif defined(BIT_MSC_INSTR) || defined(BIT_MSC_ARM) || defined(BIT_MSC_ARM64)

	unsigned int _cdecl btol_dw(unsigned int uValue)
	{
	return _byteswap_ulong(uValue);
	}

#elif defined(BIT_MSC_VC60)

	unsigned int _cdecl btol_dw(unsigned int uValue)
	{
	_asm{
		mov		eax,uValue
		bswap	eax;				;// Tausche Bytes in EAX
		}
	}

#elif defined(BIT_GNU_ARM) || defined(BIT_GNU_ARM64) || defined(BIT_IAR_ARM)

	unsigned int _cdecl btol_dw(unsigned int uValue)
	{
	unsigned 	uRet;

    asm volatile( "REV %0,%1"
                  : "=r"(uRet)
                  : "r" (uValue));

	return uRet;
	}

#elif defined(BIT_KEIL_ARM)

	unsigned int _cdecl btol_dw(unsigned int uValue)
	{
	unsigned 	uRet;

    __asm
		{
		REV 	uRet,uValue
		}

	return uRet;
	}

#else

	unsigned int _cdecl btol_dw(unsigned int uValue)
	{
	unsigned int  uVal;

	uVal  =            btol_w((unsigned short)(uValue >>16));
	uVal |= ((unsigned)btol_w((unsigned short)(uValue)))<<16;

	return uVal;
	}

#endif


//*****************************************************************************
//*
//*		btol_qw
//*
//*****************************************************************************

#ifndef	_DEBUG

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	unsigned ibt64 _cdecl btol_qw(unsigned ibt64 uValue)
	{
	unsigned ibt64 uVal;

	uVal  =                  btol_dw((unsigned int)(uValue>>32));
	uVal |= ((unsigned ibt64)btol_dw((unsigned int)(uValue)))<<32;

	return uVal;
	}

#elif defined(BIT_GNU_4_3)

	unsigned ibt64 _cdecl btol_qw(unsigned ibt64 uValue)
	{
	return __builtin_bswap64(uValue);
	}

#elif defined(BIT_GNU_X64)

	unsigned ibt64 _cdecl btol_qw(unsigned ibt64 uValue)
	{
    ibt64 iRet;

	asm volatile (
        "bswap     %%rax  \n"
        : "=a" (iRet)
        : "a"  (uValue)
        );

	return iRet;
	}

#elif defined(BIT_MSC_INSTR) || defined(BIT_MSC_ARM) || defined(BIT_MSC_ARM64)

	unsigned ibt64 _cdecl btol_qw(unsigned ibt64 uValue)
	{
	return _byteswap_uint64(uValue);
	}

#elif defined(BIT_MSC_VC60)

	unsigned ibt64 _cdecl btol_qw(unsigned ibt64 uValue)
	{
	_asm{
		mov		edx,dword ptr uValue+0
		mov		eax,dword ptr uValue+4
		bswap	edx;
		bswap	eax;
		}
	}

#else

	unsigned ibt64 _cdecl btol_qw(unsigned ibt64 uValue)
	{
	unsigned ibt64  uVal;

	uVal  =         btol_dw((unsigned int)(uValue >>32));
	uVal |= ((ibt64)btol_dw((unsigned int)(uValue)))<<32;

	return uVal;
	}

#endif


//*****************************************************************************
//*
//*		btol
//*
//*****************************************************************************
//	Big-Endial zu Little-Endian
#ifdef	_DEBUG

	void	_cdecl btol(		 ibt64 &iValue){iValue=btol_qw(iValue);}
	void	_cdecl btol(unsigned ibt64 &uValue){uValue=btol_qw(uValue);}
	void	_cdecl btol(		 short &iValue){iValue=btol_w (iValue);}
	void	_cdecl btol(unsigned short &uValue){uValue=btol_w (uValue);}
	void	_cdecl btol(		 int   &iValue){iValue=btol_dw(iValue);}
	void	_cdecl btol(unsigned int   &uValue){uValue=btol_dw(uValue);}

	#if defined(BIT_GNU_X64) || defined(BIT_GNU_ARM64)
	void	_cdecl btol(		 long  &iValue){iValue=btol_qw(iValue);}
	void	_cdecl btol(unsigned long  &uValue){uValue=btol_qw(uValue);}
	#else
	void	_cdecl btol(		 long  &iValue){iValue=btol_dw(iValue);}
	void	_cdecl btol(unsigned long  &uValue){uValue=btol_dw(uValue);}
	#endif

#endif

//*****************************************************************************
//*
//*		bit_rol
//*
//*****************************************************************************
//	Bits nach links rollen
#ifdef	_DEBUG

	void	_cdecl bit_rol(		    ibt64 &iValue,int iCount){iValue=bit_rol_qw(iValue,iCount);}
	void	_cdecl bit_rol(unsigned ibt64 &uValue,int iCount){uValue=bit_rol_qw(uValue,iCount);}
	void	_cdecl bit_rol(		    short &iValue,int iCount){iValue=bit_rol_w (iValue,iCount);}
	void	_cdecl bit_rol(unsigned short &uValue,int iCount){uValue=bit_rol_w (uValue,iCount);}
	void	_cdecl bit_rol(		    int   &iValue,int iCount){iValue=bit_rol_dw(iValue,iCount);}
	void	_cdecl bit_rol(unsigned int   &uValue,int iCount){uValue=bit_rol_dw(uValue,iCount);}

	#if defined(BIT_GNU_X64) || defined(BIT_GNU_ARM64)
	void	_cdecl bit_rol(		    long  &iValue,int iCount){iValue=bit_rol_qw(iValue,iCount);}
	void 	_cdecl bit_rol(unsigned long  &uValue,int iCount){uValue=bit_rol_qw(uValue,iCount);}
	#else
	void 	_cdecl bit_rol(		    long  &iValue,int iCount){iValue=bit_rol_dw(iValue,iCount);}
	void 	_cdecl bit_rol(unsigned long  &uValue,int iCount){uValue=bit_rol_dw(uValue,iCount);}
	#endif

#endif


//*****************************************************************************
//*
//*		bit_ror
//*
//*****************************************************************************
//	Bits nach rechts rollen
#ifdef	_DEBUG

	void	_cdecl bit_ror(		    ibt64 &iValue,int iCount){iValue=bit_ror_qw(iValue,iCount);}
	void	_cdecl bit_ror(unsigned ibt64 &uValue,int iCount){uValue=bit_ror_qw(uValue,iCount);}
	void	_cdecl bit_ror(		    short &iValue,int iCount){iValue=bit_ror_w (iValue,iCount);}
	void	_cdecl bit_ror(unsigned short &uValue,int iCount){uValue=bit_ror_w (uValue,iCount);}
	void	_cdecl bit_ror(		    int   &iValue,int iCount){iValue=bit_ror_dw(iValue,iCount);}
	void	_cdecl bit_ror(unsigned int   &uValue,int iCount){uValue=bit_ror_dw(uValue,iCount);}

	#if defined(BIT_GNU_X64) || defined(BIT_GNU_ARM64)
	void	_cdecl bit_ror(		    long  &iValue,int iCount){iValue=bit_ror_qw(iValue,iCount);}
	void 	_cdecl bit_ror(unsigned long  &uValue,int iCount){uValue=bit_ror_qw(uValue,iCount);}
	#else
	void 	_cdecl bit_ror(		    long  &iValue,int iCount){iValue=bit_ror_dw(iValue,iCount);}
	void 	_cdecl bit_ror(unsigned long  &uValue,int iCount){uValue=bit_ror_dw(uValue,iCount);}
	#endif

#endif


//*****************************************************************************
//*
//*		bit_find_1
//*
//*****************************************************************************
//	Sucht das erste gesetzte Bit in einem Speicherbereich.
//	pMem		: Ist der Zeiger auf den Speicherbereich.
//	uCount		: Ist die Gre der Speicherbereiches in Bits.
//	Ergibt die Position an der das Bit gefunden wurde, bzw. die Gre des Bereiches.
#if defined(BIT_KEIL_THUMB)

__asm int bit_find_1(const void *pMem,unsigned uCount)
	{
		// R0 = pMem
		// R1 = uCount

		MOV		r2,r1						// r2(pos) = uCount
		ANDS 	r3,r0,#3					// r3      = pMem&3;
		BEQ		next_c

		PUSH	{r4}						// Register retten
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		LSL		r3,r3,#3					// r3 	  *= 8
		LSL		r4,r4,r3					// mask  <<= (pMem&3)*8
		ADD		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		ANDS	r3,r3,r4					// DWORD  &= mask
		POP		{r4}						// Register wieder herstellen
		BNE		find_c						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_c						// if(pos<=0)...

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
next_c
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		CBNZ	r3,find_c					// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_c						// if(pos<=0)...
stop_c
		MOV		r0,r1						// return uCount;
		BLX		lr							//		*
find_c
		RBIT	r3,r3						// Bitreihenfolge umkehren
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_c						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		BLX		lr							//		*
	}

#elif defined(BIT_KEIL_ARM)

__asm int bit_find_1(const void *pMem,unsigned uCount)
	{
		// R0 = pMem
		// R1 = uCount

		MOV		r2,r1						// r2(pos) = uCount
		ANDS 	r3,r0,#3					// r3      = pMem&3;
		BEQ		next_c

		PUSH	{r4}						// Register retten
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		LSL		r3,r3,#3					// r3 	  *= 8
		LSL		r4,r4,r3					// mask  <<= (pMem&3)*8
		ADD		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		ANDS	r3,r3,r4					// DWORD  &= mask
		POP		{r4}						// Register wieder herstellen
		BNE		find_c						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_c						// if(pos<=0)...

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
next_c
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
	    CMP     r3,#0                       // if(DWORD==0)...
	    BNE     find_c						//      *
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_c						// if(pos<=0)...
stop_c
		MOV		r0,r1						// return uCount;
		BLX		lr							//		*
find_c
		RBIT	r3,r3						// Bitreihenfolge umkehren
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_c						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		BLX		lr							//		*
	}

#elif defined(BIT_IAR_ARM)

	// this functions are defined in "./Assembler/BitFunctionsIar[Thumb].s"

#elif !defined(_DEBUG)

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	int _cdecl bit_find_1(const void *pMem,unsigned uCount)
	{
	return bit_find_1(pMem,uCount,0);
	}

#elif defined(BIT_GNU_X86)

	int _cdecl bit_find_1(const void *pMem,unsigned uCount)
	{
	int iRet;

    asm volatile(
        "xorl   %%eax,%%eax     \n"
        "jmp    2f              \n"
        "1:                     \n"
        "addl   $4   ,%%esi     \n"
        "addl   $32  ,%%eax     \n"
        "2:                     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "bsfl   (%%esi),%%ecx   \n"
        "jz     1b              \n"
        "add    %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
        :"=a" (iRet)
        :"S" (pMem), "D" (uCount)
        :"ecx"
        );

	return  iRet;
	}

#elif defined(BIT_GNU_X64)

	int _cdecl bit_find_1(const void *pMem,unsigned uCount)
	{
	int iRet;

    asm volatile(
        "xorl   %%eax,%%eax     \n"
        "jmp    2f              \n"
        "1:                     \n"
        "addq   $8   ,%%rsi     \n"
        "addl   $64  ,%%eax     \n"
        "2:                     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "bsfq   (%%rsi),%%rcx   \n"
        "jz     1b              \n"
        "add    %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
        :"=a" (iRet)
        :"S" (pMem), "D" (uCount)
        :"ecx"
        );

    return iRet;
	}

#elif defined(BIT_MSC_X86)

	int _cdecl bit_find_1(const void *pMem,unsigned uCount)
	{
	_asm{
		mov		esi,pMem			;// ESI = Adresse
		xor		eax,eax				;// EAX = Position
		mov		edi,uCount			;// EDI = Anzahl
		jmp		start
	loop1:
		add		esi,4				;// Die nchsten 32-Bit
		add		eax,32				;// 		*
	start:
		cmp		eax,edi				;// 		*
		jnb     ende1				;// berlauf ?
		bsf	    ecx,[esi]			;// Suche Bit
		jz	    loop1				;// 		*

		add     eax,ecx				;// Ergebnis berechen
      	cmp		eax,edi				;// berlauf ?
		jbe     ende2				;// 		*
	ende1:
		mov		eax,edi				;// 		*
	ende2:
		}
	}

#else

	int _cdecl bit_find_1(const void *pMem,unsigned uCount)
	{
	return bit_find_1(pMem,uCount,0);
	}

#endif

//*****************************************************************************
//*
//*		bit_find_1x
//*
//*****************************************************************************
//	Sucht das erste gesetzte Bit in einem Speicherbereich.
//	Die Bits werden von 0x80 in Richtung 0x01 gesucht.
//	pMem		: Ist der Zeiger auf den Speicherbereich.
//	uCount		: Ist die Gre der Speicherbereiches in Bits.
//	Ergibt die Position an der das Bit gefunden wurde, bzw. die Gre des Bereiches.
#if defined(BIT_KEIL_THUMB)

__asm int bit_find_1x(const void *pMem,unsigned uCount)
	{
		// R0 = pMem
		// R1 = uCount

		MOV		r2,r1						// r2(pos) = uCount
		ANDS 	r3,r0,#3					// r3      = pMem&3;
		BEQ		next_e

		PUSH	{r4}						// Register retten
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		LSL		r3,r3,#3					// r3 	  *= 8
		LSR		r4,r4,r3					// mask  >>= (pMem&3)*8
		ADD		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		ANDS	r3,r3,r4					// DWORD  &= mask
		POP		{r4}						// Register wieder herstellen
		BNE		find_e						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_e						// if(pos<=0)...

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
next_e
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		CBNZ	r3,find_e					// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_e						// if(pos<=0)...
stop_e
		MOV		r0,r1						// return uCount;
		BLX		lr							//		*
find_e
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_e						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		BLX		lr							//		*
	}

#elif defined(BIT_KEIL_ARM)

__asm int bit_find_1x(const void *pMem,unsigned uCount)
	{
		// R0 = pMem
		// R1 = uCount

		MOV		r2,r1						// r2(pos) = uCount
		ANDS 	r3,r0,#3					// r3      = pMem&3;
		BEQ		next_e

		PUSH	{r4}						// Register retten
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		LSL		r3,r3,#3					// r3 	  *= 8
		LSR		r4,r4,r3					// mask  >>= (pMem&3)*8
		ADD		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		ANDS	r3,r3,r4					// DWORD  &= mask
		POP		{r4}						// Register wieder herstellen
		BNE		find_e						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_e						// if(pos<=0)...

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
next_e
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
	    CMP     r3,#0                       // if(DWORD==0)...
	    BNE     find_e						//      *
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_e						// if(pos<=0)...
stop_e
		MOV		r0,r1						// return uCount;
		BLX		lr							//		*
find_e
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_e						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		BLX		lr							//		*
	}

#elif defined(BIT_IAR_ARM)

	// this functions are defined in "./Assembler/BitFunctionsIar[Thumb].s"

#elif !defined(_DEBUG)

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount)
	{
	return bit_find_1x(pMem,uCount,0);
	}

#elif defined(BIT_GNU_X86)

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount)
	{
	int iRet;

    asm volatile(
        "xorl   %%eax,%%eax     \n"
        "jmp    2f              \n"
        "1:                     \n"
        "addl   $4   ,%%esi     \n"
        "addl   $32  ,%%eax     \n"
        "2:                     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
		"movl   (%%esi),%%ecx   \n"
		"bswap  %%ecx           \n"
        "bsrl   %%ecx,%%ecx     \n"
        "jz     1b              \n"
		"xorl   $31  ,%%ecx     \n"
        "add    %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
        :"=a" (iRet)
        :"S" (pMem), "D" (uCount)
        :"ecx"
        );

	return  iRet;
	}

#elif defined(BIT_GNU_X64)

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount)
	{
	int iRet;

    asm volatile(
        "xorl   %%eax,%%eax     \n"
        "jmp    2f              \n"
        "1:                     \n"
        "addq   $8   ,%%rsi     \n"
        "addl   $64  ,%%eax     \n"
        "2:                     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
		"movq   (%%rsi),%%rcx   \n"
		"bswap  %%rcx           \n"
        "bsrq   %%rcx,%%rcx     \n"
        "jz     1b              \n"
		"xorl   $63  ,%%ecx     \n"
        "add    %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
        :"=a" (iRet)
        :"S" (pMem), "D" (uCount)
        :"ecx"
        );

    return iRet;
	}

#elif defined(BIT_MSC_X86)

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount)
	{
	_asm{
		mov		esi,pMem			;// ESI = Adresse
		xor		eax,eax				;// EAX = Position
		mov		edi,uCount			;// EDI = Anzahl
		jmp		start
	loop1:
		add		esi,4				;// Die nchsten 32-Bit
		add		eax,32				;// 		*
	start:
		cmp		eax,edi				;// 		*
		jnb     ende1				;// berlauf ?
		mov	    ecx,[esi]			;// Lade 32-Bit
		bswap   ecx					;// Drehe Byte Reihenfolge
		bsr	    ecx,ecx  			;// Suche Bit
		jz	    loop1				;// 		*

		xor     ecx,1Fh				;// Bitposition korrigiren
		add     eax,ecx				;// Ergebnis berechen
      	cmp		eax,edi				;// berlauf ?
		jbe     ende2				;// 		*
	ende1:
		mov		eax,edi				;// 		*
	ende2:
		}
	}

#else

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount)
	{
	return bit_find_1x(pMem,uCount,0);
	}

#endif

//*****************************************************************************
//*
//*		bit_find_0
//*
//*****************************************************************************
//	Sucht das erste gelschte Bit in einem Speicherbereich.
//	pMem		: Ist der Zeiger auf den Speicherbereich.
//	uCount		: Ist die Gre der Speicherbereiches in Bits.
//	Ergibt die Position an der das Bit gefunden wurde, bzw. die Gre des Bereiches.

#if defined(BIT_KEIL_THUMB)

__asm int bit_find_0(const void *pMem,unsigned uCount)
	{
		// R0 = pMem
		// R1 = uCount

		MOV		r2,r1						// r2(pos) = uCount
		ANDS 	r3,r0,#3					// r3      = pMem&3;
		BEQ		next_a

		PUSH	{r4}						// Register retten
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		LSL		r3,r3,#3					// r3 	  *= 8
		LSL		r4,r4,r3					// mask  <<= (pMem&3)*8
		ADD		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		BICS	r3,r4,r3					// DWORD  &= ~DWORD&mask
		POP		{r4}						// Register wieder herstellen
		BNE		find_a						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_a						// if(pos<=0)...

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
next_a
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		MVNS    r3,r3						// DWORD = ~DWORD
		BNE		find_a						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_a						// if(pos<=0)...
stop_a
		MOV		r0,r1						// return uCount;
		BLX		lr							//		*
find_a
		RBIT	r3,r3						// Bitreihenfolge umkehren
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_a						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		BLX		lr							//		*
	}

#elif defined(BIT_KEIL_ARM)

__asm int bit_find_0(const void *pMem,unsigned uCount)
	{
		// R0 = pMem
		// R1 = uCount

		MOV		r2,r1						// r2(pos) = uCount
		ANDS 	r3,r0,#3					// r3      = pMem&3;
		BEQ		next_a

		PUSH	{r4}						// Register retten
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		LSL		r3,r3,#3					// r3 	  *= 8
		LSL		r4,r4,r3					// mask  <<= (pMem&3)*8
		ADD		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		BICS	r3,r4,r3					// DWORD  &= ~DWORD&mask
		POP		{r4}						// Register wieder herstellen
		BNE		find_a						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_a						// if(pos<=0)...

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
next_a
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		MVNS    r3,r3						// DWORD = ~DWORD
		BNE		find_a						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_a						// if(pos<=0)...
stop_a
		MOV		r0,r1						// return uCount;
		BLX		lr							//		*
find_a
		RBIT	r3,r3						// Bitreihenfolge umkehren
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_a						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		BLX		lr							//		*
	}

#elif defined(BIT_IAR_ARM)

	// this functions are defined in "./Assembler/BitFunctionsIar[Thumb].s"

#elif !defined(_DEBUG)

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	int _cdecl bit_find_0(const void *pMem,unsigned uCount)
	{
	return bit_find_0(pMem,uCount,0);
	}

#elif defined(BIT_GNU_X86)

	int _cdecl bit_find_0(const void *pMem,unsigned uCount)
	{
	int iRet;

    asm volatile(
		"pushl  %%ebx   		\n"
        "xorl   %%eax,%%eax     \n"
        "jmp    2f              \n"
        "1:                     \n"
        "addl   $4   ,%%esi     \n"
        "addl   $32  ,%%eax     \n"
        "2:                     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movl   (%%esi),%%ebx   \n"
        "notl   %%ebx           \n"
        "bsfl   %%ebx,%%ecx     \n"
        "jz     1b              \n"
        "add    %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%ebx   		\n"
        :"=a" (iRet)
        :"S" (pMem), "D" (uCount)
        :"ecx"
        );

	return  iRet;
	}

#elif defined(BIT_GNU_X64)

	int _cdecl bit_find_0(const void *pMem,unsigned uCount)
	{
	int iRet;

    asm volatile(
		"pushl  %%rbx   		\n"
        "xorl   %%eax,%%eax     \n"
        "jmp    2f              \n"
        "1:                     \n"
        "addq   $8   ,%%rsi     \n"
        "addl   $64  ,%%eax     \n"
        "2:                     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movq   (%%rsi),%%rbx   \n"
        "notq   %%rbx           \n"
        "bsfq   %%rbx,%%rcx     \n"
        "jz     1b              \n"
        "add    %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%rbx	  		\n"
        :"=a" (iRet)
        :"S" (pMem), "D" (uCount)
        :"rcx","rsi"
        );

    return iRet;
	}

#elif defined(BIT_MSC_X86)

	int _cdecl bit_find_0(const void *pMem,unsigned uCount)
	{
	_asm{
		mov		esi,pMem			;// ESI = Adresse
		xor		eax,eax				;// EAX = Position
		mov		edi,uCount			;// EDI = Anzahl
		jmp		start
	loop1:
		add		esi,4				;// Die nchsten 32-Bit
		add		eax,32				;// 		*
	start:
		cmp		eax,edi				;// 		*
		jnb     ende1				;// berlauf ?

		mov     ebx,[esi]			;// 32-Bit laden
		not     ebx					;// Invertieren
		bsf	    ecx,ebx				;// Suche Bit
		jz	    loop1				;// 		*

		add     eax,ecx				;// Ergebnis berechen
      	cmp		eax,edi				;// berlauf ?
		jbe     ende2				;// 		*
	ende1:
		mov		eax,edi				;// 		*
	ende2:
		}
	}

#else

	int _cdecl bit_find_0(const void *pMem,unsigned uCount)
	{
	return bit_find_0(pMem,uCount,0);
	}

#endif

//*****************************************************************************
//*
//*		bit_find_0x
//*
//*****************************************************************************
//	Sucht das erste gelschte Bit in einem Speicherbereich.
//	pMem		: Ist der Zeiger auf den Speicherbereich.
//	uCount		: Ist die Gre der Speicherbereiches in Bits.
//	Ergibt die Position an der das Bit gefunden wurde, bzw. die Gre des Bereiches.

#if defined(BIT_KEIL_THUMB)

__asm int bit_find_0x(const void *pMem,unsigned uCount)
	{
		// R0 = pMem
		// R1 = uCount

		MOV		r2,r1						// r2(pos) = uCount
		ANDS 	r3,r0,#3					// r3      = pMem&3;
		BEQ		next_f

		PUSH	{r4}						// Register retten
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		LSL		r3,r3,#3					// r3 	  *= 8
		LSR		r4,r4,r3					// mask  <<= (pMem&3)*8
		ADD		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		BICS	r3,r4,r3					// DWORD  &= ~DWORD&mask
		POP		{r4}						// Register wieder herstellen
		BNE		find_f						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_f						// if(pos<=0)...

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
next_f
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		MVNS    r3,r3						// DWORD = ~DWORD
		BNE		find_f						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_f						// if(pos<=0)...
stop_f
		MOV		r0,r1						// return uCount;
		BLX		lr							//		*
find_f
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_f						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		BLX		lr							//		*
	}

#elif defined(BIT_KEIL_ARM)

__fsm int bit_find_0x(const void *pMem,unsigned uCount)
	{
		// R0 = pMem
		// R1 = uCount

		MOV		r2,r1						// r2(pos) = uCount
		ANDS 	r3,r0,#3					// r3      = pMem&3;
		BEQ		next_f

		PUSH	{r4}						// Register retten
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		LSL		r3,r3,#3					// r3 	  *= 8
		LSR		r4,r4,r3					// mask  <<= (pMem&3)*8
		ADD		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		BICS	r3,r4,r3					// DWORD  &= ~DWORD&mask
		POP		{r4}						// Register wieder herstellen
		BNE		find_f						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_f						// if(pos<=0)...

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
next_f
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		MVNS    r3,r3						// DWORD = ~DWORD
		BNE		find_f						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_f						// if(pos<=0)...
stop_f
		MOV		r0,r1						// return uCount;
		BLX		lr							//		*
find_f
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_f						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		BLX		lr							//		*
	}

#elif defined(BIT_IAR_ARM)

	// this functions are defined in "./Assembler/BitFunctionsIar[Thumb].s"

#elif !defined(_DEBUG)

	// function defined in inline in "BitFunctions.h"

#elif defined(NO_BIT_ASM)

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount)
	{
	return bit_find_0x(pMem,uCount,0);
	}

#elif defined(BIT_GNU_X86)

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount)
	{
	int iRet;

    asm volatile(
		"pushl  %%ebx   		\n"
        "xorl   %%eax,%%eax     \n"
        "jmp    2f              \n"
        "1:                     \n"
        "addl   $4   ,%%esi     \n"
        "addl   $32  ,%%eax     \n"
        "2:                     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movl   (%%esi),%%ebx   \n"
        "notl   %%ebx           \n"
		"bswap  %%ebx           \n"
		"bsrl   %%ebx,%%ecx     \n"
        "jz     1b              \n"
		"xorl   $31  ,%%ecx     \n"
        "add    %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%ebx   		\n"
        :"=a" (iRet)
        :"S" (pMem), "D" (uCount)
        :"ecx"
        );

	return  iRet;
	}

#elif defined(BIT_GNU_X64)

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount)
	{
	int iRet;

    asm volatile(
		"pushl  %%rbx   		\n"
        "xorl   %%eax,%%eax     \n"
        "jmp    2f              \n"
        "1:                     \n"
        "addq   $8   ,%%rsi     \n"
        "addl   $64  ,%%eax     \n"
        "2:                     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movq   (%%rsi),%%rbx   \n"
        "notq   %%rbx           \n"
		"bswap  %%rbx           \n"
        "bsrq   %%rbx,%%rcx     \n"
        "jz     1b              \n"
		"xorl   $63  ,%%ecx     \n"
        "add    %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%rbx   		\n"
        :"=a" (iRet)
        :"S" (pMem), "D" (uCount)
        :"rcx"
        );

    return iRet;
	}

#elif defined(BIT_MSC_X86)

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount)
	{
	_asm{
		mov		esi,pMem			;// ESI = Adresse
		xor		eax,eax				;// EAX = Position
		mov		edi,uCount			;// EDI = Anzahl
		jmp		start
	loop1:
		add		esi,4				;// Die nchsten 32-Bit
		add		eax,32				;// 		*
	start:
		cmp		eax,edi				;// 		*
		jnb     ende1				;// berlauf ?

		mov     ebx,[esi]			;// 32-Bit laden
		not     ebx					;// Invertieren
		bswap	ebx					;// Byte Reihenfolge drehen
		bsr	    ecx,ebx				;// Suche Bit
		jz	    loop1				;// 		*

		xor     ecx,1Fh				;// Bitposition korrigiren
		add     eax,ecx				;// Ergebnis berechen
      	cmp		eax,edi				;// berlauf ?
		jbe     ende2				;// 		*
	ende1:
		mov		eax,edi				;// 		*
	ende2:
		}
	}

#else

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount)
	{
	return bit_find_0x(pMem,uCount,0);
	}

#endif



//*****************************************************************************
//*
//*		bit_find_1
//*
//*****************************************************************************
//	Sucht das nchste gesetzte Bit in einem Speicherbereich ab einem Offset.
//	pMem		: Ist der Zeiger auf den Speicherbereich.
//	uCount		: Ist die Gre der Speicherbereiches in Bits.
//	uOffset		: Ist die Bitposition ab der gesucht wird
//	Ergibt die Position an der das Bit gefunden wurde, bzw. die Gre des Bereiches.

#if defined(BIT_KEIL_THUMB)

__asm int  bit_find_1(const void *pMem,unsigned uCount,unsigned uOffset)
	{
		// R0 = pMem
		// R1 = uCount
		// R2 = uOffset

		PUSH	{r4}						// Register retten
		CMP		r2,r1						// if(uOffset>=uCount)...
		BHS		stop_d						// 			*
		UBFX 	r3,r2,#0,#3					// r3      = BitOffset
		ADD		r0,r0,r2,LSR #3				// pMem   += uOffset/8
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		BFC		r2,#0,#3					// uOffset&= ~7
		LSL		r4,r4,r3					// r4(mask)= (1<<uOffset&7)-1

		UBFX 	r3,r0,#0,#2					// r3      =  pMem&3;
		LSL		r3,r3,#3					// r3 	  *=  8
		LSL		r4,r4,r3					// mask  <<= (pMem&3)*8
		SUB		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3
		SUB		r2,r1,r2					// r2(pos) = uCount-r2

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
		// R4 = Maske fr estes DWORD

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		AND		r3,r3,r4					// DWORD  &= mask
		CBNZ	r3,find_d					// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_d						// if(pos<=0)...
next_d
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		CBNZ	r3,find_d					// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_d						// if(pos<=0)...
stop_d
		MOV		r0,r1						// return uCount;
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
find_d
		RBIT	r3,r3						// Bitreihenfolge umkehren
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_d						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
	}

#elif defined(BIT_KEIL_ARM)

__asm int bit_find_1(const void *pMem,unsigned uCount,unsigned uOffset)
	{
		// R0 = pMem
		// R1 = uCount
		// R2 = uOffset

		PUSH	{r4}						// Register retten
		CMP		r2,r1						// if(uOffset>=uCount)...
		BHS		stop_d						// 			*
		UBFX 	r3,r2,#0,#3					// r3      = BitOffset
		ADD		r0,r0,r2,LSR #3				// pMem   += uOffset/8
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		BFC		r2,#0,#3					// uOffset&= ~7
		LSL		r4,r4,r3					// r4(mask)= (1<<uOffset&7)-1

		UBFX 	r3,r0,#0,#2					// r3      =  pMem&3;
		LSL		r3,r3,#3					// r3 	  *=  8
		LSL		r4,r4,r3					// mask  <<= (pMem&3)*8
		SUB		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3
		SUB		r2,r1,r2					// r2(pos) = uCount-r2

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
		// R4 = Maske fr estes DWORD

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		AND		r3,r3,r4					// DWORD  &= mask
		CBNZ	r3,find_d					// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_d						// if(pos<=0)...
next_d
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
	    CMP     r3,#0                       // if(DWORD==0)...
	    BNE     find_d						//      *
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_d						// if(pos<=0)...
stop_d
		MOV		r0,r1						// return uCount;
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
find_d
		RBIT	r3,r3						// Bitreihenfolge umkehren
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_d						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
	}

#elif defined(BIT_IAR_ARM)

	// this functions are defined in "./Assembler/BitFunctionsIar[Thumb].s"

#elif defined(BIT_GNU_X86) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int     iRet;

    asm volatile(
		"pushl  %%ebx           \n"
        "movl   %%eax,%%ebx     \n"
        "movl   %%eax,%%ecx     \n"
        "shrl   $5   ,%%ebx     \n"
        "andb   $31  ,%%cl      \n"
        "shll   $2   ,%%ebx     \n"
        "subb   %%cl ,%%al      \n"
        "addl   %1   ,%%ebx     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movl   (%%ebx),%%esi   \n"
        "shrl   %%cl ,%%esi     \n"
        "shll   %%cl ,%%esi     \n"
        "bsfl   %%esi,%%ecx     \n"
        "jnz    2f              \n"
        "1:                     \n"
        "addl   $4   ,%%ebx     \n"
        "addl   $32  ,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "bsfl   (%%ebx),%%ecx   \n"
        "jz     1b              \n"
        "2:                     \n"
        "addl   %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%ebx           \n"
        :"=a" (iRet)
        :"m" (pMem), "D" (uCount), "a" (uOffset)
        :"ecx","esi", "cc"
        );

    return iRet;
	}

#elif defined(BIT_GNU_X64) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int     iRet;

    asm volatile(
		"pushl  %%rbx           \n"
        "movl   %%eax,%%ebx     \n"
        "movl   %%eax,%%ecx     \n"
        "shrl   $6   ,%%ebx     \n"
        "andb   $63  ,%%cl      \n"
        "shll   $3   ,%%ebx     \n"
        "subb   %%cl ,%%al      \n"
        "andq $0xFFFFFFFF,%%rbx \n"
        "addq   %1   ,%%rbx     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movq   (%%rbx),%%rsi   \n"
        "shrq   %%cl ,%%rsi     \n"
        "shlq   %%cl ,%%rsi     \n"
        "bsfq   %%rsi,%%rcx     \n"
        "jnz    2f              \n"
        "1:                     \n"
        "addq   $8   ,%%rbx     \n"
        "addl   $64  ,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "bsfq   (%%rbx),%%rcx   \n"
        "jz     1b              \n"
        "2:                     \n"
        "addl   %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%rbx           \n"
        :"=a" (iRet)
        :"m" (pMem), "D" (uCount), "a" (uOffset)
        :"rcx","rsi", "cc"
        );

	return iRet;
	}

#elif defined(BIT_MSC_X64) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned __int64 *pMem64;
	unsigned __int64  uValue;
	unsigned long  	  uPos;
	unsigned		  uMax;
	unsigned		  uSub;

	if(uOffset>=uCount)return uCount;

	uSub	 = ((unsigned)pMem)&7;
	pMem64	 = (unsigned __int64*)(((char*)pMem)-uSub);
	uSub	*= 8;
	uCount	+= uSub;
	uOffset	+= uSub;
	uValue	 = pMem64[uOffset/64];
	uMax	 = (uCount-1)&~0x3F;

	if(uOffset&0x3F)
		{
		uValue  &= (__int64)0xFFFFFFFFFFFFFFFF<<(uOffset&0x3F);
		uOffset	&= ~0x3F;
		}

	while(uOffset<uMax)
		{
		if(_BitScanForward64(&uPos,uValue))
			{
			return uOffset+uPos-uSub;
			}

		uOffset += 64;
		uValue   = pMem64[uOffset/64];
		}

	if(uCount&0x3F)
		{
		uValue |= (__int64)0xFFFFFFFFFFFFFFFF<<(uCount&0x3F);
		}

	if(_BitScanForward64(&uPos,uValue))
		{
		return uOffset+uPos-uSub;
		}

	return uCount-uSub;
	}

#elif defined(BIT_MSC_INSTR) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned		 *pMem32;
	unsigned		  uValue;
	unsigned long  	  uPos;
	unsigned		  uMax;
	unsigned		  uSub;

	if(uOffset>=uCount)return uCount;

	uSub	 = ((unsigned)pMem)&3;
	pMem32	 = (unsigned*)(((char*)pMem)-uSub);
	uSub	*= 8;
	uCount	+= uSub;
	uOffset	+= uSub;
	uValue	 = pMem32[uOffset/32];
	uMax	 = (uCount-1)&~0x1F;

	if(uOffset&0x1F)
		{
		uValue  &= 0xFFFFFFFF<<(uOffset&0x1F);
		uOffset	&= ~0x1F;
		}

	while(uOffset<uMax)
		{
		if(_BitScanForward(&uPos,uValue))
			{
			return uOffset+uPos-uSub;
			}

		uOffset += 32;
		uValue   = pMem32[uOffset/32];
		}

	if(uCount&0x1F)
		{
		uValue |= 0xFFFFFFFF<<(uCount&0x1F);
		}

	if(_BitScanForward(&uPos,uValue))
		{
		return uOffset+uPos-uSub;
		}

	return uCount-uSub;
	}

#elif defined(BIT_MSC_VC60) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	_asm{
		mov		eax,uOffset			;// Offset laden
		mov     ebx,eax				;// 		*
		mov     ecx,eax				;// 		*

		shr     ebx,5				;// EBX = DWORD-Offset vom Start-Bit
		and		cl ,1Fh				;// ESI = Start-Bit auf 32 gerundet
		mov		edi,uCount			;// EDI = Bit-Anzahl
		shl     ebx,2				;// EBX = Byte-Offset vom Start-Bit
		sub     al ,cl				;// EAX = Bit-Position auf 32 gerundet
		add		ebx,pMem			;// EBX = Byte-Adresse

		cmp		eax,edi				;// berlauf ?
		jnb     ende1				;// 		*
		mov     esi,[ebx]			;// Die ersten 32-Bit laden
 		shr		esi,cl				;// 		*
		shl		esi,cl				;// 		*
		bsf	    ecx,esi				;// Suche Bit
		jnz		start				;// 		*
	loop1:
		add		ebx,4				;// Die nchsten 32-Bit
		add		eax,32				;// 	*
		cmp		eax,edi				;// 	*
		jnb     ende1				;// berlauf ?
		bsf	    ecx,[ebx]			;// Suche Bit
		jz	    loop1
	start:
		add     eax,ecx				;// Ergebnis berechen
      	cmp		eax,edi				;// 	*
		jbe     ende2				;// Gab es einen berlauf
	ende1:
		mov		eax,edi				;// ;berlauf
	ende2:
		}
	}

#elif defined(BIT_GNU_THUMB) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int	 iRet;

    asm volatile(
		"CMP	%3,%2			\n"
		"BHS	2f				\n"
		"UBFX 	r3,%3,#0,#3		\n"
		"ADD	%1,%1,%3,LSR #3	\n"
		"MOV	r4,#-1			\n"
		"BFC	%3,#0,#3		\n"
		"LSL	r4,r4,r3		\n"
		"						\n"
		"UBFX 	r3,%1,#0,#2		\n"
		"LSL	r3,r3,#3		\n"
		"LSL	r4,r4,r3		\n"
		"SUB	%3,%3,r3		\n"
		"BFC	%1,#0,#2		\n"
		"SUB	%3,%2,%3		\n"
		"						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"ANDS	r3,r3,r4		\n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BLE	2f				\n"
		"1:						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"CBNZ	r3,3f           \n"
		"SUBS	%3,%3,#32		\n"
		"BGT	1b				\n"
		"2:						\n"
		"MOV	%0,%2			\n"
		"B		4f				\n"
		"3:						\n"
		"RBIT	r3,r3			\n"
		"CLZ	r3,r3			\n"
		"SUBS	%3,%3,r3        \n"
		"BLT    2b				\n"
		"SUB	%0,%2,%3		\n"
		"4:						\n"
		:"=r" (iRet)
		:"r"  (pMem),
		 "r"  (uCount),
		 "r"  (uOffset)
		:"r3","r4","cc" );

	return iRet;
	}

#elif defined(BIT_GNU_ARM) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int	 iRet;

    asm volatile(
		"CMP	%3,%2			\n"
		"BHS	2f				\n"
		"UBFX 	r3,%3,#0,#3		\n"
		"ADD	%1,%1,%3,LSR #3	\n"
		"MOV	r4,#-1			\n"
		"BFC	%3,#0,#3		\n"
		"LSL	r4,r4,r3		\n"
		"						\n"
		"UBFX 	r3,%1,#0,#2		\n"
		"LSL	r3,r3,#3		\n"
		"LSL	r4,r4,r3		\n"
		"SUB	%3,%3,r3		\n"
		"BFC	%1,#0,#2		\n"
		"SUB	%3,%2,%3		\n"
		"						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"ANDS	r3,r3,r4		\n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BLE	2f				\n"
		"1:						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"CMP    r3,#0	        \n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BGT	1b				\n"
		"2:						\n"
		"MOV	%0,%2			\n"
		"B		4f				\n"
		"3:						\n"
		"RBIT	r3,r3			\n"
		"CLZ	r3,r3			\n"
		"SUBS	%3,%3,r3        \n"
		"BLT    2b				\n"
		"SUB	%0,%2,%3		\n"
		"4:						\n"
		:"=r" (iRet)
		:"r"  (pMem),
		 "r"  (uCount),
		 "r"  (uOffset)
		:"r3","r4","cc" );

	return iRet;
	}

#else

	int _cdecl bit_find_1(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned   *pData;
	unsigned	uBits;
	unsigned	uMask;
	unsigned	uEnd;
	unsigned 	uPos;
	unsigned 	uSub;

	if(uOffset>=uCount)return uCount;

	uSub	= (unsigned)pMem&3;
	pData	= (unsigned*)((char*)pMem-uSub);
	uPos	= uSub*8;
	uEnd	= uPos+uCount;
	uPos   += uOffset;
	uMask	= 0xFFFFFFFF<<(uEnd&0x1F);
	uBits	= 0xFFFFFFFF<<(uPos&0x1F);
	uOffset = (uPos&~0x1F)-uSub*8;
	uEnd	= (uEnd-1)/32;
	uPos   /= 32;
	uBits  &= pData[uPos];


	for(;;)									// Durchsuche 32 Bitpakete
		{
		if(uPos>=uEnd)
			{
			if(uMask==0xFFFFFFFF)break;
			uBits |= uMask;
			break;
			}

		if(uBits)break;

		uPos	+= 1;
		uOffset += 32;
		uBits	 = pData[uPos];
		}

	if(uBits&0xFF)return uOffset+ucBitOffset[uBits&0xFF];
	   uBits>>=8;
	if(uBits&0xFF)return uOffset+ucBitOffset[uBits&0xFF]+8;
	   uBits>>=8;
	if(uBits&0xFF)return uOffset+ucBitOffset[uBits&0xFF]+16;
	   uBits>>=8;
				  return uOffset+ucBitOffset[uBits&0xFF]+24;
	}

#endif

//*****************************************************************************
//*
//*		bit_find_1x
//*
//*****************************************************************************
//	Sucht das nchste gesetzte Bit in einem Speicherbereich ab einem Offset.
//	Die Bits werden von 0x80 in Richtung 0x01 gesucht.
//	pMem		: Ist der Zeiger auf den Speicherbereich.
//	uCount		: Ist die Gre der Speicherbereiches in Bits.
//	uOffset		: Ist die Bitposition ab der gesucht wird
//	Ergibt die Position an der das Bit gefunden wurde, bzw. die Gre des Bereiches.

#if defined(BIT_KEIL_THUMB)

__asm int  bit_find_1x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
		// R0 = pMem
		// R1 = uCount
		// R2 = uOffset

		PUSH	{r4}						// Register retten
		CMP		r2,r1						// if(uOffset>=uCount)...
		BHS		stop_g						// 			*
		UBFX 	r3,r2,#0,#3					// r3      = BitOffset
		ADD		r0,r0,r2,LSR #3				// pMem   += uOffset/8
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		BFC		r2,#0,#3					// uOffset&= ~7
		LSR		r4,r4,r3					// r4(mask)= 0xFFFFFFFF>>(uOffset&7)

		UBFX 	r3,r0,#0,#2					// r3      =  pMem&3;
		LSL		r3,r3,#3					// r3 	  *=  8
		LSR		r4,r4,r3					// mask  >>= (pMem&3)*8
		SUB		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3
		SUB		r2,r1,r2					// r2(pos) = uCount-r2

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
		// R4 = Maske fr estes DWORD

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		AND		r3,r3,r4					// DWORD  &= mask
		CBNZ	r3,find_g					// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_g						// if(pos<=0)...
next_g
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		CBNZ	r3,find_g					// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_g						// if(pos<=0)...
stop_g
		MOV		r0,r1						// return uCount;
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
find_g
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_g						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
	}

#elif defined(BIT_KEIL_ARM)

__asm int bit_find_1x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
		// R0 = pMem
		// R1 = uCount
		// R2 = uOffset

		PUSH	{r4}						// Register retten
		CMP		r2,r1						// if(uOffset>=uCount)...
		BHS		stop_g						// 			*
		UBFX 	r3,r2,#0,#3					// r3      = BitOffset
		ADD		r0,r0,r2,LSR #3				// pMem   += uOffset/8
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		BFC		r2,#0,#3					// uOffset&= ~7
		LSR		r4,r4,r3					// r4(mask)= 0xFFFFFFFF>>(uOffset&7)

		UBFX 	r3,r0,#0,#2					// r3      =  pMem&3;
		LSL		r3,r3,#3					// r3 	  *=  8
		LSR		r4,r4,r3					// mask  >>= (pMem&3)*8
		SUB		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3
		SUB		r2,r1,r2					// r2(pos) = uCount-r2

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
		// R4 = Maske fr estes DWORD

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		AND		r3,r3,r4					// DWORD  &= mask
		CBNZ	r3,find_g					// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_g						// if(pos<=0)...
next_g
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
	    CMP     r3,#0                       // if(DWORD==0)...
	    BNE     find_g						//      *
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_g						// if(pos<=0)...
stop_g
		MOV		r0,r1						// return uCount;
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
find_g
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_g						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
	}

#elif defined(BIT_IAR_ARM)

	// this functions are defined in "./Assembler/BitFunctionsIar[Thumb].s"

#elif defined(BIT_GNU_X86) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int     iRet;

    asm volatile(
		"pushl  %%ebx		    \n"
        "movl   %%eax,%%ebx     \n"
        "movl   %%eax,%%ecx     \n"
        "shrl   $5   ,%%ebx     \n"
        "andb   $31  ,%%cl      \n"
        "shll   $2   ,%%ebx     \n"
        "subb   %%cl ,%%al      \n"
        "addl   %1   ,%%ebx     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movl   (%%ebx),%%esi   \n"
		"bswap  %%esi           \n"
        "shll   %%cl ,%%esi     \n"
        "shrl   %%cl ,%%esi     \n"
        "bsrl   %%esi,%%ecx     \n"
        "jnz    2f              \n"
        "1:                     \n"
        "addl   $4   ,%%ebx     \n"
        "addl   $32  ,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
		"movl   (%%ebx),%%ecx   \n"
		"bswap  %%ecx           \n"
        "bsrl   %%ecx,%%ecx     \n"
        "jz     1b              \n"
        "2:                     \n"
		"xorl   $31  ,%%ecx     \n"
        "addl   %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%ebx		    \n"
        :"=a" (iRet)
        :"m" (pMem), "D" (uCount), "a" (uOffset)
        :"ecx","esi", "cc"
        );

    return iRet;
	}

#elif defined(BIT_GNU_X64) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int     iRet;

    asm volatile(
		"pushl  %%rbx		    \n"
        "movl   %%abx,%%ebx     \n"
        "movl   %%abx,%%ecx     \n"
        "shrl   $6   ,%%ebx     \n"
        "andb   $63  ,%%cl      \n"
        "shll   $3   ,%%ebx     \n"
        "subb   %%cl ,%%al      \n"
        "andq $0xFFFFFFFF,%%rbx \n"
        "addq   %1   ,%%rbx     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movq   (%%rbx),%%rsi   \n"
		"bswap  %%rsi           \n"
        "shlq   %%cl ,%%rsi     \n"
        "shrq   %%cl ,%%rsi     \n"
		"bsrq   %%rsi,%%rcx     \n"
        "jnz    2f              \n"
        "1:                     \n"
        "addq   $8   ,%%rbx     \n"
        "addl   $64  ,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
		"movq   (%%rbx),%%rcx   \n"
		"bswap  %%rcx           \n"
        "bsrq   %%rcx,%%rcx     \n"
        "jz     1b              \n"
        "2:                     \n"
		"xorl   $63  ,%%ecx     \n"
        "addl   %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%rbx		    \n"
        :"=a" (iRet)
        :"m" (pMem), "D" (uCount), "a" (uOffset)
        :"rcx","rsi", "cc"
        );

	return iRet;
	}

#elif defined(BIT_MSC_X64) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned __int64 *pMem64;
	unsigned __int64  uValue;
	unsigned long  	  uPos;
	unsigned		  uMax;
	unsigned		  uSub;

	if(uOffset>=uCount)return uCount;

	uSub	 = ((unsigned)pMem)&7;
	pMem64	 = (unsigned __int64*)(((char*)pMem)-uSub);
	uSub	*= 8;
	uCount	+= uSub;
	uOffset	+= uSub;
	uValue	 = pMem64[uOffset/64];
	uMax	 = (uCount-1)&~0x3F;
	uValue   = _byteswap_uint64(uValue);


	if(uOffset&0x3F)
		{
		uValue  &= (__int64)0xFFFFFFFFFFFFFFFF>>(uOffset&0x3F);
		uOffset	&= ~0x3F;
		}

	while(uOffset<uMax)
		{
		if(_BitScanReverse64(&uPos,uValue))
			{
			uPos ^= 0x3F;
			return uOffset+uPos-uSub;
			}

		uOffset += 64;
		uValue   = pMem64[uOffset/64];
		uValue   = _byteswap_uint64(uValue);
		}

	if(uCount&0x3F)
		{
		uValue |= (__int64)0xFFFFFFFFFFFFFFFF>>(uCount&0x3F);
		}

	if(_BitScanReverse64(&uPos,uValue))
		{
		uPos ^= 0x3F;
		return uOffset+uPos-uSub;
		}

	return uCount-uSub;
	}

#elif defined(BIT_MSC_INSTR) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned		 *pMem32;
	unsigned		  uValue;
	unsigned long  	  uPos;
	unsigned		  uMax;
	unsigned		  uSub;

	if(uOffset>=uCount)return uCount;

	uSub	 = ((unsigned)pMem)&3;
	pMem32	 = (unsigned*)(((char*)pMem)-uSub);
	uSub	*= 8;
	uCount	+= uSub;
	uOffset	+= uSub;
	uValue	 = pMem32[uOffset/32];
	uMax	 = (uCount-1)&~0x1F;
	uValue   = _byteswap_ulong(uValue);

	if(uOffset&0x1F)
		{
		uValue  &= 0xFFFFFFFF>>(uOffset&0x1F);
		uOffset	&= ~0x1F;
		}

	while(uOffset<uMax)
		{
		if(_BitScanReverse(&uPos,uValue))
			{
			uPos ^= 0x1F;
			return uOffset+uPos-uSub;
			}

		uOffset += 32;
		uValue   = pMem32[uOffset/32];
		uValue   = _byteswap_ulong(uValue);
		}

	if(uCount&0x1F)
		{
		uValue |= 0xFFFFFFFF>>(uCount&0x1F);
		}

	if(_BitScanReverse(&uPos,uValue))
		{
		uPos ^= 0x1F;
		return uOffset+uPos-uSub;
		}

	return uCount-uSub;
	}

#elif defined(BIT_MSC_VC60) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	_asm{
		mov		eax,uOffset			;// Offset laden
		mov     ebx,eax				;// 		*
		mov     ecx,eax				;// 		*

		shr     ebx,5				;// EBX = DWORD-Offset vom Start-Bit
		and		cl ,1Fh				;// ESI = Start-Bit auf 32 gerundet
		mov		edi,uCount			;// EDI = Bit-Anzahl
		shl     ebx,2				;// EBX = Byte-Offset vom Start-Bit
		sub     al ,cl				;// EAX = Bit-Position auf 32 gerundet
		add		ebx,pMem			;// EBX = Byte-Adresse

		cmp		eax,edi				;// berlauf ?
		jnb     ende1				;// 		*
		mov     esi,[ebx]			;// Die ersten 32-Bit laden
		bswap	esi					;// Byte Reihenfolge drehen
 		shl		esi,cl				;// 		*
		shr		esi,cl				;// 		*
		bsr	    ecx,esi				;// Suche Bit
		jnz		done				;// 		*
	loop1:
		add		ebx,4				;// Die nchsten 32-Bit
		add		eax,32				;// 	*
		cmp		eax,edi				;// 	*
		jnb     ende1				;// berlauf ?
		mov	    ecx,[ebx]			;// Lade 32-Bit
		bswap   ecx					;// Drehe Byte Reihenfolge
		bsr	    ecx,ecx  			;// Suche Bit
		jz	    loop1
	done:
		xor     ecx,1Fh				;// Bitposition korrigiren
		add     eax,ecx				;// Ergebnis berechen
      	cmp		eax,edi				;// 	*
		jbe     ende2				;// Gab es einen berlauf
	ende1:
		mov		eax,edi				;// ;berlauf
	ende2:
		}
	}

#elif defined(BIT_GNU_THUMB) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int	 iRet;

    asm volatile(
		"CMP	%3,%2			\n"
		"BHS	2f				\n"
		"UBFX 	r3,%3,#0,#3		\n"
		"ADD	%1,%1,%3,LSR #3	\n"
		"MOV	r4,#-1			\n"
		"BFC	%3,#0,#3		\n"
		"LSR	r4,r4,r3		\n"
		"						\n"
		"UBFX 	r3,%1,#0,#2		\n"
		"LSL	r3,r3,#3		\n"
		"LSR	r4,r4,r3		\n"
		"SUB	%3,%3,r3		\n"
		"BFC	%1,#0,#2		\n"
		"SUB	%3,%2,%3		\n"
		"						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ != __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"ANDS	r3,r3,r4		\n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BLE	2f				\n"
		"1:						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ != __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"CBNZ	r3,3f           \n"
		"SUBS	%3,%3,#32		\n"
		"BGT	1b				\n"
		"2:						\n"
		"MOV	%0,%2			\n"
		"B		4f				\n"
		"3:						\n"
		"CLZ	r3,r3			\n"
		"SUBS	%3,%3,r3        \n"
		"BLT    2b				\n"
		"SUB	%0,%2,%3		\n"
		"4:						\n"
		:"=r" (iRet)
		:"r"  (pMem),
		 "r"  (uCount),
		 "r"  (uOffset)
		:"r3","r4","cc" );

	return iRet;
	}

#elif defined(BIT_GNU_ARM) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int	 iRet;

    asm volatile(
		"CMP	%3,%2			\n"
		"BHS	2f				\n"
		"UBFX 	r3,%3,#0,#3		\n"
		"ADD	%1,%1,%3,LSR #3	\n"
		"MOV	r4,#-1			\n"
		"BFC	%3,#0,#3		\n"
		"LSR	r4,r4,r3		\n"
		"						\n"
		"UBFX 	r3,%1,#0,#2		\n"
		"LSL	r3,r3,#3		\n"
		"LSR	r4,r4,r3		\n"
		"SUB	%3,%3,r3		\n"
		"BFC	%1,#0,#2		\n"
		"SUB	%3,%2,%3		\n"
		"						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ != __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"ANDS	r3,r3,r4		\n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BLE	2f				\n"
		"1:						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ != __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"CMP    r3,#0	        \n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BGT	1b				\n"
		"2:						\n"
		"MOV	%0,%2			\n"
		"B		4f				\n"
		"3:						\n"
		"CLZ	r3,r3			\n"
		"SUBS	%3,%3,r3        \n"
		"BLT    2b				\n"
		"SUB	%0,%2,%3		\n"
		"4:						\n"
		:"=r" (iRet)
		:"r"  (pMem),
		 "r"  (uCount),
		 "r"  (uOffset)
		:"r3","r4","cc" );

	return iRet;
	}

#else

	int _cdecl bit_find_1x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned   *pData;
	unsigned	uBits;
	unsigned	uMask;
	unsigned	uEnd;
	unsigned 	uPos;
	unsigned 	uSub;

	if(uOffset>=uCount)return uCount;

	uSub	= (unsigned)pMem&3;
	pData	= (unsigned*)((char*)pMem-uSub);
	uPos	= uSub*8;
	uEnd	= uPos+uCount;
	uPos   += uOffset;
	uMask	= 0xFFFFFFFF>>(uEnd&0x1F);
	uBits	= 0xFFFFFFFF>>(uPos&0x1F);
	uOffset = (uPos&~0x1F)-uSub*8;
	uEnd	= (uEnd-1)/32;
	uPos   /= 32;
	uBits  &= btol_dw(pData[uPos]);


	for(;;)									// Durchsuche 32 Bitpakete
		{
		if(uPos>=uEnd)
			{
			if(uMask==0xFFFFFFFF)break;
			uBits |= uMask;
			break;
			}

		if(uBits)break;

		uPos	+= 1;
		uOffset	+= 32;
		uBits	 = btol_dw(pData[uPos]);
		}

	if(uBits&0xFF000000)return uOffset+ucBitOffsetX[(uBits>>24)&0xFF];
	if(uBits&0x00FF0000)return uOffset+ucBitOffsetX[(uBits>>16)&0xFF]+8;
	if(uBits&0x0000FF00)return uOffset+ucBitOffsetX[(uBits>> 8)&0xFF]+16;
				        return uOffset+ucBitOffsetX[(uBits    )&0xFF]+24;
	}

#endif

//*****************************************************************************
//*
//*		bit_find_0
//*
//*****************************************************************************
//	Sucht das nchste gelschte Bit in einem Speicherbereich ab einem Offset.
//	pMem		: Ist der Zeiger auf den Speicherbereich.
//	uCount		: Ist die Gre der Speicherbereiches in Bits.
//	uOffset		: Ist die Bitposition ab der gesucht wird
//	Ergibt die Position an der das Bit gefunden wurde, bzw. die Gre des Bereiches.

#if defined(BIT_KEIL_THUMB)

__asm int bit_find_0(const void *pMem,unsigned uCount,unsigned uOffset)
	{
		// R0 = pMem
		// R1 = uCount
		// R2 = uOffset

		PUSH	{r4}						// Register retten
		CMP		r2,r1						// if(uOffset>=uCount)...
		BHS		stop_b						// 			*
		UBFX 	r3,r2,#0,#3					// r3      = BitOffset
		ADD		r0,r0,r2,LSR #3				// pMem   += uOffset/8
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		BFC		r2,#0,#3					// uOffset&= ~7
		LSL		r4,r4,r3					// r4(mask)= (1<<uOffset&7)-1

		UBFX 	r3,r0,#0,#2					// r3      =  pMem&3;
		LSL		r3,r3,#3					// r3 	  *=  8
		LSL		r4,r4,r3					// mask  <<= (pMem&3)*8
		SUB		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3
		SUB		r2,r1,r2					// r2(pos) = uCount-r2

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
		// R4 = Maske fr estes DWORD

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		BICS	r3,r4,r3					// DWORD  &= ~DWORD&mask
		BNE		find_b						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_b						// if(pos<=0)...
next_b
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		MVNS    r3,r3						// DWORD = ~DWORD
		BNE		find_b						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_b						// if(pos<=0)...
stop_b
		MOV		r0,r1						// return uCount;
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
find_b
		RBIT	r3,r3						// Bitreihenfolge umkehren
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_b						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
	}

#elif defined(BIT_KEIL_ARM)

__asm int bit_find_0(const void *pMem,unsigned uCount,unsigned uOffset)
	{
		// R0 = pMem
		// R1 = uCount
		// R2 = uOffset

		PUSH	{r4}						// Register retten
		CMP		r2,r1						// if(uOffset>=uCount)...
		BHS		stop_b						// 			*
		UBFX 	r3,r2,#0,#3					// r3      = BitOffset
		ADD		r0,r0,r2,LSR #3				// pMem   += uOffset/8
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		BFC		r2,#0,#3					// uOffset&= ~7
		LSL		r4,r4,r3					// r4(mask)= (1<<uOffset&7)-1

		UBFX 	r3,r0,#0,#2					// r3      =  pMem&3;
		LSL		r3,r3,#3					// r3 	  *=  8
		LSL		r4,r4,r3					// mask  <<= (pMem&3)*8
		SUB		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3
		SUB		r2,r1,r2					// r2(pos) = uCount-r2

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
		// R4 = Maske fr estes DWORD

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		BICS	r3,r4,r3					// DWORD  &= ~DWORD&mask
		BNE		find_b						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_b						// if(pos<=0)...
next_b
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifdef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		MVNS    r3,r3                       // DWORD = ~DWORD
		BNE		find_b						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_b						// if(pos<=0)...
stop_b
		MOV		r0,r1						// return uCount;
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
find_b
		RBIT	r3,r3						// Bitreihenfolge umkehren
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_b						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
	}

#elif defined(BIT_IAR_ARM)

	// this functions are defined in "./Assembler/BitFunctionsIar[Thumb].s"

#elif defined(BIT_GNU_X86) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int     iRet;

    asm volatile(
		"pushl  %%ebx		    \n"
        "movl   %%eax,%%ebx     \n"
        "movl   %%eax,%%ecx     \n"
        "shrl   $5   ,%%ebx     \n"
        "andb   $31  ,%%cl      \n"
        "shll   $2   ,%%ebx     \n"
        "subb   %%cl ,%%al      \n"
        "addl   %1   ,%%ebx     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movl   (%%ebx),%%esi   \n"
        "notl   %%esi           \n"
        "shrl   %%cl ,%%esi     \n"
        "shll   %%cl ,%%esi     \n"
        "bsfl   %%esi,%%ecx     \n"
        "jnz    2f              \n"
        "1:                     \n"
        "addl   $4   ,%%ebx     \n"
        "addl   $32  ,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movl   (%%ebx),%%esi   \n"
        "notl   %%esi           \n"
        "bsfl   %%esi,%%ecx     \n"
        "jz     1b              \n"
        "2:                     \n"
        "addl   %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%ebx		    \n"
        :"=a" (iRet)
        :"m" (pMem), "D" (uCount), "a" (uOffset)
        :"ecx","esi", "cc"
        );

    return iRet;
	}

#elif defined(BIT_GNU_X64) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int     iRet;

    asm volatile(
		"pushl  %%rbx		    \n"
        "movl   %%eax,%%ebx     \n"
        "movl   %%eax,%%ecx     \n"
        "shrl   $6   ,%%ebx     \n"
        "andb   $63  ,%%cl      \n"
        "shll   $3   ,%%ebx     \n"
        "subb   %%cl ,%%al      \n"
        "andq $0xFFFFFFFF,%%rbx \n"
        "addq   %1   ,%%rbx     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movq   (%%rbx),%%rsi   \n"
        "notq   %%rsi           \n"
        "shrq   %%cl ,%%rsi     \n"
        "shlq   %%cl ,%%rsi     \n"
        "bsfq   %%rsi,%%ecx     \n"
        "jnz    2f              \n"
        "1:                     \n"
        "addq   $8   ,%%rbx     \n"
        "addl   $64  ,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movq   (%%rbx),%%rsi   \n"
        "notq   %%rsi           \n"
        "bsfq   %%rsi,%%rcx     \n"
        "jz     1b              \n"
        "2:                     \n"
        "addl   %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%rbx		    \n"
        :"=a" (iRet)
        :"m" (pMem), "D" (uCount), "a" (uOffset)
        :"rcx","rsi", "cc"
        );

  	return iRet;
	}

#elif defined(BIT_MSC_X64) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned __int64 *pMem64;
	unsigned __int64  uValue;
	unsigned long  	  uPos;
	unsigned		  uMax;
	unsigned		  uSub;

	if(uOffset>=uCount)return uCount;

	uSub	 = ((unsigned)pMem)&7;
	pMem64	 = (unsigned __int64*)(((char*)pMem)-uSub);
	uSub	*= 8;
	uCount	+= uSub;
	uOffset	+= uSub;
	uValue	 = ~pMem64[uOffset/64];
	uMax	 = (uCount-1)&~0x3F;

	if(uOffset&0x3F)
		{
		uValue  &= (__int64)0xFFFFFFFFFFFFFFFF<<(uOffset&0x3F);
		uOffset	&= ~0x3F;
		}

	while(uOffset<uMax)
		{
		if(_BitScanForward64(&uPos,uValue))
			{
			return uOffset+uPos-uSub;
			}

		uOffset += 64;
		uValue   = ~pMem64[uOffset/64];
		}

	if(uCount&0x3F)
		{
		uValue |= (__int64)0xFFFFFFFFFFFFFFFF<<(uCount&0x3F);
		}

	if(_BitScanForward64(&uPos,uValue))
		{
		return uOffset+uPos-uSub;
		}

	return uCount-uSub;
	}

#elif defined(BIT_MSC_INSTR) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned		 *pMem32;
	unsigned		  uValue;
	unsigned long  	  uPos;
	unsigned		  uMax;
	unsigned		  uSub;

	if(uOffset>=uCount)return uCount;

	uSub	 = ((unsigned)pMem)&3;
	pMem32	 = (unsigned*)(((char*)pMem)-uSub);
	uSub	*= 8;
	uCount	+= uSub;
	uOffset	+= uSub;
	uValue	 = ~pMem32[uOffset/32];
	uMax	 = (uCount-1)&~0x1F;

	if(uOffset&0x1F)
		{
		uValue  &= 0xFFFFFFFF<<(uOffset&0x1F);
		uOffset	&= ~0x1F;
		}

	while(uOffset<uMax)
		{
		if(_BitScanForward(&uPos,uValue))
			{
			return uOffset+uPos-uSub;
			}

		uOffset += 32;
		uValue   = ~pMem32[uOffset/32];
		}

	if(uCount&0x1F)
		{
		uValue |= 0xFFFFFFFF<<(uCount&0x1F);
		}

	if(_BitScanForward(&uPos,uValue))
		{
		return uOffset+uPos-uSub;
		}

	return uCount-uSub;
	}

#elif defined(BIT_MSC_VC60) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	_asm{
		mov		eax,uOffset			;// Offset laden
		mov     ebx,eax				;// 		*
		mov     ecx,eax				;// 		*

		shr     ebx,5				;// EBX = Byte-Offset vom Start-Bit
		and		cl ,1Fh				;// ESI = Start-Bit    auf 32 gerundet
		mov		edi,uCount			;// EDI = Bit-Anzahl
		shl     ebx,2				;// EBX = Byte-Offset vom Start-Bit
		sub     al ,cl				;// EAX = Bit-Position auf 32 gerundet
		add		ebx,pMem			;// EBX = Byte-Adresse

		cmp		eax,edi				;// berlauf ?
		jnb     ende1				;// 		*
		mov     esi,[ebx]			;// Die ersten 32-Bit laden
		not		esi					;// 		*
 		shr		esi,cl				;// 		*
		shl		esi,cl				;// 		*
		bsf	    ecx,esi				;// Suche Bit
		jnz		start				;// 		*
	loop1:
		add		ebx,4				;// Die nchsten 32-Bit
		add		eax,32				;// 	*
		cmp		eax,edi				;// 	*
		jnb     ende1				;// berlauf ?
		mov	    esi,[ebx]			;// Suche Bit
		not		esi					;// 		*
		bsf	    ecx,esi				;// Suche Bit
		jz	    loop1
	start:
		add     eax,ecx				;// Ergebnis berechen
      	cmp		eax,edi				;// 	*
		jbe     ende2				;// Gab es einen berlauf

	ende1:
		mov		eax,edi				;// berlauf
	ende2:
		}
	}

#elif defined(BIT_GNU_THUMB) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int	 iRet;

    asm volatile(
		"CMP	%3,%2			\n"
		"BHS	2f				\n"
		"UBFX 	r3,%3,#0,#3		\n"
		"ADD	%1,%1,%3,LSR #3	\n"
		"MOV	r4,#-1			\n"
		"BFC	%3,#0,#3		\n"
		"LSL	r4,r4,r3		\n"
		"						\n"
		"UBFX 	r3,%1,#0,#2		\n"
		"LSL	r3,r3,#3		\n"
		"LSL	r4,r4,r3		\n"
		"SUB	%3,%3,r3		\n"
		"BFC	%1,#0,#2		\n"
		"SUB	%3,%2,%3		\n"
		"						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"BICS	r3,r4,r3		\n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BLE	2f				\n"
		"1:						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"MVNS   r3,r3           \n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BGT	1b				\n"
		"2:						\n"
		"MOV	%0,%2			\n"
		"B		4f				\n"
		"3:						\n"
		"RBIT	r3,r3			\n"
		"CLZ	r3,r3			\n"
		"SUBS	%3,%3,r3        \n"
		"BLT    2b				\n"
		"SUB	%0,%2,%3		\n"
		"4:						\n"
		:"=r" (iRet)
		:"r"  (pMem),
		 "r"  (uCount),
		 "r"  (uOffset)
		:"r3","r4","cc" );

	return iRet;
	}

#elif defined(BIT_GNU_ARM) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int	 iRet;

    asm volatile(
		"CMP	%3,%2			\n"
		"BHS	2f				\n"
		"UBFX 	r3,%3,#0,#3		\n"
		"ADD	%1,%1,%3,LSR #3	\n"
		"MOV	r4,#-1			\n"
		"BFC	%3,#0,#3		\n"
		"LSL	r4,r4,r3		\n"
		"						\n"
		"UBFX 	r3,%1,#0,#2		\n"
		"LSL	r3,r3,#3		\n"
		"LSL	r4,r4,r3		\n"
		"SUB	%3,%3,r3		\n"
		"BFC	%1,#0,#2		\n"
		"SUB	%3,%2,%3		\n"
		"						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"BICS	r3,r4,r3		\n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BLE	2f				\n"
		"1:						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"MVNS   r3,r3           \n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BGT	1b				\n"
		"2:						\n"
		"MOV	%0,%2			\n"
		"B		4f				\n"
		"3:						\n"
		"RBIT	r3,r3			\n"
		"CLZ	r3,r3			\n"
		"SUBS	%3,%3,r3        \n"
		"BLT    2b				\n"
		"SUB	%0,%2,%3		\n"
		"4:						\n"
		:"=r" (iRet)
		:"r"  (pMem),
		 "r"  (uCount),
		 "r"  (uOffset)
		:"r3","r4","cc" );

	return iRet;
	}

#else

	int _cdecl bit_find_0(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned   *pData;
	unsigned	uBits;
	unsigned	uMask;
	unsigned	uEnd;
	unsigned 	uPos;
	unsigned 	uSub;

	if(uOffset>=uCount)return uCount;

	uSub	= (unsigned)pMem&3;
	pData	= (unsigned*)((char*)pMem-uSub);
	uPos	= uSub*8;
	uEnd	= uPos+uCount;
	uPos   += uOffset;
	uMask	= 0xFFFFFFFF<<(uEnd&0x1F);
	uBits	= 0xFFFFFFFF<<(uPos&0x1F);
	uOffset = (uPos&~0x1F)-uSub*8;
	uEnd	= (uEnd-1)/32;
	uPos   /= 32;
	uBits  &= ~pData[uPos];


	for(;;)											// Durchsuche 32 Bitpakete
		{
		if(uPos>=uEnd)
			{
			if(uMask==0xFFFFFFFF)break;
			uBits |= uMask;
			break;
			}

		if(uBits)break;

		uPos    +=  1;
		uOffset +=  32;
		uBits	 = ~pData[uPos];
		}

	if(uBits&0xFF)return uOffset+ucBitOffset[uBits&0xFF];
	   uBits>>=8;
	if(uBits&0xFF)return uOffset+ucBitOffset[uBits&0xFF]+8;
	   uBits>>=8;
	if(uBits&0xFF)return uOffset+ucBitOffset[uBits&0xFF]+16;
	   uBits>>=8;
				  return uOffset+ucBitOffset[uBits&0xFF]+24;
	}

#endif

//*****************************************************************************
//*
//*		bit_find_0x
//*
//*****************************************************************************
//	Sucht das nchste gelschte Bit in einem Speicherbereich ab einem Offset.
//	pMem		: Ist der Zeiger auf den Speicherbereich.
//	uCount		: Ist die Gre der Speicherbereiches in Bits.
//	uOffset		: Ist die Bitposition ab der gesucht wird
//	Ergibt die Position an der das Bit gefunden wurde, bzw. die Gre des Bereiches.

#if defined(BIT_KEIL_THUMB)

__asm int bit_find_0x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
		// R0 = pMem
		// R1 = uCount
		// R2 = uOffset

		PUSH	{r4}						// Register retten
		CMP		r2,r1						// if(uOffset>=uCount)...
		BHS		stop_h						// 			*
		UBFX 	r3,r2,#0,#3					// r3      = BitOffset
		ADD		r0,r0,r2,LSR #3				// pMem   += uOffset/8
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		BFC		r2,#0,#3					// uOffset&= ~7
		LSR		r4,r4,r3					// r4(mask)= 0xFFFFFFFF>>(uOffset&7)

		UBFX 	r3,r0,#0,#2					// r3      =  pMem&3;
		LSL		r3,r3,#3					// r3 	  *=  8
		LSR		r4,r4,r3					// mask  >>= (pMem&3)*8
		SUB		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3
		SUB		r2,r1,r2					// r2(pos) = uCount-r2

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
		// R4 = Maske fr estes DWORD

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		BICS	r3,r4,r3					// DWORD  &= ~DWORD&mask
		BNE		find_h						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_h						// if(pos<=0)...
next_h
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		MVNS    r3,r3						// DWORD = ~DWORD
		BNE		find_h						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_h						// if(pos<=0)...
stop_h
		MOV		r0,r1						// return uCount;
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
find_h
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_h						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
	}

#elif defined(BIT_KEIL_ARM)

__asm int bit_find_0x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
		// R0 = pMem
		// R1 = uCount
		// R2 = uOffset

		PUSH	{r4}						// Register retten
		CMP		r2,r1						// if(uOffset>=uCount)...
		BHS		stop_h						// 			*
		UBFX 	r3,r2,#0,#3					// r3      = BitOffset
		ADD		r0,r0,r2,LSR #3				// pMem   += uOffset/8
		MOV	    r4,#-1						// r4(mask)= 0xFFFFFFFF
		BFC		r2,#0,#3					// uOffset&= ~7
		LSR		r4,r4,r3					// r4(mask)= 0xFFFFFFFF>>(uOffset&7)

		UBFX 	r3,r0,#0,#2					// r3      =  pMem&3;
		LSL		r3,r3,#3					// r3 	  *=  8
		LSR		r4,r4,r3					// mask  >>= (pMem&3)*8
		SUB		r2,r2,r3					// r2(pos)-= (pMem&3)*8
		BFC		r0,#0,#2					// pMem   &= ~3
		SUB		r2,r1,r2					// r2(pos) = uCount-r2

		// R0 = pMem auf Offset und Align 4 abgerundet
		// R1 = uCount (Gesamtanzahl der Bits)
		// R2 = pos (Bits von R0 bis zum Ende)
		// R4 = Maske fr estes DWORD

		LDR		r3,[r0],#4					// Lade erste DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		BICS	r3,r4,r3					// DWORD  &= ~DWORD&mask
		BNE		find_h						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BLE		stop_h						// if(pos<=0)...
next_h
		LDR		r3,[r0],#4					// Lade nchtes DWORD
		#ifndef __BIG_ENDIAN
		REV		r3,r3						// Big zu Little-Endian
		#endif
		MVNS    r3,r3                       // DWORD = ~DWORD
		BNE		find_h						// if(DWORD==0)...
		SUBS	r2,r2,#32					// pos    -= 32
		BGT		next_h						// if(pos<=0)...
stop_h
		MOV		r0,r1						// return uCount;
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
find_h
		CLZ		r3,r3						// r3      = erstes high Bit
		SUBS	r2,r2,r3                    // pos    -= erstes high Bit
		BLT    	stop_h						// if(ret<uCount)...
		SUB		r0,r1,r2					// ret     = uCount-pos
		POP		{r4}						// Register wieder herstellen
		BLX		lr							//		*
	}

#elif defined(BIT_IAR_ARM)

	// this functions are defined in "./Assembler/BitFunctionsIar[Thumb].s"

#elif defined(BIT_GNU_X86) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int     iRet;

    asm volatile(
		"pushl  %%ebx		    \n"
        "movl   %%eax,%%ebx     \n"
        "movl   %%eax,%%ecx     \n"
        "shrl   $5   ,%%ebx     \n"
        "andb   $31  ,%%cl      \n"
        "shll   $2   ,%%ebx     \n"
        "subb   %%cl ,%%al      \n"
        "addl   %1   ,%%ebx     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movl   (%%ebx),%%esi   \n"
        "notl   %%esi           \n"
		"bswap  %%esi           \n"
        "shll   %%cl ,%%esi     \n"
        "shrl   %%cl ,%%esi     \n"
        "bsrl   %%esi,%%ecx     \n"
        "jnz    2f              \n"
        "1:                     \n"
        "addl   $4   ,%%ebx     \n"
        "addl   $32  ,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movl   (%%ebx),%%esi   \n"
        "notl   %%esi           \n"
		"bswap  %%esi           \n"
        "bsrl   %%esi,%%ecx     \n"
        "jz     1b              \n"
        "2:                     \n"
		"xorl   $31  ,%%ecx     \n"
        "addl   %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%ebx		    \n"
        :"=a" (iRet)
        :"m" (pMem), "D" (uCount), "a" (uOffset)
        :"ecx","esi", "cc"
        );

    return iRet;
	}

#elif defined(BIT_GNU_X64) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int     iRet;

    asm volatile(
		"pushl  %%rbx		    \n"
        "movl   %%eax,%%ebx     \n"
        "movl   %%eax,%%ecx     \n"
        "shrl   $6   ,%%ebx     \n"
        "andb   $63  ,%%cl      \n"
        "shll   $3   ,%%ebx     \n"
        "subb   %%cl ,%%al      \n"
        "andq $0xFFFFFFFF,%%rbx \n"
        "addq   %1   ,%%rbx     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movq   (%%rbx),%%rsi   \n"
        "notq   %%rsi           \n"
		"bswap  %%rsi           \n"
        "shlq   %%cl ,%%rsi     \n"
        "shrq   %%cl ,%%rsi     \n"
        "bsrq   %%rsi,%%ecx     \n"
        "jnz    2f              \n"
        "1:                     \n"
        "addq   $8   ,%%rbx     \n"
        "addl   $64  ,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jnb    3f              \n"
        "movq   (%%rbx),%%rsi   \n"
        "notq   %%rsi           \n"
		"bswap  %%rsi           \n"
        "bsrq   %%rsi,%%rcx     \n"
        "jz     1b              \n"
        "2:                     \n"
		"xorl   $63  ,%%ecx     \n"
        "addl   %%ecx,%%eax     \n"
        "cmpl   %%edi,%%eax     \n"
        "jbe    4f              \n"
        "3:                     \n"
        "movl   %%edi,%%eax     \n"
        "4:                     \n"
		"popl   %%rbx		    \n"
        :"=a" (iRet)
        :"m" (pMem), "D" (uCount), "a" (uOffset)
        :"rcx","rsi", "cc"
        );

  	return iRet;
	}

#elif defined(BIT_MSC_X64) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned __int64 *pMem64;
	unsigned __int64  uValue;
	unsigned long  	  uPos;
	unsigned		  uMax;
	unsigned		  uSub;

	if(uOffset>=uCount)return uCount;

	uSub	 = ((unsigned)pMem)&7;
	pMem64	 = (unsigned __int64*)(((char*)pMem)-uSub);
	uSub	*= 8;
	uCount	+= uSub;
	uOffset	+= uSub;
	uValue	 = ~pMem64[uOffset/64];
	uMax	 = (uCount-1)&~0x3F;
	uValue	 = _byteswap_uint64(uValue);

	if(uOffset&0x3F)
		{
		uValue  &= (__int64)0xFFFFFFFFFFFFFFFF>>(uOffset&0x3F);
		uOffset	&= ~0x3F;
		}

	while(uOffset<uMax)
		{
		if(_BitScanReverse64(&uPos,uValue))
			{
			uPos ^= 0x3F;
			return uOffset+uPos-uSub;
			}

		uOffset += 64;
		uValue   = ~pMem64[uOffset/64];
		uValue	 = _byteswap_uint64(uValue);
		}

	if(uCount&0x3F)
		{
		uValue |= (__int64)0xFFFFFFFFFFFFFFFF>>(uCount&0x3F);
		}

	if(_BitScanReverse64(&uPos,uValue))
		{
		uPos ^= 0x3F;
		return uOffset+uPos-uSub;
		}

	return uCount-uSub;
	}

#elif defined(BIT_MSC_INSTR) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned		 *pMem32;
	unsigned		  uValue;
	unsigned long  	  uPos;
	unsigned		  uMax;
	unsigned		  uSub;

	if(uOffset>=uCount)return uCount;

	uSub	 = ((unsigned)pMem)&3;
	pMem32	 = (unsigned*)(((char*)pMem)-uSub);
	uSub	*= 8;
	uCount	+= uSub;
	uOffset	+= uSub;
	uValue	 = ~pMem32[uOffset/32];
	uMax	 = (uCount-1)&~0x1F;
	uValue	 = _byteswap_ulong(uValue);

	if(uOffset&0x1F)
		{
		uValue  &= 0xFFFFFFFF>>(uOffset&0x1F);
		uOffset	&= ~0x1F;
		}

	while(uOffset<uMax)
		{
		if(_BitScanReverse(&uPos,uValue))
			{
			uPos ^= 0x1F;
			return uOffset+uPos-uSub;
			}

		uOffset += 32;
		uValue   = ~pMem32[uOffset/32];
		uValue	 = _byteswap_ulong(uValue);
		}

	if(uCount&0x1F)
		{
		uValue |= 0xFFFFFFFF>>(uCount&0x1F);
		}

	if(_BitScanReverse(&uPos,uValue))
		{
		uPos ^= 0x1F;
		return uOffset+uPos-uSub;
		}

	return uCount-uSub;
	}

#elif defined(BIT_MSC_VC60) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	_asm{
		mov		eax,uOffset			;// Offset laden
		mov     ebx,eax				;// 		*
		mov     ecx,eax				;// 		*

		shr     ebx,5				;// EBX = Byte-Offset vom Start-Bit
		and		cl ,1Fh				;// ESI = Start-Bit    auf 32 gerundet
		mov		edi,uCount			;// EDI = Bit-Anzahl
		shl     ebx,2				;// EBX = Byte-Offset vom Start-Bit
		sub     al ,cl				;// EAX = Bit-Position auf 32 gerundet
		add		ebx,pMem			;// EBX = Byte-Adresse

		cmp		eax,edi				;// berlauf ?
		jnb     ende1				;// 		*
		mov     esi,[ebx]			;// Die ersten 32-Bit laden
		not		esi					;// 		*
		bswap   esi					;// Byte Reihenfolge drehen
 		shl		esi,cl				;// 		*
		shr		esi,cl				;// 		*
		bsr	    ecx,esi				;// Suche Bit
		jnz		done				;// 		*
	loop1:
		add		ebx,4				;// Die nchsten 32-Bit
		add		eax,32				;// 	*
		cmp		eax,edi				;// 	*
		jnb     ende1				;// berlauf ?
		mov	    esi,[ebx]			;// Suche Bit
		not		esi					;// 		*
		bswap   esi					;// Byte Reihenfolge drehen
		bsr	    ecx,esi				;// Suche Bit
		jz	    loop1
	done:
		xor     ecx,1Fh				;// Bitposition korrigiren
		add     eax,ecx				;// Ergebnis berechen
      	cmp		eax,edi				;// 	*
		jbe     ende2				;// Gab es einen berlauf

	ende1:
		mov		eax,edi				;// berlauf
	ende2:
		}
	}

#elif defined(BIT_GNU_THUMB) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int	 iRet;

    asm volatile(
		"CMP	%3,%2			\n"
		"BHS	2f				\n"
		"UBFX 	r3,%3,#0,#3		\n"
		"ADD	%1,%1,%3,LSR #3	\n"
		"MOV	r4,#-1			\n"
		"BFC	%3,#0,#3		\n"
		"LSR	r4,r4,r3		\n"
		"						\n"
		"UBFX 	r3,%1,#0,#2		\n"
		"LSL	r3,r3,#3		\n"
		"LSR	r4,r4,r3		\n"
		"SUB	%3,%3,r3		\n"
		"BFC	%1,#0,#2		\n"
		"SUB	%3,%2,%3		\n"
		"						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ != __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"BICS	r3,r4,r3		\n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BLE	2f				\n"
		"1:						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ != __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"MVNS   r3,r3           \n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BGT	1b				\n"
		"2:						\n"
		"MOV	%0,%2			\n"
		"B		4f				\n"
		"3:						\n"
		"CLZ	r3,r3			\n"
		"SUBS	%3,%3,r3        \n"
		"BLT    2b				\n"
		"SUB	%0,%2,%3		\n"
		"4:						\n"
		:"=r" (iRet)
		:"r"  (pMem),
		 "r"  (uCount),
		 "r"  (uOffset)
		:"r3","r4","cc" );

	return iRet;
	}

#elif defined(BIT_GNU_ARM) && !defined(NO_BIT_ASM)

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	int	 iRet;

    asm volatile(
		"CMP	%3,%2			\n"
		"BHS	2f				\n"
		"UBFX 	r3,%3,#0,#3		\n"
		"ADD	%1,%1,%3,LSR #3	\n"
		"MOV	r4,#-1			\n"
		"BFC	%3,#0,#3		\n"
		"LSR	r4,r4,r3		\n"
		"						\n"
		"UBFX 	r3,%1,#0,#2		\n"
		"LSL	r3,r3,#3		\n"
		"LSR	r4,r4,r3		\n"
		"SUB	%3,%3,r3		\n"
		"BFC	%1,#0,#2		\n"
		"SUB	%3,%2,%3		\n"
		"						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ != __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"BICS	r3,r4,r3		\n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BLE	2f				\n"
		"1:						\n"
		"LDR	r3,[%1],#4		\n"
		#if __BYTE_ORDER__ != __ORDER_LITTLE_ENDIAN__
		"REV	r3,r3			\n"
		#endif
		"MVNS   r3,r3           \n"
		"BNE	3f				\n"
		"SUBS	%3,%3,#32		\n"
		"BGT	1b				\n"
		"2:						\n"
		"MOV	%1,%2			\n"
		"B		4f				\n"
		"3:						\n"
		"CLZ	r3,r3			\n"
		"SUBS	%3,%3,r3        \n"
		"BLT    2b				\n"
		"SUB	%1,%2,%3		\n"
		"4:						\n"
		:"=r" (iRet)
		:"r"  (pMem),
		 "r"  (uCount),
		 "r"  (uOffset)
		:"r3","r4","cc" );

	return iRet;
	}

#else

	int _cdecl bit_find_0x(const void *pMem,unsigned uCount,unsigned uOffset)
	{
	unsigned   *pData;
	unsigned	uBits;
	unsigned	uMask;
	unsigned	uEnd;
	unsigned 	uPos;
	unsigned 	uSub;

	if(uOffset>=uCount)return uCount;

	uSub	= (unsigned)pMem&3;
	pData	= (unsigned*)((char*)pMem-uSub);
	uPos	= uSub*8;
	uEnd	= uPos+uCount;
	uPos   += uOffset;
	uMask	= 0xFFFFFFFF>>(uEnd&0x1F);
	uBits	= 0xFFFFFFFF>>(uPos&0x1F);
	uOffset = (uPos&~0x1F)-uSub*8;
	uEnd	= (uEnd-1)/32;
	uPos   /= 32;
	uBits  &= btol_dw(~pData[uPos]);


	for(;;)											// Durchsuche 32 Bitpakete
		{
		if(uPos>=uEnd)
			{
			if(uMask==0xFFFFFFFF)break;
			uBits |= uMask;
			break;
			}

		if(uBits)break;

		uPos	+= 1;
		uOffset += 32;
		uBits	 = btol_dw(~pData[uPos]);
		}

	if(uBits&0xFF000000)return uOffset+ucBitOffsetX[(uBits>>24)&0xFF];
	if(uBits&0x00FF0000)return uOffset+ucBitOffsetX[(uBits>>16)&0xFF]+8;
	if(uBits&0x0000FF00)return uOffset+ucBitOffsetX[(uBits>> 8)&0xFF]+16;
				        return uOffset+ucBitOffsetX[(uBits    )&0xFF]+24;
	}

#endif

//*****************************************************************************
//*
//*		bit_count_0
//*
//*****************************************************************************
//	Zhlt die Low-Bits in einem Speicherbereich, beginnent bei 0x80 zu 0x01.
//	pMem	: Zeiger auf das Bit-Array
//	uSize	: Gre des Bit-Arrays in Bits
//	uOffset	: Bit ab dem die Low-Bits gesucht werden
//	Ergibt die Anzahl der gefunden Low-Bits.

#if defined(_DEBUG)

	int _cdecl bit_count_0(const void *pMem,unsigned uSize,unsigned uOffset)
	{

	if(uOffset>=uSize)return 0;

	return uSize-uOffset-bit_count_1(pMem,uSize,uOffset);
	}

#endif

//*****************************************************************************
//*
//*		bit_count_0x
//*
//*****************************************************************************
//	Zhlt die Low-Bits in einem Speicherbereich, beginnent bei 0x80 zu 0x01.
//	pMem	: Zeiger auf das Bit-Array
//	uSize	: Gre des Bit-Arrays in Bits
//	uOffset	: Bit ab dem die Low-Bits gesucht werden
//	Ergibt die Anzahl der gefunden Low-Bits.

#if defined(_DEBUG)

	int _cdecl bit_count_0x(const void *pMem,unsigned uSize,unsigned uOffset)
	{

	if(uOffset>=uSize)return 0;

	return uSize-uOffset-bit_count_1x(pMem,uSize,uOffset);
	}

#endif

//*****************************************************************************
//*
//*		bit_count_1
//*
//*****************************************************************************
//	Zhlt die High-Bits in einem Speicherbereich, beginnent bei 0x01 zu 0x80.
//	pMem	: Zeiger auf das Bit-Array
//	uSize	: Gre des Bit-Arrays in Bits
//	uOffset	: Bit ab dem die High-Bits gesucht werden
//	Ergibt die Anzahl der gefunden High-Bits.
int _cdecl bit_count_1(const void *pMem,unsigned uSize,unsigned uOffset)
{
unsigned   *pData=(unsigned*)pMem;
unsigned	uIndex;
unsigned	uCount;
unsigned	uStop;
unsigned	uBits;
unsigned	uVal;



	if(uOffset>=uSize)return 0;

	uVal = ((unsigned)pData)&3;

	if(uVal)											// Auf eine gerade Adresse
		{
		*(unsigned*)&pData &= ~3;
		uSize			   +=  8*uVal;
		uOffset			   +=  8*uVal;
		}


	uIndex 	= uOffset/32;
	uStop	= uSize  /32;


	if(uIndex==uStop)									// Bereich kleiner als 4 Byte
		{
		uBits   = B_ORDER_DW(pData[uIndex]);

		uBits >>= uOffset&0x1F;
		uBits <<= uOffset&0x1F;
		uBits <<= 32-(uSize&0x1F);
		uBits >>= 32-(uSize&0x1F);

		uVal	= 0x55555555&(uBits>>1);
		uBits  &= 0x55555555;
		uBits  += uVal;

		uVal	= 0x33333333&(uBits>>2);
		uBits  &= 0x33333333;
		uBits  += uVal;

		uVal	= 0x07070707&(uBits>>4);
		uBits  &= 0x07070707;
		uBits  += uVal;

		uVal	= 0x000F000F&(uBits>>8);
		uBits  &= 0x000F000F;
		uBits  += uVal;

		uVal	= 0x0000001F&(uBits>>16);
		uBits  &= 0x0000001F;
		uBits  += uVal;

		return uBits;
		}



	uBits = pData[uIndex];


	if(uOffset&0x1F)									// Erstes DWORD nicht voll Zhlen
		{
		uBits	= B_ORDER_DW(uBits);
		uBits >>= uOffset&0x1F;
		uBits <<= uOffset&0x1F;
		}

	uCount	= 0;


	while(uIndex<uStop)									// Bits bei einem DWORD zhlen
		{
		uVal	= 0x55555555&(uBits>>1);
		uBits  &= 0x55555555;
		uBits  += uVal;

		uVal	= 0x33333333&(uBits>>2);
		uBits  &= 0x33333333;
		uBits  += uVal;

		uVal	= 0x07070707&(uBits>>4);
		uBits  &= 0x07070707;
		uBits  += uVal;

		uVal	= 0x000F000F&(uBits>>8);
		uBits  &= 0x000F000F;
		uBits  += uVal;

		uVal	= 0x0000001F&(uBits>>16);
		uBits  &= 0x0000001F;
		uBits  += uVal;

		uCount += uBits;
		uIndex += 1;
		uBits	= pData[uIndex];
		}

	if(uSize&0x1F)										// Sind noch Bits im letzten DWORD
		{
		uBits	= B_ORDER_DW(uBits);
		uBits <<= 32-(uSize&0x1F);
		uBits >>= 32-(uSize&0x1F);

		uVal	= 0x55555555&(uBits>>1);
		uBits  &= 0x55555555;
		uBits  += uVal;

		uVal	= 0x33333333&(uBits>>2);
		uBits  &= 0x33333333;
		uBits  += uVal;

		uVal	= 0x07070707&(uBits>>4);
		uBits  &= 0x07070707;
		uBits  += uVal;

		uVal	= 0x000F000F&(uBits>>8);
		uBits  &= 0x000F000F;
		uBits  += uVal;

		uVal	= 0x0000001F&(uBits>>16);
		uBits  &= 0x0000001F;
		uBits  += uVal;

		uCount += uBits;
		}


return uCount;
}

//*****************************************************************************
//*
//*		bit_count_1x
//*
//*****************************************************************************
//	Zhlt die High-Bits in einem Speicherbereich, beginnent bei 0x80 zu 0x01.
//	pMem	: Zeiger auf das Bit-Array
//	uSize	: Gre des Bit-Arrays in Bits
//	uOffset	: Bit ab dem die High-Bits gesucht werden
//	Ergibt die Anzahl der gefunden High-Bits.
int _cdecl bit_count_1x(const void *pMem,unsigned uSize,unsigned uOffset)
{
unsigned   *pData=(unsigned*)pMem;
unsigned	uIndex;
unsigned	uCount;
unsigned	uStop;
unsigned	uBits;
unsigned	uVal;



	if(uOffset>=uSize)return 0;

	uVal = ((unsigned)pData)&3;

	if(uVal)											// Auf eine gerade Adresse
		{
		*(unsigned*)&pData &= ~3;
		uSize			   +=  8*uVal;
		uOffset			   +=  8*uVal;
		}


	uIndex 	= uOffset/32;
	uStop	= uSize  /32;


	if(uIndex==uStop)									// Bereich kleiner als 4 Byte
		{
		uBits   = B_XSWAP_DW(pData[uIndex]);

		uBits <<= uOffset&0x1F;
		uBits >>= uOffset&0x1F;
		uBits >>= 32-(uSize&0x1F);

		uVal	= 0x55555555&(uBits>>1);
		uBits  &= 0x55555555;
		uBits  += uVal;

		uVal	= 0x33333333&(uBits>>2);
		uBits  &= 0x33333333;
		uBits  += uVal;

		uVal	= 0x07070707&(uBits>>4);
		uBits  &= 0x07070707;
		uBits  += uVal;

		uVal	= 0x000F000F&(uBits>>8);
		uBits  &= 0x000F000F;
		uBits  += uVal;

		uVal	= 0x0000001F&(uBits>>16);
		uBits  &= 0x0000001F;
		uBits  += uVal;

		return uBits;
		}


	uBits = pData[uIndex];

	if(uOffset&0x1F)									// Erstes DWORD nicht voll Zhlen
		{
		uBits	= B_XSWAP_DW(uBits);
		uBits <<= uOffset&0x1F;
		uBits >>= uOffset&0x1F;
		}

	uCount	= 0;


	while(uIndex<uStop)									// Bits bei einem DWORD zhlen
		{
		uVal	= 0x55555555&(uBits>>1);
		uBits  &= 0x55555555;
		uBits  += uVal;

		uVal	= 0x33333333&(uBits>>2);
		uBits  &= 0x33333333;
		uBits  += uVal;

		uVal	= 0x07070707&(uBits>>4);
		uBits  &= 0x07070707;
		uBits  += uVal;

		uVal	= 0x000F000F&(uBits>>8);
		uBits  &= 0x000F000F;
		uBits  += uVal;

		uVal	= 0x0000001F&(uBits>>16);
		uBits  &= 0x0000001F;
		uBits  += uVal;

		uCount += uBits;
		uIndex += 1;
		uBits	= pData[uIndex];
		}

	if(uSize&0x1F)										// Sind noch Bits im letzten DWORD
		{
		uBits	=  B_XSWAP_DW(uBits);
		uBits >>=  32-(uSize&0x1F);

		uVal	= 0x55555555&(uBits>>1);
		uBits  &= 0x55555555;
		uBits  += uVal;

		uVal	= 0x33333333&(uBits>>2);
		uBits  &= 0x33333333;
		uBits  += uVal;

		uVal	= 0x07070707&(uBits>>4);
		uBits  &= 0x07070707;
		uBits  += uVal;

		uVal	= 0x000F000F&(uBits>>8);
		uBits  &= 0x000F000F;
		uBits  += uVal;

		uVal	= 0x0000001F&(uBits>>16);
		uBits  &= 0x0000001F;
		uBits  += uVal;

		uCount += uBits;
		}


return uCount;
}

